/*******************************************************************************
* FILE NAME: istpdata.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in istpdata.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #include <limits.h>
  #include <string.h>
  }

#include <istpdata.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTPDATA_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Define whiteSpace characters                                                 |
------------------------------------------------------------------------------*/
static const char    whiteSpace[] = "\t\n\v\f\r ";
static const IString emptyString;

/*------------------------------------------------------------------------------
| IStringParserData::IStringParserData                                         |
------------------------------------------------------------------------------*/
IStringParserData::IStringParserData ( const IString &text )
                  :parseText(text),
                   currentPosition(1),
                   patternPosition(1),
                   patternLength(0),
                   tokenArraySize(0),
                   tokenCount(0),
                   lastToken(0),
                   lastSkip(false),
                   usedTokens(0),
                   refs(1)
{
}

/*------------------------------------------------------------------------------
| IStringParserData::~IStringParserData                                        |
------------------------------------------------------------------------------*/
IStringParserData::~IStringParserData ()
{
  if (usedTokens)
    delete usedTokens;
}

/*------------------------------------------------------------------------------
| IStringParserData::processToken                                              |
|                                                                              |
| Parse the remaining string data into the token.  If this is a token reparse  |
| then do not save the token (since it will not be parsed again).              |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processToken(IString &token,
                                                   Boolean reparse)
{
  reparseLastToken();
  if (currentPosition > parseText.size())
    token = emptyString;
  else
    token = parseText.subString(currentPosition);
  lastToken = &token;
  lastSkip = false;
  if (!reparse)
    saveToken(lastToken);

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::processPattern (const IString &)                          |
|                                                                              |
| Searches for the specified pattern (from the end of the last pattern) and    |
| re-parses the text that preceded the match.  If the pattern was not found    |
| position at end of string text.                                              |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processPattern(const IString &pattern)
{
  unsigned long startPos = patternPosition + patternLength;
  patternPosition = parseText.indexOf(pattern, startPos);

  if (patternPosition > 0)
  {
    patternLength = pattern.size();
    if (tokenCount > 0)
      reparseTokens ( startPos, patternPosition );
  }
  else
  {
    patternPosition = parseText.size() + 1;
    patternLength = 0;
  }

  currentPosition = patternPosition + patternLength;
  clearSavedTokens();

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::processPattern (const char *)                             |
|                                                                              |
| Searches for the specified pattern (from the end of the last pattern) and    |
| re-parses the text that preceded the match.  If the pattern was not found    |
| position at end of string text.                                              |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processPattern(const char *pattern)
{
  unsigned long startPos = patternPosition + patternLength;
  patternPosition = parseText.indexOf(pattern, startPos);

  if (patternPosition > 0)
  {
    patternLength = strlen(pattern);
    if (tokenCount > 0)
      reparseTokens ( startPos, patternPosition );
  }
  else
  {
    patternPosition = parseText.size() + 1;
    patternLength = 0;
  }

  currentPosition = patternPosition + patternLength;
  clearSavedTokens();

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::processPattern (char)                                     |
|                                                                              |
| Searches for the specified pattern (from the end of the last pattern) and    |
| re-parses the text that preceded the match.  If the pattern was not found    |
| position at end of string text.                                              |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processPattern(char pattern)
{
  unsigned long startPos = patternPosition + patternLength;
  patternPosition = parseText.indexOf(pattern, startPos);

  if (patternPosition > 0)
  {
    patternLength = 1;
    if (tokenCount > 0)
      reparseTokens ( startPos, patternPosition );
  }
  else
  {
    patternPosition = parseText.size() + 1;
    patternLength = 0;
  }

  currentPosition = patternPosition + patternLength;
  clearSavedTokens();

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::processSkip                                               |
|                                                                              |
| Parse the remaining string data as if a token was specified, except in this  |
| case the token does not exist.  If this is a reparse of a skip, do not save  |
| the skip marker in the token list.                                           |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processSkip(Boolean reparse)
{
  reparseLastToken();
  lastToken = 0;
  lastSkip = true;
  if (!reparse)
    saveToken(0);

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::processTest                                               |
|                                                                              |
| Use the IStringTest object to set the parse position                         |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::processTest(const IStringTest &test)
{
  unsigned long position =
                parseText.indexOf(test,patternPosition + patternLength);
  if (position == 0)
    position = parseText.size() + 1;
  return setPosition(position);
}

/*------------------------------------------------------------------------------
| IStringParserData::setPosition                                               |
|                                                                              |
| Sets the current location for parsing to the position specified in the       |
| original parse text string.  Since positioning acts like a pattern match,    |
| the pattern position will be updated to reflect the positioning.             |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::setPosition(unsigned long position)
{
  if (position == 0)
    position = 1;

  if (position > parseText.size())
  {
    patternPosition = parseText.size() + 1;
  }
  else
  {
    unsigned long startPos = patternPosition + patternLength;
    if ((position > startPos) && (tokenCount > 0))
      reparseTokens(startPos, position);
    patternPosition = position;
  }

  patternLength = 0;
  currentPosition = patternPosition;
  clearSavedTokens();

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::changePosition (using signed value)                       |
|                                                                              |
| Sets the current location for parsing to the relative position specified.    |
| For negative values, do the processing here with underflow checking.  For    |
| positive values, use the unsigned long delta method.                         |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::changePosition(int delta)
{
  if (delta < 0)
  {
    unsigned long newPosition;
    if ((patternPosition < (-delta)) || (delta == INT_MIN))
      newPosition = 1;
    else
      newPosition = patternPosition + delta;

    if ((patternLength > 0) && (tokenCount > 0))
      reparseTokens(patternPosition,parseText.size() + 1);

    patternLength = 0;
    setPosition(newPosition);
  }
  else
  {
    changePosition((unsigned long) delta);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::changePosition (using unsigned long value)                |
|                                                                              |
| Sets the current location for parsing to the relative position specified     |
| with overflow checking.                                                      |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::changePosition(unsigned long delta)
{
  unsigned long newPosition = patternPosition + delta;
  if (patternPosition > (UINT_MAX - delta))
    newPosition = parseText.size() + 1;

  if ((patternLength > 0 ) && (tokenCount > 0) &&
      (newPosition > parseText.size()))
    reparseTokens(patternPosition, parseText.size() + 1);

  patternLength = 0;
  return setPosition(newPosition);
}

/*------------------------------------------------------------------------------
| IStringParserData::reparseLastToken                                          |
|                                                                              |
| This method is used to reparse the last parsed token so that it only         |
| contains the next word and then positions around the parsed word.  If no     |
| word is found, set the token to an empty string and position at end of text. |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::reparseLastToken()
{

  if (lastToken)
  {
  unsigned long startPos = parseText.indexOfAnyBut(whiteSpace, currentPosition);
    if (startPos > 0)
    {
      unsigned long stopPos = parseText.indexOfAnyOf(whiteSpace, startPos);
      if (stopPos > 0)
      {
         if (startPos == currentPosition)
           lastToken->remove(stopPos - startPos + 1);
         else
           *lastToken = parseText.subString(startPos, stopPos - startPos);
         currentPosition = stopPos + 1;
      }
      else
      {
         if (startPos > currentPosition)
           *lastToken = parseText.subString(startPos);
         currentPosition = parseText.size() + 1;
      }
    }
    else
    {
      *lastToken=emptyString;
      currentPosition = parseText.size() + 1;
    }
  }

  if (lastSkip)
  {
  unsigned long startPos = parseText.indexOfAnyBut(whiteSpace, currentPosition);
    if (startPos > 0)
    {
      unsigned long stopPos = parseText.indexOfAnyOf(whiteSpace, startPos);
      if (stopPos > 0)
      {
         currentPosition = stopPos + 1;
      }
      else
      {
         currentPosition = parseText.size() + 1;
      }
    }
    else
    {
      currentPosition = parseText.size() + 1;
    }
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::reparseTokens                                             |
|                                                                              |
| This method is used to reparse all saved tokens.  A reparse value of true is |
| used when calling processToken and processSkip so that the token pointer or  |
| skip marker is not saved for later parsing.                                  |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::reparseTokens(unsigned long startPos,
                                                    unsigned long stopPos)
{
  IStringParserData preText ( parseText.subString( startPos,
                              stopPos - startPos ));
  for ( int cursor = 0; cursor < tokenCount; cursor++ )
  {
    IString *token = usedTokens[cursor];
    if (token)
      preText.processToken(*token,true);
    else
      preText.processSkip(true);
  }
  clearSavedTokens();

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::saveToken                                                 |
|                                                                              |
| This method saves a token pointer or skip marker so that it can be reparsed  |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::saveToken(IString *token)
{
  if (!usedTokens)
  {
    usedTokens = new ITokenPointer[10];
    tokenArraySize = 10;
  }

  if (tokenArraySize <= tokenCount+1)
  {
     ITokenPointer *oldList = usedTokens;
     usedTokens = new ITokenPointer[tokenArraySize + 10];
     for (int tmp = 0; tmp < tokenCount; tmp++ )
       usedTokens[tmp] = oldList[tmp];
     delete oldList;
     tokenArraySize += 10;
  }

  usedTokens[tokenCount] = token;
  tokenCount++;

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::clearSavedTokens                                          |
|                                                                              |
| This method clears all references to saved tokens                            |
------------------------------------------------------------------------------*/
IStringParserData& IStringParserData::clearSavedTokens()
{

  tokenCount = 0;
  lastToken = 0;
  lastSkip = false;

  return *this;
}

/*------------------------------------------------------------------------------
| IStringParserData::addRef                                                    |
|                                                                              |
| Add a reference by incrementing the reference count                          |
------------------------------------------------------------------------------*/
void IStringParserData :: addRef ( )
{
  refs++;
}

/*------------------------------------------------------------------------------
| IStringParserData::removeRef                                                 |
|                                                                              |
| Remove a reference by decrementing the reference count. If the refernce      |
| count goes to 0 then delete this.                                            |
------------------------------------------------------------------------------*/
void IStringParserData :: removeRef ( )
{
  if ( --refs == 0 )
    delete this;
}

