/*******************************************************************************
* FILE NAME: itrace.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in itrace.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <itrace.hpp>

#ifdef IC_PMWIN
#define INCL_DOSSEMAPHORES
#define INCL_DOSINFOSEG
#define INCL_DOSPROCESS
#include <iwindefs.h>
#endif

#include <istring.hpp>
#include <istream.hpp>
#ifdef IC_WIN
#include <iplatfrm.hpp>
#endif


extern "C"
{
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include <imprintf.h>
#ifdef IC_MOTIF
  #include <unistd.h>
#endif
}


#define IC_MAX_THREADS 255


#pragma data_seg(ICLNonConst)
int ITrace::iClState = ITrace::uninitialized;

#ifdef IC_PMWIN
ITrace::Destination ITrace::iClTraceLocation = ITrace::queue;
#endif

#ifndef IC_PMWIN
ITrace::Destination ITrace::iClTraceLocation = ITrace::standardOutput;
#endif

static unsigned ITrace__uClIndent[IC_MAX_THREADS] = {0};
static unsigned long ITrace__ulClProcessId =0;
static IBase::Boolean ITrace__fCheckMemory = false;
static IBase::Boolean ITrace__fCheckStack = false;
static const unsigned INDENT_LEVEL = 2;
#ifdef IC_WIN
static unsigned long winThreadID[IC_MAX_THREADS] = {0};
#endif
#pragma data_seg()


#if defined (__SOM_ENABLED__)
#if defined (__RRBC_LIB__)
#pragma SOMAsDefault (on)
#else
#pragma SOMAsDefault (off)
#endif
#endif

class ITraceSetup
{
  public:

/*------------------------------------------------------------------------------
| ITraceSetup::traceOn                                                         |
|                                                                              |
| Returns true if tracing is on                                                |
------------------------------------------------------------------------------*/
static IBase::Boolean traceOn()
     {
#ifdef PTR7989
       if (ITrace::iClState == ITrace::uninitialized)
         initialize();
#else
       if (ITrace::iClState & ITrace::uninitialized)
         initialize();
#endif
       return (ITrace::iClState & ITrace::on);
     }
/*------------------------------------------------------------------------------
| ITraceSetup::initialize                                                      |
------------------------------------------------------------------------------*/
static void initialize();
};

#if defined (__SOM_ENABLED__)
#pragma SOMAsDefault (pop)
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ITRACE_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| ITrace::ITrace                                                               |
|                                                                              |
| Default constructor for tracing.                                             |
------------------------------------------------------------------------------*/
ITrace :: ITrace(const char* pszTraceName,
                 long lLineNo)
        : pszClTraceName(0)
{
   if(ITraceSetup::traceOn())
   {
      if (pszTraceName!=0)
      {
        pszClTraceName = (char*)pszTraceName;
        IString str(pszTraceName);
        if(isWriteLineNumberEnabled() && lLineNo>0)
        {
           str+=IString("(");
           str+=IString(lLineNo);
           str+=IString(")");
        }
        writeFormattedString(str, "+");
      }
      else
         pszClTraceName = 0;
   }
}


/*------------------------------------------------------------------------------
| ITrace::~ITrace                                                              |
|                                                                              |
| Write exit trace and flush the buffer.                                       |
------------------------------------------------------------------------------*/
ITrace :: ~ITrace()
{
   if(ITraceSetup::traceOn())
   {
      if (pszClTraceName!=0)
      {
        writeFormattedString(IString(pszClTraceName),"-");
      }
   }
}


/*------------------------------------------------------------------------------
| ITrace::write                                                                |
|                                                                              |
| Write an IString using writeFormattedString after adding the appropriate     |
| padding for indentation.                                                     |
------------------------------------------------------------------------------*/
void ITrace :: write(const IString& strString)
{
   if(ITraceSetup::traceOn())
   {
       writeFormattedString(strString, ">");
   }
#ifdef __DEBUG_ALLOC__
   if(ITrace__fCheckMemory)
     _heap_check();
#endif

}

/*------------------------------------------------------------------------------
| ITrace::write                                                                |
|                                                                              |
| Write a char array after adding the appropriate padding for indentation.     |
------------------------------------------------------------------------------*/
void ITrace::write(const char* pszText)
{
   write(IString(pszText));
}

/*------------------------------------------------------------------------------
| ITrace::writeFormattedString                                                 |
|                                                                              |
| Write a formatted IString using writeString.                                 |
------------------------------------------------------------------------------*/
void  ITrace :: writeFormattedString(const IString& strString,
                                     char* pszMarker)
{
   static unsigned long ulSequence = 0;

   unsigned long ulThreadId = threadId();
   if(*pszMarker=='-' && ITrace__uClIndent[ulThreadId] >= INDENT_LEVEL)
      ITrace__uClIndent[ulThreadId]-=INDENT_LEVEL;


   IString strOut("");
   IString strPrefix("");
   strPrefix.insert(pszMarker, ITrace__uClIndent[ulThreadId], ' ');

   if(isWritePrefixEnabled())
   {
      IString strThreadId  = IString(ulThreadId).rightJustify(2, '0');
      IString strProcId  = IString(ITrace__ulClProcessId).rightJustify(6, '0');
      IString strLineNo  = IString(ulSequence).rightJustify(8, '0');
      if(ITrace__fCheckStack)
      {
        IString strStack =
                        IString(ITrace::remainingStack()).rightJustify(7, '0');
        strPrefix = IString(" ") + strStack + " " + strPrefix;
      }
      strPrefix = strLineNo + " " + strProcId + ":" + strThreadId + " " +
          strPrefix;
   }


   // copy the string and add the prefix info where a newline found
   strOut = IString::change(strString, '\n', "\n" + strPrefix);
   strOut = strPrefix+strOut+"\n";

   writeString((char*)strOut);

   ulSequence++;
   if(*pszMarker=='+')
      ITrace__uClIndent[ulThreadId]+=INDENT_LEVEL;

}

/*------------------------------------------------------------------------------
| ITrace::threadId                                                             |
|                                                                              |
| Determine the threadId of the current thread                                 |
------------------------------------------------------------------------------*/
unsigned long ITrace::threadId()
{
#ifdef IC_PM
   unsigned long traceid;

   PTIB    ptib;      /*  Thread Information Block  */
   PPIB    ppib;      /*  Process Information Block */

   DosGetInfoBlocks(&ptib, &ppib);

   traceid = ptib->tib_ptib2->tib2_ultid;
   ITrace__ulClProcessId = ppib->pib_ulpid;
   return (traceid);
#endif
#ifdef IC_WIN
  ITrace__ulClProcessId = GetCurrentProcessId();
  unsigned long ourTid = GetCurrentThreadId();
  int index = ourTid % IC_MAX_THREADS;
  int count = IC_MAX_THREADS;
  while (count-- > 0 &&
         winThreadID[index] != ourTid && winThreadID[index] != 0)
  {
     if (++index >= IC_MAX_THREADS) index = 0;
  }
  if (count == 0)
     ourTid = 0;
  else
  {
     winThreadID[index] = ourTid;
     ourTid = index;
  }
  return (ourTid);
#endif

#ifdef IC_400
  return 1;
#endif

#ifdef IC_MVS
   unsigned long traceid = 1;
   ITrace__ulClProcessId = 1;
   return (traceid);
#endif

#ifdef IC_MOTIF
   unsigned long traceid = 1;
   ITrace__ulClProcessId = (unsigned long) getpid();
   return (traceid);
#endif


}

/*------------------------------------------------------------------------------
| ITrace::remainingStack                                                       |
|                                                                              |
| Determine the size of the remaining stack of the current thread.             |
------------------------------------------------------------------------------*/
unsigned long ITrace::remainingStack()
{

#ifdef IC_PM
   PTIB    ptib;      /*  Thread Information Block  */
   PPIB    ppib;      /*  Process Information Block */
   DosGetInfoBlocks(&ptib, &ppib);
   void
     *base = ptib->tib_pstack,
     *cur  = &ptib;
   return (char*)cur - (char*)base;
#endif

#ifdef IC_WIN
   CONTEXT context;
   context.ContextFlags = CONTEXT_CONTROL;
   GetThreadContext(GetCurrentThread(), &context);
   void
     *base = (void *)context.Esp,
     *cur  = &context;
   return (char*)cur - (char*)base;
#endif

#ifdef IC_MVS
  return 0;
#endif

#ifdef IC_400
  return 0;
#endif

#ifdef IC_MOTIF
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| ITrace::writeString                                                          |
|                                                                              |
| Write a text string to file.                                                 |
|                                                                              |
| Notes:                                                                       |
|                                                                              |
|   1) We cannot use cout or cerr. since these are C++ objects that may not    |
|      be initialized when this routine is called.                             |
|   2) For Unix, standardOutput and queue are treated identically.             |
------------------------------------------------------------------------------*/
void  ITrace :: writeString(char* pszString)
{

   switch(ITrace::iClTraceLocation)
   {
     case ITrace::standardError :
       {
#ifdef IC_WIN
         if (!IPlatform::isWinNT())
         {
           fprintf(stdout, "%s", pszString);
           fflush(stdout);
         }
         else
#endif
         {
           fprintf(stderr, "%s", pszString);
           fflush(stderr);
         }
         break;
       }

     case ITrace::standardOutput :
#if defined(IC_MVS) || defined(IC_MOTIF)
     case ITrace::queue :
#endif
       {
         fprintf(stdout, "%s", pszString);
         fflush(stdout);
         break;
       }

#if ( defined(IC_PM) || defined(IC_WIN) || defined(IC_400) )
     case ITrace::queue :
       {
         imprintf("%0.*s", IC_PRINTFMAXLEN, pszString);
         break;
       }
#endif

     case ITrace::file :
       {
         // check to see whether the environment variable
         // is defined
         char *envptr = getenv("ICLUI TRACEFILE");
         if (!envptr)
           envptr = getenv("ICLUI_TRACEFILE");

         // If (environment variable defined) {
         //   - strip the leading and trailing blanks from the value
         //   - uppercase it
         //   - use the uppercased value as filename
         //   - open the file
         //   - if (open successful) {
         //       - append the provided string the file
         //       - close the file
         //     } else {
         //       - do nothing
         //     }
         // } else {
         //   - do nothing
         // }
         if (envptr)
         {
           IString strTraceOutputFile(envptr);
           strTraceOutputFile.strip().upperCase();

           FILE *fp = fopen(strTraceOutputFile, "a");
           if (fp)
           {
             fprintf(fp, "%s", pszString);
             fclose(fp);
           }
         }
         break;
       }
   }
}



/*------------------------------------------------------------------------------
| ITraceSetup::initialize                                                      |
|                                                                              |
| Initialize uses the following environment variables to set up the tracing    |
| options.                                                                     |
|                                                                              |
| Environment variables and possible values:                                   |
|   ICLUI TRACE                                                                |
|      ON                - turns tracing on                                    |
|      NOPREFIX          - no prefixes                                         |
|   ICLUI TRACETO                                                              |
|      STDERR | ERR      - standard error for tracing output                   |
|      STDOUT | OUT      - standard output for tracing output                  |
|      QUEUE  | PMPRINTF - tracing output to queue                             |
|   ICLUI CHECKMEMORY                                                          |
|      ON | TRUE         - check memory                                        |
|   ICLUI CHECKSTACK                                                           |
|      ON | TRUE         - check stack                                         |
------------------------------------------------------------------------------*/

void ITraceSetup::initialize()
{
   char *envptr;
   /******************************************************************/
   /* Reset uninitialized bit and setup trace defaults.  OR values   */
   /* in to prevent resetting other trace bits that may have been    */
   /* set PRIOR to initialization being called.                      */
   /******************************************************************/
   ITrace::iClState &= ~ITrace::uninitialized;
   ITrace::iClState |= ITrace::writeLineNumber | ITrace::writePrefix;

   /* Decide whether tracing is on */
   if (!(envptr = getenv("ICLUI TRACE")))
      envptr = getenv("ICLUI_TRACE");
   IString strTrace(envptr);
   strTrace.strip().upperCase();
   if(strTrace== "ON")
      ITrace::iClState |=ITrace::on;
   else if(strTrace== "NOPREFIX")
   {
      ITrace::iClState |=ITrace::on;
      ITrace::iClState &= ~ITrace::writePrefix;
   }

   /* decide where to send the output */
   if (!(envptr = getenv("ICLUI TRACETO")))
      envptr = getenv("ICLUI_TRACETO");
   IString strTraceTo(envptr);
   strTraceTo.strip().upperCase();

   if(strTraceTo.length() > 2)
   {
      /* Turn tracing on */
      ITrace::iClState |=ITrace::on;

      /* Setup output location; default is PM Queue  */
      if(strTraceTo=="STDERR" || strTraceTo=="ERR")
        ITrace::iClTraceLocation =ITrace::standardError;
      else if(strTraceTo=="STDOUT" || strTraceTo=="OUT")
        ITrace::iClTraceLocation=ITrace::standardOutput;
      else if(strTraceTo=="FILE")
        ITrace::iClTraceLocation=ITrace::file;
      else if(strTraceTo=="QUEUE" || strTraceTo=="PMPRINTF")
        ITrace::iClTraceLocation=ITrace::queue;
      else
        ITrace::iClTraceLocation=ITrace::queue;
   }
#ifdef __DEBUG_ALLOC__
   /* decide whether to check memory */
   if (!(envptr = getenv("ICLUI CHECKMEMORY")))
      envptr = getenv("ICLUI_CHECKMEMORY");
   IString strCheckMemory(envptr);
   strTraceTo.strip().upperCase();
   if(strCheckMemory.length() > 1)
   {
      if(strCheckMemory=="ON" | strCheckMemory=="TRUE")
        ITrace__fCheckMemory = true;
   }
#endif

   /* decide whether to check stack usage */
   if (!(envptr = getenv("ICLUI CHECKSTACK")))
      envptr = getenv("ICLUI_CHECKSTACK");
   IString strCheckStack(envptr);
   strCheckStack.strip().upperCase();
   if(strCheckStack.length() > 1)
   {
      if(strCheckStack=="ON" | strCheckStack=="TRUE")
        ITrace__fCheckStack = true;
   }
}

// Note: The following are normally inlined but cannot be in the Windows environment
//       due to the variables they reference not being exported (Win32S private data).
#ifdef IC_WIN
void ITrace :: enableWriteLineNumber()
{
  ITrace::iClState |= ITrace::writeLineNumber;
}

void ITrace :: disableWriteLineNumber()

{
  ITrace::iClState &= ~ITrace::writeLineNumber;
}


IBase::Boolean ITrace :: isWriteLineNumberEnabled()
{
  return (ITrace::iClState & ITrace::writeLineNumber);
}


void ITrace::enableTrace()
{
  ITrace::iClState |= ITrace::on;
}

void ITrace::disableTrace()
{
  ITrace::iClState &= ~ITrace::on;
}

IBase::Boolean ITrace::isTraceEnabled()
{
  return (ITrace::iClState & ITrace::on);
}

void ITrace::writeToQueue()
{
 ITrace::iClTraceLocation = ITrace::queue;
}

void ITrace::writeToStandardError()
{
   ITrace::iClTraceLocation = ITrace::standardError;
}

void ITrace::writeToStandardOutput()
{
   ITrace::iClTraceLocation = ITrace::standardOutput;
}

void ITrace::writeToFile()
{
   ITrace::iClTraceLocation = ITrace::file;
}

ITrace::Destination ITrace::traceDestination()
{
   return(ITrace::iClTraceLocation);
}

void ITrace::enableWritePrefix()
{
  ITrace::iClState |= ITrace::writePrefix;
}

void ITrace::disableWritePrefix()
{
  ITrace::iClState &= ~ITrace::writePrefix;
}

IBase::Boolean ITrace::isWritePrefixEnabled()
{
  return (ITrace::iClState & ITrace::writePrefix);
}
#endif
