/*******************************************************************************
* FILE NAME: icompnen.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IComponent                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "icompnen.hpp"

#include <icompimp.hpp>
#include <iguibndl.hpp>
#include <iembmod.hpp>
#include <iecmpimp.hpp>
#include <istatnry.hpp>
#include <idocstor.hpp>
#include <imetatyp.hpp>
#include <iview.hpp>
#include <inotifev.hpp>
#include <ititle.hpp>
#include <itime.hpp>
#include <icmpfram.hpp>

// have our component output to the Aswan CLASS debug channel
#define CH_DOCFWK_COMPONENT CH_CLASS_OBJINST

//   Components/Models are created as follows:
//
//   1. IComponentFactory::CreateInstance results in a call to:
// 
//        IComponentStationery::createComponent()
// 
//      Default implementation creates IComponent but user can override this
//      in a stationery subclass.
// 
//   2. Next, the model is created by:
// 
//        IComponentStationery::createModel()
// 
//      followed by a call to:
// 
//        IComponent::initialize(bundle, model);
// 
//      The created model is of the type specified in the stationery 
//      template parameter.
// 
//   3. During "initialize" IComponent adopts the model.
// 
//   4. Whenever IDocumentStorage::loadStorage or loadFile attempts to
//      ::resurrect a model from storage it assigns the result to a temp
//      pointer variable.  If the resurrection is successful, it will call
//      IComponent::adoptModel, thereby deleting the exiting (possibly empty)
//      model.
// 
//      Note: This algorithm is required because the user can stream in a
//      new model object (i.e. reload the file) at any time. The user might
//      also ask us to load from the wrong kind of file, in which case we
//      must end up with an empty model object rather then no model.
//
//      Note2: since the model object can be replaced at any time, all access
//      to it should be through the component.
// 

INotificationId
IComponent::ContentsReplacedId( "IComponent::ContentsReplacedId" );

// ***************************************************************************
//                          Initialization and termination
// ***************************************************************************

// -----------------------------------------------------------------------------
// IComponent
// -----------------------------------------------------------------------------
IComponent::IComponent() :
    fpModel(NULL),
    fpBundle(NULL),
    fpStorage(NULL),
    fpImplementation(new IComponentImp(this))
{
    
    // set initial extents to 250 x 250 pixels
    setInitialSize();

    // start notification
    fNotifier.enableNotification();
}

// -----------------------------------------------------------------------------
// ~IComponent
// -----------------------------------------------------------------------------
IComponent::~IComponent()
{
    fNotifier.disableNotification();

    delete fpModel;
    delete fpStorage;
    
    fpImplementation->Shutdown( FALSE );
    delete fpImplementation;

}
    

// -----------------------------------------------------------------------------
// initialize
// -----------------------------------------------------------------------------
void
IComponent::initialize( IGUIBundle& bundle, IModel* adoptedModel )
{
    ASSERT( fpBundle == NULL );
    ASSERT( fpModel == NULL );

    CAswResult  result(S_OK, ASD_OK);
    
    // Set the bundle pointer
    fpBundle = &bundle;

    // Setup the CAswanObjectInstance
    result = fpImplementation->Setup( &bundle, adoptedModel );

    IASWASSERTSTATE( result );

    // Adopt the passed in model
    adoptModel( adoptedModel );

    // if our model is an embedder, point embModel at it
    // else set embModel to NULL
    IEmbedderModel* embModel = NULL;
    ::dynamicCastTo( embModel, fpModel );
        
    // set up a storage for this component
    if ( embModel != NULL || isStructuredStorage() )
    {
        fpStorage = new IStructuredStorage( *this, embModel );
    }
    else
    {
        fpStorage = new IFlatFileStorage( *this );
    }

    fpImplementation->SetStorage( fpStorage );

}


// -----------------------------------------------------------------------------
// adoptModel
// -----------------------------------------------------------------------------
// Sets the model; deletes any old models; attaches this component into the model
void
IComponent::adoptModel(IModel* adoptedModel)
{
    delete fpModel;
    fpModel = adoptedModel;
    fpModel->attachTo( *this );
}
    

// -----------------------------------------------------------------------------
// orphanModel
// -----------------------------------------------------------------------------
IModel*
IComponent::orphanModel()
{
    IModel* model = fpModel;
    fpModel = NULL;
    return model;
}
    

// -----------------------------------------------------------------------------
// isStructuredStorage
// -----------------------------------------------------------------------------
Boolean
IComponent::isStructuredStorage() const
{
    return false; // use flat file storage by default
}


// -----------------------------------------------------------------------------
// setDocumentName
// -----------------------------------------------------------------------------
void
IComponent::setDocumentName( const IString& name )
{
    if ( name.length() == 0 )
    {
        // "Untitled" - force a unique name for the running object table
        // The app short name and the current time should do
        ITime theTime;
        IString theString( theTime.asString() );
        fUniqueDocumentName = IComponentStationery::stationery().shortName() + theString;

        // "untitled" into the save document name
        fDocumentName = (const char*)IBase::messageText( IC_CDF_UNTITLED_NAME );
    }
    else
    {
        fDocumentName = name;
        fUniqueDocumentName = name;
    }

    assert( fpBundle != NULL );
 
    if ( fpStorage != NULL ) // don't try to set title before frame is created
    {

        ITitle title( &fpBundle->frameWindow() );
        title.setTitleText( fDocumentName );

    }

    IEmbedderModel* embModel = NULL;
    ::dynamicCastTo( embModel, fpModel );

    if ( embModel )
    {
        CAswRootContainer* rootCntr = embModel->rootContainer();

        if( rootCntr )
        {
            CAswOLEString oleDocName( fDocumentName );
            rootCntr->SetDocumentName( oleDocName.GetOLEPtr() );

        } // this is a root container

    } // this is an embedder model

}


// -----------------------------------------------------------------------------
// isContainer
// -----------------------------------------------------------------------------
Boolean
IComponent::isContainer() const
{
    IEmbedderModel* pEmbModel = NULL;
    ::dynamicCastTo( pEmbModel, model() );
    return (pEmbModel != NULL);
}

// -----------------------------------------------------------------------------
// isEmbedded
// -----------------------------------------------------------------------------
Boolean
IComponent::isEmbedded() const
{
    // since this component is registered as a single use server it is OK to
    // return the stationery's is embedded. When this becomes a multiple-use
    // server we will have to return the bundle.component's is embedded.
    return IComponentStationery::stationery().isEmbedded();
}


// -----------------------------------------------------------------------------
// handleSetArea
// -----------------------------------------------------------------------------
// Request to set the size of the object (in himetric)
// Called in response to a SetExtent message.

// Note: The framework only handles a content aspect by default;
//       This method should be implemented in the derived class for other aspects
void
IComponent::handleSetArea( const EAspect aspect, const ISize& theSize )
{

    switch( aspect )
    {
        case kContent:
        {
            fExtent = theSize;
            break;
        }
    }
}


// -----------------------------------------------------------------------------
// handleGetArea
// -----------------------------------------------------------------------------
// return the size of the object (for all aspects) in HiMetric

// The "content" aspect size returned is the one set with "handleSetArea"
// Users should override this method in their derived class if any other behaviour
// is desired (eg. icon sizes of other than 32x32 etc).
ISize
IComponent::handleGetArea( const EAspect aspect ) const
{

    SIZEL sizeHi, sizePix;

    switch( aspect )
    {
        case kContent:
            // return the content extent
            return fExtent;

        case kThumbnail:
            // a thumbnail is always 120x120 pixels
            sizePix.cx = sizePix.cy = 120;
            XformSizeInPixelsToHimetric( NULL, &sizePix, &sizeHi );
            return ISize( sizeHi );

        case kIcon:
            // an icon is always 32x32
            sizePix.cx = sizePix.cy = 32;
            XformSizeInPixelsToHimetric( NULL, &sizePix, &sizeHi );
            return ISize( sizeHi );

        case kDocPrint:
            // we don't support printing in this release
            // so just return an arbitrary size.
            return ISize(10000, 10000);
    }

    // should never get here!
    // This means that the requested aspect was not one of the 4 supported.
    assert( false );
    return ISize(0,0);

}

// -----------------------------------------------------------------------------
// contentSize
// -----------------------------------------------------------------------------
// convenience function for returning the current content extent in pixels
ISize
IComponent::contentSize() const
{

    SIZEL sizeHi, sizePix;
    sizeHi.cx = fExtent.width();
    sizeHi.cy = fExtent.height();
    XformSizeInHimetricToPixels( NULL, &sizeHi, &sizePix );
    return ISize( sizePix );

}

// -----------------------------------------------------------------------------
// setInitialSize
// -----------------------------------------------------------------------------
// initialize extents to 250 x 250 pixels
void
IComponent::setInitialSize()
{

    SIZEL sizePix, sizeHi;
    sizePix.cx = sizePix.cy = 250;
    XformSizeInPixelsToHimetric( NULL, &sizePix, &sizeHi );
    fExtent = ISize( sizeHi );

}

const IString&
IComponent::uniqueDocumentName() const
{
    return fUniqueDocumentName;
}

// SMY0221
// this method is designed to be implemented in a subclass
// return true if you have handled the verb, false otherwise
Boolean
IComponent::checkVerb( long iVerb )
{
    // default is "not handled"
    return false;

}

