/*******************************************************************************
* FILE NAME: iembmod.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IEmbedderModel                                                           *
*     IEmbeddedComponent                                                       *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*   Portions or this component are based on work originating from Lotus.       *
*   (C) Copyright Lotus Development Corporation. 1996 - All Rights Reserved.   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "iembmod.hpp"

#include <iecmpimp.hpp>
#include <istatnry.hpp>
#include <icompimp.hpp>
#include <idocstor.hpp>
#include <ibasstrm.hpp>
#include <idataex.hpp>
#include <icmpfram.hpp>

// these were included solely for notification of resizing from grab handles
#include <iguibndl.hpp>
#include <iview.hpp>
#include <inotifev.hpp>

/*******************************************************************************
*                                                                              *
*                           IEmbeddedComponent                                 *
*                                                                              *
*******************************************************************************/

TypeExtensionMacro(IEmbeddedComponent)

// -----------------------------------------------------------------------------
// IEmbeddedComponent
// -----------------------------------------------------------------------------
// default constructor
IEmbeddedComponent::IEmbeddedComponent() :
    fArea( IPoint(0, 0), IPoint(200, 200) ),
    fAspect( IComponent::kContent ),
    fNumber( 0 ),
    fpBundle( NULL ),
    fpImplementation( NULL ),
    fpGrabHandles( NULL )
{
}


// -----------------------------------------------------------------------------
// IEmbeddedComponent
// -----------------------------------------------------------------------------
// copy constructor
IEmbeddedComponent::IEmbeddedComponent( const IEmbeddedComponent& theCopy )
{
    ASSERT( FALSE );
}

// -----------------------------------------------------------------------------
// ~IEmbeddedComponent
// -----------------------------------------------------------------------------
// destructor
IEmbeddedComponent::~IEmbeddedComponent()
{
    showGrabHandles( false );
    delete fpImplementation;
    
}
        
// -----------------------------------------------------------------------------
// preinitialize
// -----------------------------------------------------------------------------
void
IEmbeddedComponent::preinitialize( IGUIBundle& inBundle,
                                   CAswStorage* pFromStorage )
{
    CAswStorage* pStorageToUse;
    
    fpBundle = &inBundle;
    fpImplementation = new IEmbeddedComponentImp( this, fAspect );

    IEmbedderModel* pModel = NULL;
    ::dynamicCastTo( pModel, inBundle.component().model() );
    assert( pModel != NULL );

    if ( pFromStorage == NULL )
    {
        // not preinitializing from stream-in -> need to assign a number
        unsigned num = 0;
        if ( ! pModel->componentList().isEmpty() )
        {
            IEmbeddedComponent* lastComp = pModel->componentList().lastElement();
            num = lastComp->number() + 1;
        }
        fNumber = num;

        pStorageToUse = inBundle.component().storage().storage();
    }
    else
        pStorageToUse = pFromStorage;

    CAswOLEString pszOLEName( name() );
    
    // Set up and add the embedded component to the Aswan container
    fpImplementation->Setup( pModel->container(), pStorageToUse, pszOLEName );
//FB check for error here???    
    
}
                                  
// -----------------------------------------------------------------------------
// isInitialized
// -----------------------------------------------------------------------------
Boolean
IEmbeddedComponent::isInitialized() const
{
    return fpImplementation && fpImplementation->isInitialized();
}

// -----------------------------------------------------------------------------
// initialize
// -----------------------------------------------------------------------------
// this method creates an Aswan aspect (contents) and aswan rendering device (screen)
// and a pres site
void
IEmbeddedComponent::initialize( Boolean bNew, Boolean bActivate )
{
    // if new then we want to launch it, if not new then don't launch
    assert( fpImplementation );
    fpImplementation->initialize( bNew, bActivate );

    if ( bActivate )
    {
        IEmbedderModel* pModel = NULL;
        ::dynamicCastTo( pModel, fpBundle->component().model() );
        assert( pModel != NULL );

        pModel->selectComponent( this );
    }

}

// -----------------------------------------------------------------------------
// draw
// -----------------------------------------------------------------------------
// called by the view paint handler
void
IEmbeddedComponent::draw( IPresSpaceHandle& hdl,
                          const IRectangle& invalidArea,
                          Boolean metaFile )
{
    // call our draw method
    
    RECTL rectl;
    rectl.left = area().minX();
    rectl.top = area().minY();
    rectl.right = area().maxX();
    rectl.bottom = area().maxY();

    LPRECTL prcOuter = NULL;
    if( metaFile )
    {
        prcOuter = new RECTL;
        prcOuter->left   = invalidArea.minX();
        prcOuter->top    = invalidArea.minY();
        prcOuter->right  = invalidArea.width(); // NOTE: Aswan requires width and height
        prcOuter->bottom = invalidArea.height();// not bottom right coordinate

    } // is metaFile
    
//FB check if this really is/(needs to be) called before presSite is setup
    if( isInitialized() )
        fpImplementation->presSite().Draw( &rectl,
                                           NULL,
                                           (HDC)hdl.asUnsigned(),
                                           FALSE,
                                           NULL,
                                           prcOuter );

    delete prcOuter;

    // draw the grab handles (if any)
    if( fpGrabHandles && !metaFile )
        fpGrabHandles->draw( hdl );

}

// -----------------------------------------------------------------------------
// showGrabHandles
// -----------------------------------------------------------------------------
// displays or turns off the sizing grab handles. 
// This is not necessary for in-place objects as they supply their own
void
IEmbeddedComponent::showGrabHandles( Boolean bShowGrabHandles )
{
    // we do not provide handles for in-place active objects;
    // that is the server's responsibility
    if( bShowGrabHandles )
    {
        // turn on the grab handles
        assert( isInitialized() );
        
        delete fpGrabHandles;
        Boolean wantDrag = fpBundle->objectView().isDragSource();
        fpGrabHandles = new IGrabHandles( *this, wantDrag );

    }

    else
    {
        // turn off grab handles
        delete fpGrabHandles;
        fpGrabHandles = NULL;

    }
}

// -----------------------------------------------------------------------------
// goActive
// -----------------------------------------------------------------------------
// makes the component active
void
IEmbeddedComponent::goActive( LONG lVerb )
{
    assert( isInitialized() );

    // SMY29 
    // before calling activate we must make sure that the in-place parent
    // window is kept current.
    HWND hWnd = (HWND)(void*)fpBundle->objectView().handle();
    fpImplementation->presSite().ChangeInPlaceWnd( hWnd );

    CAswResult ar = fpImplementation->Activate( NULL,
                                                &fpImplementation->presSite(),
                                                lVerb );
    if ( ASWFAILED(ar) )
    {
        throw IAswException( ar );
    }

}

// -----------------------------------------------------------------------------
// goInactive
// -----------------------------------------------------------------------------
// makes the component inactive
void
IEmbeddedComponent::goInactive()
{
    if ( isInitialized() )
        fpImplementation->presSite().Deactivate();

}

// -----------------------------------------------------------------------------
// isInPlaceActive
// -----------------------------------------------------------------------------
Boolean
IEmbeddedComponent::isInPlaceActive() const
{
    assert( isInitialized() );
    return fpImplementation->presSite().isInPlace();
    
}

// -----------------------------------------------------------------------------
// setAspect
// -----------------------------------------------------------------------------
// set the display aspect for the embedded component
void
IEmbeddedComponent::setAspect( const IComponent::EAspect aspect )
{
    // Release 1 - dynamic aspect change not supported (by Aswan)
    if ( !fpImplementation || fpImplementation->isInitialized() )
//      throw IInvalidRequest( "dynamic aspect change currently not supported" );
        //FB use resource
        return;

    if ( fAspect != aspect )
    {
        IAswAspect* newAspect;
        switch ( aspect ) {
        case IComponent::kContent:
            newAspect = &fpImplementation->contentAspect();
            break;
        case IComponent::kIcon:
            newAspect = &fpImplementation->iconAspect();
            break;
        default:
//          throw IInvalidRequest( "requested aspect currently not supported" );
            //FB use resource
            //FB throw (not supported) exception ???
            return;
        }

        fpImplementation->switchAspect( *newAspect );
        fAspect = aspect;
    }
}

// -----------------------------------------------------------------------------
// objectArea
// -----------------------------------------------------------------------------
// call on the object to return its current size
IRectangle
IEmbeddedComponent::objectArea() const
{
    assert( isInitialized() );

    SIZEL sizeHi, sizePix;
    fpImplementation->aspect().GetExtent( &sizeHi );
    // size is in himetric units - convert to pixels
    XformSizeInHimetricToPixels( NULL, &sizeHi, &sizePix );
    return IRectangle( fArea.topLeft(), IPoint(sizePix.cx, sizePix.cy) );

}

// -----------------------------------------------------------------------------
// setObjectArea
// -----------------------------------------------------------------------------
void
IEmbeddedComponent::setObjectArea( const IRectangle& area )
{
    assert( isInitialized() );
    
    SIZEL sizeHi, sizePix;
    sizePix.cx = area.width();
    sizePix.cy = area.height();
    XformSizeInPixelsToHimetric( NULL, &sizePix, &sizeHi );

    // call on the pres site to request the extent change
    fpImplementation->presSite().RequestExtentChange ( &sizeHi );

}

// -----------------------------------------------------------------------------
// resize
// -----------------------------------------------------------------------------
// handles programmatic sizing of embedded component
void
IEmbeddedComponent::resize( const IRectangle& newRect )
{
    // update grab handles if present
    if ( fpGrabHandles )
    {
        fpGrabHandles->changeTo( newRect );
    }

    setArea( newRect );

    // send notification -> repaint the view
    IModel& model = *bundle().component().model();
    model.notifyOfChange(
        INotificationEvent( IEmbedderModel::ComponentAreaChangedId, model ) );

    // notify the server of the size change
    // the server may (and probably will) disregard this hint
    setObjectArea( newRect );

}

// -----------------------------------------------------------------------------
// rect
// -----------------------------------------------------------------------------
IRectangle
IEmbeddedComponent::rect() const
{
    return fArea;
}

// -----------------------------------------------------------------------------
// setRect
// -----------------------------------------------------------------------------
// handle sizing change notifications from grab handles
void
IEmbeddedComponent::setRect( const IRectangle& newRect )
{
    // fix up tool bar if necessary    
    ((IComponentFrameWindow&)fpBundle->frameWindow()).adjustToolBar();

    if ( newRect == fArea )
        return;

    setArea( newRect );

    // send notification -> repaint the view
    IModel& model = *bundle().component().model();
    model.notifyOfChange(
        INotificationEvent( IEmbedderModel::ComponentAreaChangedId, model ) );

    // notify the server of the size change
    // the server may (and probably will) disregard this hint
    setObjectArea( newRect );
    
}

// -----------------------------------------------------------------------------
// inWindow
// -----------------------------------------------------------------------------
IWindow&
IEmbeddedComponent::inWindow() const
{
    return bundle().objectView();
    
}

// -----------------------------------------------------------------------------
// doDragDrop
// -----------------------------------------------------------------------------
Boolean
IEmbeddedComponent::doDragDrop( const IPoint& startPoint )
{

    IPoint ptTopLeft( area().minX(), area().minY() );
    ftDragMouseOffset = startPoint - ptTopLeft;
    
    DWORD dwEffect;
    IAswDragSource& src = IComponentStationery::stationery().dragSource();
    CAswResult ar = src.DoDragDrop( DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK, &dwEffect );
//    IASWSHOWERROR( ar );
//    CAswResult ar = ((CAswFrameWindow&)fpBundle->frameWindow()).StartDragDrop(
//                &src, 
//                DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK,
//                &dwEffect );

    // if a move was requested delete the source object
    // unless it was an internal action (which is just a "move")
    if( ar.Scode() == DRAGDROP_S_DROP  && 
        dwEffect == DROPEFFECT_MOVE  &&
        !src.IsInternalAction() )
    {
        // get embedder model pointer
        IEmbedderModel* pEmbModel = NULL;
        ::dynamicCastTo( pEmbModel, fpBundle->component().model() );

        // fix up tool bar if necessary    
        ((IComponentFrameWindow&)fpBundle->frameWindow()).adjustToolBar();

        delete pEmbModel->orphan( *this );
        return true;

    } // DROPEFFECT_MOVE

    // fix up tool bar if necessary    
    ((IComponentFrameWindow&)fpBundle->frameWindow()).adjustToolBar();
    return false;

}

// -----------------------------------------------------------------------------
// operator>>=
// -----------------------------------------------------------------------------
IBaseStream&
IEmbeddedComponent::operator>>= ( IBaseStream& toStream ) const
{
    writeVersion( toStream );
    
    fNumber >>= toStream;

    (short)fAspect >>= toStream;
    
    fArea.minX() >>= toStream;
    fArea.minY() >>= toStream;
    fArea.maxX() >>= toStream;
    fArea.maxY() >>= toStream;
    
    return toStream;
}

// -----------------------------------------------------------------------------
// operator<<=
// -----------------------------------------------------------------------------
IBaseStream&
IEmbeddedComponent::operator<<= ( IBaseStream& fromStream )
{
    readVersion( fromStream );
    
    fNumber <<= fromStream;

    short tAspect;
    tAspect <<= fromStream;
    fAspect = (IComponent::EAspect) tAspect;

    IPair::Coord minx, miny, maxx, maxy;
    minx <<= fromStream;
    miny <<= fromStream;
    maxx <<= fromStream;
    maxy <<= fromStream;
    setArea( IRectangle( IPoint( minx, miny ), IPoint( maxx, maxy) ) );
    
    return fromStream;
}

IString
IEmbeddedComponent::name() const
{
    return IString( "EmbeddedComponent" ) + IString( fNumber );
}


// *******************************************************************************
// *                                                                              *
// *                           IEmbedderModel                                     *
// *                                                                              *
// *******************************************************************************/

TypeExtensionMacro(IEmbedderModel)

INotificationId
IEmbedderModel::AddedComponentId( "IEmbedderModel::AddedComponentId" );

INotificationId
IEmbedderModel::RemovedComponentId( "IEmbedderModel::RemovedComponentId" );

INotificationId
IEmbedderModel::ComponentAreaChangedId( "IEmbedderModel::ComponentAreaChangedId" );

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
IEmbedderModel::IEmbedderModel() :
    fpContainer( NULL ),
    fpRootContainer( NULL ),
    fpSelectedComponent( NULL )
{
}

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
IEmbedderModel::IEmbedderModel( const IEmbedderModel& theCopy ) :
    IModel( theCopy ),
    fpContainer( theCopy.fpContainer ),
    fpRootContainer( theCopy.fpRootContainer ),
    fpSelectedComponent( theCopy.fpSelectedComponent )
{
    assert( false );
}

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
IEmbedderModel::~IEmbedderModel()
{
    if ( fpContainer ) {
        // remove and delete all embedded components
        deleteAll();

        // shutdown and delete aswan container
        component().objectInstance()->SetContainerClass( NULL );
        fpContainer->Shutdown();
        delete fpContainer;

    }

}

Boolean
IEmbedderModel::hasChanged() const
{
    if ( IModel::hasChanged() )
        return true;

    if ( fpContainer )
    {
        CAswResult ar = fpContainer->IsAnyObjectDirty();
        IASWSHOWERROR( ar );

        if( ar.Scode() == S_OK )
            return true;

    }

    return false;
        
}

// -----------------------------------------------------------------------------
// attachTo
// -----------------------------------------------------------------------------
// override of the base class method. This will set the component, as well
// as create and set up the container
void
IEmbedderModel::attachTo( IComponent& comp )
{
    // call base class
    IModel::attachTo( comp );

    CAswResult result;
    IComponentStationery& theStationery = IComponentStationery::stationery();
    
    IComponentImp* objectInstance = comp.objectInstance();
        
    // create an Aswan container
    // if we are a root container (ie. not embedded) we need a CAswRootContainer
    // else just a CAswContainer
    if( comp.isEmbedded() )
    { 
        // we are not a root container
        fpContainer = new CAswContainer( theStationery );
        
        // set up the container
        result = fpContainer->Setup( NULL, TRUE, objectInstance );

    }
    
    else
    {
        // we are a root container
        fpRootContainer = new CAswRootContainer( theStationery );
        fpContainer = fpRootContainer;
        
        // set up the root container without a root file storage.
        // we'll set the root file storage later using SetRootStorage()
        result = fpRootContainer->Setup( NULL, TRUE, NULL, objectInstance );

    }
    
    // tell object instance about the container
    objectInstance->SetContainerClass( fpContainer );

    ASSERT ( ASWSUCCEEDED( result ) );
    
}

// -----------------------------------------------------------------------------
// deleteAll
// -----------------------------------------------------------------------------
// shut down and delete all embedded components prior to shutting down and
// deleting the model
void
IEmbedderModel::deleteAll()
{
    ICursor& cursor = *fComponentList.newCursor();

    for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext()) {
        
        IEmbeddedComponent* pEmbeddedComp = fComponentList.elementAt(cursor);

        pEmbeddedComp->containedObject()->Close( OLECLOSE_NOSAVE );
        pEmbeddedComp->containedObject()->Unload();
        pEmbeddedComp->containedObject()->Shutdown();

        delete pEmbeddedComp;
    }

    fComponentList.removeAll();

    delete &cursor;
    
}

// -----------------------------------------------------------------------------
// createEmbeddedComponent
// -----------------------------------------------------------------------------
IEmbeddedComponent*
IEmbedderModel::createEmbeddedComponent() const
{
    return new IEmbeddedComponent;
}

// -----------------------------------------------------------------------------
// selectComponent
// -----------------------------------------------------------------------------
void
IEmbedderModel::selectComponent( IEmbeddedComponent* embComp )
{

    if ( fpSelectedComponent != embComp )
    {
        if ( fpSelectedComponent )
        {
            fpSelectedComponent->goInactive();
            fpSelectedComponent->showGrabHandles( false );
        }

        if ( embComp && ! embComp->isInPlaceActive() )
            embComp->showGrabHandles();

        fpSelectedComponent = embComp;

        // no "select all" when we have a selected component
        if ( fpSelectedComponent )
            component().bundle().handleSelectAll( false );

    }
    
}

// -----------------------------------------------------------------------------
// adopt
// -----------------------------------------------------------------------------
// adopt and add to container a newly created component (component is not
// initialized at this point)
void
IEmbedderModel::adopt( IEmbeddedComponent* embeddedComp )
{

    embeddedComp->preinitialize( component().bundle() ); 
    
    // Add the embedded component to our list
    IElemPointer<IEmbeddedComponent> elem = embeddedComp;
    fComponentList.addAsLast ( elem );

    notifyOfChange( INotificationEvent( AddedComponentId, *this ) );
//    setChanged( true );

}

// -----------------------------------------------------------------------------
// orphan
// -----------------------------------------------------------------------------
// orphan and remove from container the speicified component.
IEmbeddedComponent*
IEmbedderModel::orphan( IEmbeddedComponent& embeddedComp )
{
    ICursor& cursor = *fComponentList.newCursor();

    // Search for and then remove the component from our and Aswans list
    for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
    {
        
        IEmbeddedComponent* pEmbeddedComp = fComponentList.elementAt(cursor);

        if ( pEmbeddedComp == &embeddedComp )
        {

            // if this is the selected component then deselect it
            if( pEmbeddedComp == selectedComponent() )
                selectComponent( NULL );

            if ( pEmbeddedComp->isInitialized() )
                embeddedComp.containedObject()->PersistentlyDelete();

            // remove component from the internal list
            fComponentList.removeAt( cursor );

            notifyOfChange( INotificationEvent( RemovedComponentId, *this ) );
            break;
        }
    }

    delete &cursor;
    
    return &embeddedComp;
}


IBaseStream&
IEmbedderModel::operator>>= ( IBaseStream& toStream ) const
{
    writeVersion( toStream );
    IModel::operator>>=( toStream );

    // Stream out all the local properties of the embedded component.
    // The actual compoonent (CAswContainedObject) will be streamed
    // out to it's own stream/storage later.
    fComponentList >>= toStream;

    return toStream;
}


IBaseStream&
IEmbedderModel::operator<<= ( IBaseStream& fromStream )
{
    readVersion( fromStream );
    IModel::operator<<=( fromStream );

    // Stream in all the local properties of the embedded component.
    // The actual compoonent (CAswContainedObject) will be streamed
    // in from it's own stream/storage later.
    fComponentList <<= fromStream;
    
    return fromStream;
}


void
IEmbedderModel::fileRenamed( const IString& name )
{

    CAswOLEString pszOLEFile(name);
    fpContainer->SetFileName( pszOLEFile );

}

