/*******************************************************************************
* FILE NAME: istatnry.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IComponentStationery                                                     *
*     IComponentStationeryFor                                                  *
*     IComponentFactory                                                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*   Portions or this component are based on work originating from Lotus.       *
*   (C) Copyright Lotus Development Corporation. 1996 - All Rights Reserved.   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "istatnry.hpp"
#include "icompfac.hpp"

#include <imetatyp.hpp>
#include <iguibndl.hpp>
#include <icmpfram.hpp>
#include <icompnen.hpp>
#include <icompimp.hpp>
#include <iembmod.hpp>
#include <iview.hpp>
#include <idataex.hpp>

#include <icoordsy.hpp>

#define kLangID "EN"

void ASW_DLL_Init(void);
void ASW_DLL_Term(void);

// our one and only stationery instance
IComponentStationery*
IComponentStationery::fgTheStationery = NULL;

#if defined(ASWANDEBUG)
    CAswDebugToFile*    gDebugFile;     // Debug messages to file
//  CAswDebugToMonitor* gDebugMonitor;  // debug messages to ::OutputDebugString()
    CAswDebugToAssertDialog* gpAssertDialog;
#endif



/*******************************************************************************
*                                                                              *
*   IComponentStationery::run                                                  *
*                                                                              *
*   Main component entry point                                                 *
*                                                                              *
*******************************************************************************/

int
IComponentStationery::run( int argc, char* argv[] )
{

    int nReturnCode = -1;

    // change the coordinate system to upper-left
    // note: this works for Windows only
    ICoordinateSystem::setApplicationOrientation ( ICoordinateSystem::originUpperLeft );

    // parse the argument list
    IString file;
    for ( int cnt = 1; cnt < argc; cnt++ )
    {
        IString strTemp = argv[cnt];
        // are we embedded? If so OLE calls us with the magic parameter
        // "-Embedding" but some older programs use "/Embedding" so must
        // check for both
        // RT need to account for blanks in the file name
        if ( strTemp == "-Embedding" || strTemp == "/Embedding" )
            fbEmbedded = true;
        else
            file += " " + strTemp;
    }
    // strip out any leading blanks, if any
    file.stripLeading();

    // "do-while(0)" to allow "breaks" to exit
    do
    {
        // call the stationery to do windows-specific initialization
        // this initializes ICLUI application object, Aswan control, etc.
        if( !initWindows() )
            break;

        // Callout to derived class application-specific initialization
        if( !initialize() )
            break;

        // create a bundle if we are not embedding
        // if we ARE embedding then the bundle will be created in response to
        // the IComponentFactory::CreateInstance call
        if( !isEmbedded() )
        {
            if( !setupBundle() )
                break;

            if ( file.length() != 0 ) {
                try {
                    bundle().component().storage().loadFile( file );
                }
                catch ( IException& ) {
                    // "Unable to open: %1"
                    const char* msg = IBase::messageText(IC_CDF_OPEN_FAILED,
                                                         file);
                    messageBox( msg, IMessageBox::moveable         |
                                     IMessageBox::applicationModal |
                                     IMessageBox::errorIcon        |
                                     IMessageBox::okButton );
                }
            }

            bundle().frameWindow().setFocus();
            bundle().frameWindow().show();

            ::PostMessage( (HWND)(void*)bundle().frameWindow().handle(),
                           WM_COMMAND, kViewToolbar, 0 );

            bundle().component().objectInstance()->AdviseOnShowWindow();

        }

        IApplication::current().setArgs( argc, argv );

        // message pump
        IApplication::current().run();
        nReturnCode = 0;


    } while( FALSE );


    return nReturnCode;


} // run


// -----------------------------------------------------------------------------
// ShowAswResult
// -----------------------------------------------------------------------------
// Static global helper function to show failed return codes on the screen
// and to the debug stream.
void
ShowAswResult( CAswResult& ar, const IExceptionLocation& location )
{
    char szMsg[1024];
    sprintf( szMsg,
             "%s\nFILE = %s\nFUNCTION = %s\nLINE = %d\nSCODE = %X",
             (const char*)IBase::messageText(IC_UNEXPECTED_ERROR),
             location.fileName(),
             location.functionName(),
             location.lineNumber(),
             ar.Scode() );

    // send the message to the debug stream
    #ifdef IC_SCREEN_DEBUG
        cout << szMsg << endl;
    #endif

    IComponentStationery::stationery().messageBox( szMsg );

}


// -----------------------------------------------------------------------------
// IExcept__throwAccess
// -----------------------------------------------------------------------------
// Static global helper function. Should eventually be moved to iexcept.cpp
void IExcept__throwAccess ( const char*               errorText,
                            unsigned long             errorCode,
                            const IExceptionLocation& location )
{
    IAccessError accessErr( errorText,
                            errorCode,  // The Aswan return code
                            IException::unrecoverable );
    accessErr.setErrorCodeGroup(IException::other);
    accessErr.addLocation(location);
    accessErr.logExceptionData();

#if (INO_EXCEPTIONS_SUPPORT)
    accessErr.terminate();
#else
    throw accessErr;
#endif
}


/*******************************************************************************
*                                                                              *
*   Class IComponentStationery                                                 *
*                                                                              *
*   Abstract base class for component stationeries                             *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// constructor
// -----------------------------------------------------------------------------
IComponentStationery::IComponentStationery() :
    fpAswControl( NULL ),
    fpComponentFactory( NULL ),
    fpBundle( NULL ),
    fbEmbedded( FALSE ),
    fpClipSource( NULL ),
    fpPasteTarget( NULL ),
    fpDragSource( NULL ),
    fcfEmbeddedObject( NULL ),
    fcfObjectDescriptor( NULL ),
    fcfLinkSource( NULL ),
    fcfLinkSrcDescriptor( NULL ),
    fpSourceFormatSet( NULL ),
    fpTargetFormatSet( NULL )
{

    // There can only be 1 stationery object per component.
    // So check our static global (compiler initializes it to NULL)
    ASSERT( fgTheStationery == NULL );

    // Since the stationery object is a static it will be constructed
    // before the app starts and before the WinMain is called.
    // We stick the instance pointer into a static so that it will be
    // globally available through the call: IComponentStationery::stationery()
    fgTheStationery = this;

} // IComponentStationery



// -----------------------------------------------------------------------------
// destructor
// -----------------------------------------------------------------------------
IComponentStationery::~IComponentStationery()
{

    // destroy the bundle
    delete fpBundle;

    // destroy class factory (it is shut down in its destructor)
    delete fpComponentFactory;

    // delete clipboard and drag/drop handlers
    SHUTDOWNANDDELETE( fpClipSource );
    SHUTDOWNANDDELETE( fpPasteTarget );
    SHUTDOWNANDDELETE( fpDragSource );
    SHUTDOWNANDDELETE( fpSourceFormatSet );
    SHUTDOWNANDDELETE( fpTargetFormatSet );

    // shutdown and destroy Aswan control
    if( fpAswControl )
    {
        fpAswControl->Uninitialize();
        SHUTDOWNANDDELETE( fpAswControl );

    } // an aswan control was constructed

#if defined(ASWANDEBUG)
    // delete all other objects created by CTOR
    if (gDebugFile)
        delete gDebugFile;

//  if (gDebugMonitor)
//      delete gDebugMonitor;

    if( gpAssertDialog )
    delete gpAssertDialog;
#endif

#ifndef NONINTEGRATED_CDF
    ASW_DLL_Term();
#endif

} // ~IComponentStationery


// -----------------------------------------------------------------------------
// stationery
// -----------------------------------------------------------------------------
// Returns the global static stationery reference
IComponentStationery&
IComponentStationery::stationery()
{
    return *fgTheStationery;
}

// -----------------------------------------------------------------------------
// initWindows
// -----------------------------------------------------------------------------
// Instance specific initialization that must be done each time
// derived classes can supply "initialize" and "initComponent" methods
// for app-specific initialization
BOOL
IComponentStationery::initWindows()
{
    HINSTANCE hInstance = (HINSTANCE)(void*)IApplication::current().userResourceLibrary().handle();

#if defined(ASWANDEBUG)
    // Put the log files in the same directory as the EXE
    char szName[256];
    GetModuleFileName(hInstance,szName,256);
    LPSTR lpszT = _fstrrchr(szName,'\\');
    if (lpszT)
       lpszT++;
    else
       lpszT = szName;
    lstrcpy(lpszT,"aswan.log");
    // always delete log
    CAswOLEString oleStr( szName );
    CAswDiskFile logFile( oleStr.GetOLEPtr() );
    logFile.Delete();

    // allocate objects for debug support
    gDebugFile = new CAswDebugToFile( szName ); // Debug messages to file (all)
//  gDebugMonitor = new CAswDebugToMonitor();            // debug messages to monitor - all channels
#endif

    // Running args
    fstrAppName = IApplication::current().userResourceLibrary().
                         loadString(IC_CDF_APP_LONGNAME);
    fstrFileExt = IApplication::current().userResourceLibrary().
                         loadString(IC_CDF_FILE_EXT);
    fstrShortName = IApplication::current().userResourceLibrary().
                         loadString(IC_CDF_APP_SHORTNAME);


    // ---------------------------------------------------------
    // Initialize Aswan and OLE
    // ---------------------------------------------------------

    // Register (or update) the app info with the system registry
    // We don't have to do this if we are embedded; hey if we got this far
    // then the registry must be fine.
    if( !fbEmbedded )
        updateRegistry();

    // Create new aswan control
    // this can now be accessed either via the control() member function or
    // casting "this" into a CAswControl*

#ifndef NONINTEGRATED_CDF
    ASW_DLL_Init();
#endif

    fpAswControl = new CAswControl( hInstance );
    ASSERT( fpAswControl != NULL );

    // Setup aswan control
#ifdef NONINTEGRATED_CDF
    CAswResult ar = fpAswControl->Setup( kLangID /* , FALSE */ );
#else
    HINSTANCE hResourceDll = (HINSTANCE)(void*)IApplication::current().resourceLibrary().handle();
    CAswResult ar = fpAswControl->Setup( kLangID , FALSE, hResourceDll );
#endif
    IASWASSERTSTATE( ar );

#if defined(ASWANDEBUG)
    // Debug monitor shows all channels
//  gDebugMonitor->SetChannels( CH_ALLCHANNELS );

    // aswan.log shows all channels.
    // flush after each write
    gDebugFile->SetChannels( CH_ALLCHANNELS );
    gDebugFile->SetFlushOnWrite(TRUE);

    // Tell Control about the debug output devices we just set up
    // monitor gets some channels
    // debug file gets some channels
    // debug all gets all channels
//  fpAswControl->SetDebugOut( gDebugMonitor );
    fpAswControl->SetDebugOut( gDebugFile );

    // Enable assert dialogs
    gpAssertDialog = new CAswDebugToAssertDialog( fpAswControl );
    gpAssertDialog->SetChannels( CH_ASSERTIONS, CH_WARNINGS );
    fpAswControl->SetDebugOut( gpAssertDialog );
#endif

    // Initiazlize Aswan control - this initialized Ole
    CAswOLEString pszOLEName( fstrAppName );
    ar = fpAswControl->Initialize( pszOLEName.GetOLEPtr() );
    IASWASSERTSTATE( ar );

    // Create a class factory
    fpComponentFactory = new IComponentFactory( fpAswControl );
    ASSERT( fpComponentFactory != NULL );

    // Setup class factory - this registers our CLSID with OLE
    // we only support local servers (ie. .EXE) and SINGLE_USE
    CAswOLEString strClsId = (char*)IApplication::current().userResourceLibrary().
                                             loadString(IC_CDF_CLSID);
    CLSID clsid;
    CLSIDFromString( strClsId.GetOLEPtr(), &clsid );
    ar = fpComponentFactory->Setup( clsid,
                                    CLSCTX_LOCAL_SERVER,
                                    REGCLS_SINGLEUSE,
                                    fbEmbedded );
    IASWASSERTSTATE( ar );


    // ---------------------------------------------------------
    // Clipboard and Drag/Drop support initialization
    // ---------------------------------------------------------

    // Create and setup a clipboard source
    fpClipSource = new IAswClipboardSource( fpAswControl );
    // setup clipboard source with no format set; we will use the "set format set" method
    ar = fpClipSource->Setup();
    IASWASSERTSTATE( ar );

    // Create and setup a paste target
    fpPasteTarget = new IAswPasteTarget( fpAswControl );
    // setup paste target with no format set; we will use the "query format set" method
    ar = fpPasteTarget->Setup();
    IASWASSERTSTATE( ar );

    // Create and setup a drag source
    fpDragSource = new IAswDragSource( fpAswControl );
    // setup paste target with no format set; we will use the "query format set" method
    ar = fpDragSource->Setup();
    IASWASSERTSTATE( ar );

    // register standard OLE clipboard formats
    fcfEmbeddedObject = (CLIPFORMAT) RegisterClipboardFormat( CF_EMBEDDEDOBJECT );
    fcfObjectDescriptor = (CLIPFORMAT) RegisterClipboardFormat( CF_OBJECTDESCRIPTOR );
    fcfLinkSource = (CLIPFORMAT) RegisterClipboardFormat( CF_LINKSOURCE );
    fcfLinkSrcDescriptor = (CLIPFORMAT) RegisterClipboardFormat( CF_LINKSRCDESCRIPTOR );

    // create format set placeholders

    // source format set for embedder models
    // (source format set for models is in the component implementation)
    fpSourceFormatSet = new CAswSourceFormatSet( fpAswControl );

    const int numFormats = 5;
    FORMATETC aFormats[ numFormats ];

    // generic embedded object
    aFormats[0].cfFormat = cfEmbeddedObject();
    aFormats[0].ptd = NULL;
    aFormats[0].dwAspect = DVASPECT_CONTENT;
    aFormats[0].lindex = -1;
    aFormats[0].tymed = TYMED_ISTORAGE;

    // object descriptor
    aFormats[1].cfFormat = cfObjectDescriptor();
    aFormats[1].dwAspect = DVASPECT_CONTENT;
    aFormats[1].ptd      = NULL;
    aFormats[1].tymed    = TYMED_HGLOBAL;
    aFormats[1].lindex   = -1;

    // link source
    aFormats[2].cfFormat = cfLinkSource();
    aFormats[2].ptd = NULL;
    aFormats[2].dwAspect = DVASPECT_CONTENT;
    aFormats[2].lindex = -1;
    aFormats[2].tymed = TYMED_ISTREAM;

    // link source descriptor
    aFormats[3].cfFormat = cfLinkSrcDescriptor();
    aFormats[3].dwAspect = DVASPECT_CONTENT;
    aFormats[3].ptd      = NULL;
    aFormats[3].tymed    = TYMED_HGLOBAL;
    aFormats[3].lindex   = -1;

    // METAFILE
    // this is how we want to look in cached presentations
    aFormats[4].cfFormat = CF_METAFILEPICT;
    aFormats[4].ptd = NULL;
    aFormats[4].dwAspect = DVASPECT_CONTENT;
    aFormats[4].lindex = -1;
    aFormats[4].tymed = TYMED_MFPICT;

    fpSourceFormatSet->Setup( aFormats, numFormats );

    // target format set for embedder models
    fpTargetFormatSet = new CAswTargetFormatSet( fpAswControl );
    OLEUIPASTEENTRY   fe[2];
    UINT              rgLinkFormats[1];

    // embedded object format
    CAswOLEString  szFmtStr0("%s Embedded Object");
    CAswOLEString  szResultStr0("%s");
    fe[0].fmtetc.cfFormat = IComponentStationery::stationery().cfEmbeddedObject();
    fe[0].fmtetc.dwAspect = DVASPECT_CONTENT;
    fe[0].fmtetc.ptd      = NULL;
    fe[0].fmtetc.tymed    = TYMED_ISTORAGE;
    fe[0].fmtetc.lindex   = -1;
    fe[0].lpstrFormatName = (LPASWCOLESTR) szFmtStr0;
    fe[0].lpstrResultText = (LPASWCOLESTR) szResultStr0;
    fe[0].dwFlags         = OLEUIPASTE_PASTE | OLEUIPASTE_ENABLEICON;

    // link source format
    CAswOLEString  szFmtStr2("%s");
    CAswOLEString  szResultStr2("%s");
    fe[1].fmtetc.cfFormat = IComponentStationery::stationery().cfLinkSource();
    fe[1].fmtetc.dwAspect = DVASPECT_CONTENT;
    fe[1].fmtetc.ptd      = NULL;
    fe[1].fmtetc.tymed    = TYMED_ISTORAGE;
    fe[1].fmtetc.lindex   = -1;
    fe[1].lpstrFormatName = (LPASWCOLESTR) szFmtStr2;
    fe[1].lpstrResultText = (LPASWCOLESTR) szResultStr2;
    fe[1].dwFlags         = OLEUIPASTE_LINKTYPE1 | OLEUIPASTE_ENABLEICON;

    rgLinkFormats[0] = IComponentStationery::stationery().cfLinkSource();

    // set up the format set
    fpTargetFormatSet->Setup( fe, 2, rgLinkFormats, 1 );

    return TRUE;

} // initWindows



// -----------------------------------------------------------------------------
// setupBundle
// -----------------------------------------------------------------------------
// Creates and initialize the GUI bundle
// Called by the IComponentFactory's CreateObject method in response to a message
// from OLE
BOOL
IComponentStationery::setupBundle()
{

    BOOL bOK = FALSE;

    do
    {
        // create a new bundle
        fpBundle = createBundle();
        if( fpBundle == NULL )
            break;

        // create and adopt-in the parts of the bundle
        fpBundle->adoptComponent( createComponent() );

        // create a model and let the component adopt it
        fpBundle->component().initialize( *fpBundle, createModel() );

        // createFrameWindow will create a in-place capable frame window if this
        // is a container
        fpBundle->adoptFrameWindow( createFrameWindow() );
        // the bundle will attach the view as the client area of the frame window
        fpBundle->adoptView( createView() );

        // if we're embedded, create an in-place view as well
        if ( fpBundle->component().isEmbedded() )
            fpBundle->createInPlaceWindows();

        bOK = TRUE;

    } while( FALSE );

    return bOK;

} // setupBundle

// -----------------------------------------------------------------------------
// Overrides - default implementations
// -----------------------------------------------------------------------------

// -----------
// initialize
// -----------
// callout to allow the user to do application specific initialization
// default does nothing
BOOL
IComponentStationery::initialize()
{
    return TRUE;

} // initialize


// -----------------
// createComponent
// -----------------
// callout to allow the user to create something other than an IComponent
// default: creates an IComponent
IComponent*
IComponentStationery::createComponent() const
{
    return new IComponent();

} // createComponent

// -----------------------------------------------------------------------------
// createBundle
// -----------------------------------------------------------------------------
// Creates the GUI bundle
IGUIBundle*
IComponentStationery::createBundle() const
{
    return new IGUIBundle();
}

// -------------------
// createFrameWindow
// -------------------
// callout to allow the user to override the creation of the frame window
// default: creates a basic frame window if a server only, and an in-place capable
// frame window if this component is a container
IComponentFrameWindow*
IComponentStationery::createFrameWindow()
{
    // create a frame window
    IComponentFrameWindow* pFrame = new IComponentFrameWindow();

    // setup the frame window
    // we pass a pointer to the aswan control if this is a container, otherwise
    // we pass in NULL indicating we don't need container functionality
    CAswControl* pControl = NULL;

    if( isContainer() )
        pControl = &control();


    pFrame->setup( pControl );

    // add a default menu bar - events will be handled in the bundle
    pFrame->addDefaultMenuBar();


    return pFrame;

} // createFrameWindow


// -----------------------------------------------------------------------------
// CAswControl methods exposed to the "public"
// -----------------------------------------------------------------------------
void
IComponentStationery::enableBusyDialog( BOOL bEnable )
{
    fpAswControl->EnableBusyDialog( bEnable );
}

void
IComponentStationery::enableNotRespondingDialog( BOOL bEnable )
{
    fpAswControl->EnableNotRespondingDialog( bEnable );
}

void
IComponentStationery::rejectInComingCalls()
{
    fpAswControl->RejectInComingCalls();
}

void
IComponentStationery::allowInComingCalls()
{
    fpAswControl->AllowInComingCalls();
}

// RT should use const IResourceId& when message file is built
IMessageBox::Response
IComponentStationery::messageBox(   const char*    msg,
                                    const IMessageBox::Style& style, /*= IMessageBox::information*/
                                    IWindow*              hWnd /*= NULL */ )
{
    fpAswControl->RejectInComingCalls();

    // RT use ICLUI IMessageBox instead, hWnd is used as owner rather than parent
    IMessageBox   msgBox(hWnd);

    msgBox.setTitle(fstrAppName);
    IMessageBox::Response iRet=msgBox.show(msg, style);

    fpAswControl->AllowInComingCalls();
    return iRet;
}



// -----------------------------------------------------------------------------
// isContainer
// -----------------------------------------------------------------------------
Boolean
IComponentStationery::isContainer() const
{
    return fpBundle->component().isContainer();

}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
// Updates the registry with the current app information
// (in case it changed)
void
IComponentStationery::updateRegistry()
{

    // convert CLSID to a string
    IString strCLSID = IApplication::current().userResourceLibrary().
                                             loadString(IC_CDF_CLSID);
    // get current path by getting module name and chopping off EXE name
    char bufTemp[ MAX_PATH ];
    //HINSTANCE hInstance = (HINSTANCE)(void*)IApplication::current().userResourceLibrary().handle();
    GetModuleFileName( NULL /*hInstance*/, bufTemp, MAX_PATH );
    IString strName = bufTemp;

    // set file extension entry
    RegSetValue( HKEY_CLASSES_ROOT, fileExt(), REG_SZ, shortName(), shortName().length() );

    // set short name entries
    RegSetValue( HKEY_CLASSES_ROOT, shortName(), REG_SZ, appName(), appName().length() );
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\CLSID", REG_SZ, strCLSID, strCLSID.length() );
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\Insertable", REG_SZ, "", 0 );
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\shell\\open\\command", REG_SZ, strName + " %1", strName.length() +3 );
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\shell\\open\\ddeexec", REG_SZ, "[open(\"%1\")]", 12 );
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\protocol\\StdFileEditing\\server", REG_SZ, strName, strName.length() );

    IString name = IApplication::current().userResourceLibrary().loadString(IC_CDF_PROT_EDIT_VERB_0);
    RegSetValue( HKEY_CLASSES_ROOT, shortName() + "\\protocol\\StdFileEditing\\verb\\0", REG_SZ, name, name.length() );

    // set entries in the CLSID
    IString strKey = "CLSID\\" + strCLSID;
    RegSetValue( HKEY_CLASSES_ROOT, strKey, REG_SZ, appName(), appName().length() );

    name = IApplication::current().userResourceLibrary().loadString(IC_CDF_CLSID_AUX_3);
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\AuxUserType\\3", REG_SZ, name, name.length() );
    name = IApplication::current().userResourceLibrary().loadString(IC_CDF_CLSID_AUX_2);
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\AuxUserType\\2", REG_SZ, name, name.length() );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\DefaultIcon", REG_SZ, strName, strName.length() );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\LocalServer32", REG_SZ, strName, strName.length() );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\ProgId", REG_SZ, shortName(), shortName().length() );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\MiscStatus", REG_SZ, "32", 2 );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\Insertable", REG_SZ, "", 0 );

    name = IApplication::current().userResourceLibrary().loadString(IC_CDF_CLSID_VERB_1);
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\verb\\1", REG_SZ, name, name.length() );
    name = IApplication::current().userResourceLibrary().loadString(IC_CDF_CLSID_VERB_0);
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\verb\\0", REG_SZ, name, name.length() );
    RegSetValue( HKEY_CLASSES_ROOT, strKey + "\\InprocHandler32", REG_SZ, "ole32.dll", 9 );


}

// -----------------------------------------------------------------------------
// sourceFormatSet
// -----------------------------------------------------------------------------
// returns the source format set depending on the current selection
CAswSourceFormatSet&
IComponentStationery::sourceFormatSet()
{

    // check the current selection
    //   - either one embedded object is selected or "select all"
    IEmbedderModel* pEmbedderModel = NULL;
    IEmbeddedComponent* pComponent = NULL;
    ::dynamicCastTo( pEmbedderModel, bundle().component().model() );
    if( pEmbedderModel )
        pComponent = pEmbedderModel->selectedComponent();

    if( pComponent )
        return *fpSourceFormatSet;

    else
        // for the "select all" we can just return the model's format set
        return *bundle().component().objectInstance()->formatSet();

}

// -----------------------------------------------------------------------------
// targetFormatSet
// -----------------------------------------------------------------------------
CAswTargetFormatSet&
IComponentStationery::targetFormatSet()
{

    return *fpTargetFormatSet;
}


/*******************************************************************************
*                                                                              *
*   Class IComponentFactory                                                    *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// constructor
// -----------------------------------------------------------------------------

IComponentFactory::IComponentFactory( PASWCONTROL pControl ) :
    CAswObjectClass( pControl )
{
}

// -----------------------------------------------------------------------------
// destructor
// -----------------------------------------------------------------------------

IComponentFactory::~IComponentFactory()
{
    // Call base class shutdown method to shut down the CAswObjectClass
    Shutdown( FALSE );
}

// -----------------------------------------------------------------------------
// CreateInstance
// -----------------------------------------------------------------------------
// called by OLE to create a new instance
// delegates up to the Statonery to create a new bundle
CAswResult
IComponentFactory::CreateInstance( LPUNKNOWN pUnkOuter, PASWCOMBASE ASWPTR * ppCOMBase )
{
    CAswResult ar( S_OK, ASD_OK );

    // get pointer to the stationery
    IComponentStationery& theStationery = IComponentStationery::stationery();

    // call on the stationery to create a bundle
    theStationery.setupBundle();

    // fill in OUT parameter
    *ppCOMBase = theStationery.bundle().component().objectInstance();

    return ar;

} // CreateInstance
