/*******************************************************************************
* FILE NAME: immamix.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immamix.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )
extern "C"
   {
   #include <string.h>
   #define INCL_MCIOS2
   #define IC_MULTIMEDIA
   #include <immdefs.h>
   }

#include <itrace.hpp>
#include <immexcpt.hpp>
#include <imultsta.hpp>
#include <immowin.hpp>
#include <immamix.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMAMIX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMMAmpMixer::IMMAmpMixer                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer::IMMAmpMixer( Boolean openNow,
                          unsigned long instance,
                          Boolean openShareable)
           : IMMDevice("AMPMIX")
{
  IMODTRACE_DEVELOP("IMMAmpMixer::ctor #1");

#ifdef IC_PM
  if (openNow)
  {
     IString fullName("AMPMIX");

     if (instance)
        fullName += IString(instance).rightJustify(2, '0');

     openOnThread(fullName, openShareable);
  }
#endif
#ifdef IC_WIN
  // In Win32, the mixer device is not accessed through MCI_ calls,
  // so we must forego standard device treatment, and fake it.
  // Since we are not opening now, the default ID in IMMDeviceData will be
  // 0, whish is what we want.
  setOpenStatus(openSuccessful);
  setPassDeviceRequested(true);

  // Since the mixer is not opened as an MCI device, it cannot be closed
  // as one either.
  setCloseOnDestroy(false);
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::IMMAmpMixer                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer::IMMAmpMixer(unsigned long newDeviceId,
                         const IString &newAlias)
#ifdef IC_PM
           : IMMDevice(newDeviceId, newAlias)
#endif
#ifdef IC_WIN
           : IMMDevice(0, newAlias)   // force to 0
#endif
{
  IMODTRACE_DEVELOP("IMMAmpMixer::ctor #2");

#ifdef IC_WIN
  // In Win32, the mixer device is not accessed through MCI_ calls,
  // so we must forego standard device treatment, and fake it.
  // We want the mixer device to have ID=0, so we force it in the CTOR
  // initializer, no matter what is passed in
  setOpenStatus(openSuccessful);
  setPassDeviceRequested(true);

  // Since the mixer is not opened as an MCI device, it cannot be closed
  // as one either.
  setCloseOnDestroy(false);
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::~IMMAmpMixer                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer::~IMMAmpMixer()
{
   IMODTRACE_DEVELOP("IMMAmpMixer::dtor");
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::balance                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::balance(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::balance");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_BALANCE, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"itemStatus");
  return rc;
#endif
#ifdef IC_WIN
  unsigned long       ulDetails[2];
  unsigned long       curUniformVol;

  // cast away the const to call mixerControlValues (which is not const)
  IMMAmpMixer * const nonConstThis = (IMMAmpMixer * const) this ;

  // Get current values of each channels
  if (!nonConstThis->mixerControlValues(
              MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,
              MIXERCONTROL_CONTROLTYPE_VOLUME,
              ulDetails,
              false))       // Get uniform value
     ITHROWMMERROR(lastError(),"mixerDetails");

  ITRACE_DEVELOP("IMMAmpMixer::balance - current left channel volume: "
                  +IString(ulDetails[0]));
  ITRACE_DEVELOP("IMMAmpMixer::balance - current right channel volume: "
                  +IString(ulDetails[1]));

  if (!ulDetails[0])   // Left channel is at minimum level
     return 100;
  if (!ulDetails[1])   // Right channel is at minimum level
     return   0;
  return ( (ulDetails[1] * 100) / (ulDetails[0]+ulDetails[1]) );


#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::bass                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::bass(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::bass");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_BASS, 0, call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
#endif
#ifdef IC_WIN
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::gain                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::gain(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::gain");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_GAIN, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
#endif
#ifdef IC_WIN
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::isMonitoringEnabled                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMAmpMixer::isMonitoringEnabled(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::isMonitoringEnabled");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  Boolean rc = itemStatus(MCI_AMP_STATUS_MONITOR, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
#endif
#ifdef IC_WIN
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::pitch                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::pitch(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::pitch");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_PITCH, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
#endif
#ifdef IC_WIN
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::treble                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMMAmpMixer::treble(CallType call) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::treble");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  unsigned long rc = itemStatus(MCI_AMP_STATUS_TREBLE, 0,call);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
  return rc;
#endif
#ifdef IC_WIN
  return 0;
#endif
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setBalance                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setBalance(unsigned long balance, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::setBalance");
  ITRACE_DEVELOP("IMMAmpMixer::setBalance - balance = "
                  +IString(balance));

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)balance;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_BALANCE;
  parms.ulValue          = (unsigned long)balance;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_BALANCE | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
#ifdef IC_WIN
  unsigned long       ulDetails[2];
  unsigned long       curUniformVol;

  // Get current values of both channels, since one won't change
  if (!mixerControlValues(
              MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,
              MIXERCONTROL_CONTROLTYPE_VOLUME,
              ulDetails,
              true))       // Get uniform value
     ITHROWMMERROR(lastError(),"mixerDetails");

  curUniformVol = ulDetails[0];

  ITRACE_DEVELOP("IMMAmpMixer::setBalance - current uniform volume: "
                  +IString(curUniformVol));

  if (balance > 50)
  {
     // Moving towards full right balance, reduce left channel
     ulDetails[0] = curUniformVol * ( 2.0 * (100.0 - balance) / 100.0);

     // Don't forget to set the right channel to the current uniform volume
     ulDetails[1] = curUniformVol;

     ITRACE_DEVELOP("IMMAmpMixer::setBalance - changing left line to "
                     +IString(ulDetails[0]));
     ITRACE_DEVELOP("IMMAmpMixer::setBalance - right will remain at "
                     +IString(ulDetails[1]));
  }
  else
  {
     // Moving towards full left balance, reduce right channel
     ulDetails[1] = curUniformVol * ( (2.0 * balance) / 100.0);
     ITRACE_DEVELOP("IMMAmpMixer::setBalance - changing right line to "
                     +IString(ulDetails[1]));
     ITRACE_DEVELOP("IMMAmpMixer::setBalance - left will remain at "
                     +IString(ulDetails[0]));
  }

  if (!setMixerControlValues(
              MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,
              MIXERCONTROL_CONTROLTYPE_VOLUME,
              ulDetails,
              false))        // individually set channels
     ITHROWMMERROR(lastError(),"mixerDetails");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setBass                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setBass(unsigned long bass, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::setBass");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)bass;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_BASS;
  parms.ulValue          = (unsigned long)bass;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_BASS | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setGain                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setGain(unsigned long gain, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::setGain");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)gain;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_GAIN;
  parms.ulValue          = (unsigned long)gain;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_GAIN | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::enableMonitoring                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::enableMonitoring(Boolean enable, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::enableMonitoring");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_MONITOR;
  sendCommand(MCI_SET,
              MCI_SET_ITEM | (enable ? MCI_SET_ON : MCI_SET_OFF) | call,
              &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::disableMonitoring                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::disableMonitoring(CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::disableMonitoring");

#ifdef IC_PM
  enableMonitoring(false,call);
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setPitch                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setPitch(unsigned long pitch, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::setPitch");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)pitch;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_PITCH;
  parms.ulValue          = (unsigned long)pitch;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_PITCH | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::setTreble                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::setTreble(unsigned long treble, CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::setTreble");

#ifdef IC_PM
  IASSERTSTATE(isOpen());
  MCI_AMP_SET_PARMS parms;

  memset(&parms, 0, sizeof(parms));
  parms.ulOver           = 0;
  parms.ulAudio          = IMMDevice::all;
  parms.ulLevel          = (unsigned long)treble;
  parms.hwndCallback     = HWND(deviceWindow().handle());
  parms.ulItem           = MCI_AMP_SET_TREBLE;
  parms.ulValue          = (unsigned long)treble;
  sendCommand(MCI_SET, MCI_SET_AUDIO | MCI_AMP_SET_TREBLE | call, &parms);
  if (lastError())
     ITHROWMMERROR(lastError(),"sendCommand");
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMMAmpMixer::isOpenStringValid                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IMMAmpMixer::isOpenStringValid(const IString &deviceName) const
{
  IMODTRACE_DEVELOP("IMMAmpMixer::isOpenStringValid");

  IString tester(deviceName);
  return (tester.upperCase().indexOf("AMPMIX") == 1 ? true : false);
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IMMAmpMixer::open                                                            |
| In Win32, an AmpMixer is not treated as an MCI device, so we can't close     |
| with standard MCI_OPEN. Hence we override and simulate.                      |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::open(const IString &fileOrDevice,
                                       Boolean shareable,
                                       CallType call)
{
   IMODTRACE_DEVELOP("IMMAmpMixer::open #1");
   return *this;
} // open

/*------------------------------------------------------------------------------
| IMMAmpMixer::openOnThread                                                    |
| In Win32, an AmpMixer is not treated as an MCI device, so we can't close     |
| with standard MCI_OPEN. Hence we override and simulate.                      |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::openOnThread(const IString &fileOrDevice,
                                         Boolean shareable,
                                         CallType call)
{
   IMODTRACE_DEVELOP("IMMAmpMixer::openOnThread #1");
   return *this;
} // openOnThread

/*------------------------------------------------------------------------------
| IMMAmpMixer::open                                                            |
| In Win32, an AmpMixer is not treated as an MCI device, so we can't close     |
| with standard MCI_OPEN. Hence we override and simulate.                      |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::open(unsigned long instanceNumber,
                                 Boolean shareable,
                                 CallType call)
{
   IMODTRACE_DEVELOP("IMMAmpMixer::open #2");
   return *this;
} // open

/*------------------------------------------------------------------------------
| IMMAmpMixer::openOnThread                                                    |
| In Win32, an AmpMixer is not treated as an MCI device, so we can't close     |
| with standard MCI_OPEN. Hence we override and simulate.                      |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::openOnThread(unsigned long instanceNumber,
                                         Boolean shareable)
{
   IMODTRACE_DEVELOP("IMMAmpMixer::openOnThread #2");
   return *this;
} // openOnThread

/*------------------------------------------------------------------------------
| IMMAmpMixer::close                                                           |
| In Win32, an AmpMixer is not treated as an MCI device, so we can't close     |
| with standard MCI_CLOSE. Hence we override and fake a close.                 |
------------------------------------------------------------------------------*/
IMMAmpMixer& IMMAmpMixer::close(CallType call)
{
  IMODTRACE_DEVELOP("IMMAmpMixer::close");
  setOpenStatus(openNotAttempted);
  setPassDeviceRequested(false);
  return *this;
}
#endif
