/*******************************************************************************
* FILE NAME: icustbut.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icustbut.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINMESSAGEMGR
  #define INCL_WINBUTTONS
  #define INCL_WININPUT
  #define INCL_WINSYS
  #define INCL_WINDIALOGS
  #include <iwindefs.h>
}

#include <icustbut.hpp>
#include <icolor.hpp>
#include <ifont.hpp>
#include <istring.hpp>
#include <iexcept.hpp>
#include <inotifev.hpp>
#include <icustbhd.hpp>
#include <itrace.hpp>
#include <ikeyevt.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICUSTBUT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
class ICustomButtonLatchHandler : public IHandler
{
public:
  ICustomButtonLatchHandler ();
  ~ICustomButtonLatchHandler ();
virtual Boolean
  dispatchHandlerEvent(IEvent& event);
ICustomButton*
  buttonDown;
};

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
ICustomButtonLatchHandler :: ICustomButtonLatchHandler ()
                           : buttonDown ( 0 )
{
}

ICustomButtonLatchHandler :: ~ICustomButtonLatchHandler ()
{
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
Boolean ICustomButtonLatchHandler :: dispatchHandlerEvent(IEvent& event)
{
  Boolean rc = false;
  ICustomButton *button = (ICustomButton *) event.dispatchingWindow();

  switch ( event.eventId() )
  {
    // Check for button up message
    //
    // NOTE: This is needed to smoothly change the button latch state.
    //       We first check to see if the button was last drawn in the
    //       down state.  If so, then we need to toggle the latch state.
    //       The refreshDisplay flag is set to false since PM will
    //       already redraw the button and we do not want double drawing.

    case WM_BUTTON1UP:
         if ( buttonDown == button )
         {
           if ( button->isLatched() )
           {
             if ( button->isLatchingEnabled() )
               button->latch(false,false);
           }
           else
           {
             if ( button->isLatchingEnabled() || button->isAutoLatchEnabled() )
             {
               button->latch(true,false);
             }
           }
         }
         break;

    // Check for unlatch message
    //
    // NOTE: This is an internal message used by the class library to
    //       implement the autoLatch feature of custom buttons

    case IC_UM_UNLATCH:
         if ( button->isAutoLatchEnabled() )
           button->unlatch();
         rc = true;
         break;

    // Check draw down message
    //
    // NOTE: This is an internal message used by the class library to
    //       implement smooth latching

    case IC_UM_DRAW_DOWN:
         if ( event.parameter1().asUnsignedLong() )
         {
            buttonDown = button;
         }
         else
         {
            if ( buttonDown == button)
               buttonDown = 0;
         }
         rc = true;
         break;

#ifdef IC_WIN
    // Generate WM_COMMAND message expected by command handlers.
    case WM_CONTROL:
       if (event.parameter1().highNumber() == BN_CLICKED)
          {
          // generate the needed events due to a button click
          if (button)
             {
             IWindow* owner = button->owner();
             if ((owner == 0) || (owner == IWindow::desktopWindow()) )
                {
                owner = button->parent();
                }
             if (owner)
                {
                ITRACE_ALL( IString("owner=") +
                            IString( (unsigned long)((void*)owner) ).d2x() +
                            IString(" hwnd=") +
                            IString( owner->handle().asUnsigned() ).d2x() );
                owner->postEvent( WM_COMMAND,
                                  IEventData(
                                     event.parameter1().lowNumber(),
                                     CMDSRC_PUSHBUTTON),
                                  0 );
                rc = true;
                } // if owner
             }   // if button
          }      // if BN_CLICKED
       break;
#endif

    // Remove the DLGC_PUSHBUTTON dialog code for custom buttons because
    // we do supprt the "default button" for custom buttons and the canvas
    // logic looks at this code to determine if a button can be the
    // default

    case WM_QUERYDLGCODE:
        event.window()->dispatchRemainingHandlers(event);
        event.setResult(event.result().asUnsignedLong() &
                           (unsigned long)(~DLGC_PUSHBUTTON));
        rc = true;
        break;

    // Make the enter key and spacebar work the same as the mouse
    //
    // Note: Normally we would only want the spacebar to act this way,
    //       but we shipped with a bug regarding default pushbutton
    //       support and current applications will expect the enter
    //       key to press the button (so we are adding enter key support
    //       to try and maintain better backward compatibility)

    case WM_CHAR:
        {
          IKeyboardEvent keyevt(event);
          if ( !(keyevt.isUpTransition()) && keyevt.isVirtual() )
          {
            if (( keyevt.virtualKey() == IKeyboardEvent::space ) ||
                ( keyevt.virtualKey() == IKeyboardEvent::enter ) ||
                ( keyevt.virtualKey() == IKeyboardEvent::newLine ))
            {
              button->click();
              rc = true;
            }
          }
        }
        break;

  }
  return rc;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
static struct DefaultCustomButtonHandlers {
  ~DefaultCustomButtonHandlers ( )
    {
    if (latchHandler)
      {
      delete latchHandler;
      latchHandler = 0;
      }
    if (drawHandler)
      {
      delete drawHandler;
      drawHandler = 0;
      }
    }
ICustomButtonLatchHandler
 *latchHandler;
ICustomButtonDrawHandler
 *drawHandler;
} defaultCustomButtonHandlers;

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
class ICustomButtonData : public IBase
{
public:
  ICustomButtonData ( );
Boolean
  latched,
  halftoneLatch;
IColor
  *latchColor;
IColor
  *latchForeground;
unsigned long
  userData;
};

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
ICustomButtonData :: ICustomButtonData ( )
                   : latched ( false ),
                     halftoneLatch ( true ),
                     latchColor ( 0 ),
                     latchForeground ( 0 ),
                     userData ( 0 )
{
}

/*------------------------------------------------------------------------------
| ICustomButton styles                                                         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ICustomButton::Style
  ICustomButton::latchable         ( 0, IBS_CUSTOMLATCHABLE ),
  ICustomButton::autoLatch         ( 0, IBS_CUSTOMAUTOLATCH ),
  ICustomButton::classDefaultStyle = WS_VISIBLE;
#pragma data_seg()

#pragma data_seg(ICLNonConst)
  ICustomButton::Style ICustomButton::fgCurrentDefaultStyle = WS_VISIBLE;
#pragma data_seg()

/*------------------------------------------------------------------------------
| ICustomButton ctor                                                           |
------------------------------------------------------------------------------*/
ICustomButton :: ICustomButton (unsigned long id,
                                IWindow* parent,
                                IWindow* owner,
                                const IRectangle& initial,
                                const Style& style )
               : fCustomButtonData(0)
{
  IASSERTPARM(parent!=0);
  fCustomButtonData = new ICustomButtonData;

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  /****************************************************************/
  /* Borland has bug with embedded ? operator so below is needed. */
  /****************************************************************/
#ifdef __BORLANDC__
  IWindowHandle hOwner;
  if (owner == 0)
    hOwner = IWindowHandle(0);
  else
    hOwner = owner->handle();
#else
  IWindowHandle hOwner = (owner == 0) ? IWindowHandle(0)
                                       : owner->handle();
#endif

  IWindowHandle whButton =
      this -> create( id,
                      0,
                      convertToGUIStyle( style ),
                      WC_BUTTON,
                      parent->handle(),
                      hOwner,
                      initial,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( style, true ) );

  startHandlingEventsFor(whButton);

  if ( !defaultCustomButtonHandlers.latchHandler )
    defaultCustomButtonHandlers.latchHandler = new ICustomButtonLatchHandler;
  if ( !defaultCustomButtonHandlers.drawHandler )
    defaultCustomButtonHandlers.drawHandler = new ICustomButtonDrawHandler;


  defaultCustomButtonHandlers.latchHandler->handleEventsFor(this);
  defaultCustomButtonHandlers.drawHandler->handleEventsFor(this);
}

/*------------------------------------------------------------------------------
| ICustomButton (wrapper) ctor                                                 |
------------------------------------------------------------------------------*/
ICustomButton :: ICustomButton (const IWindowHandle& handle )
               : fCustomButtonData(0)
{
  setAutoDestroyWindow ( false );

  fCustomButtonData = new ICustomButtonData;

  startHandlingEventsFor(handle);

  if ( !defaultCustomButtonHandlers.latchHandler )
    defaultCustomButtonHandlers.latchHandler = new ICustomButtonLatchHandler;
  if ( !defaultCustomButtonHandlers.drawHandler )
    defaultCustomButtonHandlers.drawHandler = new ICustomButtonDrawHandler;

  defaultCustomButtonHandlers.latchHandler->handleEventsFor(this);
  defaultCustomButtonHandlers.drawHandler->handleEventsFor(this);

}

/*------------------------------------------------------------------------------
| ICustomButton (protected) ctor                                               |
------------------------------------------------------------------------------*/
ICustomButton :: ICustomButton ()
               : fCustomButtonData(0)
{
  fCustomButtonData = new ICustomButtonData;

  if ( !defaultCustomButtonHandlers.latchHandler )
    defaultCustomButtonHandlers.latchHandler = new ICustomButtonLatchHandler;
  if ( !defaultCustomButtonHandlers.drawHandler )
    defaultCustomButtonHandlers.drawHandler = new ICustomButtonDrawHandler;

  defaultCustomButtonHandlers.latchHandler->handleEventsFor(this);
  defaultCustomButtonHandlers.drawHandler->handleEventsFor(this);

}

/*------------------------------------------------------------------------------
| ICustomButton dtor                                                           |
------------------------------------------------------------------------------*/
ICustomButton :: ~ICustomButton()
{
  if (fCustomButtonData->latchColor)
    delete fCustomButtonData->latchColor;
  if (fCustomButtonData)
    delete fCustomButtonData;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
ICustomButton::Style ICustomButton :: defaultStyle()
{
  return fgCurrentDefaultStyle;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
void ICustomButton :: setDefaultStyle(const Style& style)
{
  fgCurrentDefaultStyle = style;
}

/*------------------------------------------------------------------------------
| ICustomButton::convertToGUIStyle                                             |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ICustomButton::convertToGUIStyle(const IBitFlag& guiStyle,
                                               Boolean bExtOnly) const
{
  // Obtain the style from the class (IButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // Set the GUI user button style
#ifdef IC_PM
    ulStyle |= BS_USERBUTTON;
#endif
#ifdef IC_WIN
    ulStyle |= BS_OWNERDRAW | WS_CHILD;
#endif
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| ICustomButton::enableLatching                                                |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: enableLatching ( Boolean enable )
{
  unsigned long ulExtStyle = extendedStyle();
  unsigned long ulOldExtStyle = ulExtStyle;

  if ( enable )
  {
    ulExtStyle |= latchable.asExtendedUnsignedLong();
  }
  else
  {
    latch(false);
    ulExtStyle &= ~latchable.asExtendedUnsignedLong();
  }

  if ( ulExtStyle != ulOldExtStyle )
  {
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::disableLatching                                               |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: disableLatching ( )
{
  enableLatching(false);
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::enableAutoLatch                                               |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: enableAutoLatch ( Boolean enable )
{
  unsigned long ulExtStyle = extendedStyle();
  unsigned long ulOldExtStyle = ulExtStyle;

  if ( enable )
  {
    ulExtStyle |= autoLatch.asExtendedUnsignedLong();
  }
  else
  {
    ulExtStyle &= ~autoLatch.asExtendedUnsignedLong();
  }

  if ( ulExtStyle != ulOldExtStyle )
  {
    setExtendedStyle(ulExtStyle);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::disableAutoLatch                                              |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: disableAutoLatch ( )
{
  enableAutoLatch(false);
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::isLatchable                                                   |
------------------------------------------------------------------------------*/
Boolean ICustomButton :: isLatchingEnabled () const
{
  return( (extendedStyle() & latchable.asExtendedUnsignedLong()) ? true
                                                                 : false );
}

/*------------------------------------------------------------------------------
| ICustomButton::isAutoLatch                                                   |
------------------------------------------------------------------------------*/
Boolean ICustomButton :: isAutoLatchEnabled () const
{
  return( (extendedStyle() & autoLatch.asExtendedUnsignedLong()) ? true
                                                                 : false );
}

/*------------------------------------------------------------------------------
| ICustomButton::latch                                                         |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: latch ( Boolean latched,
                                        Boolean refresh )
{
  unsigned long ulMask = latchable.asExtendedUnsignedLong() |
                         autoLatch.asExtendedUnsignedLong();

  if ( (extendedStyle() & ulMask) &&
       ( fCustomButtonData->latched != latched ))
  {
    fCustomButtonData->latched = latched;

    notifyObservers(INotificationEvent(ICustomButton::latchId, *this,
                                       true, (void*)latched));

    if ( refresh )
      this->refresh();
    if ( latched )
      checkAutoLatch ();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::unlatch                                                       |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: unlatch ()
{
  return latch(false);
}

/*------------------------------------------------------------------------------
| ICustomButton::isLatched                                                     |
------------------------------------------------------------------------------*/
Boolean ICustomButton :: isLatched () const
{
  return fCustomButtonData->latched;
}

/*------------------------------------------------------------------------------
| ICustomButton::latchedBackgroundColor                                        |
------------------------------------------------------------------------------*/
IColor ICustomButton :: latchedBackgroundColor () const
{
  if (fCustomButtonData->latchColor)
    return (*(fCustomButtonData->latchColor));
  else
    return(backgroundColor());
}

/*------------------------------------------------------------------------------
| ICustomButton::latchedForegroundColor                                        |
------------------------------------------------------------------------------*/
IColor ICustomButton :: latchedForegroundColor () const
{
  if (fCustomButtonData->latchForeground)
    return (*(fCustomButtonData->latchForeground));
  else
    return(foregroundColor());
}

/*------------------------------------------------------------------------------
| ICustomButton::setLatchedBackgroundColor                                     |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: setLatchedBackgroundColor ( const IColor& color,
                                                            Boolean halftone )
{
  if (fCustomButtonData->latchColor)
    *(fCustomButtonData->latchColor) = color;
  else
    fCustomButtonData->latchColor = new IColor(color);
  fCustomButtonData->halftoneLatch = halftone;
  if (fCustomButtonData->latched)
    refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::setLatchedForegroundColor                                     |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: setLatchedForegroundColor ( const IColor& color )
{
  if (fCustomButtonData->latchForeground)
    *(fCustomButtonData->latchForeground) = color;
  else
    fCustomButtonData->latchForeground = new IColor(color);
  if (fCustomButtonData->latched)
    refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::resetLatchedBackgroundColor                                   |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: resetLatchedBackgroundColor ()
{
  if (fCustomButtonData->latchColor)
  {
    delete fCustomButtonData->latchColor;
    fCustomButtonData->latchColor = 0;
  }
  fCustomButtonData->halftoneLatch = true;
  if (fCustomButtonData->latched)
    refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::resetLatchedForegroundColor                                   |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: resetLatchedForegroundColor ()
{
  if (fCustomButtonData->latchForeground)
  {
    delete fCustomButtonData->latchForeground;
    fCustomButtonData->latchForeground = 0;
  }
  if (fCustomButtonData->latched)
    refresh();
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::isLatchedBackgroundColorHalftone                              |
------------------------------------------------------------------------------*/
Boolean ICustomButton :: isLatchedBackgroundColorHalftone () const
{
  return fCustomButtonData->halftoneLatch;
}

/*------------------------------------------------------------------------------
| ICustomButton::backgroundColor                                               |
------------------------------------------------------------------------------*/
IColor ICustomButton :: backgroundColor () const
{
#ifdef IC_WIN
  return IGUIColor( IGUIColor::buttonMiddle );
#else
  return ( IWindow::color ( PP_BACKGROUNDCOLOR,
                            IGUIColor ( IGUIColor::buttonMiddle )));
#endif
}

/*------------------------------------------------------------------------------
| ICustomButton::calcMinimumSize                                               |
------------------------------------------------------------------------------*/
ISize ICustomButton :: calcMinimumSize () const
{
  ISize sz;

  // Get the current font and text
  IFont         fnt(this);
  IString       txt(text());

  // Set up temporary variables used for calculations
  unsigned long index = 0;
  unsigned long start;
  unsigned long cx;

  // For each line add to height and test width
  sz.setHeight(1);
  do
  {
    // Only add max height - 1 (so that one pel may overlap between lines)
    sz.setHeight(sz.height()+fnt.maxCharHeight()-1);
    start = index + 1;
    index = txt.indexOf('\n', (unsigned)start);
#ifdef IC_PMWIN
    if ( IBidiSettings::isBidiSupported() )
    {
      IBidiSettings settings(*this);
      if (index != 0)
        cx = fnt.textWidth(txt.subString((unsigned)start,
                                         (unsigned)(index-start)).strip(),settings);
      else
        cx = fnt.textWidth(txt.subString((unsigned)start).strip(),settings);
    }
    else
#endif
    {
      if (index != 0)
        cx = displaySize(txt.subString((unsigned)start,
                                       (unsigned)(index-start)).strip()).width();
      else
        cx = displaySize(txt.subString((unsigned)start).strip()).width();
    }
    if (cx > sz.width())
      sz.setWidth(cx);
  }
  while (index != 0);

  // Add room for border painting and pad before returning
  // There may be up to 5 borders and add 6x3 pels of pad

  unsigned long cxBorder = IQUERYSYSVALUE(SV_CXBORDER);
  unsigned long cyBorder = IQUERYSYSVALUE(SV_CYBORDER);

  sz.setWidth ( sz.width() + ( cxBorder * 5 ) + 6 );
  sz.setHeight ( sz.height() + ( cyBorder * 5 ) + 3 );

  return sz;
}

/*------------------------------------------------------------------------------
| ICustomButton::checkAutoLatch                                                |
------------------------------------------------------------------------------*/

void ICustomButton :: checkAutoLatch ()
{
  // Check if the autoLatch is set

  if ( isAutoLatchEnabled() )
  {
    // Unlatch all other buttons in the current group

    IWindow *top = parent();
    IWindow *first = 0;
    IWindow *current = 0;

    IWindow::ChildCursor cursor(*top);

    // Find the first window in the same group

    cursor.setToFirst();
    while ( cursor.isValid() )
    {
      current = IWindow::windowWithHandle(top->childAt(cursor));
      if (( first == 0 ) || (( current != 0 ) && ( current->isGroup() )))
      {
        first = current;
      }
      if ( current == this )
        cursor.invalidate();
      else
        cursor.setToNext();
    }

    // Send IC_UM_UNLATCH to all other buttons in the ground

    if ( first )
    {
      cursor.setToFirst();
      while ( cursor.isValid() )
      {
        current = IWindow::windowWithHandle(top->childAt(cursor));
        if ( current != 0 )
        {
          if ( first == current )
          {
            if ( current != this )
              current->sendEvent(IC_UM_UNLATCH);
            first = 0;
            cursor.setToNext();
          }
          else
          {
            if ( first == 0 )
            {
              if (current->isGroup())
                cursor.invalidate();
              else
              {
                if ( current != this )
                  current->sendEvent(IC_UM_UNLATCH);
                cursor.setToNext();
              }
            }
            else
              cursor.setToNext();
          }
        }
        else
          cursor.setToNext();
      }
    }
  }
}

/*------------------------------------------------------------------------------
| ICustomButton::setUserData                                                   |
------------------------------------------------------------------------------*/
ICustomButton& ICustomButton :: setUserData ( unsigned long data )
{
  fCustomButtonData->userData = data;
  return *this;
}

/*------------------------------------------------------------------------------
| ICustomButton::userData                                                      |
------------------------------------------------------------------------------*/
unsigned long ICustomButton :: userData ( ) const
{
  return fCustomButtonData->userData;
}

