/*******************************************************************************
* FILE NAME: iglist.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iglist.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <iglist.hpp>
#include <igrafctx.hpp>
#include <ipoint.hpp>
#include <itrnsfrm.hpp>
#include <iwindow.hpp>
#include <iseq.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGLIST_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGListSeq member functions                                                   |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IGListSeq class declaration                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
class IGListSeq : public ISequence<IGraphic*>
{
public:
  IGListSeq( unsigned long items = 25 );
  ~IGListSeq( );
};

/*------------------------------------------------------------------------------
| IGListSeq::IGListSeq                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGListSeq::IGListSeq( unsigned long items )
  : ISequence<IGraphic*>(items)
{}

/*------------------------------------------------------------------------------
| IGListSeq::~IGListSeq                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGListSeq::~IGListSeq()
{}


/*------------------------------------------------------------------------------
| ObjectCursor member functions                                                |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList& list )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::regular ), fId(0), fGc(0), fCtm(0), fBtm(0)
{
}

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList& list, unsigned long objectIdentifier )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::id ), fId(objectIdentifier),
    fGc(0), fCtm(0), fBtm(0)
{}

/*------------------------------------------------------------------------------
| IGList::Cursor::Cursor                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::Cursor( const IGList&    list,
                        IGraphicContext& graphicContext,
                        const IPoint&    hitPoint )
  : fCursorList(&list), fCursor(new IGListSeq::Cursor(*list.fList)),
    fCursorType( IGList::Cursor::hit ), fId(0), fGc(&graphicContext), fCtm(0),
    fBtm(0)
{

  fGc->setHitPoint( hitPoint );
  if (list.hasTransformMatrix())
  {
    fCtm = new ITransformMatrix(fGc->worldTransformMatrix());
    fGc->setWorldTransformMatrix(list.transformMatrix(), list.transformMethod());
  }
}

/*------------------------------------------------------------------------------
| IGList::Cursor::~Cursor                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::Cursor::~Cursor()
{
  if (fCursorType == IGList::Cursor::hit)
  {
    if (fCursorList->hasTransformMatrix())
    {
      fGc->setWorldTransformMatrix(*fCtm, IGraphic::replace );
      delete fCtm;
    }
    if (fBtm)
      delete fBtm;
  }
  delete (IGListSeq::Cursor*)fCursor;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToFirst                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToFirst()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToFirst());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    case IGList::Cursor::id:
      if (rc)
      {
        IGraphic& graphic(fCursorList->graphicAt( *this ));
        if (graphic.id() == fId)
          rc = true;
        else
          rc = setToNext();
      }
    break;
    case IGList::Cursor::hit:
      if (rc)
      {
        fGc->startHitTesting();

        IGraphic& graphic(fCursorList->graphicAt(*this));
        graphic.setHitSelected(false);

        if (graphic.isHitSelectable())
          graphic.drawOn(*fGc);
        if (!graphic.isHitSelected())
        {
          rc = setToNext();
          if (!rc)
            fGc->stopHitTesting();
        }
        else
        {
          if (fBtm)
            delete fBtm;
          fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
          fGc->stopHitTesting();
        }
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToNext                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToNext()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToNext());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    case IGList::Cursor::id:
      if (rc)
      {
        rc = false;
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToNext())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          if (graphic.id() == fId)
          {
            rc = true;
            break;
          }
        }
      }
    break;
    case IGList::Cursor::hit:
      if (rc)
      {
        rc = false;

        fGc->startHitTesting();

        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToNext())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          graphic.setHitSelected(false);

          if (graphic.isHitSelectable())
            graphic.drawOn(*fGc);
          if (graphic.isHitSelected())
          {
            rc = true;
            if (fBtm)
              delete fBtm;
            fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
            break;
          }
        } /* endfor */
        fGc->stopHitTesting();
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToPrevious                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToPrevious()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToPrevious());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    case IGList::Cursor::id:
      if (rc)
      {
        rc = false;
        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToPrevious())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          if (graphic.id() == fId)
          {
            rc = true;
            break;
          }
        }
      }
    break;
    case IGList::Cursor::hit:
      if (rc)
      {
        rc = false;
        fGc->startHitTesting();

        for (;isValid();((IGListSeq::Cursor*)fCursor)->setToPrevious())
        {
          IGraphic& graphic(fCursorList->graphicAt( *this ));
          graphic.setHitSelected(false);

          if (graphic.isHitSelectable())
            graphic.drawOn(*fGc);
          if (graphic.isHitSelected())
          {
            rc = true;
            if (fBtm)
              delete fBtm;
            fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
            break;
          }
        } /* endfor */
        fGc->stopHitTesting();
      }
    break;
  }
  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::setToLast                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::setToLast()
{
  IBase::Boolean rc(((IGListSeq::Cursor*)fCursor)->setToLast());

  switch ( fCursorType )
  {
    case IGList::Cursor::regular:
    break;
    case IGList::Cursor::id:
      if (rc)
      {
        IGraphic& graphic(fCursorList->graphicAt( *this ));
        if (graphic.id() == fId)
          rc = true;
        else
          rc = setToPrevious();
      }
    break;
    case IGList::Cursor::hit:
      if (rc)
      {
        fGc->startHitTesting();

        IGraphic& graphic(fCursorList->graphicAt(*this));
        graphic.setHitSelected(false);

        if (graphic.isHitSelectable())
          graphic.drawOn(*fGc);
        if (!graphic.isHitSelected())
        {
          rc = setToPrevious();
          if (!rc)
            fGc->stopHitTesting();
        }
        else
        {
          if (fBtm)
            delete fBtm;
          fBtm = new ITransformMatrix(fGc->worldTransformMatrix());
        fGc->stopHitTesting();
        }
      }
    break;
  } /* endswitch */

  return rc;
}

/*------------------------------------------------------------------------------
| IGList::Cursor::isValid                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::Cursor::isValid() const
{
  return ((IGListSeq::Cursor*)fCursor)->isValid();
}

/*------------------------------------------------------------------------------
| IGList::Cursor::invalidate                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGList::Cursor::invalidate()
{
  ((IGListSeq::Cursor*)fCursor)->invalidate();
}


/*------------------------------------------------------------------------------
| IGList member functions                                                      |
------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList ( )
  : IGList::Inherited(), fList(new IGListSeq())
{
  fIsList = true;
}

/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList( IGraphic& graphic )
  : IGList::Inherited( ), fList(new IGListSeq())
{
  fIsList = true;
  fList->add(&graphic);
}

/*------------------------------------------------------------------------------
| IGList::IGList                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::IGList ( const IGList& list )
  : IGList::Inherited( list ), fList(new IGListSeq())
{
  fIsList = true;
  fList->addAllFrom(*list.fList);
}

/*------------------------------------------------------------------------------
| IGList::~IGList                                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList::~IGList ( )
{
  delete fList;
}

/*------------------------------------------------------------------------------
| IGList::drawOn                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::drawOn( IGraphicContext& fGc )
{
  fGc.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::numberOfGraphics                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGList::numberOfGraphics( ) const
{
  return fList->numberOfElements();
}

/*------------------------------------------------------------------------------
| IGList::isEmpty                                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::isEmpty( ) const
{
  return fList->isEmpty();
}

/*------------------------------------------------------------------------------
| IGList::isIGList                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGList::isIGList( const Cursor& cursor ) const
{
  IGraphic& graphic(graphicAt( cursor ));
  return graphic.fIsList;
}

/*------------------------------------------------------------------------------
| IGList::firstGraphic                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::firstGraphic( ) const
{
  IGraphic* graphic(fList->firstElement());
  return graphic;
}

/*------------------------------------------------------------------------------
| IGList::lastGraphic                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::lastGraphic( ) const
{
  IGraphic* graphic(fList->lastElement());
  return graphic;
}

/*------------------------------------------------------------------------------
| IGList::graphicAt                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic& IGList::graphicAt( const Cursor& cursor ) const
{
  IGraphic* graphic(fList->elementAt(*((IGListSeq::Cursor*)cursor.fCursor)));
  return *graphic;
}

/*------------------------------------------------------------------------------
| IGList::graphicAtPosition                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic& IGList::graphicAtPosition( unsigned long position ) const
{
  IGraphic* graphic(fList->elementAtPosition(position));
  return *graphic;
}

/*------------------------------------------------------------------------------
| IGList::boundingRectAt                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGList::boundingRectAt( const Cursor& cursor ) const
{
  if (cursor.fGc)
  {
    cursor.fGc->startBoundaryAccumulation();
    IGraphic& graphic(graphicAt(cursor));
    graphic.drawOn( *cursor.fGc );
    return cursor.fGc->stopBoundaryAccumulation();
  }
  else
  {
    IGraphicContext gc( IWindow::desktopWindow()->handle() );
    IGraphic& graphic(graphicAt(cursor));
    return graphic.boundingRect( gc );
  }
}

/*------------------------------------------------------------------------------
| IGList::addAsFirst                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::addAsFirst( IGraphic& graphic )
{
  fList->addAsFirst(&graphic);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsLast                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::addAsLast( IGraphic& graphic )
{
  fList->addAsLast(&graphic);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsNext                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::addAsNext( const Cursor& cursor, IGraphic& graphic )
{
  fList->addAsNext( &graphic, *((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAsPrevious                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::addAsPrevious( const Cursor& cursor, IGraphic& graphic )
{
  fList->addAsPrevious( &graphic, *((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::addAtPosition                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::addAtPosition( unsigned long position, IGraphic& graphic )
{
  fList->addAtPosition( position, &graphic );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::replaceAt                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::replaceAt( const Cursor& cursor, IGraphic& graphic )
{
  fList->replaceAt(*((IGListSeq::Cursor*)cursor.fCursor), &graphic );
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeFirst                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeFirst()
{
  fList->removeFirst();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeLast                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeLast()
{
  fList->removeLast();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAtPosition                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeAtPosition( unsigned long position )
{
  fList->removeAtPosition(position);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAt                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeAt( const Cursor& cursor )
{
  fList->removeAt(*((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAll                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeAll( )
{
  fList->removeAll();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::removeAllWithId                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::removeAllWithId( unsigned long identifier )
{
  IGList::Cursor cursor( *this, identifier );
  for( cursor.setToFirst(); cursor.isValid(); )
  {
    removeAt( cursor );
    cursor.setToFirst();
  }
  return *this;
}

#ifdef NOTYET
/*------------------------------------------------------------------------------
| IGList::deleteFirst                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteFirst()
{
  IGraphic* graphic(firstGraphic());
  if (graphic)
  {
    delete graphic;
    fList->removeFirst();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::deleteLast                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteLast()
{
  IGraphic* graphic(lastGraphic());
  if (graphic)
  {
    delete graphic;
    fList->removeLast();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::deleteAtPosition                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteAtPosition( unsigned long position )
{
  IGraphic* graphic(&graphicAtPosition( position ));
  delete graphic;
  fList->removeAtPosition(position);
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::deleteAt                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteAt( const Cursor& cursor )
{
  IGraphic* graphic(&graphicAt( cursor ));
  delete graphic;
  fList->removeAt(*((IGListSeq::Cursor*)cursor.fCursor));
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::deleteAll                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteAll( )
{
  IGList::Cursor cursor(*this);

  for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
    IGraphic* graphic(&graphicAt( cursor ));
    delete graphic;
  }
  fList->removeAll();
  return *this;
}

/*------------------------------------------------------------------------------
| IGList::deleteAllWithId                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::deleteAllWithId( unsigned long identifier )
{
  IGList::Cursor cursor( *this, identifier );
  for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
    IGraphic* graphic(&graphicAt( cursor ));
    delete graphic;
    removeAt( cursor );
  }
  return *this;
}
#endif

/*------------------------------------------------------------------------------
| IGList::topGraphicUnderPoint                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::topGraphicUnderPoint( const IPoint& point,
                                        IGraphicContext& graphicContext )
{
  IGraphic* hit(0);
  IGList::Cursor cursor( *this, graphicContext, point );

  if (cursor.setToLast())
    hit = &graphicAt( cursor );

  return hit;
}

/*------------------------------------------------------------------------------
| IGList::bottomGraphicUnderPoint                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphic* IGList::bottomGraphicUnderPoint( const IPoint& point,
                                           IGraphicContext& graphicContext )
{
  IGraphic* hit(0);
  IGList::Cursor cursor( *this, graphicContext, point );

  if (cursor.setToFirst())
    hit = &graphicAt( cursor );

  return hit;
}


/*------------------------------------------------------------------------------
| IGList::sort                                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGList& IGList::sort( long (*comparisonFunction)(IGraphic* const&,
                                                 IGraphic* const&))
{
  fList->sort( comparisonFunction );
  return *this;
}
