/*******************************************************************************
* FILE NAME: igrafctx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igrafctx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
  #include <iwindefs.h>
  #include <math.h>
}

#include <igrafctx.hpp>
#include <igcdata.hpp>
#include <iwindow.hpp>
#include <itrnsfrm.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <igline.hpp>
#include <igpyline.hpp>
#include <igrect.hpp>
#include <igelipse.hpp>
#include <igarc.hpp>
#include <igpie.hpp>
#include <igregion.hpp>
#include <igstring.hpp>
#include <iglist.hpp>
#include <istring.hpp>
#include <ithread.hpp>
#include <itrace.hpp>
#include <iplatfrm.hpp>
#include <icoordsy.hpp>
#include <ireslib.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGRAFCTX_CPP_
  #include <ipagetun.h>
#endif


#define      LONG2DBL(x)      ((double) (((double)x)/(double)65536))
#define      DEGREES_TO_RADS  57.29577951
#define MPOINT2POINT(mpt, pt)   ((pt).x = (mpt).x, (pt).y = (mpt).y)
#define POINT2MPOINT(pt, mpt)   ((mpt).x = (SHORT)(pt).x, (mpt).y = (SHORT)(pt).y)
#define POINTS2VECTOR2D(pt0, pt1, vect) ((vect).x = (double)((pt1).x - (pt0).x), \
                                         (vect).y = (double)((pt1).y - (pt0).y))

/*------------------------------------------------------------------------------
| IGCData::IGCData                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGCData::IGCData( IGCData::Mode mode,
                  IGraphicBundle::DrawOperation drawOperation,
                  const IColor& penColor,
                  const IColor& fillColor,
                  const IColor& backgroundColor,
                  IGraphicBundle::MixMode mixMode,
                  IGraphicBundle::BackgroundMixMode backgroundMixMode,
                  unsigned long penWidth,
                  IGraphicBundle::PenType penType,
                  IGraphicBundle::PenEndingStyle penEndingStyle,
                  IGraphicBundle::PenJoiningStyle penJoiningStyle,
                  unsigned long penPattern,
                  unsigned long fillPattern )
  : drawingMode(mode),
    fDrawOperation(drawOperation),
    fPenColor(penColor),
    fFillColor(fillColor),
    fBackgroundColor(backgroundColor),
    fMixMode(mixMode),
    fBackgroundMixMode(backgroundMixMode),
    fPenWidth(penWidth),
    fPenType(penType),
    fPenEndingStyle(penEndingStyle),
    fPenJoiningStyle(penJoiningStyle),
    fPenPattern(penPattern),
    fFillPattern(fillPattern),
    deferCreation(false),
    hitPoint(IPoint(0,0)),
    hitSize(ISize(12,20)),
    fTransformMatrix(),
    fBoundingRect(),
    recoordHeight(0),
    fRecoordinate(false),
    originalBrush(0),
    originalPen(0),
    originalBmp(0)
{ }

/*------------------------------------------------------------------------------
| IGCData::~IGCData                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGCData::~IGCData( )
{ }

/*------------------------------------------------------------------------------
| IGCData::createPen                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGCData& IGCData::createPen( const IPresSpaceHandle& ps )
{
  DWORD    penStyle(0);
  LOGBRUSH logBrush;
  HPEN     newPen;
  DWORD    dashLength[4];
  DWORD    numDashes(0);

  if (!deferCreation)
  {
    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: One pen style, no PS_GEOMETRIC or PS_COSMETIC");
    }
    else
    {
      if (fPenWidth > 1)
        penStyle |= PS_GEOMETRIC;
      else
        penStyle |= PS_COSMETIC;
    }

    switch (fPenType)
    {
      case IGraphicBundle::solid:
        penStyle |= PS_SOLID;
      break;
      case IGraphicBundle::dot:
        if (IPlatform::isWin95() && (penStyle & PS_GEOMETRIC))
          penStyle |= PS_SOLID;
        else
          penStyle |= PS_DOT;
      break;
      case IGraphicBundle::shortDash:
        if (IPlatform::isWin95() && (penStyle & PS_GEOMETRIC))
          penStyle |= PS_SOLID;
        else
          penStyle |= PS_DASH;
      break;
      case IGraphicBundle::dashDot:
        if (IPlatform::isWin95() && (penStyle & PS_GEOMETRIC))
          penStyle |= PS_SOLID;
        else
          penStyle |= PS_DASHDOT;
      break;
      case IGraphicBundle::doubleDot:
        if (IPlatform::isWin32S() || (IPlatform::isWin95() && !(penStyle & PS_GEOMETRIC)))
        {
          ITRACE_DEVELOP("Win32s: doubleDot pen not supported, default to dot");
          penStyle |= PS_DOT;
        }
        else if (!IPlatform::isWin95())
        {
          penStyle |= PS_USERSTYLE;
          dashLength[0] = fPenWidth;
          dashLength[1] = fPenWidth;
          dashLength[2] = fPenWidth;
          dashLength[3] = 2 * fPenWidth;
          numDashes = 4;
        }
      break;
      case IGraphicBundle::longDash:
        if (IPlatform::isWin95() && (penStyle & PS_GEOMETRIC))
          penStyle |= PS_SOLID;
        else
          penStyle |= PS_DASH;
      break;
      case IGraphicBundle::dashDoubleDot:
        if (IPlatform::isWin95() && (penStyle & PS_GEOMETRIC))
          penStyle |= PS_SOLID;
        else
          penStyle |= PS_DASHDOTDOT;
      break;
      case IGraphicBundle::alternate:
        if (IPlatform::isWin32S() || IPlatform::isWin95() ||
           (IPlatform::isWinNT() && fPenWidth > 1))
        {
          ITRACE_DEVELOP("Win32s: alternate pen not supported, use IGraphicBundle::solid");
          penStyle |= PS_SOLID;
        }
        else
        {
          penStyle |= PS_ALTERNATE;
        }
      break;
      case IGraphicBundle::invisible:
        penStyle |= PS_NULL;
      break;
    }

    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: PS_GEOMETRIC not supported, skip styles");
    }
    else
    {
      if (penStyle & PS_GEOMETRIC)
      {
        switch (fPenEndingStyle)
        {
          case IGraphicBundle::flat:
            penStyle |= PS_ENDCAP_FLAT;
          break;
          case IGraphicBundle::square:
            penStyle |= PS_ENDCAP_SQUARE;
          break;
          case IGraphicBundle::rounded:
            penStyle |= PS_ENDCAP_ROUND;
          break;
        }
      }

      if (IPlatform::isWin32S())
      {
        ITRACE_DEVELOP("Win32s: PS_GEOMETRIC joining styles not supported, skip");
      }
      else
      {
        switch (fPenJoiningStyle)
        {
          case IGraphicBundle::bevel:
            penStyle |= PS_JOIN_BEVEL;
          break;
          case IGraphicBundle::miter:
            penStyle |= PS_JOIN_MITER;
          break;
          case IGraphicBundle::round:
            penStyle |= PS_JOIN_ROUND;
          break;
        }
      }
    }

    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s: Extended pen styles not supported, skip");
    }
    else
    {
      if ((fPenWidth > 1) && !IPlatform::isWin95())
      {
        if (fPenPattern == IGraphicBundle::dense1)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense1)
            IGCData::fgBmpDense1 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE1);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense1;
        }
        else if (fPenPattern == IGraphicBundle::dense2)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense2)
            IGCData::fgBmpDense2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE2);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense2;
        }
        else if (fPenPattern == IGraphicBundle::dense3)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense3)
            IGCData::fgBmpDense3 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE3);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense3;
        }
        else if (fPenPattern == IGraphicBundle::dense4)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense4)
            IGCData::fgBmpDense4 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE4);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense4;
        }
        else if (fPenPattern == IGraphicBundle::dense5)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense5)
            IGCData::fgBmpDense5 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE5);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense5;
        }
        else if (fPenPattern == IGraphicBundle::dense6)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense6)
            IGCData::fgBmpDense6 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE6);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense6;
        }
        else if (fPenPattern == IGraphicBundle::dense7)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense7)
            IGCData::fgBmpDense7 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE7);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense7;
        }
        else if (fPenPattern == IGraphicBundle::dense8)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpDense8)
            IGCData::fgBmpDense8 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE8);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpDense8;
        }
        else if (fPenPattern == IGraphicBundle::vertical)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_VERTICAL;
        }
        else if (fPenPattern == IGraphicBundle::horizontal)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_HORIZONTAL;
        }
        else if (fPenPattern == IGraphicBundle::forewardDiag1)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_FDIAGONAL;
        }
        else if (fPenPattern == IGraphicBundle::forewardDiag2)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpfDiag2)
            IGCData::fgBmpfDiag2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPFDIAG2);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpfDiag2;
        }
        else if (fPenPattern == IGraphicBundle::backwardDiag1)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_BDIAGONAL;
        }
        else if (fPenPattern == IGraphicBundle::backwardDiag2)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpbDiag2)
            IGCData::fgBmpbDiag2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPBDIAG2);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpbDiag2;
        }
        else if (fPenPattern == IGraphicBundle::hollow)
          logBrush.lbStyle = BS_HOLLOW;
        else if (fPenPattern == IGraphicBundle::filled)
          logBrush.lbStyle = BS_SOLID;
        else if (fPenPattern == IGraphicBundle::halftone)
        {
          logBrush.lbStyle = BS_PATTERN;
          if (!IGCData::fgBmpHalftone)
            IGCData::fgBmpHalftone = IApplication::current().resourceLibrary().loadBitmap(IC_BMPHALFTONE);
          logBrush.lbHatch = (long)(void*)IGCData::fgBmpHalftone;
        }
        else if (fPenPattern == IGraphicBundle::hatched)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_CROSS;
        }
        else if (fPenPattern == IGraphicBundle::hatchedDiag)
        {
          logBrush.lbStyle = BS_HATCHED;
          logBrush.lbHatch = HS_DIAGCROSS;
        }
      }
      else
      {
        logBrush.lbStyle = BS_SOLID;
        logBrush.lbHatch = 0;
      }
    }

    COLORREF cColor  = fPenColor.asRGBLong();
    logBrush.lbColor = cColor;

    if (IPlatform::isWin32S())
    {
      ITRACE_DEVELOP("Win32s:  Issue CreatePen instead of ExtCreatePen");
      newPen = CreatePen( penStyle, (penStyle & PS_SOLID ? fPenWidth : 1), cColor);
      if (newPen == 0)
      {
        ITHROWGUIERROR("CreatePen");
      }
    }
    else
    {
      if (IPlatform::isWin95())
      {
        newPen = ExtCreatePen( penStyle, fPenWidth, &logBrush, 0, 0 );
      }
      else
      {
        if ((penStyle & PS_STYLE_MASK) == PS_USERSTYLE)
          newPen = ExtCreatePen( penStyle, fPenWidth, &logBrush, numDashes, dashLength );
        else
          newPen = ExtCreatePen( penStyle, fPenWidth, &logBrush, 0, 0 );
      }

      if (newPen == 0)
      {
        ITHROWGUIERROR("ExtCreatePen");
      }
    }
    if (!originalPen)
    {
      originalPen = SelectObject(ps, newPen);
    }
    else
    {
      DeleteObject(SelectObject(ps, newPen));
    }

    SetTextColor( ps, NEARESTPALETTECOLOR(fPenColor.asRGBLong() ));
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGCData::createBrush                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGCData& IGCData::createBrush( const IPresSpaceHandle& ps )
{
  HBRUSH newBrush;
  COLORREF cColor = fFillColor.asRGBLong();

  if (!deferCreation)
  {
    if (fFillPattern == IGraphicBundle::dense1)
    {
      if (!IGCData::fgBmpDense1)
       IGCData::fgBmpDense1 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE1);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense1 );
    }
    else if (fFillPattern == IGraphicBundle::dense2)
    {
      if (!IGCData::fgBmpDense2)
        IGCData::fgBmpDense2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE2);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense2 );
    }
    else if (fFillPattern == IGraphicBundle::dense3)
    {
      if (!IGCData::fgBmpDense3)
        IGCData::fgBmpDense3 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE3);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense3 );
    }
    else if (fFillPattern == IGraphicBundle::dense4)
    {
      if (!IGCData::fgBmpDense4)
        IGCData::fgBmpDense4 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE4);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense4 );
    }
    else if (fFillPattern == IGraphicBundle::dense5)
    {
      if (!IGCData::fgBmpDense5)
        IGCData::fgBmpDense5 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE5);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense5 );
    }
    else if (fFillPattern == IGraphicBundle::dense6)
    {
      if (!IGCData::fgBmpDense6)
        IGCData::fgBmpDense6 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE6);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense6 );
    }
    else if (fFillPattern == IGraphicBundle::dense7)
    {
      if (!IGCData::fgBmpDense7)
        IGCData::fgBmpDense7 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE7);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense7 );
    }
    else if (fFillPattern == IGraphicBundle::dense8)
    {
      if (!IGCData::fgBmpDense8)
        IGCData::fgBmpDense8 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPDENSE8);
      newBrush = CreatePatternBrush( IGCData::fgBmpDense8 );
    }
    else if (fFillPattern == IGraphicBundle::vertical)
      newBrush = CreateHatchBrush( HS_VERTICAL, NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::horizontal)
      newBrush = CreateHatchBrush( HS_HORIZONTAL, NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::forewardDiag1)
      newBrush = CreateHatchBrush( HS_FDIAGONAL, NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::forewardDiag2)
    {
      if (!IGCData::fgBmpfDiag2)
        IGCData::fgBmpfDiag2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPFDIAG2);
      newBrush = CreatePatternBrush( IGCData::fgBmpfDiag2 );
    }
    else if (fFillPattern == IGraphicBundle::backwardDiag1)
      newBrush = CreateHatchBrush( HS_BDIAGONAL, NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::backwardDiag2)
    {
      if (!IGCData::fgBmpbDiag2)
        IGCData::fgBmpbDiag2 = IApplication::current().resourceLibrary().loadBitmap(IC_BMPBDIAG2);
      newBrush = CreatePatternBrush( IGCData::fgBmpbDiag2 );
    }
    else if (fFillPattern == IGraphicBundle::hollow)
      newBrush = (HBRUSH)GetStockObject( HOLLOW_BRUSH );
    else if (fFillPattern == IGraphicBundle::filled)
      newBrush = CreateSolidBrush( NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::halftone)
    {
      if (!IGCData::fgBmpHalftone)
        IGCData::fgBmpHalftone = IApplication::current().resourceLibrary().loadBitmap(IC_BMPHALFTONE);
      newBrush = CreatePatternBrush( IGCData::fgBmpHalftone );
    }
    else if (fFillPattern == IGraphicBundle::hatched)
      newBrush = CreateHatchBrush( HS_CROSS, NEARESTPALETTECOLOR( cColor ));
    else if (fFillPattern == IGraphicBundle::hatchedDiag)
      newBrush = CreateHatchBrush( HS_DIAGCROSS, NEARESTPALETTECOLOR( cColor ));
    else
      newBrush = CreatePatternBrush( (HBITMAP)fFillPattern );

    if (!originalBrush)
    {
      originalBrush = SelectObject(ps, newBrush);
    }
    else
    {
      DeleteObject(SelectObject(ps, newBrush));
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGCData::usingCustomFillPat                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGCData::usingCustomFillPat( unsigned long fillPattern )
{
  return (fillPattern < IGraphicBundle::vertical);
}


/*------------------------------------------------------------------------------
| IGCData::angleNormalize                                                      |
| Normalizes any angle to an positive angle within 0 to 360 degree.            |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::angleNormalize( double& degree )
{
  if (degree < 0)
  {
    degree = -degree;
    degree = 360.0 - fmod(degree, 360.0);
  }
  else
  {
    degree = fmod(degree, 360.0);
  } /* endif */
}

/*------------------------------------------------------------------------------
| IGCData::radialPoint                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGCData::radialPoint  ( Boolean textMapMode,
                               const IRectangle& rect, float angle )
{
  long radialX;
  long radialY;
  float tanOfAngle;

  if (angle < 0)
  {
    angle = -angle;
    angle = 360.0 - fmod(angle, 360.0);
  }
  else
  {
    angle = fmod(angle, 360.0);
  } /* endif */

  tanOfAngle = tan( (double)angle/(double)DEGREES_TO_RADS );

  if ((int)angle == 90)
  {
    radialX = rect.center().x();
    if (textMapMode)
      radialY = rect.minY();
    else
      radialY = rect.maxY();
  }
  else if ((int)angle == 270)
  {
    radialX = rect.center().x();
    if (textMapMode)
      radialY = rect.maxY();
    else
      radialY = rect.minY();
  }
  else if ((int)angle < 45)
  {
    float adj(rect.maxX() - rect.center().x());
    radialX = rect.maxX();
    if (textMapMode)
      radialY = rect.center().y() - (int)(adj * tanOfAngle);
    else
      radialY = rect.center().y() + (int)(adj * tanOfAngle);
  }
  else if ((int)angle < 135 )
  {
    float opp;
    if (textMapMode)
    {
      opp = rect.minY() - rect.center().y();
      radialY = rect.minY();
    }
    else
    {
      opp = rect.maxY() - rect.center().y();
      radialY = rect.maxY();
    }
    radialX = rect.center().x() - (int)(opp / tanOfAngle);
  }
  else if ((int)angle < 225 )
  {
    float adj(rect.center().x() - rect.minX());
    radialX = rect.minX();
    if (textMapMode)
      radialY = rect.center().y() + (int)(adj * tanOfAngle);
    else
      radialY = rect.center().y() - (int)(adj * tanOfAngle);
  }
  else if ((int)angle < 315)
  {
    float opp;
    if (textMapMode)
    {
      opp = rect.center().y() - rect.maxY();
      radialY = rect.maxY();
    }
    else
    {
      opp = rect.center().y() - rect.minY();
      radialY = rect.minY();
    }
    radialX = rect.center().x() + (int)(opp / tanOfAngle);
  }
  else
  {
    float adj(rect.maxX() - rect.center().x());
    radialX = rect.maxX();
    if (textMapMode)
      radialY = rect.center().y() - (int)(adj * tanOfAngle);
    else
      radialY = rect.center().y() + (int)(adj * tanOfAngle);
  }
  return IPoint(radialX, radialY);
}

/*------------------------------------------------------------------------------
| IGCData::transformPoint                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGCData::transformPoint( const ITransformMatrix& mat, const IPoint& point )
{
  IPoint pt(point);
  if (mat != ITransformMatrix())
  {
    pt.setX( (double)point.x() * mat.element11() + (double)point.y() * mat.element21() + mat.element31() );
    pt.setY( (double)point.x() * mat.element12() + (double)point.y() * mat.element22() + mat.element32() );
  }
  return (pt);
}

/*------------------------------------------------------------------------------
| IGCData::transformPoint                                                      |
| double precision tranformation.                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::transformPoint( const ITransformMatrix& mat, double& x, double& y)
{
  if (mat != ITransformMatrix())
  {
    x = x * mat.element11() + y * mat.element21() + mat.element31();
    y = x * mat.element12() + y * mat.element22() + mat.element32();
  }
}

/*------------------------------------------------------------------------------
| IGCData::arcCenterPoint                                                      |
|                                                                              |
| Calculate the center point of a circular arc                                 |
|                                                                              |
| Parameters:  dp3Coord[4] - Array of 4 points: [0] - start pt                 |
|                                               [1] - end pt                   |
|                                               [2] - point on arc             |
|                                      Returned [3] - center pt                |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::arcCenterPoint(POINTD ptdPts[4])
{
  POINTD     ptdPoints[5];

  // check for the special case of a circle
  if ((ptdPts[0].x == ptdPts[1].x) && (ptdPts[0].y == ptdPts[1].y))
  {
    // the centre is the mid point between the mid point on arc and
    // the start/end point
    ptdPts[3].x = ptdPts[0].x + (ptdPts[2].x - ptdPts[0].x) / 2;
    ptdPts[3].y = ptdPts[0].y + (ptdPts[2].y - ptdPts[0].y) / 2;
    return;
  }

  // find mid point of the first chord of the arc
  ptdPoints[0].x = ptdPts[0].x + (ptdPts[2].x - ptdPts[0].x) / 2;
  ptdPoints[0].y = ptdPts[0].y + (ptdPts[2].y - ptdPts[0].y) / 2;

  // define a line at 90 degrees to this point
  perpLineEnd(&ptdPoints[0], &ptdPts[0], &ptdPts[2]);

  // now find mid point on 2nd chord of arc
  ptdPoints[2].x = ptdPts[2].x + (ptdPts[1].x - ptdPts[2].x) / 2;
  ptdPoints[2].y = ptdPts[2].y + (ptdPts[1].y - ptdPts[2].y) / 2;

  // define a line at 90 degrees to this point
  perpLineEnd(&ptdPoints[2], &ptdPts[2], &ptdPts[1]);

  // now calculate the point of intersection between these two lines
  // which are at right angles to the arc
  pointAtIntersect(ptdPoints);

  // return the new centre point of arc
  ptdPts[3].x = ptdPoints[4].x;
  ptdPts[3].y = ptdPoints[4].y;
}

/*------------------------------------------------------------------------------
| IGCData::perpLineEnd                                                         |
|                                                                              |
| Calculates an end point for a perpindicular line given a point and a line    |
|                                                                              |
| Parameters:  ptdPoint[2]- [0] Point to calculate from                        |
|                           [1] Returned end pt of perp. line                  |
|              pptdStart  - Start point of line                                |
|              pptdEnd    - End   point of line                                |
------------------------------------------------------------------------------*/
void IGCData::perpLineEnd(POINTD ptdPoint[2], POINTD *pptdStart, POINTD *pptdEnd)
{
  double dX1;           // Delta X of line
  double dY1;           // Delta Y of line
  double dX2;           // Delta X of perpendicular line
  double dY2;           // Delta Y of perpindicular line

  dX1 = pptdStart->x - pptdEnd->x;
  dY1 = pptdStart->y - pptdEnd->y;

  if (fabs(dX1) < 1.0E-10)    // If slope undefined, then perp is flat
  {
    dX2 = 1.0;
    dY2 = 0.0;
  }
  else
  {
    if (fabs(dY1) < 1.0E-10)  // If slope flat, then perp is undefined
    {
      dX2 = 0.0;
      dY2 = 1.0;
    }
    else
    {
      dX2 = 1.0;          // Set constant delta X for perpindicular
      dY2 = -(dX1 / dY1); // And solve for delta Y of perpindicular
    }
  }

  //
  // Calculate a point on the perpendicular line
  //
  dX1  = fabs(dX2) > fabs(dY2) ? fabs(dX2): fabs(dY2); // Get maximum delta
  dX2 /= (dX1 / 1.0E-2);                 // And convert deltas to a
  dY2 /= (dX1 / 1.0E-2);                 // range between 0 and 0.01

  ptdPoint[1].x = ptdPoint[0].x + dX2;
  ptdPoint[1].y = ptdPoint[0].y + dY2;
}

/*------------------------------------------------------------------------------
| IGCData::pointAtIntersect                                                    |
|                                                                              |
| Calculate the point of intersection between two lines.                       |
|                                                                              |
| Parameters:  ptdPts[5] - array of points: [0] line 1 start                   |
|                                           [1] line 1 end                     |
|                                           [2] line 2 start                   |
|                                           [3] line 2 end                     |
|                                           [4] returned intersect             |
|                                               point.                         |
------------------------------------------------------------------------------*/
void IGCData::pointAtIntersect(POINTD ptdPts[5])
{
  double dXIntersect;
  double dYIntersect;
  double dc1, dc2, dm1, dm2;

  if ((ptdPts[0].x == ptdPts[1].x) && (ptdPts[2].x == ptdPts[3].x))
  {
  }
  else if (ptdPts[0].x == ptdPts[1].x) /* first line is vertical */
  {
    if (ptdPts[2].x == ptdPts[3].x)
    {
      // Intersection does not exist (both vertical lines)
      dXIntersect = 0.0;
      dYIntersect = 0.0;
    }
    else // only first line is vertical
    {
      dXIntersect = ptdPts[0].x;  // X Intersect is known
      // calculate gradient of the other line
      dm2 = ((ptdPts[2].y - ptdPts[3].y) / (ptdPts[2].x - ptdPts[3].x));
      // calculate the constant in the equation of the other line
      dc2 = ptdPts[2].y - (dm2 * ptdPts[2].x);
      dYIntersect =  (dm2 * dXIntersect) + dc2;
    }
  }
  else if (ptdPts[2].x == ptdPts[3].x) // second line is vertical
  {
    dXIntersect = ptdPts[2].x;  // X Intersect is known
    dm1 = ((ptdPts[0].y - ptdPts[1].y) / (ptdPts[0].x - ptdPts[1].x));
    // calculate the constant in the equation of line 1
    dc1 = ptdPts[0].y - (dm1 * ptdPts[0].x);
    dYIntersect = (dm1 * dXIntersect) + dc1;
  }
  else // neither line is vertical
  {
    // calculate gradient of line 1
    dm1 = ((ptdPts[0].y - ptdPts[1].y) / (ptdPts[0].x - ptdPts[1].x));

    // calculate gradient of line 2
    dm2 = ((ptdPts[2].y - ptdPts[3].y) / (ptdPts[2].x - ptdPts[3].x));

    // calculate the constant in the equation of line 1
    dc1 = ptdPts[0].y - (dm1 * ptdPts[0].x);

    // calculate the constant in the equation of line 2
    dc2 = ptdPts[2].y - (dm2 * ptdPts[2].x);

    // at the point of intersection, the x,y values are equal,
    // ie. dm1x + dc1   =  dm2x + dc2
    //     x(dm1 - dm2) =  dc2 - dc1
    //     x            = (dc2 - dc1)/(dm1 - dm2)

    // if the gradients are equal there is no point of intersection
    if (dm1 == dm2)
    {
      dXIntersect = 0.0;
      dYIntersect = 0.0;
    }
    else
    {
      dXIntersect = (dc2 - dc1)/(dm1 - dm2);
      // this x value can be used in either equation to give the
      // y intersect
      dYIntersect = (dm1 * dXIntersect) + dc1;
    }
  }

  // store the point of intersection for return to caller
  ptdPts[4].x = dXIntersect;
  ptdPts[4].y = dYIntersect;
}

/*------------------------------------------------------------------------------
| IGCData::determinant                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGCData::determinant( double e11, double e12, double e21, double e22 )
{
  return ( e11 * e22 - e21 * e12 );
}

/*------------------------------------------------------------------------------
| IGCData::subtractVectors                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
PVECTOR2D IGCData::subtractVectors(PVECTOR2D v0, PVECTOR2D v1, PVECTOR2D v)
{
  if (v0 == NULL || v1 == NULL)
    v = (PVECTOR2D)NULL;
  else
  {
    v->x = v0->x - v1->x;
    v->y = v0->y - v1->y;
  }
  return(v);
}

/*------------------------------------------------------------------------------
| IGCData::vectorSquared                                                       |
|                                                                              |
|                                                                              |
-----------------------------------------------------------------------------*/
double IGCData::vectorSquared(PVECTOR2D v0)
{
  double dSqLen;

  if (v0 == NULL)
    dSqLen = 0.0;
  else
    dSqLen = (double)(v0->x * v0->x) + (double)(v0->y * v0->y);
  return (dSqLen);
}

/*------------------------------------------------------------------------------
| IGCData::vectorMagnitude                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGCData::vectorMagnitude(PVECTOR2D v0)
{
  double dMagnitude;

  if (v0 == NULL)
    dMagnitude = 0.0;
  else
    dMagnitude = sqrt(IGCData::vectorSquared(v0));
  return (dMagnitude);
}

/*------------------------------------------------------------------------------
| IGCData::dotProduct                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGCData::dotProduct(PVECTOR2D v0, PVECTOR2D v1)
{
  return ((v0 == NULL || v1 == NULL) ? 0.0
                                     : (v0->x * v1->x) + (v0->y * v1->y));
}

/*------------------------------------------------------------------------------
| IGCData::projectAndResolve                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::projectAndResolve(PVECTOR2D v0, PVECTOR2D v1, PPROJECTION ppProj)
{
  VECTOR2D ttProjection, ttOrthogonal;
  double proj1;
  //
  //obtain projection vector
  //
  //c = a * b
  //    ----- b
  //    |b|^2
  //
  proj1 = IGCData::dotProduct(v0, v1)/IGCData::dotProduct(v1, v1);
  ttProjection.x = v1->x * proj1;
  ttProjection.y = v1->y * proj1;
  //
  //obtain perpendicular projection : e = a - c
  //
  IGCData::subtractVectors(v0, &ttProjection, &ttOrthogonal);
  //
  //fill PROJECTION structure with appropriate values
  //
  ppProj->LenProjection = IGCData::vectorMagnitude(&ttProjection);
  ppProj->LenPerpProjection = IGCData::vectorMagnitude(&ttOrthogonal);

  ppProj->ttProjection.x = ttProjection.x;
  ppProj->ttProjection.y = ttProjection.y;
  ppProj->ttPerpProjection.x = ttOrthogonal.x;
  ppProj->ttPerpProjection.y = ttOrthogonal.y;
}

/*------------------------------------------------------------------------------
| IGCData::distance                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGCData::distance(PVECTOR2D v0, PVECTOR2D v1)
{
  return(sqrt( pow((v0->x - v1->x), 2) + pow((v0->y - v1->y), 2)));
}

/*------------------------------------------------------------------------------
| IGCData::distFromPointToLine                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
double IGCData::distFromPointToLine(LPPOINT pt0, LPPOINT pt1, LPPOINT ptTest)
{
  VECTOR2D ttLine, ttTest;
  PROJECTION pProjection;

  POINTS2VECTOR2D(*pt0, *pt1, ttLine);
  POINTS2VECTOR2D(*pt0, *ptTest, ttTest);

  IGCData::projectAndResolve(&ttTest, &ttLine, &pProjection);

  return(pProjection.LenPerpProjection);
}

/*------------------------------------------------------------------------------
| IGCData::hitTestLine                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGCData::hitTestLine(POINT pt0, POINT pt1, POINT ptMouse, int nWidth)
{
  POINT PtM;
  VECTOR2D tt0, tt1;
  double dist;
  int nHalfWidth;

  nHalfWidth = (nWidth/2 < 1) ? 1 : nWidth/2;
  //
  //convert the line into a vector
  //
  POINTS2VECTOR2D(pt0, pt1, tt0);
  //
  //convert the mouse points (short) into POINT (long)
  //
  MPOINT2POINT(ptMouse ,PtM);
  POINTS2VECTOR2D(pt0, PtM, tt1);
  //
  //if the mouse click is past the endpoints of
  //a line segment return FALSE
  //
  if (pt0.x <= pt1.x)
  {
    if (PtM.x < (pt0.x - nWidth) || PtM.x > (pt1.x + nWidth))
          return false;
  }
  else
  {
    if (PtM.x > (pt0.x + nWidth) || PtM.x < (pt1.x - nWidth))
          return false;
  }

  //this is the call to the function that does the work
  //of obtaining the distance of the point to the line
  //
  dist = IGCData::distFromPointToLine(&pt0, &pt1, &PtM);

  //
  //TRUE if the distance is within the width of the pen about the
  //line otherwise FALSE
  //

  return (dist >= -nHalfWidth && dist <= nHalfWidth);
}

/*------------------------------------------------------------------------------
| IGCData::angleWithin                                                         |
| Returns true if angle is inside the open angle from angle1 to angle2,        |
| otherwise, returns false.                                                    |
------------------------------------------------------------------------------*/
Boolean IGCData::angleWithin( double angle1, double angle2, double angle )
{
  // Normalization of angle2 and angle relative to angle1.
  // so that angle1 is always adjusted to 0 degree.
  angle2 -= angle1;
  angle -= angle1;
  IGCData::angleNormalize(angle2);
  IGCData::angleNormalize(angle);

  return ( (0<=angle) && (angle<=angle2) );
}

/*------------------------------------------------------------------------------
| IGCData::rectTransformedBy                                                   |
| double precision scaling.                                                    |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::rectTransformedBy(IRectangle& rect, const ITransformMatrix& mat)
{
  double x0 = rect.minX();
  double y0 = rect.minY();
  double x1 = rect.maxX();
  double y1 = rect.maxY();

  IGCData::transformPoint(mat, x0, y0);
  IGCData::transformPoint(mat, x1, y1);

  rect = IRectangle(IPoint(x0, y0), IPoint(x1, y1));
}

/*------------------------------------------------------------------------------
| IGCData::ellipseContains                                                     |
| Using mathematical approach to detect whether IGCData's hitPoint is          |
| contained in an ellipse. Used in Windows95 where such support is not         |
| available.                                                                   |
------------------------------------------------------------------------------*/
Boolean IGCData::ellipseContains(const IGraphicContext& gc, const IRectangle& rect,
                                 Boolean filled)
{
  double aperture = ( hitSize.width() + hitSize.height() ) / 2;
  double w = fPenWidth;
  IPoint center = rect.center();

  // get two radius of the ellipse.
  double a = (rect.maxX() - rect.minX())/2;
  double b = (rect.maxY() - rect.minY())/2;
  Boolean hit = false;

  // map hitPt from device space to logical space.
  // and normalize it relative the center of the
  // ellipse. (in hit testing, the orgin is moved
  // to the center of the ellipse.)
  POINT hitPt;
  hitPt.x = hitPoint.x();
  hitPt.y = hitPoint.y();
  DPtoLP(gc.handle(), &hitPt, 1);
  double xNormal = hitPt.x - center.x();
  double yNormal = hitPt.y - center.y();

  // add aperture and pen width.
  a += ( (aperture+w)/2 );
  b += ( (aperture+w)/2 );
  if ( (fabs(xNormal) > a) || (fabs(yNormal) > b) )
    return false;

  if ( !filled )
  {
    if ( hit = (b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0)  )
    {
      a -= (aperture+w);
      b -= (aperture+w);
      hit = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b >= 0;
    }
  }
  else
  {
    hit = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0;
  }

  return (hit);
}

/*------------------------------------------------------------------------------
| IGCData::arcContains                                                         |
| Using mathematical approach to detect whether IGCData's hitPoint is          |
| contained in an arc. Used in Windows95 where such support is not available.  |
------------------------------------------------------------------------------*/
Boolean IGCData::arcContains(const IGraphicContext& gc, const IGArc& arc, const IRectangle& rect)
{
  double aperture = ( hitSize.width() + hitSize.height() ) / 2;
  double w = fPenWidth;
  IPoint center = rect.center();

  // get two radius of the ellipse.
  double a = (rect.maxX() - rect.minX()) / 2;
  double b = (rect.maxY() - rect.minY()) / 2;
  Boolean onArc = false;

  // map hitPt from device space to logical space.
  // and normalize it relative the center of the
  // ellipse. (in hit testing, the orgin is moved
  // to the center of the ellipse.)
  POINT hitPt;
  hitPt.x = hitPoint.x();
  hitPt.y = hitPoint.y();
  DPtoLP(gc.handle(), &hitPt, 1);
  double xNormal = hitPt.x - center.x();
  double yNormal = hitPt.y - center.y();

  // add aperture and pen width.
  a += ( (aperture+w)/2 );
  b += ( (aperture+w)/2 );
  if ( (fabs(xNormal) > a) || (fabs(yNormal) > b) )
    return false;
  if ( (xNormal==0) && (yNormal==0) )
    return true;

  if ( onArc = (b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0) )
  {
    a -= (aperture+w);
    b -= (aperture+w);
    onArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b >= 0;
  }

  if (onArc)
  {
    double angle1, angle2;

    // Windows95 only supports counterclockwise drawing.
    // So in the case of closewise, starting and ending
    // angle has to be switched.
    if (arc.direction() == IGArc::clockwise)
    {
      angle2 = arc.startAngle();
      angle1 = arc.startAngle() - arc.sweepAngle();
    }
    else
    {
      angle1 = arc.startAngle();
      angle2 = arc.startAngle() + arc.sweepAngle();
    }

    IGCData::angleNormalize(angle1);
    IGCData::angleNormalize(angle2);

    double angle = atan2(yNormal, xNormal);

    //reverse the y-axial for Windows coordinate system.
    if ( gc.mappingMode() == IGraphicContext::pels )
      angle = -angle;

    // transfer radius to degree.
    angle *= (double)DEGREES_TO_RADS;
    if (angle <0)
      angle += 360;
    onArc = onArc && IGCData::angleWithin(angle1, angle2, angle);
  }
  return (onArc);
}

/*------------------------------------------------------------------------------
| IGCData::arc3PointContains                                                   |
| Using mathematical approach to detect whether IGCData's hitPoint is          |
| contained in a 3 point arc. Used in Windows95 where such support is not      |
| available. Notice! from startPt to endPt has to be counterclosewise in       |
| order for this function to work properly.                                    |
------------------------------------------------------------------------------*/
Boolean IGCData::arc3PointContains(const IGraphicContext& gc, const IRectangle& rect,
                                   const IPoint& startPt, const IPoint& endPt)
{
  double aperture = ( hitSize.width() + hitSize.height() ) / 2;
  double w = fPenWidth;
  IPoint center = rect.center();

  // get two radius of the ellipse.
  double a = (rect.maxX() - rect.minX()) / 2;
  double b = (rect.maxY() - rect.minY()) / 2;
  Boolean onArc = false;

  // map hitPt from device space to logical space.
  // and normalize it relative the center of the
  // ellipse. (in hit testing, the orgin is moved
  // to the center of the ellipse.)
  POINT hitPt;
  hitPt.x = hitPoint.x();
  hitPt.y = hitPoint.y();
  DPtoLP(gc.handle(), &hitPt, 1);
  double xNormal = hitPt.x - center.x();
  double yNormal = hitPt.y - center.y();

  // add aperture and pen width.
  a += ( (aperture+w)/2 );
  b += ( (aperture+w)/2 );
  if ( (fabs(xNormal) > a) || (fabs(yNormal) > b) )
    return false;
  if ( (xNormal==0) && (yNormal==0) )
    return true;

  if ( onArc = (b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0) )
  {
    a -= (aperture+w);
    b -= (aperture+w);
    onArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b >= 0;
  }

  if (onArc)
  {
    // calculating the starting angle: angle1 and the ending angle: angle2
    // for the 3 point arc.
    double angle1 = atan2((startPt.y() - center.y()), (startPt.x() - center.x()));

    //reverse the y-axial for Windows coordinate system.
    if ( gc.mappingMode() == IGraphicContext::pels )
     angle1 = -angle1;

    // transfer radius to degree.
    angle1 *= (double)DEGREES_TO_RADS;
    if (angle1 <0)
      angle1 += 360;

    double angle2 = atan2((endPt.y() - center.y()), (endPt.x() - center.x()));
    if ( gc.mappingMode() == IGraphicContext::pels )
      angle2 = -angle2;
    angle2 *= (double)DEGREES_TO_RADS;
    if (angle2 <0)
      angle2 += 360;

    double angle = atan2(yNormal, xNormal);
    if ( gc.mappingMode() == IGraphicContext::pels )
      angle = -angle;
    angle *= (double)DEGREES_TO_RADS;
    if (angle <0)
      angle += 360;

    onArc = onArc && IGCData::angleWithin(angle1, angle2, angle);
  }
  return (onArc);
}

/*------------------------------------------------------------------------------
| IGCData::pieContains                                                         |
| Using mathematical approach to detect whether IGCData's hitPoint is          |
| contained in a pie. Used in Windows95 where such support is not available.   |
------------------------------------------------------------------------------*/
Boolean IGCData::pieContains(const IGraphicContext& gc, const IGPie& pie,
                             const IRectangle& rect, Boolean filled )
{
  double aperture = ( hitSize.width() + hitSize.height() ) / 2;
  double w = fPenWidth;
  IPoint center = rect.center();

  // get two radius of the ellipse.
  double a = (rect.maxX() - rect.minX()) / 2;
  double b = (rect.maxY() - rect.minY()) / 2;
  Boolean onArc = false;
  Boolean onSide = false;

  // map hitPt from device space to logical space.
  // and normalize it relative the center of the
  // ellipse. (in hit testing, the orgin is moved
  // to the center of the ellipse.)
  POINT hitPt;
  hitPt.x = hitPoint.x();
  hitPt.y = hitPoint.y();
  DPtoLP(gc.handle(), &hitPt, 1);
  double xNormal = hitPt.x - center.x();
  double yNormal = hitPt.y - center.y();

  // add aperture and pen width.
  a += ( (aperture+w)/2 );
  b += ( (aperture+w)/2 );
  if ( (fabs(xNormal) > a) || (fabs(yNormal) > b) )
    return false;
  if ( (xNormal==0) && (yNormal==0) )
    return true;

  double angle = atan2(yNormal, xNormal);  // minus for Windows coordinate
  if ( gc.mappingMode() == IGraphicContext::pels )
    angle = -angle;
  angle *= (double)DEGREES_TO_RADS;
  if (angle <0)
    angle += 360;

  double angle1 = pie.startAngle();
  double angle2 = pie.startAngle() + pie.sweepAngle();
  IGCData::angleNormalize(angle1);
  IGCData::angleNormalize(angle2);

  if ( !filled )
  {
    if ( onArc = (b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0) )
    {
      a -= (aperture+w);
      b -= (aperture+w);
      onArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b >= 0;
    }
    onArc = onArc && IGCData::angleWithin(angle1, angle2, angle);

    onSide = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0;
    onSide = onSide && ( (angle == angle1) || (angle == angle2) );
  }
  else
  {
    onArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0;
    onArc = onArc && IGCData::angleWithin(angle1, angle2, angle);
  }

  return(onArc || onSide);
}

/*------------------------------------------------------------------------------
| IGCData::chordContains                                                       |
| Using mathematical approach to detect whether IGCData's hitPoint is          |
| contained in a chord. Used in Windows95 where such support is not available. |
------------------------------------------------------------------------------*/
Boolean IGCData::chordContains(const IGraphicContext& gc, const IGChord& chord,
                               const IRectangle& rect, Boolean filled)
{
  double aperture = ((double)( hitSize.width() + hitSize.height() )) / 2;
  double w = fPenWidth;
  IPoint center = rect.center();

  // get two radius of the ellipse.
  double a = ((double)(rect.maxX() - rect.minX())) / 2;
  double b = ((double)(rect.maxY() - rect.minY())) / 2;

  // map hitPt from device space to logical space.
  // and normalize it relative the center of the
  // ellipse. (in hit testing, the orgin is moved
  // to the center of the ellipse.)
  POINT hitPt;
  hitPt.x = hitPoint.x();
  hitPt.y = hitPoint.y();
  DPtoLP(gc.handle(), &hitPt, 1);
  double xNormal = (double)(hitPt.x - center.x());
  double yNormal = (double)(hitPt.y - center.y());
  Boolean inArc=false, inside=false, inChord=false;

  if ( (fabs(xNormal) > a + (aperture+w)/2) || (fabs(yNormal) > b + (aperture+w)/2) )
    return false;

  double angle1 = chord.startAngle();
  double angle2 = chord.startAngle() + chord.sweepAngle();
  IGCData::angleNormalize(angle1);
  IGCData::angleNormalize(angle2);

  // get the starting and ending points of the chord;
  // the Ellipse is represented in polar system which
  // makes calculation mush easier.
  // sqr(r) = sqr(ab)/(sqr(a*sin(theta)) + sqr(b*cos(theta)))
  // here r is always positive and theta decides the sign of x1, y1.
  double x1, y1, x2, y2, r, theta, base1, base2;
  theta = angle1 / (double)DEGREES_TO_RADS;
  base1 = a*a - b*b;
  base2 = cos(theta);
  r = a*b/(sqrt(a*a - base1 * base2*base2));
  x1 = r*base2;
  y1 = r*sin(theta);

  theta = angle2 / (double)DEGREES_TO_RADS;
  base2 = cos(theta);
  r = a*b/(sqrt(a*a - base1 * base2*base2));
  x2 = r*base2;
  y2 = r*sin(theta);

  double position;
  if (x1==x2)
    position = x1 - xNormal;
  else
  {
    if ( gc.mappingMode()       == IGraphicContext::pels )
      position = - yNormal - y1 - (xNormal-x1)*(y1-y2)/(x1-x2);
    else
      position = yNormal - y1 - (xNormal-x1)*(y1-y2)/(x1-x2);
  }

  if ( (xNormal==0) && (yNormal==0) )
    return (position==0);

  double angle = atan2(yNormal, xNormal);
  if ( gc.mappingMode() == IGraphicContext::pels )
    angle = -angle;
  angle *= (double)DEGREES_TO_RADS;
  if (angle <0)
    angle += 360;

  if ( !filled )
  {
    // add aperture and pen width.
    a += ( (aperture+w)/2 );
    b += ( (aperture+w)/2 );
    if ( inArc = (b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0) )
    {
      a -= (aperture+w);
      b -= (aperture+w);
      inArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b >= 0;
    }

    inside = IGCData::angleWithin(angle1, angle2, angle);
    POINT pt1, pt2, ptHit;
    pt1.x = x1;
    pt1.y = y1;
    pt2.x = x2;
    pt2.y = y2;
    ptHit.x = xNormal;
    ptHit.y = yNormal;
    inChord = IGCData::hitTestLine(pt1, pt2, ptHit, aperture+w);

    return ( (inArc || inChord) && inside );
  }
  else
  {
    a += ( (aperture+w)/2 );
    b += ( (aperture+w)/2 );
    inArc = b*b * xNormal*xNormal + a*a * yNormal*yNormal -  a*a*b*b <= 0;

    if (x1>=x2)
      inChord = (position>=0);
    else
      inChord = (position<=0);

    return (inArc && inChord);
  }
}

/*------------------------------------------------------------------------------
| IGCData::boundingRectangle                                                   |
| rect is the bounding rectangle BEFORE any transformations.                   |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGCData::boundingRectangle(const IGraphicContext& gc, const IRectangle& rect,
                                      double startAngle, double sweepAngle,
                                      Boolean isIGPie, Boolean clockwise)
{
  double r, theta, base1, base2, angle1, angle2;
  double borderArrayX[7], borderArrayY[7];   // accumulating the end points.
  int borderCount=0, i;

  if (isIGPie) // if IGPie, count in the center point. I.e. the origin.
  {
    borderArrayX[borderCount] = 0;
    borderArrayY[borderCount++] = 0;
  }

  if (clockwise)
  {
    angle2 = startAngle;
    angle1 = startAngle - sweepAngle;
  }
  else
  {
    angle1 = startAngle;
    angle2 = startAngle + sweepAngle;
  }

  IGCData::angleNormalize(angle1);
  IGCData::angleNormalize(angle2);

  // get two radius of the ellipse.
  double a = ((double)(rect.maxX() - rect.minX())) / 2;
  double b = ((double)(rect.maxY() - rect.minY())) / 2;

  // get the two interaction points or ends of the arc.
  // the Ellipse is represented in polar system which
  // makes calculation mush easier.
  // sqr(r) = sqr(ab)/(sqr(a*sin(theta)) + sqr(b*cos(theta)))
  // here r is always positive and theta decides the sign of x, y.
  theta = angle1 / (double)DEGREES_TO_RADS;
  base1 = a*a - b*b;
  base2 = cos(theta);
  r = a*b/(sqrt(a*a - base1 * base2*base2));
  borderArrayX[borderCount] = r*base2;
  double yCoord = r*sin(theta);
  if ( gc.mappingMode() == IGraphicContext::pels )
    borderArrayY[borderCount++] = -yCoord;
  else
    borderArrayY[borderCount++] = yCoord;

  theta = angle2 / (double)DEGREES_TO_RADS;
  base2 = cos(theta);
  r = a*b/(sqrt(a*a - base1 * base2*base2));
  borderArrayX[borderCount] = r*base2;
  yCoord = r*sin(theta);
  if ( gc.mappingMode() == IGraphicContext::pels )
    borderArrayY[borderCount++] = -yCoord;
  else
    borderArrayY[borderCount++] = yCoord;

  //reverse the y-axial for Windows coordinate system.
  double yRadius = b;
  if ( gc.mappingMode() == IGraphicContext::pels )
    yRadius = -yRadius;

  if (IGCData::angleWithin(angle1, angle2, 0))
  {
    borderArrayX[borderCount] = a;
    borderArrayY[borderCount++] = 0;
  }
  if (IGCData::angleWithin(angle1, angle2, 90))
  {
    borderArrayX[borderCount] = 0;
    borderArrayY[borderCount++] = yRadius;
  }
  if (IGCData::angleWithin(angle1, angle2, 180))
  {
    borderArrayX[borderCount] = -a;
    borderArrayY[borderCount++] = 0;
  }
  if (IGCData::angleWithin(angle1, angle2, 270))
  {
    borderArrayX[borderCount] = 0;
    borderArrayY[borderCount++] = -yRadius;
  }

  // We get all the "ends" of the figure in borderArray[]
  // when it is normalize to the origin to simplify the algorithm.
  // then, the figure is moved back to its original position,
  // then apply the whole transformation to those points in
  // the borderArray[]. Then we get the final bounding rect
  // from those transferedPt[].

  double centerX = ((double) (rect.maxX() + rect.minX())) / 2;
  double centerY = ((double) (rect.maxY() + rect.minY())) / 2;
  for (i=0; i<borderCount; i++)
  {
    borderArrayX[i] = borderArrayX[i] + centerX;
    borderArrayY[i] = borderArrayY[i] + centerY;
    IGCData::transformPoint(gc.worldTransformMatrix( ),
                            borderArrayX[i], borderArrayY[i]);
  }

  double left=borderArrayX[0], bottom = borderArrayY[0],
         right=borderArrayX[0], top = borderArrayY[0];
  for (i=1; i<borderCount; i++)
  {
    if (borderArrayX[i] < left)
      left = borderArrayX[i];
    else if (right < borderArrayX[i])
      right = borderArrayX[i];

    if (borderArrayY[i] < bottom)
      bottom = borderArrayY[i];
    else if (top < borderArrayY[i])
      top = borderArrayY[i];
  }

  return (IRectangle(IPoint(left, bottom), IPoint(right, top)));
}

/*------------------------------------------------------------------------------
| IGCData::hitTestRegion                                                       |
|                                                                              |
| hitPoint in device space, regionHandle in logical space                      |
------------------------------------------------------------------------------*/
Boolean IGCData::hitTestRegion( const IGraphicContext& gc, const IRegionHandle regionHandle, Boolean inLogicalSpace )
{
  Boolean          fReturn(false);
  POINT            mousePt;

  mousePt.x = this->hitPoint.x();
  mousePt.y = this->hitPoint.y();
  // move mousePt and region to same space
  if ( inLogicalSpace )
    DPtoLP( gc.handle(), &mousePt, 1 );

  IRectangle tempRect(hitSize.width(), hitSize.height());
  tempRect.centerAt( IPoint( mousePt.x, mousePt.y ));

  RECT hitRect;
  hitRect.left   = tempRect.left();
  hitRect.right  = tempRect.right();
  hitRect.top    = tempRect.top();
  hitRect.bottom = tempRect.bottom();

  if (RectInRegion( regionHandle, &hitRect ))
    fReturn=true;

  return fReturn;
}

/*------------------------------------------------------------------------------
| IGCData::hitTestPointArray                                                   |
|                                                                              |
| hitPoint in device space, PointArray in logical space                        |
------------------------------------------------------------------------------*/
Boolean IGCData::hitTestPointArray( const IGraphicContext& gc, POINT* lpPt, unsigned long numPoints )
{
  Boolean          fReturn(false);
  POINT            mousePt;

  mousePt.x = this->hitPoint.x();
  mousePt.y = this->hitPoint.y();
  // Convert from device space to logical space
  DPtoLP( gc.handle(), &mousePt, 1 );

  if (numPoints == 1)
  {
  double  dist = sqrt((double)((mousePt.x - lpPt[0].x) * (mousePt.x - lpPt[0].x)
                     + (mousePt.y - lpPt[0].y) * (mousePt.y - lpPt[0].y)));
  int     nHalfWidth = ((this->hitSize.width()/2 < 1)
                        ? 1 : this->hitSize.width()/2);
  fReturn = (dist >= -nHalfWidth && dist <= nHalfWidth);
  }
  else if (numPoints > 1)
  {
    for ( int nPt = 0; nPt < numPoints - 1; nPt++)
    {
      if (IGCData::hitTestLine(lpPt[nPt],
                               lpPt[nPt + 1],
                               mousePt,
                               this->hitSize.width()))
      {
        fReturn = true;
        break;
      }
    } /* endfor */
  } /* endif */

  return fReturn;
}

/*------------------------------------------------------------------------------
| IGCData::hitTestPath                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGCData::hitTestPath( const IGraphicContext& gc )
{
  int              numPoints;
  LPPOINT          lpPt;
  LPBYTE           lpB;
  Boolean          fReturn(false);

  if (!EndPath(gc.handle()))
    return false;

  if (!FlattenPath(gc.handle()))
    return false;

  numPoints = GetPath(gc.handle(), (LPPOINT)NULL, (LPBYTE)NULL, 0);
  if (numPoints > 0)
  {
    lpPt = (LPPOINT)malloc((DWORD)(sizeof(POINT) * numPoints));
    lpB  = (LPBYTE)malloc((DWORD)(sizeof(BYTE) * numPoints));
    numPoints = GetPath(gc.handle(), lpPt, lpB, numPoints);

    fReturn = hitTestPointArray( gc, lpPt, numPoints );

    free(lpPt);
    free(lpB);
  }
  return fReturn;
}

/*------------------------------------------------------------------------------
| IGCData::setWorldTransform                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGCData::setWorldTransform(const ITransformMatrix& tmx,
                                IGraphic::TransformMethod tmd)
{
  if ( tmd == IGraphic::replace )
  {
    fTransformMatrix.setElement11(tmx.element11());
    fTransformMatrix.setElement12(tmx.element12());
    fTransformMatrix.setElement21(tmx.element21());
    fTransformMatrix.setElement22(tmx.element22());
    fTransformMatrix.setElement31(tmx.element31());
    fTransformMatrix.setElement32(tmx.element32());
  }
  else if ( tmd == IGraphic::rightMultiply )
  {
    double tmp = fTransformMatrix.element11()*tmx.element12() +
                 fTransformMatrix.element12()*tmx.element22();
    fTransformMatrix.setElement11(fTransformMatrix.element11()*tmx.element11() +
                                  fTransformMatrix.element12()*tmx.element21());
    fTransformMatrix.setElement12(tmp);

    tmp = fTransformMatrix.element21()*tmx.element12() +
          fTransformMatrix.element22()*tmx.element22();
    fTransformMatrix.setElement21(fTransformMatrix.element21()*tmx.element11() +
                                  fTransformMatrix.element22()*tmx.element21());
    fTransformMatrix.setElement22(tmp);

    tmp = fTransformMatrix.element31()*tmx.element12() +
          fTransformMatrix.element32()*tmx.element22();
    fTransformMatrix.setElement31(fTransformMatrix.element31()*tmx.element11() +
                                  fTransformMatrix.element32()*tmx.element21());
    fTransformMatrix.setElement32(tmp);

    fTransformMatrix.setElement31(fTransformMatrix.element31() + tmx.element31());
    fTransformMatrix.setElement32(fTransformMatrix.element32() + tmx.element32());
  }
  else
  {
    double tmp1 = fTransformMatrix.element11();
    double tmp2 = fTransformMatrix.element21();

    fTransformMatrix.setElement11(tmx.element11()*tmp1 + tmx.element12()*tmp2);
    fTransformMatrix.setElement21(tmx.element21()*tmp1 + tmx.element22()*tmp2);
    fTransformMatrix.setElement31(fTransformMatrix.element31() +
                                  tmx.element31()*tmp1 + tmx.element32()*tmp2);

    tmp1 = fTransformMatrix.element12();
    tmp2 = fTransformMatrix.element22();

    fTransformMatrix.setElement12(tmx.element11()*tmp1 + tmx.element12()*tmp2);
    fTransformMatrix.setElement22(tmx.element21()*tmp1 + tmx.element22()*tmp2);
    fTransformMatrix.setElement32(fTransformMatrix.element32() +
                                  tmx.element31()*tmp1 + tmx.element32()*tmp2);
  }
}

/*------------------------------------------------------------------------------
| IGCData::notRotated                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGCData::notRotated( const ITransformMatrix& matrix )
{
  return ((matrix.element12() == 0) && (matrix.element21() == 0));
//return (((matrix.element12() == 0) && (matrix.element21() == 0)) &&
//        ((matrix.element11() != 1) || (matrix.element22() != 1)));
}

#pragma data_seg(ICLNonConst)
IColor                            IGraphicContext::fgCDPenColor(IColor::black);
IColor                            IGraphicContext::fgCDFillColor(IColor::black);
IColor                            IGraphicContext::fgCDBackgroundColor(IColor::white);
IGraphicBundle::DrawOperation     IGraphicContext::fgCDDrawOperation(IGraphicBundle::fillAndFrame);
IGraphicBundle::MixMode           IGraphicContext::fgCDMixMode(IGraphicBundle::overPaint);
IGraphicBundle::BackgroundMixMode IGraphicContext::fgCDBackgroundMixMode(IGraphicBundle::backLeaveAlone);
unsigned long                     IGraphicContext::fgCDPenPattern(IGraphicBundle::filled);
unsigned long                     IGraphicContext::fgCDFillPattern(IGraphicBundle::filled);
IGraphicBundle::PenType           IGraphicContext::fgCDPenType(IGraphicBundle::solid);
IGraphicBundle::PenEndingStyle    IGraphicContext::fgCDPenEndingStyle(IGraphicBundle::flat);
IGraphicBundle::PenJoiningStyle   IGraphicContext::fgCDPenJoiningStyle(IGraphicBundle::bevel);
unsigned long                     IGraphicContext::fgCDPenWidth(1);
IPoint                            IGraphicContext::fgCDPatternOrigin(IPoint(0,0));
IBitmapHandle                     IGCData::fgBmpHalftone(0);
IBitmapHandle                     IGCData::fgBmpDense1(0);
IBitmapHandle                     IGCData::fgBmpDense2(0);
IBitmapHandle                     IGCData::fgBmpDense3(0);
IBitmapHandle                     IGCData::fgBmpDense4(0);
IBitmapHandle                     IGCData::fgBmpDense5(0);
IBitmapHandle                     IGCData::fgBmpDense6(0);
IBitmapHandle                     IGCData::fgBmpDense7(0);
IBitmapHandle                     IGCData::fgBmpDense8(0);
IBitmapHandle                     IGCData::fgBmpfDiag2(0);
IBitmapHandle                     IGCData::fgBmpbDiag2(0);
#pragma data_seg()

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenColor                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::defaultPenColor()
{
  return fgCDPenColor;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultFillColor                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::defaultFillColor()
{
  return fgCDFillColor;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultBackgroundColor                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::defaultBackgroundColor()
{
  return fgCDBackgroundColor;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultDrawOperation                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::DrawOperation IGraphicContext::defaultDrawOperation()
{
  return fgCDDrawOperation;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultMixMode                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::MixMode IGraphicContext::defaultMixMode()
{
  return fgCDMixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultBackgroundMixMode                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::BackgroundMixMode IGraphicContext::defaultBackgroundMixMode()
{
  return fgCDBackgroundMixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenPattern                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::defaultPenPattern()
{
  return fgCDPenPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultFillPattern                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::defaultFillPattern()
{
  return fgCDFillPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenType                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenType IGraphicContext::defaultPenType()
{
  return fgCDPenType;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenEndingStyle                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenEndingStyle IGraphicContext::defaultPenEndingStyle()
{
  return fgCDPenEndingStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenJoin                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenJoiningStyle IGraphicContext::defaultPenJoiningStyle()
{
  return fgCDPenJoiningStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPenWidth                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::defaultPenWidth()
{
  return fgCDPenWidth;
}

/*------------------------------------------------------------------------------
| IGraphicContext::defaultPatternOrigin                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGraphicContext::defaultPatternOrigin()
{
  return fgCDPatternOrigin;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultBackPenColor                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenColor( const IColor& color )
{
  fgCDPenColor = color;
}
/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultBackFillColor                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultFillColor( const IColor& color )
{
  fgCDFillColor = color;
}
/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultBackgroundColor                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultBackgroundColor( const IColor& color )
{
  fgCDBackgroundColor = color;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultDrawOperation                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultDrawOperation( IGraphicBundle::DrawOperation drawOperation )
{
  fgCDDrawOperation = drawOperation;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultMixMode                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultMixMode( IGraphicBundle::MixMode mixMode )
{
  fgCDMixMode = mixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultBackgroundMixMode                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultBackgroundMixMode(
                           IGraphicBundle::BackgroundMixMode backgroundMixMode )
{
  fgCDBackgroundMixMode = backgroundMixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPenPattern                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenPattern( unsigned long penPattern )
{
  fgCDPenPattern = penPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultFillPattern                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultFillPattern( unsigned long fillPattern )
{
  fgCDFillPattern = fillPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPenType                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenType( IGraphicBundle::PenType penType )
{
  fgCDPenType = penType;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPenEndingStyle                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenEndingStyle(
                                 IGraphicBundle::PenEndingStyle penEndingStyle )
{
  fgCDPenEndingStyle = penEndingStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPenJoin                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenJoiningStyle(
                               IGraphicBundle::PenJoiningStyle penJoiningStyle )
{
  fgCDPenJoiningStyle = penJoiningStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPenWidth                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPenWidth( unsigned long penWidth )
{
  fgCDPenWidth = penWidth;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultPatternOrigin                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IGraphicContext::setDefaultPatternOrigin( const IPoint& point )
{
  fgCDPatternOrigin = point;
}

/*------------------------------------------------------------------------------
| IGraphicContext::initialize                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::initialize()
{
  fGCData = new IGCData( IGCData::drawing,
                         defaultDrawOperation(),
                         defaultPenColor(),
                         defaultFillColor(),
                         defaultBackgroundColor(),
                         defaultMixMode(),
                         defaultBackgroundMixMode(),
                         defaultPenWidth(),
                         defaultPenType(),
                         defaultPenEndingStyle(),
                         defaultPenJoiningStyle(),
                         defaultPenPattern(),
                         defaultFillPattern() );

  setBackgroundMixMode( defaultBackgroundMixMode() );
  setPatternOrigin( defaultPatternOrigin() );
  fGCData->createPen( fPs );
  fGCData->createBrush( fPs );

  // If SetGraphicsMode succeed, NT mode, if not Windows95 or Win32s mode.
  fGCData->fAdvancedMode = SetGraphicsMode( fPs, GM_ADVANCED );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::IGraphicContext                                             |
|                                                                              |
| Create a memory PS                                                           |
------------------------------------------------------------------------------*/
IGraphicContext::IGraphicContext( )
  : fGCData(0),fFillColor(defaultFillColor()),
    fFillPattern(defaultFillPattern()), fDrawOp(defaultDrawOperation()),
    fFont(0), fCm( Graphics ), fDc(0), fBitmap(0)

{
  fPs = CreateCompatibleDC( 0 );
  initialize();
}

/*------------------------------------------------------------------------------
| IGraphicContext::IGraphicContext                                             |
|                                                                              |
| Wrapper an existing PS                                                       |
------------------------------------------------------------------------------*/
IGraphicContext::IGraphicContext( const IPresSpaceHandle& presSpaceHandle )
  : fGCData(0),fPs( presSpaceHandle ), fFillColor(defaultFillColor()),
    fFillPattern(defaultFillPattern()), fDrawOp(defaultDrawOperation()),
    fFont(0), fCm( None ), fDc(0), fBitmap(0)
{
  initialize();
}

/*------------------------------------------------------------------------------
| IGraphicContext::IGraphicContext                                             |
|                                                                              |
| Get a screen PS associated with the given window handle.                     |
------------------------------------------------------------------------------*/
IGraphicContext::IGraphicContext( const IWindowHandle& windowHandle )
  : fGCData(0),fFillColor(defaultFillColor()),
    fFillPattern(defaultFillPattern()), fDrawOp(defaultDrawOperation()),
    fFont(0), fCm( WindowHandle ), fDc(0), fBitmap(0)
{
  fPs = GetDC( windowHandle );
  if (!fPs)
    ITHROWGUIERROR("GetDC");

  initialize();
  fGCData->dcWindow = windowHandle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::IGraphicContext                                             |
|                                                                              |
| Create a memory PS with the specified size for temporary drawing             |
------------------------------------------------------------------------------*/
IGraphicContext::IGraphicContext(const ISize& contextSize)
  : fGCData(0),fFillColor(defaultFillColor()),
    fFillPattern(defaultFillPattern()), fDrawOp(defaultDrawOperation()),
    fFont(0), fCm( Graphics ), fDc(0), fBitmap(0)

{
  fPs = CreateCompatibleDC( 0 );
  if (!fPs)
    ITHROWGUIERROR("CreateCompatibleDC");

  HDC hdcDesktop = GetDC( IWindow::desktopWindow()->handle() );
  fBitmap = CreateCompatibleBitmap( hdcDesktop, contextSize.width(), contextSize.height() );
  ReleaseDC( IWindow::desktopWindow()->handle(), hdcDesktop );

  if (!fBitmap)
    ITHROWGUIERROR("CreateCompatibleBitmap");

  initialize();

  fGCData->originalBmp = SelectObject( fPs, fBitmap );
}

/*------------------------------------------------------------------------------
| IGraphicContext::setMappingMode                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setMappingMode( MappingMode mappingMode )
{
  int mapMode;

  switch (mappingMode)
  {
    case isotropic:
      mapMode = MM_ISOTROPIC;
    break;
    case pels:
      mapMode = MM_TEXT;
    break;
    case lowMetric:
      mapMode = MM_LOMETRIC;
    break;
    case highMetric:
      mapMode = MM_HIMETRIC;
    break;
    case lowEnglish:
      mapMode = MM_LOENGLISH;
    break;
    case highEnglish:
      mapMode = MM_HIENGLISH;
    break;
    case twips:
      mapMode = MM_TWIPS;
    break;
  } /* endswitch */
  if (!SetMapMode( fPs, mapMode ))
  {
    ITHROWGUIERROR2("SetMapMode",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::mappingMode                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext::MappingMode IGraphicContext::mappingMode( ) const
{
  int map(GetMapMode( fPs ));

  switch (map)
  {
    case MM_ISOTROPIC:
      return isotropic;
    case MM_TEXT:
      return pels;
    case MM_LOMETRIC:
      return lowMetric;
    case MM_HIMETRIC:
      return highMetric;
    case MM_LOENGLISH:
      return lowEnglish;
    case MM_HIENGLISH:
      return highEnglish;
    case MM_TWIPS:
      return twips;
    default:
      return pels;
  } /* endswitch */
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPageSize                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPageSize( const ISize& pageSize )
{
  SetWindowExtEx( fPs, pageSize.width(), pageSize.height(), 0 );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::pageSize                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ISize IGraphicContext::pageSize( ) const
{
  SIZE page;
  GetWindowExtEx( fPs, &page );
  return ISize(page.cx, page.cy);
}

/*------------------------------------------------------------------------------
| IGraphicContext::setViewPortRect                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setViewPortRect( const IRectangle& viewportRect )
{
  SetViewportExtEx( fPs, viewportRect.width(), viewportRect.height(), 0 );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::viewPortRect                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGraphicContext::viewPortRect( ) const
{
  SIZE viewPort;
  GetViewportExtEx( fPs, &viewPort );
  return IRectangle(IPoint(0,0), ISize( viewPort.cx, viewPort.cy ));
}

/*------------------------------------------------------------------------------
| IGraphicContext::~IGraphicContext                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext::~IGraphicContext( )
{
  if (fGCData)
  {
    Boolean success;
    if (fGCData->originalBrush)
    {
      HBRUSH temp;
      temp = SelectObject(fPs, fGCData->originalBrush);
      DeleteObject(temp);
    }
    if (fGCData->originalPen)
    {
      HPEN temp;
      temp = SelectObject(fPs, fGCData->originalPen);
      DeleteObject(temp);
    }
  }

  // delete the IFont set for the gc
  if (fFont)
  {
    fFont->endUsingFont( fPs );
    delete fFont;
  }

  // Delete temporary drawing bitmap (if needed)
  if (fBitmap)
  {
    SelectObject( fPs, fGCData->originalBmp );
    DeleteObject( fBitmap );
  }

  if ( fCm == Graphics )
    DeleteDC( fPs );
  else if ( fCm == WindowHandle )
    ReleaseDC( fGCData->dcWindow, fPs );

  if (fGCData)
    delete fGCData;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT, Windows95, and Win32s.          |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGLine& line )
{
  IPoint    start(line.startingPoint());
  IPoint    end(line.endingPoint());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  long      rc;

  if (fGCData->fRecoordinate        &&
      !isAccumulatingBoundingRect() &&
      !isHitTesting())
  {
    start.setY( fGCData->recoordHeight - start.y() );
    end.setY( fGCData->recoordHeight - end.y() );
  }

  apply( line, currentTransform, currentBundle );

  if (isAccumulatingBoundingRect())
  {
    IRectangle bounding(IPoint( IGCData::transformPoint( worldTransformMatrix(),
                                                         start )),
                        IPoint( IGCData::transformPoint( worldTransformMatrix(),
                                                         end )));
    if (penWidth() > 1)
      bounding.expandBy(penWidth());

    addToBoundingRect( bounding );
  }
  else if ( IPlatform::isWin32S() && isHitTesting())
  {
    IPoint  sPt( IGCData::transformPoint(worldTransformMatrix(), start));
    IPoint  ePt( IGCData::transformPoint(worldTransformMatrix(), end));
    POINTL  pline[2];
    pline[0] = sPt.asPOINTL();
    pline[1] = ePt.asPOINTL();
    if ( fGCData->hitTestPointArray( *this, (POINT*)pline, 2 ))
      ((IGLine&)line).setHitSelected(true);
  }
  else
  {
    if (fGCData->fAdvancedMode)   // NT mode
      setCurrentDrawingPosition( start );
    else // Windows95 or Win32s
      setCurrentDrawingPosition( IGCData::transformPoint(worldTransformMatrix(), start) );

    if (isHitTesting())
    {
      BeginPath( fPs );
    }

    if (fGCData->fAdvancedMode)   // NT mode
      LineTo( fPs, end.x(), end.y() );
    else   // Windows95 or Win32s
    {
      IPoint ep(IGCData::transformPoint(worldTransformMatrix(), end));
      LineTo( fPs, ep.x(), ep.y() );
    }

    if (isHitTesting())
    {
      if (fGCData->hitTestPath(*this))
        ((IGLine&)line).setHitSelected(true);
    }
  }

  remove( line, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT, Windows95, and Win32s.          |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGPolyline& polyline )
{
  if (polyline.numberOfPoints() > 1)
  {
    ITransformMatrix currentTransform;
    IGraphicBundle   currentBundle;
    Boolean          rc;
    unsigned long    arraySize(polyline.pointArray().size());
    POINT*           pPtl(new POINT[arraySize]);

    apply( polyline, currentTransform, currentBundle );

    if ( !IPlatform::isWin32S() && isHitTesting())
    {
      BeginPath( fPs );
    }

    for (int i=0; i < arraySize; i++)
    {
      // replace with tagPoint accessor when IPoint is updated.
      pPtl[i].x = polyline.point(i).x();

      if (fGCData->fRecoordinate        &&
          !isAccumulatingBoundingRect() &&
          !isHitTesting())
         pPtl[i].y = fGCData->recoordHeight - polyline.point(i).y();
      else
        pPtl[i].y = polyline.point(i).y();
    }

    if (!fGCData->fAdvancedMode || isAccumulatingBoundingRect())
    {
      for (int i=0; i < arraySize; i++)
      {
        IPoint pt = IGCData::transformPoint(worldTransformMatrix( ),
                                            IPoint( pPtl[i].x, pPtl[i].y ));
        pPtl[i].x = pt.x();
        pPtl[i].y = pt.y();
      }
    } /* endfor */

    if (isAccumulatingBoundingRect())
    {
      int maxX, minX, maxY, minY;
      maxX = minX = pPtl[0].x;
      maxY = minY = pPtl[0].y;
      for (i = 1;i < arraySize; i++)
      {
        if (minX > pPtl[i].x)
          minX = pPtl[i].x;
        if (maxX < pPtl[i].x)
          maxX = pPtl[i].x;
        if (minY > pPtl[i].y)
          minY = pPtl[i].y;
        if (maxY < pPtl[i].y)
          maxY = pPtl[i].y;
      }
      IRectangle bounding(minX, minY, maxX, maxY);
      if (penWidth() > 1)
        bounding.expandBy(penWidth());

      addToBoundingRect(bounding);
    }
    else if ( IPlatform::isWin32S() && isHitTesting())
    {
      if ( fGCData->hitTestPointArray( *this, pPtl, arraySize ))
        ((IGPolyline&)polyline).setHitSelected(true);
    }
    else
    {
      if (!Polyline( fPs, pPtl, polyline.numberOfPoints() ))
      {
        ITHROWGUIERROR2("Polyline",
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
      }

      if (fGCData->fAdvancedMode) // NT mode
        setCurrentDrawingPosition( polyline.pointArray()[polyline.numberOfPoints()-1] );
      else  // WIndows95
        setCurrentDrawingPosition(IGCData::transformPoint(worldTransformMatrix(),
                                  polyline.pointArray()[polyline.numberOfPoints()-1]));

      if (isHitTesting())
      {
        if (fGCData->hitTestPath(*this))
          ((IGPolyline&)polyline).setHitSelected(true);
      }
    }

    delete [] pPtl;

    remove( polyline, currentTransform, currentBundle );
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT, Windows95, and Win32s.          |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGPolygon& polygon )
{
  if (polygon.numberOfPoints() > 1)
  {
    HBRUSH           tempBrush(0);
    HPEN             tempPen(0);
    ITransformMatrix currentTransform;
    IGraphicBundle   currentBundle;
    unsigned long    arraySize(polygon.pointArray().size());
    POINT*           pPtl(new POINT[arraySize+1]);
    long             fillMode(polygon.fillMode() == IGPolygon::alternate
                                                  ? ALTERNATE
                                                  : WINDING);

    apply( polygon, currentTransform, currentBundle );

    for (int i=0; i < arraySize; i++)
    {
      // replace with tagPoint accessor when IPoint is updated.
      pPtl[i].x = polygon.point(i).x();

      if (fGCData->fRecoordinate        &&
          !isAccumulatingBoundingRect() &&
          !isHitTesting())
        pPtl[i].y = fGCData->recoordHeight - polygon.point(i).y();
      else
        pPtl[i].y = polygon.point(i).y();
    }

    if (!fGCData->fAdvancedMode || isAccumulatingBoundingRect())
    {
      for (int i=0; i < arraySize; i++)
      {
        IPoint pt = IGCData::transformPoint(worldTransformMatrix( ),
                                            IPoint( pPtl[i].x, pPtl[i].y ));
        pPtl[i].x = pt.x();
        pPtl[i].y = pt.y();
      }
    } /* endfor */

    if (isAccumulatingBoundingRect())
    {
      int maxX, minX, maxY, minY;
      maxX = minX = pPtl[0].x;
      maxY = minY = pPtl[0].y;
      for (i = 1;i < arraySize; i++)
      {
        if (minX > pPtl[i].x)
          minX = pPtl[i].x;
        if (maxX < pPtl[i].x)
          maxX = pPtl[i].x;
        if (minY > pPtl[i].y)
          minY = pPtl[i].y;
        if (maxY < pPtl[i].y)
          maxY = pPtl[i].y;
      }
      IRectangle bounding(minX, minY, maxX, maxY);
      if (penWidth() > 1)
        bounding.expandBy(penWidth());

      addToBoundingRect(bounding);
    }
    else if ( IPlatform::isWin32S() && isHitTesting())
    {
      if ( drawOperation() == IGraphicBundle::frame )
      {
        if (( pPtl[0].x != pPtl[arraySize-1].x ) ||
            ( pPtl[0].y != pPtl[arraySize-1].y ) )   //closed figure?
        {
           pPtl[arraySize] = pPtl[0];
           ++arraySize;
        }
        if ( fGCData->hitTestPointArray( *this, pPtl, arraySize ))
          ((IGPolygon&)polygon).setHitSelected(true);
      }
      else      //used for drawOp==fill || fillAndFrame
      {
        HRGN pathRegion = CreatePolygonRgn( pPtl, arraySize, fillMode );
        if ( fGCData->hitTestRegion( *this, pathRegion, true ))
          ((IGPolygon&)polygon).setHitSelected(true);
        DeleteObject( pathRegion );
      }
    }
    else
    {
      if ( drawOperation() == IGraphicBundle::frame )
        tempBrush = SelectObject( fPs, GetStockObject( HOLLOW_BRUSH ));
      else if (drawOperation() == IGraphicBundle::fill )
        tempPen   = SelectObject( fPs, GetStockObject( NULL_PEN ));

      SetPolyFillMode( fPs, fillMode );

      if (isHitTesting())
      {
        BeginPath( fPs );
      }

      COLORREF oldTextColor;
      Boolean usingCustomPattern((!tempBrush) &&
                                 IGCData::usingCustomFillPat( fillPattern()));

      if (usingCustomPattern)
        oldTextColor = SetTextColor( fPs, NEARESTPALETTECOLOR( fillColor().asRGBLong() ));

      if (!Polygon( fPs, pPtl, arraySize ))
      {
        ITHROWGUIERROR2("Polygon",
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
      }

      if (usingCustomPattern)
        SetTextColor( fPs, oldTextColor );

      if (isHitTesting())
      {
        EndPath( fPs );
        HRGN pathRegion = PathToRegion( fPs );
        if ( fGCData->hitTestRegion( *this, pathRegion ))
          ((IGPolygon&)polygon).setHitSelected(true);
        DeleteObject( pathRegion );
      }

      if ( drawOperation() == IGraphicBundle::frame )
        DeleteObject( SelectObject( fPs, tempBrush ));
      else if (drawOperation() == IGraphicBundle::fill )
        DeleteObject( SelectObject( fPs, tempPen ));
    }

    delete [] pPtl;

    remove( polygon, currentTransform, currentBundle );
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT. Only scaling and translation              |
| are supported in Windows95, and Win32s.                                      |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGEllipse& ellipse )
{
  IRectangle       enclosingRect(ellipse.enclosingRect());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  HBRUSH           tempBrush;
  HPEN             tempPen;

  apply( ellipse, currentTransform, currentBundle );

  if (fGCData->fRecoordinate        &&
      !isAccumulatingBoundingRect() &&
      !isHitTesting())
  {
    IPoint minXminY(enclosingRect.minXMinY());
    IPoint maxXmaxY(enclosingRect.maxXMaxY());
    minXminY.setY( fGCData->recoordHeight - minXminY.y() );
    maxXmaxY.setY( fGCData->recoordHeight - maxXmaxY.y() );
    enclosingRect = IRectangle( minXminY, maxXmaxY );
  }

  if (!fGCData->fAdvancedMode || // WIndows95, win32s, or boundary accumulation
      isAccumulatingBoundingRect())
  {
    IGCData::rectTransformedBy(enclosingRect, worldTransformMatrix( ));
  }

  if (isAccumulatingBoundingRect())
  {
    if (!fGCData->fAdvancedMode)
    {
      if (penWidth() > 1)
        enclosingRect.expandBy(penWidth());
      addToBoundingRect( enclosingRect );
    }
    else
    {
      IPointArray rectPts(4);
      IRectangle  rect(ellipse.enclosingRect());
      rectPts[0] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomLeft());
      rectPts[1] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomRight());
      rectPts[2] = IGCData::transformPoint(worldTransformMatrix( ), rect.topRight());
      rectPts[3] = IGCData::transformPoint(worldTransformMatrix( ), rect.topLeft());

      int maxX, minX, maxY, minY;
      maxX = minX = rectPts[0].x();
      maxY = minY = rectPts[1].y();

      for (int i = 1;i < 4; i++)
      {
        if (minX > rectPts[i].x())
          minX = rectPts[i].x();
        if (maxX < rectPts[i].x())
          maxX = rectPts[i].x();
        if (minY > rectPts[i].y())
          minY = rectPts[i].y();
        if (maxY < rectPts[i].y())
          maxY = rectPts[i].y();
      }
      IRectangle bounding(minX, minY, maxX, maxY);
      if (penWidth() > 1)
        bounding.expandBy(penWidth());

      addToBoundingRect(bounding);
    }
  }
  else
  {
    if ( drawOperation() == IGraphicBundle::frame )
      tempBrush = SelectObject( fPs, GetStockObject( HOLLOW_BRUSH ));
    else if (drawOperation() == IGraphicBundle::fill )
      tempPen   = SelectObject( fPs, GetStockObject( NULL_PEN ));

    if ( (isHitTesting())  && (fGCData->fAdvancedMode) ) //hit testing in NT
    {
      BeginPath( fPs );
    }

    if ( (isHitTesting()) && (!fGCData->fAdvancedMode) )  // hit testing in Windows95
    {
      ((IGEllipse&)ellipse).setHitSelected(
      fGCData->ellipseContains(*this, enclosingRect, ( drawOperation() != IGraphicBundle::frame )));

      if ( drawOperation() == IGraphicBundle::frame )
        DeleteObject( SelectObject( fPs, tempBrush ));
      remove( ellipse, currentTransform, currentBundle );
      return *this;
    }
    else
    {
      COLORREF oldTextColor;
      Boolean usingCustomPattern((!tempBrush) &&
                                 IGCData::usingCustomFillPat( fillPattern()));

      if (usingCustomPattern)
        oldTextColor = SetTextColor( fPs, NEARESTPALETTECOLOR( fillColor().asRGBLong() ));

      if (!IPlatform::isWin32S() ||
               (enclosingRect.maxY()-enclosingRect.minY() > 2 &&
                enclosingRect.maxX()-enclosingRect.minX() > 2 ))
      {
        if (!Ellipse( fPs, enclosingRect.minX(),
                           enclosingRect.maxY(),
                           enclosingRect.maxX(),
                           enclosingRect.minY() ))
        {
          ITHROWGUIERROR2("Ellipse",
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
        }
      }
      else  //Win32S, didn't meet minimum height and/or width
      {
        MoveToEx(fPs, enclosingRect.minX(), enclosingRect.minY(), 0);
        LineTo(fPs, enclosingRect.maxX(), enclosingRect.maxY());
      }
      if (usingCustomPattern)
        SetTextColor( fPs, oldTextColor );
    }

    if (isHitTesting())
    {
      EndPath( fPs );
      HRGN pathRegion = PathToRegion( fPs );
      if ( fGCData->hitTestRegion( *this, pathRegion ))
        ((IGEllipse&)ellipse).setHitSelected(true);
      DeleteObject( pathRegion );
    }

    if ( drawOperation() == IGraphicBundle::frame )
      DeleteObject( SelectObject( fPs, tempBrush ));
    else if (drawOperation() == IGraphicBundle::fill )
      DeleteObject( SelectObject( fPs, tempPen ));
  }

  remove( ellipse, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT, Windows95, and Win32s. When     |
| rotate, a round rectangle will become a regular rectangle in Windows95 and   |
| Win32s.                                                                      |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGRectangle& graphicRectangle )
{
  IRectangle       rect(graphicRectangle.enclosingRect());
  IPair            rounding(graphicRectangle.rounding());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  HBRUSH           tempBrush(0);
  HPEN             tempPen(0);
  POINTL           pPtl[5];     //allow a closed figure

  if (fGCData->fRecoordinate        &&
      !isAccumulatingBoundingRect() &&
      !isHitTesting())
  {
    IPoint minXminY(rect.minXMinY());
    IPoint maxXmaxY(rect.maxXMaxY());
    minXminY.setY( fGCData->recoordHeight - minXminY.y() );
    maxXmaxY.setY( fGCData->recoordHeight - maxXmaxY.y() );
    rect = IRectangle( minXminY, maxXmaxY );
  }

  apply( graphicRectangle, currentTransform, currentBundle );

  // Windows95, win32s. Prepare drawing a tilted rectangle, here actually a polygon.
  if (!fGCData->fAdvancedMode ||
      isAccumulatingBoundingRect())
  {
    if (fGCData->fTransformMatrix != ITransformMatrix())
    {
      IPoint newPt[4];
      newPt[0] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomLeft());
      newPt[1] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomRight());
      newPt[2] = IGCData::transformPoint(worldTransformMatrix( ), rect.topRight());
      newPt[3] = IGCData::transformPoint(worldTransformMatrix( ), rect.topLeft());

      for (int i=0; i<4; i++)
      {
        pPtl[i] = newPt[i].asPOINTL();
      }
    }
    else
    {
      pPtl[0] = rect.bottomLeft().asPOINTL();
      pPtl[1] = rect.bottomRight().asPOINTL();
      pPtl[2] = rect.topRight().asPOINTL();
      pPtl[3] = rect.topLeft().asPOINTL();
    }
  }

  if (isAccumulatingBoundingRect())
  {
    int maxX, minX, maxY, minY;
    if (fGCData->fAdvancedMode)
    {
      IPointArray rectPts(4);
      rectPts[0] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomLeft());
      rectPts[1] = IGCData::transformPoint(worldTransformMatrix( ), rect.bottomRight());
      rectPts[2] = IGCData::transformPoint(worldTransformMatrix( ), rect.topRight());
      rectPts[3] = IGCData::transformPoint(worldTransformMatrix( ), rect.topLeft());

      maxX = minX = rectPts[0].x();
      maxY = minY = rectPts[0].y();

      for (int i = 1;i < 4; i++)
      {
        if (minX > rectPts[i].x())
          minX = rectPts[i].x();
        if (maxX < rectPts[i].x())
          maxX = rectPts[i].x();
        if (minY > rectPts[i].y())
          minY = rectPts[i].y();
        if (maxY < rectPts[i].y())
          maxY = rectPts[i].y();
      }
    }
    else
    {
      maxX = minX = pPtl[0].x;
      maxY = minY = pPtl[0].y;
      for (int i = 1;i < 4; i++)
      {
        if (minX > pPtl[i].x)
          minX = pPtl[i].x;
        if (maxX < pPtl[i].x)
          maxX = pPtl[i].x;
        if (minY > pPtl[i].y)
          minY = pPtl[i].y;
        if (maxY < pPtl[i].y)
          maxY = pPtl[i].y;
      }
    }
    IRectangle bounding(minX, minY, maxX, maxY);
    if (penWidth() > 1)
      bounding.expandBy(penWidth());

    addToBoundingRect(bounding);
  }
  else if ( IPlatform::isWin32S() && isHitTesting())
  {
    if ( drawOperation() == IGraphicBundle::frame )
    {
      pPtl[4] = pPtl[0];        //make this a closed figure
      if ( fGCData->hitTestPointArray( *this, (POINT*)pPtl, 5 ))
        ((IGRectangle&)graphicRectangle).setHitSelected(true);
    }
    else      //used for drawOp==fill || fillAndFrame
    {
      HRGN rectRegion = CreatePolygonRgn( (POINT*)pPtl, 4, WINDING );
      if ( fGCData->hitTestRegion( *this, rectRegion, true ))
        ((IGRectangle&)graphicRectangle).setHitSelected(true);
      DeleteObject( rectRegion );
    }
  }
  else
  {
    if ( drawOperation() == IGraphicBundle::frame )
      tempBrush = SelectObject( fPs, GetStockObject( HOLLOW_BRUSH ));
    else if (drawOperation() == IGraphicBundle::fill )
      tempPen   = SelectObject( fPs, GetStockObject( NULL_PEN ));

    if (isHitTesting())
    {
      BeginPath( fPs );
    }

    if (rect.minX() == rect.maxX() || rect.minY() == rect.maxY())
    {
      if (fGCData->fAdvancedMode) // NT mode
      {
        MoveToEx(fPs, rect.minX(), rect.minY(), 0);
        LineTo( fPs, rect.maxX(), rect.maxY() );
      }
      else     // Windows95, win32s
      {
        MoveToEx(fPs, pPtl[0].x, pPtl[0].y, 0);
        LineTo( fPs, pPtl[2].x, pPtl[2].y );
      }
    }
    else
    {
      COLORREF oldTextColor;
      Boolean usingCustomPattern((!tempBrush) &&
                                 IGCData::usingCustomFillPat( fillPattern()));

      if (usingCustomPattern)
        oldTextColor = SetTextColor( fPs, NEARESTPALETTECOLOR( fillColor().asRGBLong() ));

      if (fGCData->fAdvancedMode) // NT mode
      {
        if (rounding.coord1() == 0 || rounding.coord2() == 0)
        {
          if (!Rectangle( fPs, rect.minX(), rect.minY(), rect.maxX(), rect.maxY()))
          {
            ITHROWGUIERROR2("Rectangle",
                            IBaseErrorInfo::invalidRequest,
                            IException::recoverable);
          }
        }
        else
        {
          if (!RoundRect( fPs,
                          rect.minX(), rect.minY(), rect.maxX(), rect.maxY(),
                          rounding.coord1(), rounding.coord2()))
          {
            ITHROWGUIERROR2("RoundRect",
                            IBaseErrorInfo::invalidRequest,
                            IException::recoverable);
          }
        }
      }  // end NT mode.
      else // Windows95, Win32s. Draw a polygon, a tilted rectangle.
      {
        if ( (!isHitTesting()) && (fGCData->notRotated(worldTransformMatrix()))  &&
             !(rounding.coord1() == 0 || rounding.coord2() == 0) )
        {
          if ( !IPlatform::isWin32S() ||
             ( pPtl[2].x - pPtl[0].x > 2 && pPtl[2].y - pPtl[0].y > 2 ))
          {
            if (!RoundRect( fPs,
                            pPtl[0].x, pPtl[0].y, pPtl[2].x, pPtl[2].y,
                            rounding.coord1(), rounding.coord2()))
            {
              ITHROWGUIERROR2("RoundRect",
                              IBaseErrorInfo::invalidRequest,
                              IException::recoverable);
            }
          }
          else  //Win32S, didn't meet minimum height and/or width
          {
            MoveToEx(fPs, pPtl[0].x, pPtl[0].y, 0);
            LineTo(fPs, pPtl[2].x, pPtl[2].y);
          }
        }
        else
        {
          if (!Polygon( fPs, (PPOINT)pPtl, 4 ))
          {
            ITHROWGUIERROR2("Rectangle",
                            IBaseErrorInfo::invalidRequest,
                            IException::recoverable);
          }
        }
      }

      if (usingCustomPattern)
        SetTextColor( fPs, oldTextColor );
    }

    if (isHitTesting())
    {
      if (rect.minX() == rect.maxX() || rect.minY() == rect.maxY())
      {
        if (fGCData->hitTestPath(*this))
          ((IGRectangle&)graphicRectangle).setHitSelected(true);
      }
      else
      {
        EndPath( fPs );
        HRGN pathRegion = PathToRegion( fPs );
        if ( fGCData->hitTestRegion( *this, pathRegion ))
          ((IGRectangle&)graphicRectangle).setHitSelected(true);
        DeleteObject( pathRegion );
      }
    }

    if ( drawOperation() == IGraphicBundle::frame )
      DeleteObject( SelectObject( fPs, tempBrush ));
    else if (drawOperation() == IGraphicBundle::fill )
      DeleteObject( SelectObject( fPs, tempPen ));
  }

  remove( graphicRectangle, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
| Support all graphical transformations in NT. Only scaling and translation    |
| are supported in Windows95, and Win32s.                                      |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGArc& arc )
{
  IRectangle       rect(arc.enclosingRect());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  Boolean          textMode;
  float            startAngle(arc.startAngle());
  float            sweepAngle(arc.sweepAngle());

  if (fGCData->fRecoordinate)
  {
    if (isAccumulatingBoundingRect() || isHitTesting())
    {
      textMode = false;
      float temp(startAngle);
      startAngle = sweepAngle;
      sweepAngle = temp;
    }
    else
    {
      IPoint minXminY(rect.minXMinY());
      IPoint maxXmaxY(rect.maxXMaxY());
      minXminY.setY( fGCData->recoordHeight - minXminY.y() );
      maxXmaxY.setY( fGCData->recoordHeight - maxXmaxY.y() );
      rect = IRectangle( minXminY, maxXmaxY );
      textMode = mappingMode() == IGraphicContext::pels;
    }
  }
  else
  {
    textMode = mappingMode() == IGraphicContext::pels;
  }

  apply( arc, currentTransform, currentBundle );

  int arcDirection(arc.direction() == IGArc::clockwise
                                      ? AD_CLOCKWISE
                                      : AD_COUNTERCLOCKWISE );

  IPoint firstRadial( IGCData::radialPoint( textMode, rect, startAngle ));
  IPoint secondRadial;
  // getting starting and ending points using the sweeping angle
  // and the drawing direction.
  if ( arcDirection == AD_CLOCKWISE )
  {
    secondRadial = IGCData::radialPoint( textMode,
                                         rect,
                                         startAngle - sweepAngle );
  }
  else
  {
    secondRadial = IGCData::radialPoint( textMode,
                                         rect,
                                         startAngle + sweepAngle );
  }

  if (!fGCData->fAdvancedMode) // WIndows95, win32s
  {
    // all double precision transform.
    IGCData::rectTransformedBy(rect, fGCData->fTransformMatrix);

    double tempX = firstRadial.x();
    double tempY = firstRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    firstRadial.setX(tempX);
    firstRadial.setY(tempY);

    tempX = secondRadial.x();
    tempY = secondRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    secondRadial.setX(tempX);
    secondRadial.setY(tempY);
  }

  // Accumulating Boundary Information
  if (isAccumulatingBoundingRect())
  {
    IRectangle bounding(
      fGCData->boundingRectangle(*this,
                                 arc.enclosingRect(),
                                 arc.startAngle(),
                                 arc.sweepAngle(),
                                 false,
                                 (fGCData->fRecoordinate ?
                                    !(arcDirection == AD_CLOCKWISE) :
                                    (arcDirection == AD_CLOCKWISE))));
    if (penWidth() > 1 )
      bounding.expandBy(penWidth());
    addToBoundingRect(bounding);
  }
  else
  {
    int gcArcDirection;
    if (fGCData->fAdvancedMode)  //  in  NT
    {
      gcArcDirection = SetArcDirection( fPs, arcDirection );
    }
    else
    {
      // Windows95 does not support SetArcDirection()
      // the drawing direction is always counterclockwise,
      // so in the case of closewise, firstRadial and secondRadial
      // has to be switched.
      if ( arcDirection == AD_CLOCKWISE )
      {
        IPoint tmp;
        tmp = firstRadial;
        firstRadial = secondRadial;
        secondRadial = tmp;
      }
    }

    if ( (isHitTesting()) && (fGCData->fAdvancedMode) ) // hit testing in  NT
    {
      BeginPath( fPs );
    }

    if ( (isHitTesting()) && (!fGCData->fAdvancedMode) )  // hit testing in Windows95
    {
      ((IGArc&)arc).setHitSelected(fGCData->arcContains(*this, arc, rect));

      remove( arc, currentTransform, currentBundle );
      return *this;
    }
    else
    {
      if ( !IPlatform::isWin32S() ||
         ( rect.maxX() - rect.minX() > 2 && rect.maxY() - rect.minY() > 2 ))
      {
        if (!Arc( fPs, rect.minX(), rect.minY(),
                  rect.maxX(), rect.maxY(),
                  firstRadial.x(), firstRadial.y(), secondRadial.x(), secondRadial.y() ))
        {
          ITHROWGUIERROR2("Arc",
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
        }
      }
      else  //Win32S, didn't meet minimum height and/or width
      {
        MoveToEx(fPs, rect.minX(), rect.minY(), 0);
        LineTo(fPs, rect.maxX(), rect.maxY());
      }
    }

    if (isHitTesting())
    {
      if (fGCData->hitTestPath(*this))
        ((IGArc&)arc).setHitSelected(true);
    }

    if (fGCData->fAdvancedMode)  //  in  NT
      SetArcDirection( fPs, gcArcDirection );
  }

  remove( arc, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
| Support all graphical transformations in NT. Only scaling and translation    |
| are supported in Windows95, and Win32s.                                      |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IG3PointArc& arc )
{
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  IPoint           start(arc.startingPoint());
  IPoint           inter(arc.intermediatePoint());
  IPoint           end(arc.endingPoint());

  if (fGCData->fRecoordinate        &&
      !isAccumulatingBoundingRect() &&
      !isHitTesting())
  {
    start.setY( fGCData->recoordHeight - start.y() );
    inter.setY( fGCData->recoordHeight - inter.y() );
    end.setY( fGCData->recoordHeight - end.y() );
  }
  double determinant(IGCData::determinant( (double)(inter.x() - start.x()),
                                           (double)(inter.y() - start.y()),
                                           (double)(end.x() - inter.x()),
                                           (double)(end.y() - inter.y()) ));

  POINTD arcPoints[4] = { start.x(), start.y(), inter.x(), inter.y(),
                          end.x(), end.y() };

  // Find the center of the arc.
  IGCData::arcCenterPoint( arcPoints );

  // Find the radius of the arc.
  int radius( (int)IGCData::distance( (PVECTOR2D)&arcPoints[3],
                                      (PVECTOR2D)&arcPoints[0] ));

  // Determine the bounding rectangle of the arc.
  IRectangle rect(IPoint(0,0), ISize( 2*radius, 2*radius ));
  rect.centerAt( IPoint((long)arcPoints[3].x, (long)arcPoints[3].y ));
  IRectangle enclosingRect = rect;

  apply( arc, currentTransform, currentBundle );

  if (isAccumulatingBoundingRect())
  {
    IPoint center = enclosingRect.center();
    // calculating the starting angle: angle1 and the ending angle: angle2
    // for the 3 point arc.
    double angle1 = atan2((start.y() - center.y()),
                          (start.x() - center.x()));

    //reverse the y-axial for Windows coordinate system.
    if ( mappingMode() == IGraphicContext::pels )
     angle1 = -angle1;

    // transfer radius to degree.
    angle1 *= (double)DEGREES_TO_RADS;
    if (angle1 <0)
      angle1 += 360;

    double angle2 = atan2((end.y() - center.y()),
                          (end.x() - center.x()));
    if ( mappingMode() == IGraphicContext::pels )
      angle2 = -angle2;
    angle2 *= (double)DEGREES_TO_RADS;
    if (angle2 <0)
      angle2 += 360;

    if ( determinant >= 0 )
    {
      double tmp = angle1;
      angle1 = angle2;
      angle2 = tmp;
    }

    IRectangle bounding(fGCData->boundingRectangle(*this, enclosingRect,
                       angle1, angle2-angle1, false, false) );
    if (penWidth() > 1)
      bounding.expandBy(penWidth());
    addToBoundingRect( bounding );
  }
  else
  {
    IPoint arcStart, arcEnd;

    if (fGCData->fAdvancedMode) // for NT
    {
      arcStart = start;
      arcEnd   = end;
    }
    else // Windows95, win32s
    {
      arcStart = IGCData::transformPoint(fGCData->fTransformMatrix, start);
      arcEnd   = IGCData::transformPoint(fGCData->fTransformMatrix, end);
      IGCData::rectTransformedBy(rect, fGCData->fTransformMatrix);
    }

    if ( (isHitTesting()) && (fGCData->fAdvancedMode) ) // hit testing in NT
    {
      BeginPath( fPs );
    }

    int gcArcDirection;
    if (fGCData->fAdvancedMode) // in NT
      gcArcDirection = GetArcDirection(fPs);

    if (fGCData->fAdvancedMode) // in NT
    {
      if (determinant < 0)
        SetArcDirection( fPs, AD_COUNTERCLOCKWISE );
      else
        SetArcDirection( fPs, AD_CLOCKWISE );
    }
    else
    {
      // Windows95 does not support SetArcDirection()
      // the drawing direction is always counterclockwise,
      // so in the case of closewise, firstRadial and secondRadial
      // has to be switched.
      if ( determinant >= 0 )
      {
        IPoint tmp = arcStart;
        arcStart = arcEnd;
        arcEnd = tmp;
      }
    }

    if ( (isHitTesting()) && (!fGCData->fAdvancedMode) )  // hit testing in Windows95
    {
      ((IGArc&)arc).setHitSelected(fGCData->arc3PointContains(*this, rect, arcStart, arcEnd));

      remove( arc, currentTransform, currentBundle );
      return *this;
    }
    else
    {
      if ( !IPlatform::isWin32S() ||
         ( rect.maxX() - rect.minX() > 2 && rect.maxY() - rect.minY() > 2 ))
      {
        if (!Arc( fPs, rect.minX(), rect.minY(),
                  rect.maxX(), rect.maxY(),
                  arcStart.x(),
                  arcStart.y(),
                  arcEnd.x(),
                  arcEnd.y() ))
        {
          ITHROWGUIERROR2("Arc",
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
        }
      }
      else  //Win32S, didn't meet minimum height and/or width
      {
        MoveToEx(fPs, rect.minX(), rect.minY(), 0);
        LineTo(fPs, rect.maxX(), rect.maxY());
      }
    }

    if (isHitTesting())
    {
      if (fGCData->hitTestPath(*this))
        ((IG3PointArc&)arc).setHitSelected(true);
    }

    if (fGCData->fAdvancedMode) // in NT
      SetArcDirection( fPs, gcArcDirection );
  }

  remove( arc, currentTransform, currentBundle );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
| Support all graphical transformations in NT. Only scaling and translation    |
| are supported in Windows95, and Win32s.                                      |
| Drawing direction is counterclockwise only.                                  |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGPie& pie )
{
  IRectangle       rect(pie.enclosingRect());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  HBRUSH           tempBrush(0);
  HPEN             tempPen(0);
  Boolean          textMode(true);
  float            startAngle(pie.startAngle());
  float            sweepAngle(pie.startAngle() + pie.sweepAngle());


  if (fGCData->fRecoordinate)
  {
    if (isAccumulatingBoundingRect() || isHitTesting())
    {
      textMode = false;
      float temp(startAngle);
      startAngle = sweepAngle;
      sweepAngle = temp;
    }
    else
    {
      IPoint minXminY(rect.minXMinY());
      IPoint maxXmaxY(rect.maxXMaxY());
      minXminY.setY( fGCData->recoordHeight - minXminY.y() );
      maxXmaxY.setY( fGCData->recoordHeight - maxXmaxY.y() );
      rect = IRectangle( minXminY, maxXmaxY );
      textMode = mappingMode() == IGraphicContext::pels;
    }
  }
  else
  {
    textMode = mappingMode() == IGraphicContext::pels;
  }

  apply( pie, currentTransform, currentBundle );

  IPoint firstRadial( IGCData::radialPoint( textMode, rect, startAngle ));
  IPoint secondRadial( IGCData::radialPoint( textMode, rect, sweepAngle ));

  if (!fGCData->fAdvancedMode) // WIndows95, win32s
  {
    // all double precision transform.
    IGCData::rectTransformedBy(rect, fGCData->fTransformMatrix);

    double tempX = firstRadial.x();
    double tempY = firstRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    firstRadial.setX(tempX);
    firstRadial.setY(tempY);

    tempX = secondRadial.x();
    tempY = secondRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    secondRadial.setX(tempX);
    secondRadial.setY(tempY);
  }

  if (isAccumulatingBoundingRect())
  {
    IRectangle bounding(
       fGCData->boundingRectangle(*this,
                                  pie.enclosingRect(),
                                  pie.startAngle(),
                                  pie.sweepAngle(),
                                  true,
                                  (fGCData->fRecoordinate ? true : false )));
    if (penWidth() > 1)
      bounding.expandBy(penWidth());
    addToBoundingRect( bounding );
  }
  else
  {
    if ( drawOperation() == IGraphicBundle::frame )
      tempBrush = SelectObject( fPs, GetStockObject( HOLLOW_BRUSH ));
    else if (drawOperation() == IGraphicBundle::fill )
      tempPen   = SelectObject( fPs, GetStockObject( NULL_PEN ));

    if ( (isHitTesting()) && (fGCData->fAdvancedMode) )  // hit testing in NT
    {
      BeginPath( fPs );
    }

    if ( (isHitTesting()) && (!fGCData->fAdvancedMode) )  // hit testing in Windows95
    {
      ((IGPie&)pie).setHitSelected(
      fGCData->pieContains(*this, pie, rect, (drawOperation() != IGraphicBundle::frame) ));
      if ( drawOperation() == IGraphicBundle::frame )
        DeleteObject( SelectObject( fPs, tempBrush ));
      else if (drawOperation() == IGraphicBundle::fill )
        DeleteObject( SelectObject( fPs, tempPen ));

      remove( pie, currentTransform, currentBundle );
      return *this;
    }
    else
    {
      if (!IPlatform::isWin32S())
        SetArcDirection(fPs, AD_COUNTERCLOCKWISE);      //Win3.1 always CCW

      COLORREF oldTextColor;
      Boolean usingCustomPattern((!tempBrush) &&
                                 IGCData::usingCustomFillPat( fillPattern()));

      if (usingCustomPattern)
        oldTextColor = SetTextColor( fPs, NEARESTPALETTECOLOR( fillColor().asRGBLong() ));

      if ( !IPlatform::isWin32S() ||
         ( rect.maxX() - rect.minX() > 2 && rect.maxY() - rect.minY() > 2 ))
      {
        if (!Pie( fPs,
                  rect.minX(), rect.minY(),
                  rect.maxX(), rect.maxY(),
                  firstRadial.x(), firstRadial.y(),
                  secondRadial.x(), secondRadial.y() ))
        {
          ITHROWGUIERROR2("Pie",
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable);
        }
      }
      else  //Win32S, didn't meet minimum height and/or width
      {
        MoveToEx(fPs, rect.minX(), rect.minY(), 0);
        LineTo(fPs, rect.maxX(), rect.maxY());
      }

      if (usingCustomPattern)
        SetTextColor( fPs, oldTextColor );
    }

    if (isHitTesting())
    {
      EndPath( fPs );
      HRGN pathRegion = PathToRegion( fPs );
      if ( fGCData->hitTestRegion( *this, pathRegion ))
        ((IGPie&)pie).setHitSelected(true);
      DeleteObject( pathRegion );
    }

    if ( drawOperation() == IGraphicBundle::frame )
      DeleteObject( SelectObject( fPs, tempBrush ));
    else if (drawOperation() == IGraphicBundle::fill )
      DeleteObject( SelectObject( fPs, tempPen ));
  }

  remove( pie, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
| Support all graphical transformations in NT. Only scaling and translation    |
| are supported in Windows95, and Win32s.                                      |
| Drawing direction is counterclockwise only.                                  |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGChord& chord )
{
  IRectangle       rect(chord.enclosingRect());
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;
  HBRUSH           tempBrush(0);
  HPEN             tempPen(0);
  Boolean          textMode;
  float            startAngle(chord.startAngle());
  float            sweepAngle(chord.startAngle() + chord.sweepAngle());

  if (fGCData->fRecoordinate)
  {
    if (isAccumulatingBoundingRect() || isHitTesting())
    {
      textMode = false;
      float temp(startAngle);
      startAngle = sweepAngle;
      sweepAngle = temp;
    }
    else
    {
      IPoint minXminY(rect.minXMinY());
      IPoint maxXmaxY(rect.maxXMaxY());
      minXminY.setY( fGCData->recoordHeight - minXminY.y() );
      maxXmaxY.setY( fGCData->recoordHeight - maxXmaxY.y() );
      rect = IRectangle( minXminY, maxXmaxY );
      textMode = mappingMode() == IGraphicContext::pels;
    }
  }
  else
  {
    textMode = mappingMode() == IGraphicContext::pels;
  }

  apply( chord, currentTransform, currentBundle );

  IPoint firstRadial( IGCData::radialPoint( textMode, rect, startAngle ));
  IPoint secondRadial( IGCData::radialPoint( textMode, rect, sweepAngle ));

  if (!fGCData->fAdvancedMode) // WIndows95, win32s
  {
    // all double precision transform.
    IGCData::rectTransformedBy(rect, fGCData->fTransformMatrix);

    double tempX = firstRadial.x();
    double tempY = firstRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    firstRadial.setX(tempX);
    firstRadial.setY(tempY);

    tempX = secondRadial.x();
    tempY = secondRadial.y();
    IGCData::transformPoint(fGCData->fTransformMatrix, tempX, tempY);
    secondRadial.setX(tempX);
    secondRadial.setY(tempY);
  }

  if (isAccumulatingBoundingRect())
  {
    IRectangle bounding(
      fGCData->boundingRectangle(*this,
                                 chord.enclosingRect(),
                                 chord.startAngle(),
                                 chord.sweepAngle(),
                                 false,
                                 (fGCData->fRecoordinate ? true : false)));
    if (penWidth() > 1)
      bounding.expandBy(penWidth());
    addToBoundingRect( bounding );
  }
  else
  {
    if ( drawOperation() == IGraphicBundle::frame )
      tempBrush = SelectObject( fPs, GetStockObject( HOLLOW_BRUSH ));
    else if (drawOperation() == IGraphicBundle::fill )
      tempPen   = SelectObject( fPs, GetStockObject( NULL_PEN ));

    if (isHitTesting()  && (fGCData->fAdvancedMode) )  // hit testing in NT
    {
      BeginPath( fPs );
    }

    if ( (isHitTesting()) && (!fGCData->fAdvancedMode) )  // hit testing in Windows95
    {
      ((IGChord&)chord).setHitSelected(
      fGCData->chordContains(*this, chord, rect, (drawOperation() != IGraphicBundle::frame) ));
      if ( drawOperation() == IGraphicBundle::frame )
        DeleteObject( SelectObject( fPs, tempBrush ));
      else if (drawOperation() == IGraphicBundle::fill )
        DeleteObject( SelectObject( fPs, tempPen ));

      remove( chord, currentTransform, currentBundle );
      return *this;
    }
    else
    {
      if (!IPlatform::isWin32S())
        SetArcDirection(fPs, AD_COUNTERCLOCKWISE);      //Win3.1 always CCW

      COLORREF oldTextColor;
      Boolean usingCustomPattern((!tempBrush) &&
                                 IGCData::usingCustomFillPat( fillPattern()));

      if (usingCustomPattern)
        oldTextColor = SetTextColor( fPs, NEARESTPALETTECOLOR( fillColor().asRGBLong() ));

      if ( !IPlatform::isWin32S() ||
         ( rect.maxX() - rect.minX() > 2 && rect.maxY() - rect.minY() > 2 ))
      {
        if (!Chord( fPs,
                    rect.minX(), rect.minY(),
                    rect.maxX(), rect.maxY(),
                    firstRadial.x(), firstRadial.y(),
                    secondRadial.x(), secondRadial.y() ))
        {
          ITHROWGUIERROR2("Pie",
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable);
        }
      }
      else  //Win32S, didn't meet minimum height and/or width
      {
        MoveToEx(fPs, rect.minX(), rect.minY(), 0);
        LineTo(fPs, rect.maxX(), rect.maxY());
      }
      if (usingCustomPattern)
        SetTextColor( fPs, oldTextColor );
    }

    if (isHitTesting())
    {
      EndPath( fPs );
      HRGN pathRegion = PathToRegion( fPs );
      if ( fGCData->hitTestRegion( *this, pathRegion ))
        ((IGChord&)chord).setHitSelected(true);
      DeleteObject( pathRegion );
    }

    if ( drawOperation() == IGraphicBundle::frame )
      DeleteObject( SelectObject( fPs, tempBrush ));
    else if (drawOperation() == IGraphicBundle::fill )
      DeleteObject( SelectObject( fPs, tempPen ));
  }

  remove( chord, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
| Support all graphical transformations in NT,                                 |
| Only scaling and translation are supported in Windows95, and Win32s.         |
| Drawing and hit testing is not supported when IGString using a scaled bitmap |
| font                                                                         |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw ( const IGString& graphicString )
{
  HFONT            hFont(0), oldHFont(0);
  ITransformMatrix currentTransform;
  IGraphicBundle   currentBundle;
  SIZE             stringSize;
  RECTL            clipRect;
  unsigned         clipText(0);
  IPoint           refPt(graphicString.position());
  Boolean          isBitmap(false);
  Boolean          createdNewFont(false);

  if (fGCData->fRecoordinate      &&
      !isAccumulatingBoundingRect() &&
      !isHitTesting())
  {
    refPt.setY( fGCData->recoordHeight - refPt.y() - graphicString.font().maxCharHeight());
  }

  apply( graphicString, currentTransform, currentBundle );

  if (graphicString.hasFont())
  {
    isBitmap = graphicString.font().isBitmap();
  }
  else
  {
    TEXTMETRIC textMetric;
    GetTextMetrics( fPs, &textMetric );
    isBitmap = !(textMetric.tmPitchAndFamily & TMPF_VECTOR);
  }

  // System font should take this path as well.  Other bitmap
  // fonts appear to be scalable on NT.
  if (!fGCData->fAdvancedMode                     &&
      fGCData->notRotated(worldTransformMatrix()) &&
      graphicString.hasTransformMatrix())
  {
    createdNewFont = true;
    // ************ Directly using Windows API for scaled font **************  .
    const tagLOGFONTA* lf;
    if (graphicString.hasFont())
      lf = graphicString.font().logfont();
    else
      lf = currentFont().logfont();

    unsigned long height = lf->lfHeight * fGCData->fTransformMatrix.element22();
    unsigned long width = lf->lfWidth * fGCData->fTransformMatrix.element11();

    // create font
    hFont = CreateFont(height, width,lf->lfEscapement, lf->lfOrientation,
                       lf->lfWeight, lf->lfItalic, lf->lfUnderline, lf->lfStrikeOut,
                       lf->lfCharSet, lf->lfOutPrecision, lf->lfClipPrecision,
                       lf->lfQuality, lf->lfPitchAndFamily, lf->lfFaceName);
    oldHFont = SelectObject(fPs, hFont);
    // **********************************************************************
  }
  else
  {
    if (graphicString.hasFont())
    {
      graphicString.font().beginUsingFont(fPs);
    }
  }

  //currDrawingFont is the font used for drawing now!
  if (isHitTesting() || isAccumulatingBoundingRect())
  {
    ITransformMatrix currentWorld(worldTransformMatrix());
    Boolean          isIdentity(currentWorld==ITransformMatrix());

    //hit testing using region for bitmap font is not supported.
    //and have to be done in different way using IGRectangle's hit
    //testing capability.  If an attempt is made to draw the string
    //using a bitmap font inside a path bracket, the text out function
    //will fail on W95 and NT.  If the bitmap font is rotated,
    //then the PathToRgn function works.
    if (IPlatform::isWin32S() ||
       (isBitmap && fGCData->notRotated(currentWorld)) ||
       (isIdentity))
    {
      if (IPlatform::isWin32S())
      {
        GetTextExtentPoint(fPs, (LPSTR)graphicString.text(),
                                  graphicString.text().length(), &stringSize);
      }
      else
      {
        GetTextExtentPoint32(fPs, (LPSTR)graphicString.text(),
                                  graphicString.text().length(), &stringSize);
      }

      IPoint cornerPt;

      cornerPt.setX(refPt.x() + stringSize.cx);
      cornerPt.setY(refPt.y() + stringSize.cy);
      UINT alignment = GetTextAlign(fPs);
      if ( ((alignment & TA_BASELINE) == TA_BASELINE) ||
           ((alignment & TA_BOTTOM) == TA_BOTTOM) )
      {
        cornerPt.setY(cornerPt.y() - 2*stringSize.cy);
      }

      if ((alignment & TA_CENTER) == TA_CENTER)
      {
        cornerPt.setX(cornerPt.x() - stringSize.cx/2);
        refPt.setX(refPt.x() - stringSize.cx/2);
      }
      else if ((alignment & TA_RIGHT) == TA_RIGHT)
      {
        cornerPt.setX(cornerPt.x() - 2*stringSize.cx);
      }

      IRectangle bounding(refPt, cornerPt);

      if (isAccumulatingBoundingRect())
      {
        if (createdNewFont)
        {
          refPt = IGCData::transformPoint(currentWorld, refPt );
          bounding.moveTo( refPt );
          addToBoundingRect( bounding );
        }
        else
        {
          if (isIdentity)
          {
            addToBoundingRect( bounding );
          }
          else
          {
            IGRectangle tempRect(bounding);
            addToBoundingRect( tempRect.boundingRect( *this ));
          }
        }
      }
      else //isHitTesting
      {
        // Create a temporary IGRectangle with the same dimensions
        // of the bitmap.
        IGraphicBundle bundle;
        bundle.setDrawOperation( IGraphicBundle::fill );

        IGRectangle tempRect(bounding);
        tempRect.setGraphicBundle( bundle );
        tempRect.drawOn( *this );
        if (tempRect.isHitSelected())
          ((IGString&)graphicString).setHitSelected( true );
      }

      // NT or Window95, win32s without scaling font.
      if (oldHFont)
      {
        SelectObject(fPs, oldHFont);
        DeleteObject( hFont );
      }
      else // Window95, win32s with scaling   font.
      {
        if (graphicString.hasFont())
        {
          graphicString.font().endUsingFont(fPs);
        }
      }

      remove( graphicString, currentTransform, currentBundle );

      return *this;
    }
    else
    {
      if (!BeginPath( fPs ))
        ITHROWGUIERROR("BeginPath");
    }
  }

  if (!fGCData->fAdvancedMode)   // for Windows95, win32s
  {
    refPt = IGCData::transformPoint(fGCData->fTransformMatrix, refPt);
  }

  // The graphics strings clipping rect should not be set when
  // doing hit testing or boundary accumulation.  The PathToRegion
  // returns a null region if this flag is set.
  if (graphicString.isClippingRectSet() &&
      !(isHitTesting() || isAccumulatingBoundingRect()))
  {
    // Note: It appears that setting the ETO_CLIPPED flag will cause
    //       a small GDI memory leak.
    IRectangle temp(graphicString.clippingRect().asRECTL());
    clipText = ETO_CLIPPED;
    if (fGCData->fRecoordinate)
    {
      temp.moveBy(IPair(0, fGCData->recoordHeight - temp.height()));
    }
    clipRect = temp.asRECTL();
  }

  if (!ExtTextOut( fPs,
                   refPt.x(),
                   refPt.y(),
                   clipText,
                   (RECT*)&clipRect,
                   (LPSTR)graphicString.text(),
                   graphicString.text().length(),
                   0 ))
  {
    ITHROWGUIERROR2("TextOut",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }

  if (isHitTesting() || isAccumulatingBoundingRect())
  {
    EndPath( fPs );
    HRGN pathRegion = PathToRegion( fPs );
    if (!pathRegion)
      ITHROWGUIERROR("PathToRegion");
    if (isAccumulatingBoundingRect())
    {
      RECT regionRect;
      if (GetRgnBox( pathRegion, &regionRect ) != 0)
      {
        addToBoundingRect( IRectangle( regionRect.left,
                                       regionRect.top,
                                       regionRect.right,
                                       regionRect.bottom ));
      }
      else
      {
        ITHROWGUIERROR("GetRgnBox");
      }
    }
    else //isHitTesting()==true
    {
      if ( fGCData->hitTestRegion( *this, pathRegion ))
        ((IGString&)graphicString).setHitSelected(true);
    }
    DeleteObject( pathRegion );
  }

  if (oldHFont)
  {
    SelectObject(fPs, oldHFont);
    DeleteObject( hFont );
  }
  else
  {
    if (graphicString.hasFont())
    {
      graphicString.font().endUsingFont(fPs);
    }
  }

  remove( graphicString, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw( const IGList& list )
{
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;

  apply( list, currentTransform, currentBundle );

  IGList::Cursor cursor( list );

  for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
    IGraphic& graphic(list.graphicAt(cursor));
    graphic.setHitSelected(false);
    graphic.drawOn(*this);
    if (graphic.isHitSelected() && list.isHitSelectable())
      ((IGList&)list).setHitSelected(true);
  }

  remove( list, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::draw                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::draw( const IGRegion& region )
{
  IGraphicBundle   currentBundle;
  ITransformMatrix currentTransform;

  apply( region, currentTransform, currentBundle );

  if (isAccumulatingBoundingRect())
  {
    RECT regionRect;
    if (GetRgnBox((IRegionHandle)region, &regionRect) != 0)
    {
      addToBoundingRect( IRectangle( regionRect.left,
                                     regionRect.top,
                                     regionRect.right,
                                     regionRect.bottom ));
    }
  }
  else if (isHitTesting())
  {
    if ( fGCData->hitTestRegion( *this, region, true ))
      ((IGRegion&)region).setHitSelected(true);
  }
  else
  {
    unsigned long regionHeight(0);
    RECTL         regionRect;

    // Query the region rect.  If it is empty, then do not paint
    // the region.  For some reason, painting an empty region sends
    // W95 into a tail spin.
    GetRgnBox((IRegionHandle)region, (RECT*)&regionRect);

    IRectangle tempRect( regionRect );

    if (tempRect.area() != 0)
    {
      // Set the fill color and fill pattern to the pen color
      // and pen pattern.  This is how regions are painted in
      // OS/2 version of Open Class.
      IColor fillColor(this->fillColor());
      unsigned long fillPattern(this->fillPattern());

      fGCData->deferCreation = true;
      this->setFillColor( this->penColor() );
      fGCData->deferCreation = false;
      this->setFillPattern( this->penPattern() );


      // If recoordinating, move the region.
      if (fGCData->fRecoordinate)
      {
        regionHeight = regionRect.bottom - regionRect.top;
        OffsetRgn( (IRegionHandle)region, 0, fGCData->recoordHeight - regionHeight );
      }

      if (!PaintRgn( fPs, (IRegionHandle)region ))
      {
        ITHROWGUIERROR2("PaintRgn",
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
      }

      // Now move the region back to its original location
      if (fGCData->fRecoordinate)
      {
        OffsetRgn( (IRegionHandle)region, 0, regionHeight - fGCData->recoordHeight );
      }

      fGCData->deferCreation = true;
      this->setFillColor( fillColor );
      fGCData->deferCreation = false;
      this->setFillPattern( fillPattern );
    }
  }
  remove( region, currentTransform, currentBundle );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setRecoordinationHeight                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setRecoordinationHeight( unsigned long height )
{
  fGCData->recoordHeight = height;
  if (!ICoordinateSystem::isConversionNeeded())
  {
    fGCData->fRecoordinate = false;
  }
  else
  {
    fGCData->fRecoordinate = true;
    SetTextAlign( fPs, TA_TOP | TA_LEFT );
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::recoordinationHeight                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::recoordinationHeight( ) const
{
  return fGCData->recoordHeight;
}

/*------------------------------------------------------------------------------
| IGraphicContext::handle                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPresSpaceHandle IGraphicContext::handle( ) const
{
  return fPs;
}


/*------------------------------------------------------------------------------
| IGraphicContext::currentAttributes                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle IGraphicContext::graphicBundle( ) const
{
  IGraphicBundle bundle(*this);
  return bundle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setGraphicBundle                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setGraphicBundle(const IGraphicBundle& bundle)
{
  fGCData->deferCreation = true;

  if (bundle.hasPenColor())
    setPenColor(bundle.penColor());
  if (bundle.hasFillColor())
    setFillColor(bundle.fillColor());
  if (bundle.hasBackgroundColor())
    setBackgroundColor(bundle.backgroundColor());
  if (bundle.hasDrawOperation())
    setDrawOperation(bundle.drawOperation());
  if (bundle.hasMixMode())
    setMixMode(bundle.mixMode());
  if (bundle.hasBackgroundMixMode())
    setBackgroundMixMode(bundle.backgroundMixMode());
  if (bundle.hasPenType())
    setPenType(bundle.penType());
  if (bundle.hasPenEndingStyle())
    setPenEndingStyle(penEndingStyle());
  if (bundle.hasPenJoiningStyle())
    setPenJoiningStyle(bundle.penJoiningStyle());
  if (bundle.hasPenWidth())
    setPenWidth(bundle.penWidth());
  if (bundle.hasPenPattern())
    setPenPattern(bundle.penPattern());
  if (bundle.hasFillPattern())
    setFillPattern(bundle.fillPattern());
  if (bundle.hasPatternOrigin())
    setPatternOrigin(bundle.patternOrigin());

  fGCData->deferCreation = false;
  fGCData->createPen( fPs );
  fGCData->createBrush( fPs );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setCurrentDrawingPosition                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setCurrentDrawingPosition( const IPoint& point )
{
  if (!MoveToEx(fPs, point.x(), point.y(), 0))
  {
    ITHROWGUIERROR2("MoveToEx",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::currentDrawingPosition                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGraphicContext::currentDrawingPosition( ) const
{
  POINT pt;
  MoveToEx( fPs, 0, 0, &pt );
  MoveToEx( fPs, pt.x, pt.y, 0 );
  return IPoint( pt );
}

/*------------------------------------------------------------------------------
| IGraphicContext::setFont                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setFont( const IFont& font )
{
  if ( fFont )
  {
    fFont->endUsingFont( fPs );
    delete fFont;
  }
  fFont = new IFont( font );

  fFont->beginUsingFont( fPs );

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDefaultFont                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::resetFont( )
{
  if ( fFont )
  {
    fFont->endUsingFont( fPs );
    delete fFont;
    fFont = 0;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::currentFont                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont IGraphicContext::currentFont( ) const
{
  if ( fFont )
  {
    return IFont( *fFont );
  }
  else
  {
    return IFont( fPs );
  }
}

/*------------------------------------------------------------------------------
| IGraphicContext::hasFont                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IGraphicContext::hasFont( ) const
{
//  HFONT systemFont = GetStockObject( SYSTEM_FONT );
//  HFONT currentFont = SelectObject( fPs, systemFont );

  TEXTMETRIC textmetric;
  GetTextMetrics( fPs, (LPTEXTMETRIC)&textmetric );
  if ( textmetric.tmCharSet == DEFAULT_CHARSET )
    return false;
  else
    return true;
}

/*------------------------------------------------------------------------------
| IGraphicContext::drawOperation                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::DrawOperation IGraphicContext::drawOperation() const
{
  return fGCData->fDrawOperation;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setDrawOperation                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setDrawOperation(
                                          IGraphicBundle::DrawOperation drawOperation )
{
  fGCData->fDrawOperation = drawOperation;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penColor                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::penColor() const
{
  return fGCData->fPenColor;
}

/*------------------------------------------------------------------------------
| IGraphicContext::backgroundColor                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::backgroundColor() const
{
  COLORREF background(GetBkColor(fPs));
  if (background == CLR_INVALID)
  {
    ITHROWGUIERROR2("GetBkColor",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return IColor(GetRValue(background),
                GetGValue(background),
                GetBValue(background));
}

/*------------------------------------------------------------------------------
| IGraphicContext::fillColor                                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IColor IGraphicContext::fillColor() const
{
  return fGCData->fFillColor;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setFillColor                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setFillColor( const IColor& color )
{
  fGCData->fFillColor = color;
  fGCData->createBrush( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenColor                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenColor( const IColor& color )
{
  fGCData->fPenColor = color;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setBackgroundColor                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setBackgroundColor( const IColor& background )
{
  SetBkColor( fPs, NEARESTPALETTECOLOR(background.asRGBLong()));
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setMixMode                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setMixMode(IGraphicBundle::MixMode mixMode)
{
  int mix;
  switch (mixMode)
  {
    case IGraphicBundle::or:
      mix = R2_MERGEPEN;
    break;
    case IGraphicBundle::overPaint:
      mix = R2_COPYPEN;
    break;
    case IGraphicBundle::xor:
      mix = R2_XORPEN;
    break;
    case IGraphicBundle::leaveAlone:
      mix = R2_NOP;
    break;
    case IGraphicBundle::and:
      mix = R2_MASKPEN;
    break;
    case IGraphicBundle::subtract:
      mix = R2_MASKNOTPEN;
    break;
    case IGraphicBundle::maskSourceNot:
      mix = R2_MERGEPENNOT;
    break;
    case IGraphicBundle::black:
      mix = R2_BLACK;
    break;
    case IGraphicBundle::notMergeSource:
      mix = R2_NOTMERGEPEN;
    break;
    case IGraphicBundle::notXorSource:
      mix = R2_NOTXORPEN;
    break;
    case IGraphicBundle::invert:
      mix = R2_NOT;
    break;
    case IGraphicBundle::mergeSourceNot:
      mix = R2_MASKPENNOT;
    break;
    case IGraphicBundle::notCopySource:
      mix = R2_NOTCOPYPEN;
    break;
    case IGraphicBundle::mergeNotSource:
      mix = R2_MERGENOTPEN;
    break;
    case IGraphicBundle::notMaskSource:
      mix = R2_NOTMASKPEN;
    break;
    case IGraphicBundle::white:
      mix = R2_WHITE;
    break;
  } /* endswitch */
  if (!SetROP2( fPs, mix ))
  {
    ITHROWGUIERROR2("SetROP2",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setBackgroundMixMode                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setBackgroundMixMode(
                            IGraphicBundle::BackgroundMixMode backgroundMixMode)
{
  int backgroundMode;
  switch (backgroundMixMode)
  {
    case IGraphicBundle::backOverPaint:
      backgroundMode = OPAQUE;
    break;
    case IGraphicBundle::backLeaveAlone:
    case IGraphicBundle::backOr:
    case IGraphicBundle::backXor:
      backgroundMode = TRANSPARENT;
    break;
  } /* endswitch */
  if (!SetBkMode( fPs, backgroundMode ))
  {
    ITHROWGUIERROR2("SetBkMode",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::mixMode                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::MixMode IGraphicContext::mixMode( ) const
{
  IGraphicBundle::MixMode mixMode;
  int mix(GetROP2( fPs ));
  if (!mix)
  {
    ITHROWGUIERROR2("GetROP2",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  switch ( mix )
  {
    case R2_MERGEPEN:
      mixMode = IGraphicBundle::or;
    break;
    case R2_COPYPEN:
      mixMode = IGraphicBundle::overPaint;
    break;
    case R2_XORPEN:
      mixMode = IGraphicBundle::xor;
    break;
    case R2_NOP:
      mixMode = IGraphicBundle::leaveAlone;
    break;
    case R2_MASKPEN:
      mixMode = IGraphicBundle::and;
    break;
    case R2_MASKNOTPEN:
      mixMode = IGraphicBundle::subtract;
    break;
    case R2_MERGEPENNOT:
      mixMode = IGraphicBundle::maskSourceNot;
    break;
    case R2_BLACK:
      mixMode = IGraphicBundle::black;
    break;
    case R2_NOTMERGEPEN:
      mixMode = IGraphicBundle::notMergeSource;
    break;
    case R2_NOTXORPEN:
      mixMode = IGraphicBundle::notXorSource;
    break;
    case R2_NOT:
      mixMode = IGraphicBundle::invert;
    break;
    case R2_MASKPENNOT:
      mixMode = IGraphicBundle::mergeSourceNot;
    break;
    case R2_NOTCOPYPEN:
      mixMode = IGraphicBundle::notCopySource;
    break;
    case R2_MERGENOTPEN:
      mixMode = IGraphicBundle::mergeNotSource;
    break;
    case R2_NOTMASKPEN:
      mixMode = IGraphicBundle::notMaskSource;
    break;
    case R2_WHITE:
      mixMode = IGraphicBundle::white;
    break;
  } /* endswitch */
  return mixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::backgroundMixMode                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::BackgroundMixMode IGraphicContext::backgroundMixMode( ) const
{
  IGraphicBundle::BackgroundMixMode backgroundMixMode;
  int backgroundMode(GetBkMode(fPs));

  switch (backgroundMode)
  {
    case OPAQUE:
      backgroundMixMode = IGraphicBundle::backOverPaint;
    break;
    case TRANSPARENT:
    default:
      backgroundMixMode = IGraphicBundle::backLeaveAlone;
    break;
  }
  return backgroundMixMode;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penPattern                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::penPattern( ) const
{
  return fGCData->fPenPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::fillPattern                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::fillPattern( ) const
{
  return fGCData->fFillPattern;
}

/*------------------------------------------------------------------------------
| IGraphicContext::patternOrigin                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGraphicContext::patternOrigin( ) const
{
  POINT pt;

  if (!GetBrushOrgEx( fPs, &pt ))
  {
    ITHROWGUIERROR2("GetBrushOrgEx",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return IPoint( pt );
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenType                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenType(IGraphicBundle::PenType penType)
{
  fGCData->fPenType = penType;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penType                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenType IGraphicContext::penType( ) const
{
  return fGCData->fPenType;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenWidth                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenWidth(unsigned long width )
{
  fGCData->fPenWidth = width;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penWidth                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IGraphicContext::penWidth( ) const
{
  return fGCData->fPenWidth;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenEndingStyle                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenEndingStyle(
                                 IGraphicBundle::PenEndingStyle penEndingStyle )
{
  fGCData->fPenEndingStyle = penEndingStyle;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penEndingStyle                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenEndingStyle IGraphicContext::penEndingStyle( ) const
{
  return fGCData->fPenEndingStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenJoiningStyle                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenJoiningStyle(
                                IGraphicBundle::PenJoiningStyle penJoiningStyle)
{
  fGCData->fPenJoiningStyle = penJoiningStyle;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::penJoin                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicBundle::PenJoiningStyle IGraphicContext::penJoiningStyle( ) const
{
  return fGCData->fPenJoiningStyle;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPatternOrigin                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPatternOrigin( const IPoint& point )
{
  SetBrushOrgEx( fPs, point.x(), point.y(), 0 );
  // Do not check for error condition because this function returns
  // false when called from a device context created with CreateMetaFile.
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setPenPattern                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setPenPattern( unsigned long penPattern )
{
  fGCData->fPenPattern = penPattern;
  fGCData->createPen( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setFillPattern                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setFillPattern( unsigned long fillPattern )
{
  fGCData->fFillPattern = fillPattern;
  fGCData->createBrush( fPs );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setClipRegion                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setClipRegion( const IRegionHandle& region )
{
  int rc(SelectClipRgn( fPs, region ));
  if (rc == ERROR )
  {
    ITHROWGUIERROR2("SelectClipRgn",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setClipRegion                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setClipRegion( const IGRegion& region )
{
  setClipRegion( (IRegionHandle)region );
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::clearClipRegion                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::clearClipRegion( )
{
  int rc(SelectClipRgn( fPs, 0 ));
  if (rc == ERROR )
  {
    ITHROWGUIERROR2("SelectClipRgn",
                    IBaseErrorInfo::invalidRequest,
                    IException::recoverable);
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IGraphicContext::clipRegion                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IRegionHandle IGraphicContext::clipRegion( ) const
{
  HRGN clipRegion;
  if (IPlatform::isWin32S())
     {
     ITHROWLIBRARYERROR( IC_MEMBER_FUN_NOT_SUPPORTED_ON_WIN32S,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
     }
  else
     {
     clipRegion = CreateRectRgn(0,0,0,0);
     int rc(GetClipRgn( fPs, clipRegion ));
     if (rc == -1)
        {
        ITHROWGUIERROR2("GetClipRgn",
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
        }
     else if (rc == 0)
        {
        DeleteObject( clipRegion );
        clipRegion = 0;
        }
     }
  return IRegionHandle( clipRegion );
}

/*------------------------------------------------------------------------------
| IGraphicContext::setWorldTransformMatrix                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setWorldTransformMatrix(
                                       const ITransformMatrix& transformMatrix,
                                       IGraphic::TransformMethod tm )
{
  XFORM mat(transformMatrix.asTagXFORM());

  if (fGCData->fAdvancedMode)   // NT mode
  {
    if ( tm == IGraphic::replace )
    {
      SetWorldTransform( fPs, &mat );
    }
    else if ( tm == IGraphic::rightMultiply )
    {
      ModifyWorldTransform( fPs, &mat, MWT_RIGHTMULTIPLY );
    }
    else
    {
      ModifyWorldTransform( fPs, &mat, MWT_LEFTMULTIPLY );
    }
  }
  else   // for Windows95 and win32s
  {
    fGCData->setWorldTransform(transformMatrix, tm);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::worldTransformMatrix                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ITransformMatrix IGraphicContext::worldTransformMatrix( ) const
{
  if (fGCData->fAdvancedMode)   // NT mode
  {
    XFORM mat;
    if (!GetWorldTransform(fPs, &mat))
    {
      return ITransformMatrix();
    }
    return ITransformMatrix( mat );
  }
  else   // for Windows95 and win32s
  {
    return (fGCData->fTransformMatrix);
  }
}

/*------------------------------------------------------------------------------
| IGraphicContext::startBoundaryAccumulation                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::startBoundaryAccumulation ( )
{
  fGCData->drawingMode = IGCData::accumBoundary;
  fGCData->fBoundingRect = IRectangle(0,0,0,0);
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::stopBoundaryAccumulation                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGraphicContext::stopBoundaryAccumulation ( )
{
  fGCData->drawingMode = IGCData::drawing;
  return fGCData->fBoundingRect;
}

/*------------------------------------------------------------------------------
| IGraphicContext::addToBoundingRect                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::addToBoundingRect( const IRectangle& rectangle )
{
  if (fGCData->fBoundingRect == IRectangle(0,0,0,0))
    fGCData->fBoundingRect = rectangle;
  else
    fGCData->fBoundingRect |= rectangle;

  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::isAccumulatingBoundingRect                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphicContext::isAccumulatingBoundingRect( ) const
{
  return (IBase::Boolean)(fGCData->drawingMode == IGCData::accumBoundary);
}

/*------------------------------------------------------------------------------
| IGraphicContext::startHitTesting                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::startHitTesting( )
{
  fGCData->drawingMode = IGCData::hitTesting;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::stopHitTesting                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::stopHitTesting( )
{
  fGCData->drawingMode = IGCData::drawing;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setHitPoint                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setHitPoint( const IPoint& point )
{
  fGCData->hitPoint = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::setHitApertureSize                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::setHitApertureSize( const ISize& hitSize )
{
  fGCData->hitSize = hitSize;
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::isHitTesting                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGraphicContext::isHitTesting( ) const
{
  return (IBase::Boolean)(fGCData->drawingMode == IGCData::hitTesting);
}

/*------------------------------------------------------------------------------
| IGraphicContext::hitPoint                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGraphicContext::hitPoint( ) const
{
  return fGCData->hitPoint;
}

/*------------------------------------------------------------------------------
| IGraphicContext::hitApertureSize                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
ISize IGraphicContext::hitApertureSize( ) const
{
  return fGCData->hitSize;
}

/*------------------------------------------------------------------------------
| IGraphicContext::apply                                                       |
| Preserve the current ITransformMatrix and IGraphicBundle of IGraphicContext, |
| apply the ITransformMatrix and IGraphicBundle of graphic to IGraphicContext. |
| If the graphic already has an ITransformMatrix and IGraphicBundle,           |
| make a temporary copy to *transformMatrix and *graphicBundle for drawing.    |
|-----------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::apply( const IGraphic&    graphic,
                                         ITransformMatrix&  transformMatrix,
                                         IGraphicBundle&    graphicBundle )
{
  if (graphic.hasTransformMatrix() |
      graphic.transformMethod() == IGraphic::replace)
  {
    transformMatrix = worldTransformMatrix();
    setWorldTransformMatrix(graphic.transformMatrix(),
                            graphic.transformMethod());
  }
  if (graphic.hasGraphicBundle())
  {
    graphicBundle = this->graphicBundle();
    setGraphicBundle( graphic.graphicBundle() );
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGraphicContext::remove                                                      |
| Restore back the old ITransformMatrix and IGraphicBundleContext to           |
| the IGraphicContext.                                                         |
------------------------------------------------------------------------------*/
IGraphicContext& IGraphicContext::remove( const IGraphic&    graphic,
                                          ITransformMatrix&  transformMatrix,
                                          IGraphicBundle&    graphicBundle )
{
  if (graphic.hasGraphicBundle())
  {
    setGraphicBundle(graphicBundle);
  }
  if (graphic.hasTransformMatrix() |
      graphic.transformMethod() == IGraphic::replace)
  {
    setWorldTransformMatrix(transformMatrix, IGraphic::replace );
  }
  return *this;
}
