/*******************************************************************************
* FILE NAME: iobjwin.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Definition of the class(es):                                               *
*     IObjectWindow -  Invisible Window Class                                  *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINWINDOWMGR
  #define INCL_WINERRORS
  #include <iwindefs.h>
}

#include <iobjwin.hpp>
#include <iexcept.hpp>
#include <ihandler.hpp>
#include <irect.hpp>
#ifdef IC_WIN
#include <ireslib.hpp>
#endif
#include <ithread.hpp>
#include <itrace.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IOBJWIN_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IObjectWindowCloseHandler                                                    |
|                                                                              |
| Close the object window upon receiving an IC_UM_CLOSE message.               |
------------------------------------------------------------------------------*/
class IObjectWindowCloseHandler : public IHandler {
typedef IHandler
  Inherited;
public:
  IObjectWindowCloseHandler ( );
virtual
 ~IObjectWindowCloseHandler ( );
virtual IObjectWindowCloseHandler
 &handleEventsFor       ( IObjectWindow* objectWindow ),
 &stopHandlingEventsFor ( IObjectWindow* objectWindow );
protected:
virtual Boolean
  dispatchHandlerEvent ( IEvent& event );
private:
virtual IHandler
 &handleEventsFor       ( IWindow* window ),
 &stopHandlingEventsFor ( IWindow* window );
}; // IObjectWindowCloseHandler


/*------------------------------------------------------------------------------
| IObjectWindowStatics                                                         |
|                                                                              |
| This class manages the lifetime of the static object window close handler.   |
| An object of this class is constructed during static object initialization.  |
| The handler object is constructed only if needed, and is deleted at          |
| termination of the application.                                              |
------------------------------------------------------------------------------*/
class IObjectWindowStatics : public IBase {
typedef IBase
  Inherited;
public:
 ~IObjectWindowStatics ( );
static IObjectWindowCloseHandler
 &closeHandler ( );
private:
static IObjectWindowCloseHandler
 *fCloseHandler;
}; // IObjectWindowStatics


#pragma data_seg(ICLNonConst)
static IObjectWindowStatics objectWindowStatics;
IObjectWindowCloseHandler* IObjectWindowStatics::fCloseHandler = 0;
#pragma data_seg()


/*************************************************************************/
/* IObjectWindow :: IObjectWindow                                        */
/*  Constructor for a minimal object window                              */
/*************************************************************************/
IObjectWindow :: IObjectWindow()
{
  static char* pszObjClass=0;

  IThread::current().initializeGUI();    // insure the GUI is started
#ifdef IC_WIN
  IWindowHandle focusWindow = IQUERYFOCUS(0);
#endif

  if(pszObjClass==0) {
    pszObjClass="ICLWC_OBJECT";
#ifdef IC_WIN
    HINSTANCE  hinst = GetModuleHandle(0);
    WNDCLASS   wndclass;
    // Check for it already registered.  Need this in some static
    // linking scenarios.
    Boolean    bSuc = GetClassInfo( hinst, pszObjClass, &wndclass);
    if (!bSuc)
       {
       // have to register it
       wndclass.style       = CS_NOCLOSE;
       wndclass.lpfnWndProc = (WNDPROC)IDEFWINDOWPROC;
       wndclass.cbClsExtra  = 0;
       wndclass.cbWndExtra  = 0;
       wndclass.hInstance   = hinst;
       wndclass.hIcon       = LoadIcon( (HINSTANCE)NULL, IDI_APPLICATION );
       wndclass.hCursor     = LoadCursor( (HINSTANCE)NULL, IDC_ARROW );
       wndclass.hbrBackground = (HBRUSH)(COLOR_BACKGROUND+1);
       wndclass.lpszMenuName  = NULL;
       wndclass.lpszClassName = pszObjClass;
       if (RegisterClass(&wndclass))
          bSuc = true;
       }
#else
    Boolean bSuc=WinRegisterClass((HAB)0,
                                  pszObjClass,
                                  WinDefWindowProc,
                                  0,    // no class style
                                  0);   // no extra window words
#endif
    if(!bSuc) {
     ITHROWGUIERROR("WinRegisterClass");
    }/* end if */
  }/* end if */

  IWindowHandle whObjWin =
      this -> create( 0,          // Parent=desktop so it looks "primary".
                      0,
                      0,
                      pszObjClass,
#ifdef IC_WIN
                      GetDesktopWindow(),
#else
                      HWND_DESKTOP,
#endif
                      0,
                      IRectangle(),
                      0,
                      0,
                      IWindow::onTopOfSiblings );


  startHandlingEventsFor(whObjWin);
  objectWindowStatics.closeHandler().handleEventsFor( this );
#ifdef IC_WIN
  ISETPARENT( whObjWin, NULL, true ); // Set real parent.
  // restore focus back to original window
  ISETFOCUS(0, focusWindow);
#else
  ISETPARENT( whObjWin, HWND_OBJECT, true ); // Set real parent.
#endif
}

/*************************************************************************/
/* IObjectWindow :: ~IObjectWindow                                       */
/*  Destructor for a minimal object window                               */
/*************************************************************************/
IObjectWindow :: ~IObjectWindow ( )
{
  objectWindowStatics.closeHandler().stopHandlingEventsFor( this );
}

/*************************************************************************/
/* IObjectWindow :: close                                                */
/*  Easy way to close an object window.                                  */
/*************************************************************************/
IObjectWindow& IObjectWindow::close ( )
{
  this->postEvent( IC_UM_CLOSE );
  return *this;
}


/*************************************************************************/
/* IObjectWindowStatics :: ~IObjectWindowStatics                         */
/*  Destructor to clean up.                                              */
/*************************************************************************/
IObjectWindowStatics::~IObjectWindowStatics ( )
{
  if ( fCloseHandler )
  {
     delete fCloseHandler;
  }
}

/*************************************************************************/
/* IObjectWindowStatics :: closeHandler                                  */
/*  Destructor to clean up.                                              */
/*************************************************************************/
IObjectWindowCloseHandler& IObjectWindowStatics::closeHandler ( )
{
  if ( fCloseHandler == 0 )
  {
     fCloseHandler = new IObjectWindowCloseHandler;
  }
  return *fCloseHandler;
}


/*************************************************************************/
/* IObjectWindowCloseHandler :: IObjectWindowCloseHandler                */
/*  Default constructor for page tuning.                                 */
/*************************************************************************/
IObjectWindowCloseHandler::IObjectWindowCloseHandler ( )
  : IObjectWindowCloseHandler::Inherited ( )
{ }

/*************************************************************************/
/* IObjectWindowCloseHandler :: ~IObjectWindowCloseHandler               */
/*  Empty destructor for page tuning.                                    */
/*************************************************************************/
IObjectWindowCloseHandler::~IObjectWindowCloseHandler ( )
{ }

/*************************************************************************/
/* IObjectWindowCloseHandler :: dispatchHandlerEvent                     */
/*  Handle request to destroy an object window.                          */
/*************************************************************************/
Boolean IObjectWindowCloseHandler::dispatchHandlerEvent ( IEvent& event )
{
  Boolean stopProcessingEvent = false;
  if ( event.eventId() == IC_UM_CLOSE )
  {
     IDESTROYWINDOW( event.window()->handle() );
     stopProcessingEvent = true;
  }
  return stopProcessingEvent;
}

/*************************************************************************/
/* IObjectWindowCloseHandler :: handleEventsFor                          */
/*  Attach handler to an object window.                                  */
/*************************************************************************/
IObjectWindowCloseHandler
 &IObjectWindowCloseHandler::handleEventsFor ( IObjectWindow* objectWindow )
{
  IASSERTPARM( objectWindow != 0 );
  Inherited::handleEventsFor( objectWindow );
  return *this;
}

/*************************************************************************/
/* IObjectWindowCloseHandler :: stopHandlingEventsFor                    */
/*  Detach handler from an object window.                                */
/*************************************************************************/
IObjectWindowCloseHandler
 &IObjectWindowCloseHandler::stopHandlingEventsFor ( IObjectWindow* objectWindow )
{
  IASSERTPARM( objectWindow != 0 );
  Inherited::stopHandlingEventsFor( objectWindow );
  return *this;
}

/*************************************************************************/
/* IObjectWindowCloseHandler :: handleEventsFor                          */
/*  Hide general version in IHandler.                                    */
/*************************************************************************/
IHandler& IObjectWindowCloseHandler::handleEventsFor ( IWindow* )
{
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}

/*************************************************************************/
/* IObjectWindowCloseHandler :: stopHandlingEventsFor                    */
/*  Hide general version in IHandler.                                    */
/*************************************************************************/
IHandler& IObjectWindowCloseHandler::stopHandlingEventsFor ( IWindow* )
{
  ITHROWLIBRARYERROR( IC_MEMBER_ACCESS_ERROR,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable );
  return *this;
}
