/*******************************************************************************
* FILE NAME: ipopmenu.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ipopmenu.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
   #define INCL_WINPOINTERS
   #define INCL_WINWINDOWMGR
   #define INCL_WINMENUS
   #define INCL_WININPUT
   #include <iwindefs.h>
}

#include <ipopmenu.hpp>

#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <imenuprv.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <iwindow.hpp>
#include <iplatfrm.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IPOPMENU_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IPopUpMenu::IPopUpMenu                                                       |
|  Constructor. Empty pop-up for dynamic menu.                                 |
------------------------------------------------------------------------------*/
IPopUpMenu::IPopUpMenu( IWindow*           owner,
                        unsigned long      windowId,
                        const Style&       style )
                      : fPopUpMenuData( 0 )
{
   IASSERTPARM(owner != 0);
#ifdef IC_WIN
   fhMenu = CreatePopupMenu();
   fowner = owner->handle();
   fmenuFrame = IMenuPrivate::locateFrame(owner);

   /*******************************************************************/
   /* Add an entry for the main popup, identified by the window id    */
   /*******************************************************************/
   IMenuPrivate::addToLookUpTable( fmenuFrame, fhMenu, windowId );
#endif
#ifdef IC_PM
   IWindowHandle whMenu =
      this -> create( windowId,
                      0,
                      style.asUnsignedLong(),
                      WC_MENU,
                      HWND_DESKTOP,
                      owner->handle(),
                      IRectangle(),
                      0,
                      0,
                      IWindow::onTopOfSiblings );

   startHandlingEventsFor(whMenu);
#endif
}


/*------------------------------------------------------------------------------
| IPopUpMenu::IPopUpMenu                                                       |
|  Constructor. Pop-up from .rc.                                               |
------------------------------------------------------------------------------*/
IPopUpMenu::IPopUpMenu( const IResourceId& menuResId,
                        IWindow* owner )
                      : fPopUpMenuData( 0 )
{
#ifdef IC_WIN
   char  menuText[IC_MENUTEXTLENGTH];

   /*******************************************************************/
   /* Since Windows does not allow you to load a popup directly, we   */
   /* must create an empty popup, load the menu resource and then     */
   /* copy each item over to the new menu                             */
   /*******************************************************************/
   fhMenu = CreatePopupMenu();
   HMENU menuLoaded = menuResId.resourceLibrary().loadMenu(menuResId, owner);
   fowner = owner->handle();
   fmenuFrame = IMenuPrivate::locateFrame(owner);

   /*******************************************************************/
   /* Begin with first item (0 based) and get count of items          */
   /*******************************************************************/
   int itemPos = 0;
   int itemCount = GetMenuItemCount( menuLoaded );

   /*******************************************************************/
   /* Loop through menu, transferring each item found to the new menu */
   /*******************************************************************/
   while ( itemPos < itemCount )
   {
     /*****************************************************************/
     /* If on Windows 95 or new shell, use extended menu api to xfer  */
     /*****************************************************************/
     if (IPlatform::isWin95() || IPlatform::isNTNewShell())
     {
       MENUITEMINFO miItem;
       miItem.cbSize = sizeof(MENUITEMINFO);
       miItem.fMask = MIIM_STATE | MIIM_ID | MIIM_SUBMENU | MIIM_TYPE |
                      MIIM_DATA;
       miItem.dwTypeData = menuText;
       miItem.cch = IC_MENUTEXTLENGTH;

       /***************************************************************/
       /* Get all of the information about the menuitem.              */
       /***************************************************************/
       GetMenuItemInfo( menuLoaded, itemPos, TRUE, &miItem );

       /***************************************************************/
       /* Insert the item as is into the newly created popup menu     */
       /***************************************************************/
       InsertMenuItem( fhMenu, itemPos, TRUE, &miItem );
     }
     /*****************************************************************/
     /* otherwise use old menu API to xfer items over                 */
     /*****************************************************************/
     else
     {
       ULONG ulItem = (ULONG)GetSubMenu( menuLoaded, itemPos );
       ULONG uState = GetMenuState( menuLoaded, itemPos, MF_BYPOSITION );
       if (ulItem)
         uState &= 0x000000FF;

       if (!ulItem)
         ulItem = GetMenuItemID( menuLoaded, itemPos );
       GetMenuString( menuLoaded, itemPos, menuText, IC_MENUTEXTLENGTH,
                      MF_BYPOSITION );
       InsertMenu( fhMenu, 0xFFFFFFFF, (UINT)(MF_BYPOSITION | uState),
                   (UINT)ulItem, menuText );
     }
     itemPos++;
   }

   /*******************************************************************/
   /* Loop through menu again to remove submenus so they aren't       */
   /*  destroyed (which only is a problem on Windows 95)              */
   /*******************************************************************/
   for ( itemPos = 0; itemPos < itemCount; itemPos++ )
     RemoveMenu( menuLoaded, ((itemCount-1)-itemPos), MF_BYPOSITION );

   /*******************************************************************/
   /* Destroy loaded menu once items are transferred                  */
   /*******************************************************************/
   DestroyMenu( menuLoaded );

   /*******************************************************************/
   /* For popups, always add to lookup table for either menu system   */
   /*******************************************************************/
   IMenuPrivate::buildLookUpTableFromResourceLoad( fmenuFrame, fhMenu,
                                                   menuResId );
#endif
#ifdef IC_PM
   IWindowHandle handleClPopUp =
       menuResId.resourceLibrary().loadMenu(menuResId, owner);
   WinSetParent(handleClPopUp, HWND_DESKTOP, FALSE);
   WinSetWindowUShort(handleClPopUp,           //PM load with FID_MENU
                      QWS_ID, menuResId);      //Change it.
   startHandlingEventsFor(handleClPopUp);
#endif
}


/*------------------------------------------------------------------------------
| IPopUpMenu::~IPopUpMenu                                                      |
|  Define to override IMenu abstract class definition.                         |
------------------------------------------------------------------------------*/
IPopUpMenu::~IPopUpMenu()
{
#ifdef IC_WIN
  if (fhMenu)
    DestroyMenu( fhMenu );
#endif
}


/*------------------------------------------------------------------------------
| IPopUpMenu::show                                                             |
------------------------------------------------------------------------------*/
IPopUpMenu& IPopUpMenu::show( const IPoint& ptAt )
{
   IPoint nativePoint = ICoordinateSystem::convertToNative(
                           ptAt, owner()->size() );
#ifdef IC_WIN
   POINTL scrPoint = nativePoint.asPOINTL();
   ClientToScreen( fowner, (LPPOINT)&scrPoint );

   // Show the popup - Align topleft corner of menu with mouse pointer
   if (!TrackPopupMenu( fhMenu,
                        TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RIGHTBUTTON,
                        (int)scrPoint.x, (int)scrPoint.y,
                        0, fowner, NULL ))
   {
     ITHROWGUIERROR("TrackPopupMenu");
   }
#endif
#ifdef IC_PM
   IWindowHandle ownerHwnd = owner()->handle();
   if (!WinPopupMenu(ownerHwnd, ownerHwnd,
                     handle(),
                     nativePoint.x(), nativePoint.y(),
                     0,
                     PU_HCONSTRAIN | PU_VCONSTRAIN |
                     PU_KEYBOARD | PU_MOUSEBUTTON1 | PU_MOUSEBUTTON2) )
   {
     ITHROWGUIERROR("WinPopUpMenu");
   }
#endif
   return *this;
}
