/*******************************************************************************
* FILE NAME: ithreads.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of certain IThread member funcions,  *
*   removing the dependency on beginthread for single-threaded applications.   *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// If the compiler were to support:
//    #pragma options /Gm+
// then we wouldn't need the declaration of _beginthread below.
// The declaration is lifted from stdlib.h, because it won't get
// generated unless /Gm+ has been set for the compile.
// If the declaration changes in stdlib.h, then we have to update here.

#include <iwindefs.h>

extern "C" {
  int  _Optlink _beginthread( void ( * _Optlink )( void * ), void *, unsigned,
                              void * );

}

extern "C" {
  #include <stdlib.h>
}

#include <ithread.hpp>

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#ifndef _ICRITSEC_
  #include <icritsec.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

#ifndef _IPLATFRM_
  #include <iplatfrm.hpp>
#endif

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ITHREADS_CPP_
  #include <ipagetun.h>
#endif


class IStartedThread;

/*------------------------------------------------------------------------------
| initThread                                                                   |
------------------------------------------------------------------------------*/
extern "C" void initThread ( void *aThread );

IThread :: IThread ( const IReference<IThreadFn> &aFnObjRef,
                     Boolean                      autoInitPM )
  : thread( IThread::newStartedThread() ),
    fThreadData( 0 )
  {
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    {
     ITHROWLIBRARYERROR( IC_THREAD_NOT_STARTED_WIN32S,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
    }
#endif
  this->start( aFnObjRef, autoInitPM );
  }

/*------------------------------------------------------------------------------
| IThread::IThread                                                             |
|                                                                              |
| IThread constructor for functions that use the _Optlink linkage              |
------------------------------------------------------------------------------*/
IThread :: IThread ( OptlinkFnPtr  pfn,
                     void         *anArg,
                     Boolean       autoInitPM )
  : thread( IThread::newStartedThread() ),
    fThreadData( 0 )
  {
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    {
     ITHROWLIBRARYERROR( IC_THREAD_NOT_STARTED_WIN32S,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
    }
#endif
  this->start( pfn, anArg, autoInitPM );
  }

/*------------------------------------------------------------------------------
| IThread::IThread                                                             |
|                                                                              |
| IThread constructor for functions that use _System linkage                   |
------------------------------------------------------------------------------*/
IThread :: IThread ( SystemFnPtr   pfn,
                     unsigned long anArg,
                     Boolean       autoInitPM )
  : thread( IThread::newStartedThread() ),
    fThreadData( 0 )
  {
#ifdef IC_WIN
  if (IPlatform::isWin32S())
    {
     ITHROWLIBRARYERROR( IC_THREAD_NOT_STARTED_WIN32S,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
    }
#endif
  this->start( pfn, anArg, autoInitPM );
  }

/*------------------------------------------------------------------------------
| IThread::start                                                               |
|                                                                              |
| All versions are basically implemented via the function that accepts         |
| as argument a reference to an IThread::Function object.  The other           |
| versions construct such an object from the other arguments.                  |
|                                                                              |
| The function is implemented by invoking _beginthread against the             |
| static member function IStartedThread::initThread, passing as argument       |
| the IStartedThread address.                                                  |
|                                                                              |
| The IStartedThread referenced by the receiver is added to the                |
| static "started thread" array.                                               |
------------------------------------------------------------------------------*/
void IThread :: start ( const IReference<IThreadFn> &aFnObjRef,
                        Boolean autoInitPM )
  {
  IMODTRACE_DEVELOP( "IThread::start" );

#ifdef IC_WIN
  if (IPlatform::isWin32S())
    {
     ITHROWLIBRARYERROR( IC_THREAD_NOT_STARTED_WIN32S,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
    }
#endif

  IStartedThread
   *p = this->prepareToStart( aFnObjRef, autoInitPM );

  ITRACE_DEVELOP( IString( "\tStarting thread=" ) +
                  IString::d2x( IString( (int)p )));

  int rc = _beginthread( initThread,
                         0,
                         this->stackSize(),
                         p );

  this->threadStarted( p, rc );
  }


/*------------------------------------------------------------------------------
| OptlinkFn class : An IThreadFn subclass to dispatch an optlink function      |
------------------------------------------------------------------------------*/
struct OptlinkFn : public IThreadFn {
  OptlinkFn ( IThread::OptlinkFnPtr pfn,
              void*                 p );
 ~OptlinkFn ( );

void
  run ( );
IThread::OptlinkFnPtr
  function;
void
 *arg;
private:
  OptlinkFn ( const OptlinkFn& );
OptlinkFn
 &operator= ( const OptlinkFn& );
};

/*------------------------------------------------------------------------------
| OptlinkFn::OptlinkFn                                                         |
------------------------------------------------------------------------------*/
OptlinkFn :: OptlinkFn ( IThread::OptlinkFnPtr pfn, void *p )
  : function( pfn ), arg( p )
  {
  }

/*------------------------------------------------------------------------------
| OptlinkFn::~OptlinkFn                                                        |
------------------------------------------------------------------------------*/
OptlinkFn :: ~OptlinkFn ( )
  {
  }

/*------------------------------------------------------------------------------
| OptlinkFn::run                                                               |
------------------------------------------------------------------------------*/
void OptlinkFn :: run ( )
{
  function( arg );
}

/*------------------------------------------------------------------------------
| IThread::start                                                               |
------------------------------------------------------------------------------*/
void IThread :: start ( OptlinkFnPtr  pfn,
                        void         *anArg,
                        Boolean       autoInitPM )
  {
  this->start( new OptlinkFn( pfn, anArg ), autoInitPM );
  }

/*------------------------------------------------------------------------------
| SystemFn class : An IThreadFn subclass to dispatch a system function.        |
------------------------------------------------------------------------------*/
struct SystemFn : public IThreadFn {
  SystemFn ( IThread::SystemFnPtr pfn,
             unsigned long        p );
 ~SystemFn ( );

void
  run ( );
IThread::SystemFnPtr
  function;
unsigned long
  arg;

  SystemFn  ( const SystemFn& );
SystemFn
 &operator= ( const SystemFn& );
};

/*------------------------------------------------------------------------------
| SystemFn::SystemFn                                                           |
------------------------------------------------------------------------------*/
SystemFn :: SystemFn ( IThread::SystemFnPtr pfn, unsigned long p )
  : function( pfn ), arg( p )
  {
  }

/*------------------------------------------------------------------------------
| SystemFn::~SystemFn                                                          |
------------------------------------------------------------------------------*/
SystemFn :: ~SystemFn ( )
  {
  }

/*------------------------------------------------------------------------------
| SystemFn::run                                                                |
------------------------------------------------------------------------------*/
void SystemFn :: run ( )
{
  function( arg );
}

/*------------------------------------------------------------------------------
| IThread::start                                                               |
------------------------------------------------------------------------------*/
void IThread :: start ( SystemFnPtr   pfn,
                        unsigned long anArg,
                        Boolean       autoInitPM )
  {
  this->start( new SystemFn( pfn, anArg ), autoInitPM );
  }
