/*******************************************************************************
* FILE NAME: itimer.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     ITimer                                                                   *
*     ITimerFn                                                                 *
*     ITimerMemberFn                                                           *
*     ITimerMemberFn0                                                          *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINTIMER
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
}

#include <itimer.hpp>

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

#ifndef _IRESLOCK_
  #include <ireslock.hpp>
#endif

#ifndef _IHANDLER_
  #include <ihandler.hpp>
#endif

#ifndef _IOBJWIN_
  #include <iobjwin.hpp>
#endif

#ifndef _ITHREAD_
  #include <ithread.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ITIMER_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| TimerRecord                                                                  |
|                                                                              |
| This class is used to store information active (or referenced) timers.       |
------------------------------------------------------------------------------*/

class TimerRecord {
public:
  TimerRecord ();
unsigned long
  interval;
ITimerFn*
  fn;
#ifndef IC_WIN
IObjectWindow*
  win;
#else
unsigned long
  timerId;
#endif
unsigned long
  refCnt;
};

TimerRecord :: TimerRecord ()
             : interval ( 1000 ),
               fn ( 0 ),
#ifndef IC_WIN
               win ( 0 ),
#else
               timerId( 0 ),
#endif
               refCnt ( 0 )
{
}

static IPrivateResource semTimer;       // Semaphore to access timer list
static TimerRecord* timerList = 0;      // List of referenced timers
static unsigned long timerListSize = 0; // Current size of timer list


#ifdef IC_WIN

// The windows timer class uses a timer call back procedure.  The original
// implementation used an object window like OS/2.  The problem with that
// approach is that when the object window is re-parented, it would steal
// the focus from the application.
VOID CALLBACK TimerProc( HWND hwnd, UINT uMsg, UINT timerId, DWORD time)
{
  unsigned long appTimerId;

  ITimerFn* fn = 0;
  {
    int fIndex;

    IResourceLock  lock(semTimer);

    // Iterate through the timer list searching for the timer id
    // that was returned from the SetTimer api.
    for (fIndex = 1; fIndex < timerListSize; fIndex++ )
    {
      if ( timerList[fIndex].timerId == timerId &&
           timerList[fIndex].fn != 0 )
      {
        fn = timerList[fIndex].fn;
        appTimerId = (unsigned long)(fIndex + IC_RESERVED_TIMERS);
        break;
      }
    }
  }

  // If the timer was found in the list.
  if (fn)
  {
     // Call the correct timer expired function.  Add a reference to
     // the function while we are calling it to be sure that it does
     // not get deleted while in use.

     fn->addRef();
     fn->timerExpired(appTimerId);
     fn->removeRef();
  }
}

#else

/*------------------------------------------------------------------------------
| ITimerHandler                                                                |
|                                                                              |
| This class is used to handler WM_TIMER messages and dispatch them to the     |
| correct ITimerFn.                                                            |
------------------------------------------------------------------------------*/
class ITimerHandler : public IHandler {
typedef IHandler
  Inherited;
public:
  ITimerHandler ( );
  ~ITimerHandler ( );
protected:
Boolean
  dispatchHandlerEvent ( IEvent& event );
};

ITimerHandler :: ITimerHandler ( )
{
}

ITimerHandler :: ~ITimerHandler ( )
{
}

Boolean ITimerHandler :: dispatchHandlerEvent ( IEvent& event )
{
   if ( event.eventId() == WM_TIMER )
   {
      unsigned long timerId = (unsigned long) event.parameter1();
      ITimerFn* fn = 0;
      {
        IResourceLock  lock(semTimer);
        if (( timerId > IC_RESERVED_TIMERS ) &&
           (( timerId - IC_RESERVED_TIMERS ) < timerListSize ))
          fn = timerList[timerId-IC_RESERVED_TIMERS].fn;
      }
      if (fn)
      {
         // Call the correct timer expired function.  Add a reference to
         // the function while we are calling it to be sure that it does
         // not get deleted while in use.

         fn->addRef();
         fn->timerExpired(timerId);
         fn->removeRef();

         return true;
      }
   }
   return false;
}

/*------------------------------------------------------------------------------
| DefaultTimerHandler                                                          |
|                                                                              |
| This class wrappers an object pointer and destroys the pointer to the object |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultTimerHandler {
  operator ITimerHandler* ();
ITimerHandler
 *operator -> ()
    {
    return *this;
    }
  ~DefaultTimerHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    if (timerList)
      {
      delete [] timerList;
      timerList = 0;
      }
    }
ITimerHandler
 *ptr;
} defaultTimerHandler;

DefaultTimerHandler :: operator ITimerHandler* ( )
  {
  if ( !ptr )
    ptr = new ITimerHandler;
  return ptr;
  }
#endif

/*------------------------------------------------------------------------------
| ITimer::ITimer                                                               |
|                                                                              |
| get free entry from timer list and add reference                             |
------------------------------------------------------------------------------*/
ITimer::ITimer ()
{
  fIndex = 0;
  IResourceLock lock(semTimer);
  int i;

  for ( i = 1; (( i < timerListSize ) && ( fIndex == 0 )); i++ )
  {
     if (( timerList[i].refCnt == 0 ) &&
         ( timerList[i].fn == 0 ))
     {
        fIndex = i;
     }
  }

  if ( fIndex == 0 )
  {
     if (( timerListSize + 10 + IC_RESERVED_TIMERS ) >= TID_USERMAX )
     {
       ITHROWLIBRARYERROR( IC_TIMER_ID_UNAVAILABLE,
                           IBaseErrorInfo::outOfSystemResource,
                           IException::recoverable );
     }
     else
     {
        // Increase the timer list by 10 records

        if (timerListSize > 0)
          fIndex = timerListSize;
        else
          fIndex = 1;

        TimerRecord* oldList = timerList;
        timerList = new TimerRecord[timerListSize + 10];
        for ( i = 1; i < timerListSize; i++ )
        {
           timerList[i] = oldList[i];
        }
        delete [] oldList;
        timerListSize += 10;
     }
  }

  timerList[fIndex].refCnt = 1;
}

/*------------------------------------------------------------------------------
| ITimer::ITimer                                                               |
|                                                                              |
| find entry in timer list and add reference                                   |
------------------------------------------------------------------------------*/
ITimer::ITimer ( unsigned long timerIdentifer )
{
  fIndex = timerIdentifer - IC_RESERVED_TIMERS;
  IResourceLock lock(semTimer);

  if (( fIndex == 0 ) || ( fIndex >= timerListSize ))
  {
     ITHROWLIBRARYERROR( IC_INVALID_TIMER_ID,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable );
  }

  if ( timerList[fIndex].fn == 0 )
  {
     ITHROWLIBRARYERROR( IC_INVALID_TIMER_ID,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable );
  }

  timerList[fIndex].refCnt++;
}

/*------------------------------------------------------------------------------
| ITimer::ITimer                                                               |
|                                                                              |
| add reference to timer list entry                                            |
------------------------------------------------------------------------------*/
ITimer::ITimer ( const ITimer &timer )
{
  fIndex = timer.fIndex;
  IResourceLock lock(semTimer);
  timerList[fIndex].refCnt++;
}

/*------------------------------------------------------------------------------
| ITimer::ITimer                                                               |
|                                                                              |
| get free entry from timer list, add reference, and start timer               |
------------------------------------------------------------------------------*/
ITimer::ITimer ( const IReference<ITimerFn> &timerFunction,
                 unsigned long timerInterval )
{
  ITimer newTimer;                      // Use default ctor to get free fIndex
  fIndex = newTimer.fIndex;

  // Block for modifying the timer list
  {
    IResourceLock lock(semTimer);
    timerList[fIndex].refCnt++;
  }

  start ( timerFunction, timerInterval );
}

/*------------------------------------------------------------------------------
| ITimer::operator=                                                            |
|                                                                              |
| assignement operator for ITimer                                              |
------------------------------------------------------------------------------*/
ITimer& ITimer::operator= ( const ITimer& timer )
{
  IResourceLock lock(semTimer);

  // Remove the reference to the current timer index
  timerList[fIndex].refCnt--;

  // Add reference to the new timer index
  fIndex = timer.fIndex;
  timerList[fIndex].refCnt++;

  return *this;
}

/*------------------------------------------------------------------------------
| ITimer::~ITimer                                                              |
|                                                                              |
| remove reference from timer list                                             |
------------------------------------------------------------------------------*/

ITimer::~ITimer ()
{
  IResourceLock lock(semTimer);
  timerList[fIndex].refCnt--;
}

/*------------------------------------------------------------------------------
| ITimer::start                                                                |
------------------------------------------------------------------------------*/
ITimer& ITimer :: start ( const IReference<ITimerFn> &timerFunction )
{
  if ( isStarted() )
  {
    ITHROWLIBRARYERROR( IC_TIMER_STARTED,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable );
  }

  unsigned long timerInterval;
#ifndef IC_WIN
  IObjectWindow* win = 0;
#endif

  // Block for working with the timer list
  {
    IResourceLock lock(semTimer);
    timerInterval = timerList[fIndex].interval;
    timerList[fIndex].fn = timerFunction;
#ifndef IC_WIN
    timerList[fIndex].win = new IObjectWindow();

    timerList[fIndex].win->setParent(IWindow::objectWindow());

    win = timerList[fIndex].win;
  }

  ITimerHandler* hdr = defaultTimerHandler;
  hdr->handleEventsFor(win);

  if ( ISTARTTIMER ( IThread::current().anchorBlock(), win->handle(),
                     fIndex + IC_RESERVED_TIMERS, timerInterval ) == 0 )
  {
    ITHROWGUIERROR("WinStartTimer");
  }
#else
    // In windows when you create a timer using a call back proc, the system
    // determines the timer id and is returned.  This value is saved as part
    // of the timer record and is used to uniquely identify the timer.
    timerList[fIndex].timerId = SetTimer( 0,
                                          (unsigned)(fIndex + IC_RESERVED_TIMERS),
                                          (unsigned)timerInterval,
                                          (TIMERPROC)TimerProc );
  }
#endif

  IMODTRACE_ALL( IString( "ITtimer::start(" ) +
                 IString( fIndex + IC_RESERVED_TIMERS ) +
                 IString( ')' ) );

  timerFunction->addRef();

  return *this;
}

/*------------------------------------------------------------------------------
| ITimer::start                                                                |
------------------------------------------------------------------------------*/

ITimer& ITimer :: start ( const IReference<ITimerFn> &timerFunction,
                          unsigned long timerInterval )
{
  setInterval ( timerInterval );
  return start ( timerFunction );
}

/*------------------------------------------------------------------------------
| ITimer::stop                                                                 |
------------------------------------------------------------------------------*/
ITimer& ITimer :: stop ()
{
  if ( isStarted() )
  {
#ifndef IC_WIN
    ITimerHandler* hdr = defaultTimerHandler;
    IObjectWindow* win = 0;

    { // Block for reading from the list
      IResourceLock lock(semTimer);
      win = timerList[fIndex].win;
    }

    // Check it the window still exists before calling PM.  The window may
    // have been destroyed by a terminating thread without stopping the
    // timer.

    if ( win->isValid() )
    {
      if ( ISTOPTIMER ( IThread::current().anchorBlock(), win->handle(),
                       fIndex + IC_RESERVED_TIMERS ) == FALSE )
      {
        ITHROWGUIERROR("WinStopTimer");
      }
    }
#endif

    { // Block for removing data from the list
      IResourceLock lock(semTimer);
#ifdef IC_WIN
      KillTimer( 0, (unsigned)timerList[fIndex].timerId );
      timerList[fIndex].timerId = 0;
#endif
      if (timerList[fIndex].fn)
      {
        timerList[fIndex].fn->removeRef();
        timerList[fIndex].fn = 0;
      }
#ifndef IC_WIN
      timerList[fIndex].win = 0;
#endif
    }

#ifndef IC_WIN
    if (win)
    {
      hdr->stopHandlingEventsFor(win);
      if ( win->isValid() )
      {
        // Destory the window and automatically delete the instance
        // when it is safe (ie. when we are not nested)
        win->setAutoDeleteObject();
        IDESTROYWINDOW(win->handle());
      }
      else
      {
        // The PM window does not exist, so we are sure this is not
        // a nested call for the window and it is safe to delete the
        // object instance.
        delete win;
      }
    }
#endif

    IMODTRACE_ALL( IString( "ITtimer::stop(" ) +
                   IString( fIndex + IC_RESERVED_TIMERS ) +
                   IString( ')' ) );
  }

  return *this;
}

/*------------------------------------------------------------------------------
| ITimer::id                                                                   |
------------------------------------------------------------------------------*/
unsigned long ITimer :: id () const
{
  return (fIndex + IC_RESERVED_TIMERS);
}

/*------------------------------------------------------------------------------
| ITimer::asString                                                             |
------------------------------------------------------------------------------*/
IString ITimer :: asString () const
{
  return ( IString("ITimer(") + IString(fIndex+IC_RESERVED_TIMERS) + ")" );
}

/*------------------------------------------------------------------------------
| ITimer::asDebugInfo                                                          |
------------------------------------------------------------------------------*/
IString ITimer :: asDebugInfo () const
{
  IString str;
  str = IString("ITimer: id = ") + IString(fIndex+IC_RESERVED_TIMERS);
  str += IString(", interval = ") + IString(interval());
  if (isStarted())
    str += ", timer started";
  return str;
}

/*------------------------------------------------------------------------------
| ITimer::interval                                                             |
------------------------------------------------------------------------------*/
unsigned long ITimer :: interval () const
{
  IResourceLock lock(semTimer);
  return timerList[fIndex].interval;
}

/*------------------------------------------------------------------------------
| ITimer::setInterval                                                          |
------------------------------------------------------------------------------*/
ITimer& ITimer :: setInterval ( unsigned long timerInterval )
{
  if ( isStarted() )
  {
    ITimerFn* fn;
    Boolean change;
    {
      IResourceLock lock(semTimer);
      if ( timerList[fIndex].interval == timerInterval )
      {
        change = false;
      }
      else
      {
        change = true;
        fn = timerList[fIndex].fn;
        timerList[fIndex].interval = timerInterval;
      }
    }
    if ( change )
    {
      fn->addRef();
      stop();
      start(fn);
    }
  }
  else
  {
    IResourceLock lock(semTimer);
    timerList[fIndex].interval = timerInterval;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ITimer::isStarted                                                            |
------------------------------------------------------------------------------*/
Boolean ITimer::isStarted () const
{
  IResourceLock lock(semTimer);
  if ( timerList[fIndex].fn == 0 )
    return false;
  return true;
}

/*------------------------------------------------------------------------------
| ITimerFn::ITimerFn                                                           |
|                                                                              |
| Empty ctor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
ITimerFn :: ITimerFn ( )
{
}

/*------------------------------------------------------------------------------
| ITimerFn::~ITimerFn                                                          |
|                                                                              |
| Empty dtor to prevent generated static version.                              |
------------------------------------------------------------------------------*/
ITimerFn :: ~ITimerFn ( )
{
}

/*------------------------------------------------------------------------------
| ITimer::Cursor::Cursor                                                       |
------------------------------------------------------------------------------*/
ITimer::Cursor :: Cursor ()
                : fIndex ( 0 )
{
}

/*------------------------------------------------------------------------------
| ITimer::Cursor::~Cursor                                                      |
------------------------------------------------------------------------------*/
ITimer::Cursor :: ~Cursor ( )
{
}

/*------------------------------------------------------------------------------
| ITimer::Cursor::setToFirst                                                   |
|                                                                              |
| Point to first element of started timer array                                |
------------------------------------------------------------------------------*/
Boolean ITimer::Cursor :: setToFirst ( )
{
  IResourceLock lock(semTimer);
  for ( fIndex = 1; fIndex < timerListSize; fIndex++ )
    if ( timerList[fIndex].fn != 0 )
      return true;
  fIndex = 0;
  return false;
}

/*------------------------------------------------------------------------------
| Iimer::Cursor::setToNext                                                     |
|                                                                              |
| Advance cursor and return whether it's still valid.                          |
------------------------------------------------------------------------------*/
Boolean ITimer::Cursor :: setToNext ( )
{
  fIndex++;
  IResourceLock lock(semTimer);
  for (; fIndex < timerListSize; fIndex++ )
    if ( timerList[fIndex].fn != 0 )
      return true;
  fIndex = 0;
  return false;
}

/*------------------------------------------------------------------------------
| ITimer::Cursor::isValid                                                      |
------------------------------------------------------------------------------*/
Boolean ITimer::Cursor :: isValid ( ) const
{
  if ( fIndex == 0 )
    return false;

  IResourceLock lock(semTimer);
  if ( timerList[fIndex].fn == 0 )
    return false;

  return true;
}

/*------------------------------------------------------------------------------
| Iimer::Cursor::invalidate                                                    |
------------------------------------------------------------------------------*/
void ITimer::Cursor :: invalidate ( )
{
   fIndex = 0;
}

/*------------------------------------------------------------------------------
| ITimer::timerAt                                                              |
|                                                                              |
| Check cusor validity and return timer at cursor.                             |
------------------------------------------------------------------------------*/
ITimer ITimer :: timerAt ( const ITimer::Cursor& cursor )
{
  IASSERTPARM( cursor.isValid() );
  return ITimer(cursor.fIndex + IC_RESERVED_TIMERS);
}

/*------------------------------------------------------------------------------
| ITimer::operator ==                                                          |
|                                                                              |
| Check if two timers are identical.                                           |
------------------------------------------------------------------------------*/
Boolean ITimer :: operator == ( const ITimer timer )
{
   if ( timer.id() == this->id() )
   {
     return true;
   }
   return false;
}

