/*******************************************************************************
* FILE NAME: ifont.cpp                                                         *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ifont.hpp.                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
   #define INCL_GPILCIDS
   #define INCL_GPIPRIMITIVES
   #define INCL_WINMLE
   #define INCL_WINSTDFONT
   #define INCL_WINSYS
   #define INCL_WINERRORS
   #define IUSING_OLE
   #ifdef WIN32_LEAN_AND_MEAN
     #undef WIN32_LEAN_AND_MEAN
   #endif
   #include <iwindefs.h>
   #include <string.h>
   #include <stdlib.h>
   #include <math.h>
   #include <ctype.h>
}

#include <ifont.hpp>
#include <ifntpriv.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iexcept.hpp>
#include <iwindow.hpp>
#include <iwinpriv.hpp>
#include <ifontdlg.hpp>
#include <ithread.hpp>
#include <iwcname.hpp>
#ifdef IC_PM
  #include <ipmbidi.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFONT_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
extern unsigned long propagationInProgress;
#endif

// Local non-PM flags used by this class

#define IC_BITMAP_FONTS_ONLY              0x01
#define IC_VECTOR_FONTS_ONLY              0x02
#define IC_NONPROPORTIONAL_FONTS_ONLY     0x04
#define IC_MUSTUSEVECTORFONT              0x08

const int iCharArrayWidth = 256;

/*------------------------------------------------------------------------------
| Note:                                                                        |
|                                                                              |
| Because of reference counting, any new or existing member function which     |
| causes a change to the font should call copyPrivate().  This ensures that    |
| existing font objects which point to the same font private data object       |
| will not be changed.                                                         |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IFontPrivateData::IFontPrivateData                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFontPrivateData::IFontPrivateData() : flClStyles(0),
                                       lClFontDir(CHDIRN_DEFAULT),
                                       pntClAngle(1,0),
                                       pntClShear(0,1),
                                       fxClPointSize(0),
#ifdef IC_PM
                                       pfontmetCl(0),
                                       pfattrsCl(0),
                                       plClCharWidths(0),
                                       fxClHeight(0),
                                       fxClWidth(0),
                                       lClLCId(0),
#endif
#ifdef IC_WIN
                                       fpFontHandle(new IFontHandle()),
                                       prevFont(NULL),
                                       systemFont(GetStockObject(SYSTEM_FONT)),
                                       fcharSet(DEFAULT_CHARSET),
                                       plogfontCl(0),
                                       ptextmetricCl(0),
                                       inUseCount(0),
#endif
                                       refs(1)
{
#ifdef IC_WIN
  /********************************************************************/
  /* Request information for the system font (Stock object) and if    */
  /* retrieved, use character set from it, otherwise leave as default */
  /* Only use it if it is not ANSI, since using ANSI will force the   */
  /* font mapper to select True type fonts over vector or symbol fonts*/
  /********************************************************************/
  LOGFONT logfont;
  if ( GetObject( systemFont, sizeof(LOGFONT), &logfont ) )
    if ( logfont.lfCharSet )
      fcharSet = logfont.lfCharSet;
#endif
}


/*------------------------------------------------------------------------------
| IFontPrivateData::IFontPrivateData                                           |
|                                                                              |
| Construct font private data based on input parameters.                       |
------------------------------------------------------------------------------*/
IFontPrivateData::IFontPrivateData( const IFontPrivateData& fpd )
{
  flClStyles     = fpd.flClStyles;
  lClFontDir     = fpd.lClFontDir;
  pntClAngle     = fpd.pntClAngle;
  pntClShear     = fpd.pntClShear;
  fxClPointSize  = fpd.fxClPointSize;
#ifdef IC_PM
  lClLCId        = 0;
  fxClHeight     = fpd.fxClHeight;
  fxClWidth      = fpd.fxClWidth;
  plClCharWidths = 0;

  /* Initialize the font metrics structure. */
  if (fpd.pfontmetCl)
  {
    pfontmetCl = new FONTMETRICS;
    *pfontmetCl = *(fpd.pfontmetCl);
  }
  else
  {
    pfontmetCl = 0;
  }

  /* Initialize font attributes structure */
  if (fpd.pfattrsCl)
  {
    pfattrsCl = new FATTRS;
    *pfattrsCl = *(fpd.pfattrsCl);
  }
  else
  {
    pfattrsCl = 0;
  }
#endif
#ifdef IC_WIN
  fpd.fpFontHandle->addRef();
  fpFontHandle = fpd.fpFontHandle;
  inUseCount   = 0;
  refs         = 1;
  prevFont     = 0;
  systemFont   = fpd.systemFont;
  fcharSet     = fpd.fcharSet;
  /* Initialize the log font structure. */
  if (fpd.plogfontCl)
  {
    plogfontCl = new LOGFONT;
    *plogfontCl = *(fpd.plogfontCl);
  }
  else
  {
    plogfontCl = 0;
  }
  if (fpd.ptextmetricCl)
  {
    ptextmetricCl = new TEXTMETRIC;
    *ptextmetricCl = *(fpd.ptextmetricCl);
  }
  else
  {
    ptextmetricCl = 0;
  }
#endif
}

/*------------------------------------------------------------------------------
| IFontPrivateData::removeRef                                                  |
|                                                                              |
| Decrement the ref count and delete the object if refcount is zero.           |
------------------------------------------------------------------------------*/
void IFontPrivateData::removeRef()
{
  if ( --refs == 0 )
  {
#ifdef IC_PM
    delete pfontmetCl;
    delete [] plClCharWidths;
    delete pfattrsCl;
#endif
#ifdef IC_WIN
    /* Get rid of stock font requested */
    DeleteObject(systemFont);

    /* Delete structures allocated by us for storing font information */
    delete plogfontCl;
    delete ptextmetricCl;
    fpFontHandle->removeRef();
#endif
    delete this;
  }
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IFontHandle::IFontHandle                                                     |
|                                                                              |
| Default ctor for HFONT wrapper class                                         |
------------------------------------------------------------------------------*/
IFontHandle::IFontHandle( Value hFont, Boolean allocated )
                 : IHandle( hFont ),
                   fAllocatedByIFont(allocated),
                   refs(1)
{ }


/*------------------------------------------------------------------------------
| IFontHandle::~IFontHandle                                                    |
------------------------------------------------------------------------------*/
IFontHandle::~IFontHandle()
{ }


/*------------------------------------------------------------------------------
| IFontHandle::removeRef                                                       |
|                                                                              |
| Decrement the ref count and delete the object if refcount is zero.           |
| Also delete the font object if this was allocated by the IFont code          |
------------------------------------------------------------------------------*/
void IFontHandle::removeRef()
{
  if ( --refs == 0 )
  {
    if (fAllocatedByIFont)
    {
      DeleteObject( (HFONT)this->asUnsigned() );
    }
    delete this;
  }
}


/*------------------------------------------------------------------------------
| CalcPointSize                                                                |
|                                                                              |
| Calculate the pointsize based on the height of the font                      |
| requested and the resolution of the device passed in                         |
------------------------------------------------------------------------------*/
long CalcPointSize( IPresSpaceHandle hps, long tmHeight, long tmIntlLeading )
{
  /*------------------------------------------------------------------*/
  /* Get device resolution for device passed for conversion to points */
  /*------------------------------------------------------------------*/
  int vyPelsPerLogInch = GetDeviceCaps( hps, LOGPIXELSY );

  /*------------------------------------------------------------------*/
  /* if the return value is zero, the problem may be caused by using  */
  /* a metafile device context.                                       */
  /*------------------------------------------------------------------*/
  if (vyPelsPerLogInch == 0)
  {
    IPresSpaceHandle desktopPs = IGETPS(HWND_DESKTOP);

    /*----------------------------------------------------------------*/
    /* Get device resolution for desktop for conversion to point size */
    /*----------------------------------------------------------------*/
    vyPelsPerLogInch = GetDeviceCaps( desktopPs, LOGPIXELSY );

    /*----------------------------------------------------------------*/
    /* Release any device context acquired                            */
    /*----------------------------------------------------------------*/
    IRELEASEPS( HWND_DESKTOP, desktopPs );
  }

  /*------------------------------------------------------------------*/
  /* compute the twip size                                            */
  /*------------------------------------------------------------------*/
  DWORD wTwipSize = (1440 * (tmHeight - tmIntlLeading)) / vyPelsPerLogInch;

  /*------------------------------------------------------------------*/
  /* Set the point size internally, rounding to the nearest point     */
  /*------------------------------------------------------------------*/
  return (long)((wTwipSize + 10) / 20);
}


/*------------------------------------------------------------------------------
| CalcFontHeight                                                               |
|                                                                              |
| Calculate the height of the font (in pixels) based on the point size         |
| requested and the resolution of the device passed in                         |
------------------------------------------------------------------------------*/
int CalcFontHeight( IPresSpaceHandle presSpaceHandle, unsigned long ulPointSize )
{
  /*------------------------------------------------------------------*/
  /* Get device context for desktop if no context passed              */
  /*------------------------------------------------------------------*/
  IPresSpaceHandle hps;
  if (presSpaceHandle == 0)
    hps = IGETPS(HWND_DESKTOP);
  else
    hps = presSpaceHandle;

  /*------------------------------------------------------------------*/
  /* Get device resolution for desktop for conversion to point size   */
  /*------------------------------------------------------------------*/
  int vyPelsPerLogInch = GetDeviceCaps( hps, LOGPIXELSY );

  /*------------------------------------------------------------------*/
  /* Release any device context acquired                              */
  /*------------------------------------------------------------------*/
  if (presSpaceHandle == 0)
    IRELEASEPS( HWND_DESKTOP, hps );

  /*------------------------------------------------------------------*/
  /* Calculate the height, converting based on device resolution and  */
  /* adding in a roundoff value.  Change the sign of final result to  */
  /* negative to indicate that this is a character height value, not  */
  /* a cell height value (i.e. no allowance for internal leading)     */
  /*------------------------------------------------------------------*/
  return -((((int)ulPointSize*vyPelsPerLogInch)+36)/72);
}
#endif


/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
|                                                                              |
| Construct a font for the font being used in an IWindow                       |
------------------------------------------------------------------------------*/
IFont :: IFont (const IWindow* pwindow)
         : ppd(new IFontPrivateData())
{
   IMODTRACE_DEVELOP("IFont-ctor(1)");

   IWindowHandle wh;
   if (pwindow == 0)
#ifdef IC_PM
      wh = HWND_DESKTOP;
#else
      wh = GetDesktopWindow();
#endif
   else
      wh = pwindow->handle();

   IPresSpaceHandle hps = IGETPS(wh);
#ifdef IC_PM
   /*******************************************************************/
   /* For most windows, we can get the font information needed from   */
   /* the presentation space handle                                   */
   /*******************************************************************/
   initialize(hps);

   IString fontName( 0, 120 );
   unsigned long ulPresParams;

   /*******************************************************************/
   /* If the window passed is non-null, then do some additional       */
   /* processing to set the font up correctly                         */
   /*******************************************************************/
   if (pwindow != 0)
   {
     /*****************************************************************/
     /* First check the presentation parameters for any additional    */
     /* attributes being used by the graphics engine to display font  */
     /*****************************************************************/
     if (WinQueryPresParam(pwindow->handle(),
                           PP_FONTNAMESIZE,
                           0,
                           &ulPresParams,
                           (ULONG)120,
                           (PVOID)(char*)fontName,
                           QPF_NOINHERIT) != 0)
     {
       if (fontName.includes(".Bold"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_BOLD;
       if (fontName.includes(".Italic"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_ITALIC;
       if (fontName.includes(".Outline"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_OUTLINE;
       if (fontName.includes(".Underscore"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_UNDERSCORE;
       if (fontName.includes(".Strikeout"))
         ppd->pfattrsCl->fsSelection |= FATTR_SEL_STRIKEOUT;
     }

     /*****************************************************************/
     /* Then check if window is an MLE control                        */
     /*****************************************************************/
     IWindowHandle hMle = pwindow->handle();
     IWindowClassName className( hMle );
     if ( className == WC_MLE )
     {
       /***************************************************************/
       /* If it is MLE, need to do some special case processing to    */
       /* create an IFont with the correct information since you      */
       /* cannot get at the hps for an MLE (as info set above is bad) */
       /***************************************************************/
       hMle.sendEvent( MLM_QUERYFONT,
                       IEventParameter1( ppd->pfattrsCl ),
                       IEventParameter2(0) );
       HDC hdc(GpiQueryDevice(hps));
       long vRes;
       DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &vRes);

       // Add 0.5 before doing integer conversion to avoid truncation errors
       unsigned long currPointSize = FIXEDINT(
        (( MAKEFIXED(ppd->pfattrsCl->lMaxBaselineExt, 0 ) * 72) / vRes)+0x8000);
       ppd->fxClPointSize = MAKEFIXED( currPointSize, 0 );
     }
   }
#endif
#ifdef IC_WIN
   // Do a check for a canvas window (since Windows does not allow fonts
   // to be set for this window class)
   IEventResult er = wh.sendEvent( IC_UM_REQUEST_CANVASFONT,
                                   IEventParameter1( 0 ),
                                   IEventParameter2( 0 ) );
   IFont* canvasFont = (IFont*)er.asUnsignedLong();

   // If so, then just copy the private data over from the canvas object
   if (canvasFont)
   {
     canvasFont->ppd->addRef();
     ppd->removeRef();
     ppd = canvasFont->ppd;
   }
   else
   {
     /*---------------------------------------------------------------*/
     /* Allocate a structure for storing the font information         */
     /*---------------------------------------------------------------*/
     ppd->plogfontCl = new LOGFONT;

     /*---------------------------------------------------------------*/
     /* Query the window for the font handle in use                   */
     /*---------------------------------------------------------------*/
     HFONT hFont = wh.sendEvent( WM_GETFONT,
                                 IEventParameter1( 0 ),
                                 IEventParameter2( 0 ) );

     /*---------------------------------------------------------------*/
     /* If NULL, then system font is in use, so get the handle for it */
     /*---------------------------------------------------------------*/
     if (!hFont)
       hFont = ppd->systemFont;

     /*---------------------------------------------------------------*/
     /* Finally get the font information for this handle and then     */
     /*  create a new font with this information                      */
     /*---------------------------------------------------------------*/
     int rc = GetObject( hFont, sizeof(LOGFONT), ppd->plogfontCl );
     if (!rc)
       GetObject( ppd->systemFont, sizeof(LOGFONT),
                  ppd->plogfontCl );
     makeFontMetrics( hps );
     ITRACE_ALL( IString("Font handle is: ") +
                 IString( *(ppd->fpFontHandle) ));
   }
#endif
   IRELEASEPS(wh,hps);
}

/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
|                                                                              |
| Construct font based an input parameters.                                    |
------------------------------------------------------------------------------*/
IFont :: IFont         (const char*             pszFaceName,
                        unsigned long           ulPointSize,
                        Boolean                 useFixedFont,
                        Boolean                 useVectorFont,
                        const IPresSpaceHandle& presSpaceHandle)
         : ppd(new IFontPrivateData())
{
   IMODTRACE_DEVELOP("IFont-ctor(2)");
#ifdef IC_PM
   /* initialize font attributes structure */
   ppd->pfattrsCl = new FATTRS;
   memset(ppd->pfattrsCl, 0, sizeof(FATTRS));
   (ppd->pfattrsCl)->usRecordLength = sizeof(FATTRS);

// If setName(), setPointSize(), useNonPropOnly(), useBitmapOnly(),
// and useVectorOnly() are called from here we'll end up calling
// checkForMatch excesively.
// So just do what these functions do, but don't call fontChanged()
// and checkForMatch() until the end.

     strcpy((ppd->pfattrsCl)->szFacename, pszFaceName);

     ppd->fxClPointSize = MAKEFIXED(ulPointSize, 0);
     if(useVectorFont) {
       setCharHeight(ulPointSize, presSpaceHandle);
       setCharWidth(ulPointSize, presSpaceHandle);
     }

     if (useFixedFont)
        ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;
     else
        ppd->flClStyles = ppd->flClStyles & ~IC_NONPROPORTIONAL_FONTS_ONLY;

   if (useVectorFont)
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
   else
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;

   if (!useVectorFont) {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      mustUseVector(FALSE);
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      mustUseVector(TRUE);
   } /* endif */

   checkForMatch(presSpaceHandle);

   // Insure that the fontmetrics information matches the font selected
   // by creating the font temporarily into the hps passed and then
   // querying for the accurate information (not the general info)
   if ( ppd->pfontmetCl )
   {
     delete ppd->pfontmetCl;
     ppd->pfontmetCl = 0;
   }
   makeFontMetrics( presSpaceHandle );
#endif
#ifdef IC_WIN
   /*------------------------------------------------------------------*/
   /* Create a new logfont structure, null it out and copy facename    */
   /*------------------------------------------------------------------*/
   ppd->plogfontCl = new LOGFONT;
   memset(ppd->plogfontCl, 0, sizeof(LOGFONT));
   strcpy((ppd->plogfontCl)->lfFaceName, pszFaceName);

   /*------------------------------------------------------------------*/
   /* Calculate height from point size, and set other values to default*/
   /*------------------------------------------------------------------*/
   (ppd->plogfontCl)->lfHeight =
                         CalcFontHeight( presSpaceHandle, ulPointSize );
   (ppd->plogfontCl)->lfWeight = FW_DONTCARE;
   (ppd->plogfontCl)->lfQuality = DEFAULT_QUALITY;
   (ppd->plogfontCl)->lfPitchAndFamily = FF_DONTCARE;

   /*------------------------------------------------------------------*/
   /* If fixed pitch specified, then set flag to denote this           */
   /*------------------------------------------------------------------*/
   if (useFixedFont)
     (ppd->plogfontCl)->lfPitchAndFamily |= FIXED_PITCH;

   /*------------------------------------------------------------------*/
   /* If vector font specified, then set flag to denote this           */
   /*------------------------------------------------------------------*/
   if (useVectorFont)
     (ppd->plogfontCl)->lfPitchAndFamily |= TMPF_TRUETYPE;

   /*------------------------------------------------------------------*/
   /* Finally, create the font from the logfont structure set-up       */
   /*------------------------------------------------------------------*/
   makeFontMetrics( presSpaceHandle );
   ITRACE_ALL( IString("Font handle is: ") +
               IString( *(ppd->fpFontHandle) ));
#endif
}

/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
| Construct font to one being used in Presentation Space                       |
------------------------------------------------------------------------------*/
IFont :: IFont(const IPresSpaceHandle& presSpace)
         : ppd(new IFontPrivateData())
{
   IMODTRACE_DEVELOP("IFont-ctor(3)");
   initialize(presSpace);
}

/*------------------------------------------------------------------------------
| IFont::IFont                                                                 |
| Copy constructor                                                             |
------------------------------------------------------------------------------*/
IFont :: IFont (const IFont& fntmCopy)
{
  IMODTRACE_DEVELOP("IFont-copy ctor");
  fntmCopy.ppd->addRef();
  ppd = fntmCopy.ppd;
}


/*------------------------------------------------------------------------------
| IFont::~IFont                                                                |
------------------------------------------------------------------------------*/
IFont :: ~IFont ()
{
   IMODTRACE_DEVELOP("IFont-dtor");
   ppd->removeRef();
   ppd = 0;
}

/*------------------------------------------------------------------------------
| IFont::operator=                                                             |
------------------------------------------------------------------------------*/
IFont& IFont::operator= (const IFont &aFont)
{
  if (this == &aFont )
    return *this;
  aFont.ppd->addRef();
  ppd->removeRef();
  ppd = aFont.ppd;
  return (*this);
}

/*------------------------------------------------------------------------------
| IFont::useNonPropOnly                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: useNonPropOnly (Boolean nonProportionalOnly,
                                const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (nonProportionalOnly) {
      ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;
      if (!isFixed()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_NONPROPORTIONAL_FONTS_ONLY;
      fontChanged();
   } /* endif */
#ifdef IC_PM
   checkForMatch(presSpaceHandle);
#else
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::useBitmapOnly                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: useBitmapOnly (Boolean bitmapOnly,
                              const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (bitmapOnly) {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      mustUseVector(FALSE);
      if (!isBitmap()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      fontChanged();
      mustUseVector(FALSE);
   } /* endif */
#ifdef IC_PM
   checkForMatch(presSpaceHandle);
#else
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::useVectorOnly                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: useVectorOnly (Boolean vectorOnly,
                               const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   if (vectorOnly) {
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      mustUseVector(TRUE);
      if (isBitmap()) {
         fontChanged();
      } /* endif */
   } else {
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      fontChanged();
      mustUseVector(FALSE);
   } /* endif */
#ifdef IC_PM
   checkForMatch(presSpaceHandle);
#else
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setDirection                                                          |
|                                                                              |
| This method only affects text drawn using beginUsingFont.                    |
------------------------------------------------------------------------------*/
IFont& IFont :: setDirection( Direction lFontDir )
{
   copyPrivate();
   switch (lFontDir) {
      case leftToRight:
         ppd->lClFontDir = CHDIRN_LEFTRIGHT;
         break;
      case rightToLeft:
         ppd->lClFontDir = CHDIRN_RIGHTLEFT;
         break;
      case topToBottom:
         ppd->lClFontDir = CHDIRN_TOPBOTTOM;
         break;
      case bottomToTop:
         ppd->lClFontDir = CHDIRN_BOTTOMTOP;
         break;
      case defaultDir:
         ppd->lClFontDir = CHDIRN_DEFAULT;
         break;
   } /* endswitch */
#ifdef IC_WIN
   (ppd->plogfontCl)->lfEscapement = ppd->lClFontDir;
    fontChanged();
    makeFontMetrics();
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::setName                                                               |
| Set the faceName of the font.                                                |
------------------------------------------------------------------------------*/
IFont& IFont :: setName (const char* pszFaceName,
                         const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
#ifdef IC_PM
   strcpy((ppd->pfattrsCl)->szFacename, pszFaceName);
#else
   strcpy((ppd->plogfontCl)->lfFaceName, pszFaceName);
#endif
   fontChanged();
#ifdef IC_PM
   checkForMatch(presSpaceHandle);
#else
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setPointSize                                                          |
------------------------------------------------------------------------------*/
IFont& IFont ::  setPointSize (unsigned long ulSize,
                               const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
#ifdef IC_PM
   ppd->fxClPointSize = MAKEFIXED(ulSize, 0);
   if (isBitmapOnly()) {
      fontChanged();
      checkForMatch(presSpaceHandle);
   } else {
      // As a performance boost for a vector font just need
      // to change the size and not try querying all fonts
      // for a closest match
      setCharHeight(ulSize, presSpaceHandle);
      setCharWidth(ulSize, presSpaceHandle);
   } /* endif */
#else
   (ppd->plogfontCl)->lfHeight = CalcFontHeight( presSpaceHandle, ulSize );
   (ppd->plogfontCl)->lfWidth = 0;
   fontChanged();
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


#ifdef IC_PM
/*------------------------------------------------------------------------------
| IFont::setSizeFromHps                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: setSizeFromHps (unsigned long ulSize,
                                const IPresSpaceHandle& presSpaceHandle)
{
   copyPrivate();
   ppd->fxClPointSize = MAKEFIXED(ulSize, 0);
   setWidthHeight(ulSize, presSpaceHandle);
   makeCharWidths(presSpaceHandle);
   return *this;
}
#endif // IC_PM

/*------------------------------------------------------------------------------
| IFont::setBold                                                               |
------------------------------------------------------------------------------*/
IFont& IFont :: setBold(Boolean bold)
{
    copyPrivate();
    if (bold)
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_BOLD;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_BOLD;
#endif
#ifdef IC_WIN
       ppd->plogfontCl->lfWeight = FW_BOLD;
    else
       ppd->plogfontCl->lfWeight = FW_REGULAR;
    fontChanged();
    makeFontMetrics();
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setItalic                                                             |
------------------------------------------------------------------------------*/
IFont& IFont :: setItalic(Boolean italics)
{
    copyPrivate();
    if (italics)
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_ITALIC;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_ITALIC;
#endif
#ifdef IC_WIN
       ppd->plogfontCl->lfItalic = 0xFF;
    else
       ppd->plogfontCl->lfItalic = 0;
    fontChanged();
    makeFontMetrics();
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setUnderscore                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: setUnderscore(Boolean underscore)
{
    copyPrivate();
    if (underscore)
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_UNDERSCORE;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_UNDERSCORE;
#endif
#ifdef IC_WIN
       ppd->plogfontCl->lfUnderline = 0xFF;
    else
       ppd->plogfontCl->lfUnderline = 0;
    fontChanged();
    makeFontMetrics();
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setStrikeout                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setStrikeout(Boolean strikeout)
{
    copyPrivate();
    if (strikeout)
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_STRIKEOUT;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_STRIKEOUT;
#endif
#ifdef IC_WIN
       ppd->plogfontCl->lfStrikeOut = 0xFF;
    else
       ppd->plogfontCl->lfStrikeOut = 0;
    fontChanged();
    makeFontMetrics();
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setOutline                                                            |
------------------------------------------------------------------------------*/
IFont& IFont :: setOutline(Boolean outline)
{
#ifdef IC_PM
    copyPrivate();
    if (outline)
       ppd->pfattrsCl->fsSelection |= FATTR_SEL_OUTLINE;
    else
       ppd->pfattrsCl->fsSelection &= ~FATTR_SEL_OUTLINE;
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::setAllEmphasis                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: setAllEmphasis(Boolean setOn)
{
    copyPrivate();
    if (setOn)
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection = (FATTR_SEL_BOLD |
                                      FATTR_SEL_ITALIC |
                                      FATTR_SEL_UNDERSCORE |
                                      FATTR_SEL_STRIKEOUT |
                                      FATTR_SEL_OUTLINE);
#endif
#ifdef IC_WIN
    {
      ppd->plogfontCl->lfWeight = FW_BOLD;
      ppd->plogfontCl->lfItalic = 0xFF;
      ppd->plogfontCl->lfUnderline = 0xFF;
      ppd->plogfontCl->lfStrikeOut = 0xFF;
    }
#endif
    else
#ifdef IC_PM
       ppd->pfattrsCl->fsSelection = 0;
#endif
#ifdef IC_WIN
    {
      ppd->plogfontCl->lfWeight = FW_REGULAR;
      ppd->plogfontCl->lfItalic = 0;
      ppd->plogfontCl->lfUnderline = 0;
      ppd->plogfontCl->lfStrikeOut = 0;
    }
    fontChanged();
    makeFontMetrics();
#endif
    return *this;
}

/*------------------------------------------------------------------------------
| IFont::isBitmap                                                              |
------------------------------------------------------------------------------*/
Boolean IFont :: isBitmap () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (((ppd->pfontmetCl)->fsDefn & FM_DEFN_OUTLINE) == 0);
#endif
#ifdef IC_WIN
   return (!((ppd->ptextmetricCl)->tmPitchAndFamily & TMPF_VECTOR));
#endif
}

/*------------------------------------------------------------------------------
| IFont::isFixed                                                               |
------------------------------------------------------------------------------*/
Boolean IFont :: isFixed () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (Boolean)((ppd->pfontmetCl)->fsType & FM_TYPE_FIXED);
#endif
#ifdef IC_WIN
   return (!((ppd->ptextmetricCl)->tmPitchAndFamily & TMPF_FIXED_PITCH));
#endif
}


/*------------------------------------------------------------------------------
| IFont::name                                                                  |
------------------------------------------------------------------------------*/
IString IFont :: name() const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   IString strTemp((char*)((ppd->pfontmetCl)->szFacename));
#endif
#ifdef IC_WIN
   IString strTemp((char*)((ppd->plogfontCl)->lfFaceName));
#endif
   return strTemp;
}


/*------------------------------------------------------------------------------
| IFont::isNonPropOnly                                                         |
------------------------------------------------------------------------------*/
Boolean IFont :: isNonPropOnly () const
{
   return (ppd->flClStyles & IC_NONPROPORTIONAL_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::isBitmapOnly                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isBitmapOnly () const
{
   return (ppd->flClStyles & IC_BITMAP_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::isVectorOnly                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isVectorOnly () const
{
   return (ppd->flClStyles & IC_VECTOR_FONTS_ONLY);
}


/*------------------------------------------------------------------------------
| IFont::pointSize                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: pointSize () const
{
#ifdef IC_PM
   return FIXEDINT(ppd->fxClPointSize);
#endif
#ifdef IC_WIN
  return (ppd->fxClPointSize);
#endif
}


/*------------------------------------------------------------------------------
| IFont::isBold                                                                |
------------------------------------------------------------------------------*/
Boolean IFont :: isBold ( ) const
{
#ifdef IC_PM
   return (ppd->pfattrsCl->fsSelection & FATTR_SEL_BOLD);
#endif
#ifdef IC_WIN
   return ((ppd->ptextmetricCl)->tmWeight > FW_MEDIUM );
#endif
}

/*------------------------------------------------------------------------------
| IFont::isItalic                                                              |
------------------------------------------------------------------------------*/
Boolean IFont :: isItalic ( ) const
{
#ifdef IC_PM
   return (ppd->pfattrsCl->fsSelection & FATTR_SEL_ITALIC);
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmItalic;
#endif
}

/*------------------------------------------------------------------------------
| IFont::isUnderscore                                                          |
------------------------------------------------------------------------------*/
Boolean IFont :: isUnderscore ( ) const
{
#ifdef IC_PM
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_UNDERSCORE ;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmUnderlined;
#endif
}

/*------------------------------------------------------------------------------
| IFont::isStrikeout                                                           |
------------------------------------------------------------------------------*/
Boolean IFont :: isStrikeout ( ) const
{
#ifdef IC_PM
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_STRIKEOUT ;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmStruckOut;
#endif
}

/*------------------------------------------------------------------------------
| IFont::isOutline                                                             |
------------------------------------------------------------------------------*/
Boolean IFont :: isOutline ( ) const
{
#ifdef IC_PM
   return ppd->pfattrsCl->fsSelection & FATTR_SEL_OUTLINE ;
#endif
#ifdef IC_WIN
   return false;
#endif
}


/*------------------------------------------------------------------------------
| IFont::avgCharWidth                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgCharWidth() const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lAveCharWidth;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmAveCharWidth;
#endif
}


/*------------------------------------------------------------------------------
| IFont::maxCharHeight                                                         |
------------------------------------------------------------------------------*/
unsigned long  IFont :: maxCharHeight () const
{
#ifdef IC_PM
   return maxSize().height();
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmHeight;
#endif
}

/*------------------------------------------------------------------------------
| IFont::maxSize                                                               |
------------------------------------------------------------------------------*/
ISize IFont :: maxSize () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lMaxCharInc, ppd->pfontmetCl->lMaxBaselineExt);
#endif
#ifdef IC_WIN
   ISize size( (ppd->ptextmetricCl)->tmMaxCharWidth,
               (ppd->ptextmetricCl)->tmHeight );
#endif
   return size;
}

/*------------------------------------------------------------------------------
| IFont::maxUppercaseSize                                                      |
------------------------------------------------------------------------------*/
ISize IFont :: maxUppercaseSize( ) const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lEmInc, ppd->pfontmetCl->lMaxBaselineExt);
   return size;
#endif
#ifdef IC_WIN
   return maxSize();   // Set this the same as maxSize since we cannot query
                       // the max uppercase width (but assume it is largest)
#endif
}

#ifdef IC_PM
/*------------------------------------------------------------------------------
| IFont::superscriptSize                                                       |
------------------------------------------------------------------------------*/
ISize IFont :: superscriptSize( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSuperscriptXSize,
              ppd->pfontmetCl->lSuperscriptYSize);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::subscriptSize                                                         |
------------------------------------------------------------------------------*/
ISize IFont :: subscriptSize( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSubscriptXSize,
              ppd->pfontmetCl->lSubscriptYSize);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::superscriptOffset                                                     |
------------------------------------------------------------------------------*/
ISize IFont :: superscriptOffset( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSuperscriptXOffset,
              ppd->pfontmetCl->lSuperscriptYOffset);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::subscriptOffset                                                       |
------------------------------------------------------------------------------*/
ISize IFont :: subscriptOffset( ) const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   ISize size(ppd->pfontmetCl->lSubscriptXOffset,
              ppd->pfontmetCl->lSubscriptYOffset);
   return size;
}

/*------------------------------------------------------------------------------
| IFont::avgLowercase                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgLowercase () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lXHeight;

}

/*------------------------------------------------------------------------------
| IFont::avgUppercase                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: avgUppercase () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lEmHeight;

}

/*------------------------------------------------------------------------------
| IFont::maxLowercaseAscender                                                  |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxLowercaseAscender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lLowerCaseAscent;

}

/*------------------------------------------------------------------------------
| IFont::maxLowercaseDescender                                                 |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxLowercaseDescender () const
{
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return ppd->pfontmetCl->lLowerCaseDescent;

}
#endif

/*------------------------------------------------------------------------------
| IFont::maxDescender                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxDescender () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lMaxDescender;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmDescent;
#endif
}

/*------------------------------------------------------------------------------
| IFont::maxAscender                                                           |
------------------------------------------------------------------------------*/
unsigned long IFont :: maxAscender () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lMaxAscender;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmAscent;
#endif
}

/*------------------------------------------------------------------------------
| IFont::externalLeading                                                       |
------------------------------------------------------------------------------*/
unsigned long IFont :: externalLeading () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lExternalLeading;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmExternalLeading;
#endif
}

/*------------------------------------------------------------------------------
| IFont::internalLeading                                                       |
------------------------------------------------------------------------------*/
unsigned long IFont :: internalLeading () const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeFontMetrics();
   return (ppd->pfontmetCl)->lInternalLeading;
#endif
#ifdef IC_WIN
   return (ppd->ptextmetricCl)->tmInternalLeading;
#endif
}


/*------------------------------------------------------------------------------
| IFont::charWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: charWidth (char c) const
{
#ifdef IC_PM
   IFont* const localThis = (IFont* const)this;
   localThis->makeCharWidths();
   return ppd->plClCharWidths[c];
#endif
#ifdef IC_WIN
   HDC hdc = IGETPS(HWND_DESKTOP);
   int buffer = 0;

   HFONT tempFont = SelectObject( hdc, *(ppd->fpFontHandle) );
   GetCharWidth( hdc, c, c, &buffer );
   SelectObject( hdc, tempFont );

   IRELEASEPS(HWND_DESKTOP,hdc);
   return buffer;
#endif
}


/*------------------------------------------------------------------------------
| IFont::textWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long  IFont :: textWidth (const char* pszText)  const
{
#ifdef IC_PM
   unsigned long ulWidth = 0;
   if (pszText) {
      IFont* const localThis = (IFont* const)this;
      localThis->makeCharWidths();
      const char* Temp = pszText;
      while (*Temp) {
         ulWidth += ppd->plClCharWidths[*Temp];
         Temp++;
      } /* endwhile */
   } /* endif */
   return ulWidth;
#endif
#ifdef IC_WIN
   HDC hdc = IGETPS(HWND_DESKTOP);
   IString tempString(pszText);
   SIZE size;

   HFONT tempFont = SelectObject( hdc, *(ppd->fpFontHandle) );
   GetTextExtentPoint( hdc, tempString, tempString.length(), &size );
   SelectObject( hdc, tempFont );

   IRELEASEPS(HWND_DESKTOP,hdc);
   return size.cx;
#endif
}

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IFont::textWidth                                                             |
------------------------------------------------------------------------------*/
unsigned long  IFont :: textWidth (const char*    pszText,
                                   IBidiSettings& settings)  const
{
   IString tmpText(pszText);
#ifdef IC_PM
   LAYOUT_OBJECT layout = bidiCreateLayoutObject();
   unsigned long tilde = 0;
   bidiPrepareForDisplay(layout,tmpText,settings,tilde);
   bidiDeleteLayoutObject(layout);
#endif
   return textWidth(tmpText);
}
#endif

/*------------------------------------------------------------------------------
| IFont::textLines                                                             |
------------------------------------------------------------------------------*/
unsigned long IFont :: textLines    (const char* pszLine,
                                     unsigned long ulLineWidth) const
{
   unsigned long ulLines = 0;
   if (ulLineWidth  &&
       this->minTextWidth( pszLine ) <= ulLineWidth)
   {
      ulLines = 1;
      unsigned long ulWidth = 0;
      char* LastSpace = 0;
      char* Ptr = (char*)pszLine;
      while (*Ptr) {
         if (ulWidth + charWidth( *Ptr ) > ulLineWidth)
         {                   // More text than will fit.
            if (isspace( *Ptr ))
            {                // Space between lines.
               LastSpace = Ptr;
            }
            else if (LastSpace)
            {                // Back up to start of word.
               Ptr = LastSpace;
            }
            Ptr++;
            ulLines++;       // Need another line.
            ulWidth = 0;
         }
         else if (*Ptr == '\n')
         {                   // New line character (fits on this line).
            Ptr++;
            if (*Ptr)
            {                // More characters for another line.
               ulLines++;    // Need another line.
               ulWidth = 0;
            }
         }
         else
         {                   // Adding to current line.
            ulWidth += charWidth(*Ptr);
            if (isspace(*Ptr)) {
               LastSpace = Ptr;
            }
            Ptr++;
         }
      } /* endwhile */
   } /* endif */
   return ulLines;
}


/*------------------------------------------------------------------------------
| IFont::minTextWidth                                                          |
------------------------------------------------------------------------------*/
unsigned long IFont :: minTextWidth     (const char* pszLine) const
{
   if (pszLine) {
      char* Temp = new char[strlen(pszLine)+1];
      strcpy(Temp, pszLine);
      char* Start = Temp;
      unsigned long ulWidth = 0;
      while (*Start) {
         while (*Start && isspace(*Start)) {
            Start++;
         } /* endwhile */
         char* End = Start;
         while (*End && !isspace(*End)) {
            End++;
         } /* endwhile */
         char SaveCh = *End;
         *End = 0;
         unsigned long ulChkWidth = textWidth(Start);
         if (ulChkWidth > ulWidth) {
            ulWidth = ulChkWidth;
         } /* endif */
         *End = SaveCh;
         Start = End;
      } /* endwhile */
      delete [] Temp;
      return ulWidth;
   } else {
      return 0;
   } /* endif */
}


/*------------------------------------------------------------------------------
| IFont::beginUsingFont                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: beginUsingFont (const IPresSpaceHandle& hps)
{
   IMODTRACE_DEVELOP("IFont::beginUsingFont");
#ifdef IC_PM
   if (!ppd->lClLCId) {
      ppd->lClLCId = 1;

      // Have to make sure our local Id is not already in use
      // First, get the total number of local Ids in use
      long NumLcId = GpiQueryNumberSetIds(hps);

      // Check if there are any in use
      if (NumLcId) {
         long Types[255];
         STR8 Names[255];
         long LcIds[255];

         // Get information about all local Ids
         Boolean RC = GpiQuerySetIds(hps, NumLcId, Types, Names, LcIds);
         if (!RC) {
            ITRACE_DEVELOP("Error getting ids.");
            ITHROWGUIERROR("GpiQuerySetIds");
         } /* endif */

         // Check all used Ids to see if ours is already used
         for (unsigned long LCNum = 0; LCNum < NumLcId ; LCNum++ ) {
            if (LcIds[LCNum] == ppd->lClLCId) {
               // Ours is already used, increment to next Id and restart search
               ppd->lClLCId++;
               LCNum = -1;

               // NewLCId must be in the range 1 to 254
               if (ppd->lClLCId > 254) {
                  // Error! No more free ids!
                  ITRACE_DEVELOP("Error no free ids.");
                  ITHROWGUIERROR("GpiQuerySetIds");
               } /* endif */
            } /* endif */
         } /* endfor */

      } /* endif */
   } /* endif */

   LONG match = GpiCreateLogFont(hps, 0, ppd->lClLCId, ppd->pfattrsCl);
   if (match==GPI_ERROR) {
      ITHROWGUIERROR("GpiCreateLogFont");
   } /* endif */
   if (!GpiSetCharSet(hps, ppd->lClLCId)) {
      ITHROWGUIERROR("GpiSetCharSet");
   }
          /* Check if the font found is an outline (vector) font      */
   if ((ppd->pfattrsCl)->fsFontUse & FATTR_FONTUSE_OUTLINE) {
            /* Set the Character mode so that the characters will     */
            /* be stretched properly.                                 */
      GpiSetCharMode(hps, CM_MODE2);

      SIZEF sizef;
      sizef.cx = ppd->fxClWidth;
      sizef.cy = ppd->fxClHeight;
            /* Set character box for the outline font to size above   */
      GpiSetCharBox( hps, &sizef );
      GRADIENTL gradlAngle;
      gradlAngle.x = ppd->pntClAngle.x();
      gradlAngle.y = ppd->pntClAngle.y();
      GpiSetCharAngle(hps, &gradlAngle);
      POINTL shearAngle = ppd->pntClShear.asPOINTL();
      GpiSetCharShear(hps, &shearAngle);
   } /* endif */
   else if ( ppd->pfontmetCl )
   {
      GpiSetCharMode(hps, CM_MODE2);
      SIZEF sizef = { MAKEFIXED(ppd->pfontmetCl->lEmInc,0),
                      MAKEFIXED(ppd->pfontmetCl->lEmHeight,0) };
      GpiSetCharBox( hps, &sizef );
   }
   GpiSetCharDirection(hps, ppd->lClFontDir);
#endif
#ifdef IC_WIN
   /*******************************************************************/
   /* Select our font into the device context for drawing operations, */
   /* saving the previous selected font to be restored on completion  */
   /*******************************************************************/
   HFONT prevFont = SelectObject( hps, *(ppd->fpFontHandle) );
   // If the font returned from SelectObject is the same as the font
   // being set into the device context, then another IFont reference
   // has already done a beginUsingFont.  In this case, increment the
   // inUseCount.  Otherwise, save the device context's previous font
   // handle.
   ppd->inUseCount++;
   if (prevFont != *(ppd->fpFontHandle))
   {
      ppd->prevFont = prevFont;
   }
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::endUsingFont                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: endUsingFont (const IPresSpaceHandle& hps)
{
   IMODTRACE_DEVELOP("IFont::endUsingFont");
#ifdef IC_PM
   GpiSetCharSet(hps, LCID_DEFAULT);
   if (ppd->lClLCId) {
      GpiDeleteSetId(hps, ppd->lClLCId);
      ppd->lClLCId = 0;
   } /* endif */
#endif
#ifdef IC_WIN
   /*******************************************************************/
   /* Reselect the original font back into the device context if      */
   /* the inUseCount is zero.  Otherwise, decrement the inUseCount.   */
   /*******************************************************************/
   ppd->inUseCount--;
   if (ppd->inUseCount == 0)
   {
      SelectObject( hps, ppd->prevFont );
   }
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IFont::makeFontMetrics                                                       |
------------------------------------------------------------------------------*/
IFont& IFont::makeFontMetrics( const IPresSpaceHandle& presSpaceHandle,
                               Boolean fUpdateCharSet )
{
   IMODTRACE_DEVELOP("IFont::makeFontMetrics");
#ifdef IC_PM
   if (!ppd->pfontmetCl) {
      IPresSpaceHandle hps;
      if(presSpaceHandle)
        hps = presSpaceHandle;
      else
        hps = WinGetPS(HWND_DESKTOP);
      beginUsingFont(hps);
      ppd->pfontmetCl = new FONTMETRICS;
      GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), ppd->pfontmetCl);
      endUsingFont(hps);
      if (presSpaceHandle == 0)
        WinReleasePS(hps);
   } /* endif */
#endif
#ifdef IC_WIN
   if (fUpdateCharSet)
     ppd->plogfontCl->lfCharSet = ppd->fcharSet;
   HFONT hFont = CreateFontIndirect( ppd->plogfontCl );
   ppd->fpFontHandle->removeRef();
   ppd->fpFontHandle = new IFontHandle( hFont, true );
   ppd->ptextmetricCl = new TEXTMETRIC;

   IPresSpaceHandle hps;
   if (presSpaceHandle == 0)
      hps = IGETPS(HWND_DESKTOP);
   else
      hps = presSpaceHandle;

   HFONT tempFont = SelectObject( hps, *(ppd->fpFontHandle) );
   GetTextMetrics( hps, ppd->ptextmetricCl );
   SelectObject( hps, tempFont );

   /*------------------------------------------------------------------*/
   /* Calculate and save the new point size for the font               */
   /*------------------------------------------------------------------*/
   ppd->fxClPointSize = CalcPointSize( hps, ppd->ptextmetricCl->tmHeight,
                                  ppd->ptextmetricCl->tmInternalLeading);

   if(presSpaceHandle == 0)
     IRELEASEPS(HWND_DESKTOP,hps);
#endif
   return *this;
}


#ifdef IC_PM
/*------------------------------------------------------------------------------
| IFont::makeCharWidths                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: makeCharWidths (const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::makeCharWidths");
   if (!ppd->plClCharWidths) {
      IPresSpaceHandle hps;
      if (presSpaceHandle)
        hps = presSpaceHandle;
      else
        hps = WinGetPS(HWND_DESKTOP);
      beginUsingFont(hps);
      ppd->plClCharWidths = new long[iCharArrayWidth];
      GpiQueryWidthTable(hps, 0, iCharArrayWidth, ppd->plClCharWidths);
      endUsingFont(hps);
      if (presSpaceHandle == 0)
        WinReleasePS(hps);
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::checkForMatch                                                         |
|                                                                              |
| Search for the closest match in all installed fonts.                         |
------------------------------------------------------------------------------*/
IFont& IFont :: checkForMatch (const IPresSpaceHandle& presSpaceHandle)
{

   unsigned long ulPtSize = FIXEDINT(ppd->fxClPointSize);
   long lNumFonts;
   long lReqFonts = 0;
   long lBitmapMatch = 0;
   long lNextBitmap = 0;
   long lVectorMatch = 0;
   long lHowClose = 99999;
   unsigned long
     Index = 0,
     BitmapIndex = 0,
     VectorIndex = 0,
     NextBitmapIndex = 0;

   IPresSpaceHandle hps;
   if (presSpaceHandle) {
      hps = presSpaceHandle;
   } else {
      hps = WinGetPS(HWND_DESKTOP);
   } /* endif */

   lNumFonts = GpiQueryFonts(hps, QF_PUBLIC | QF_PRIVATE,
                  (PSZ)(ppd->pfattrsCl)->szFacename, &lReqFonts, 0L, 0);

   if (lNumFonts) {
      ITRACE_DEVELOP(IString(lNumFonts) + IString("Fonts found."));

      FONTMETRICS* pfm;

      pfm = new FONTMETRICS[lNumFonts];

      GpiQueryFonts(hps, QF_PUBLIC | QF_PRIVATE, (PSZ)(ppd->pfattrsCl)->szFacename,
         &lNumFonts, sizeof(FONTMETRICS), pfm);
      ITRACE_DEVELOP((IString(lNumFonts) + IString("Fonts found for ")) +
                      IString(ppd->pfattrsCl->szFacename) );
      long lHorzRes, lVertRes;
      HDC hdc = GpiQueryDevice(hps);
      DevQueryCaps(hdc, CAPS_HORIZONTAL_FONT_RES, 1L, &lHorzRes);
      DevQueryCaps(hdc, CAPS_VERTICAL_FONT_RES, 1L, &lVertRes);
      ITRACE_DEVELOP((IString("H&V resolution ") + IString(lHorzRes)) +
                      IString(lVertRes));

      for (Index = 0; Index < lNumFonts ; Index++)
      {
         // Skip font if it doesn't match the Proportional/Fixed specification
         // or the Vector/Bitmap specification
         if ( ((isNonPropOnly()) && (pfm[Index].fsType & FM_TYPE_FIXED))   ||
              ((!isNonPropOnly()) && (!(pfm[Index].fsType & FM_TYPE_FIXED))) ||
              ((isVectorOnly()) && (pfm[Index].fsDefn & FM_DEFN_OUTLINE))  ||
              ((isBitmapOnly()) && (!(pfm[Index].fsDefn & FM_DEFN_OUTLINE))) )
         {
            // Vector font doesn't need to check device resolution and point size
            if (pfm[Index].fsDefn & FM_DEFN_OUTLINE)
            {
               lVectorMatch = pfm[Index].lMatch;
               VectorIndex = Index;
            }
            else
              // Check bitmap for device resolution and pointSize
              if ((pfm[Index].sXDeviceRes == lHorzRes) &&
                  (pfm[Index].sYDeviceRes == lVertRes))
              {
                ITRACE_DEVELOP((IString(pfm[Index].sNominalPointSize) +
                                IString(" vs. ")) + IString(ulPtSize*10) );
                if (pfm[Index].sNominalPointSize == (ulPtSize*10))
                {
                   lBitmapMatch = pfm[Index].lMatch;
                   BitmapIndex = Index;
                }
                else
                {
                   long currentHowClose =
                       abs((ulPtSize*10) - pfm[Index].sNominalPointSize);
                   if (!lBitmapMatch &&
                       currentHowClose <= lHowClose)
                   {
                       lNextBitmap = pfm[Index].lMatch;
                       NextBitmapIndex = Index;
                       lHowClose = currentHowClose;
                   } /* endif */
                }
              }
         } /* endif for match on specification */

      } /* endfor */

      // We'll reset Index below to the position in pfm that's the best match.
      // That way we can save the FONTMETRICS.
      if (isBitmapOnly()) {
         if (!lBitmapMatch) {
            lBitmapMatch = lNextBitmap;
         } /* endif */
         (ppd->pfattrsCl)->lMatch = lBitmapMatch;
         ITRACE_DEVELOP("Exact bitmap match found.");
         (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
         Index = BitmapIndex;
      } else {
         if (isVectorOnly() || (ppd->flClStyles & IC_MUSTUSEVECTORFONT)) {
            (ppd->pfattrsCl)->lMatch = lVectorMatch;
            ITRACE_DEVELOP("Exact vector match found.");
            (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse | FATTR_FONTUSE_OUTLINE;
            Index = VectorIndex;
         } else {
            if (lBitmapMatch) {
               (ppd->pfattrsCl)->lMatch = lBitmapMatch;
               (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
               Index = BitmapIndex;
            } else {
               if (lVectorMatch) {
                  (ppd->pfattrsCl)->lMatch = lVectorMatch;
                  (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse | FATTR_FONTUSE_OUTLINE;
                  Index = VectorIndex;
               } else {
                  (ppd->pfattrsCl)->lMatch = lNextBitmap;
                  (ppd->pfattrsCl)->fsFontUse = (ppd->pfattrsCl)->fsFontUse & ~FATTR_FONTUSE_OUTLINE;
                  Index = NextBitmapIndex;
               } /* endif */
            } /* endif */
         } /* endif */
      } /* endif */

      // Ok, now let's save the FONTMETRICS
      delete ppd->pfontmetCl;
      ppd->pfontmetCl = new FONTMETRICS;
      memcpy(ppd->pfontmetCl, &pfm[Index], sizeof(FONTMETRICS));

      // Set some FATTRS values from the FONTMETRICS
      (ppd->pfattrsCl)->idRegistry = (ppd->pfontmetCl)->idRegistry;
      (ppd->pfattrsCl)->usCodePage = (ppd->pfontmetCl)->usCodePage;
      if (!(ppd->pfontmetCl->fsDefn & FM_DEFN_OUTLINE))
      {
         (ppd->pfattrsCl)->lMaxBaselineExt = (ppd->pfontmetCl)->lMaxBaselineExt;
         (ppd->pfattrsCl)->lAveCharWidth = (ppd->pfontmetCl)->lAveCharWidth;
      }
      (ppd->pfattrsCl)->fsFontUse |= FATTR_FONTUSE_NOMIX;

      delete [] pfm;
   } /* endif */

   if ((isNonPropOnly() || isBitmapOnly() || isVectorOnly())
         && !lBitmapMatch && !lVectorMatch) {
      if (strcmp((ppd->pfattrsCl)->szFacename, (char*)"Courier")) {
         strcpy((ppd->pfattrsCl)->szFacename, (char*)"Courier");
         ITRACE_DEVELOP("Checking for match again.");
         checkForMatch(hps);
      } /* endif */
   } /* endif */

   if(presSpaceHandle == 0)
     WinReleasePS(hps);

   return *this;
}
#endif // IC_PM


/*------------------------------------------------------------------------------
| IFont::mustUseVector                                                         |
------------------------------------------------------------------------------*/
IFont& IFont :: mustUseVector (Boolean fDoIt)
{
   if (isBitmapOnly()) {
      ppd->flClStyles = ppd->flClStyles & ~IC_MUSTUSEVECTORFONT;
      return *this;
   } /* endif */
   if (ppd->pntClAngle.x() != 1 || ppd->pntClAngle.y() != 0 ||
       ppd->pntClShear.x() != 0 || ppd->pntClShear.y() != 1 || isVectorOnly()) {
      ppd->flClStyles = ppd->flClStyles | IC_MUSTUSEVECTORFONT;
      return *this;
   } /* endif */

   if (fDoIt) {
      ppd->flClStyles = ppd->flClStyles | IC_MUSTUSEVECTORFONT;
   } else {
      ppd->flClStyles = ppd->flClStyles & ~IC_MUSTUSEVECTORFONT;
   } /* endif */
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::initialize                                                            |
------------------------------------------------------------------------------*/
IFont& IFont :: initialize(const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::initialize");
   IPresSpaceHandle hps;
   if (presSpaceHandle == 0)
      hps = IGETPS(HWND_DESKTOP);
   else
      hps = presSpaceHandle;

#ifdef IC_PM
   ppd->pfattrsCl = new FATTRS;
   ppd->pfontmetCl = new FONTMETRICS;
   GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), ppd->pfontmetCl);

   // Set our internal flags for this font so that checkForMatch
   // will work properly.
   // Don't call the useNonPropOnly() type function since they'll
   // cause us to call checkForMatch() prematurely.

   if (ppd->pfontmetCl->fsType & FM_TYPE_FIXED)
     ppd->flClStyles = ppd->flClStyles | IC_NONPROPORTIONAL_FONTS_ONLY;

   if (ppd->pfontmetCl->fsDefn & FM_DEFN_OUTLINE)
     ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
   else
     ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;

   /* initialize font attributes structure */
   (ppd->pfattrsCl)->usRecordLength = sizeof(FATTRS);
   (ppd->pfattrsCl)->fsSelection = (ppd->pfontmetCl)->fsSelection;
   (ppd->pfattrsCl)->lMatch = (ppd->pfontmetCl)->lMatch;
   (ppd->pfattrsCl)->idRegistry = (ppd->pfontmetCl)->idRegistry;
   (ppd->pfattrsCl)->usCodePage = (ppd->pfontmetCl)->usCodePage;
   (ppd->pfattrsCl)->fsType = 0;
   (ppd->pfattrsCl)->fsFontUse = FATTR_FONTUSE_NOMIX;
   strcpy((ppd->pfattrsCl)->szFacename, (ppd->pfontmetCl)->szFacename);

   long currPointSize;
   if (ppd->flClStyles & IC_VECTOR_FONTS_ONLY) {
     (ppd->pfattrsCl)->lMaxBaselineExt = 0;
     (ppd->pfattrsCl)->lAveCharWidth = 0;

     long lyFontResolution;
     SIZEF sizef;
     HDC hDC = GpiQueryDevice(hps);
     DevQueryCaps( hDC, CAPS_VERTICAL_FONT_RES,
                   (long)1, &lyFontResolution);
     GpiQueryCharBox (hps, &sizef);
     // Add 0.5 before doing integer conversion to avoid truncation errors
     currPointSize = FIXEDINT((((sizef.cy) * 72) / lyFontResolution)+0x8000);
     }
   else {
     (ppd->pfattrsCl)->lMaxBaselineExt = (ppd->pfontmetCl)->lMaxBaselineExt;
     (ppd->pfattrsCl)->lAveCharWidth = (ppd->pfontmetCl)->lAveCharWidth;
     currPointSize = (ppd->pfontmetCl)->sNominalPointSize/10;
     }

   setSizeFromHps(currPointSize, hps);
#endif
#ifdef IC_WIN
   /*------------------------------------------------------------------*/
   /* Allocate private data to store the logfont                       */
   /*------------------------------------------------------------------*/
   ppd->plogfontCl = new LOGFONT;

   /*------------------------------------------------------------------*/
   /* Use a temp font (stock) to use to get current selection with     */
   /*------------------------------------------------------------------*/
   HFONT hFont = SelectObject( hps, ppd->systemFont );

   /*------------------------------------------------------------------*/
   /* Select it into back into the hps and use the handle to retrieve  */
   /* the log font information for that font                           */
   /*------------------------------------------------------------------*/
   SelectObject( hps, hFont );
   GetObject( hFont, sizeof(LOGFONT), ppd->plogfontCl );

   /*------------------------------------------------------------------*/
   /* Construct a font using that information                          */
   /*------------------------------------------------------------------*/
   makeFontMetrics( hps );
#endif
   if(presSpaceHandle == 0)
     IRELEASEPS(HWND_DESKTOP,hps);

   return *this;
}


/*------------------------------------------------------------------------------
| IFont::fontChanged                                                           |
|                                                                              |
| Will cause a subsequent call to makeFontMetrics()                            |
------------------------------------------------------------------------------*/
IFont& IFont :: fontChanged ()
{
// IMODTRACE_DEVELOP("IFont::fontChanged");
#ifdef IC_PM
   delete [] ppd->plClCharWidths;
   ppd->plClCharWidths = 0;
   delete ppd->pfontmetCl;
   ppd->pfontmetCl = 0;
#endif
#ifdef IC_WIN
    delete ppd->ptextmetricCl;
    ppd->ptextmetricCl = 0;
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setWindowFont                                                         |
|                                                                              |
| Used by IControl::setFont()
------------------------------------------------------------------------------*/
Boolean IFont :: setWindowFont(IWindow* pwindow) const
{
   IMODTRACE_DEVELOP("IFont::setWindowFont");
   Boolean result = True;
#ifdef IC_PM
   // Build string to pass on WinSetPresParam call
   IString presParm(pointSize());
   presParm += ".";
   presParm += name();

   if (isBold())
     presParm += ".Bold";

   if (isItalic())
     presParm += ".Italic";

   if (isOutline())
     presParm += ".Outline";
   if (isUnderscore())
     presParm += ".Underscore";
   if (isStrikeout())
     presParm += ".Strikeout";

   result = WinSetPresParam(pwindow->handle(),
                   PP_FONTNAMESIZE,
                   presParm.length()+1,
                   (char*)presParm);

   if (result == false)
     ITHROWGUIERROR("WinSetPresParam");
#endif
#ifdef IC_WIN
   if (!propagationInProgress)
   {
     pwindow->pWindowData->state |= pwindow->fontWasSet;
     if ( pwindow->pWindowData->fFont )
     {             // Clean up the prior font.
        delete pwindow->pWindowData->fFont;
     }
     pwindow->pWindowData->fFont = new IFont(*this);
   }

   pwindow->sendEvent( WM_SETFONT,
                       IEventParameter1( *(ppd->fpFontHandle) ),
                       IEventParameter2( true ) );
#endif
   return result;
}


/*------------------------------------------------------------------------------
| IFont::setCharWidth                                                          |
|                                                                              |
| Only affects chars draw after beginUsingFont                                 |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharWidth     (unsigned long ulWidth,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   // Only works for Vector Fonts
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
#ifdef IC_PM
   long  lxFontResolution;
   HDC   hDC;
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   IPresSpaceHandle hps;
   if (presSpaceHandle)
     hps = presSpaceHandle;
   else
     hps = WinGetPS(HWND_DESKTOP);
   hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_HORIZONTAL_FONT_RES,
                 (long)1, &lxFontResolution);
         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClWidth = (FIXED)((MAKEFIXED(ulWidth, 0) / 72 ) *
                     lxFontResolution );
   delete [] ppd->plClCharWidths;
   ppd->plClCharWidths = 0;
   makeCharWidths(hps);

   if (presSpaceHandle == 0)
     WinReleasePS(hps);
#endif
#ifdef IC_WIN
   (ppd->plogfontCl)->lfWidth = ulWidth;

   fontChanged();
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setCharHeight                                                         |
|                                                                              |
| Only affects chars draw after beginUsingFont                                 |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharHeight    (unsigned long ulHeight,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   // only valid for Vector fonts
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
#ifdef IC_PM
   long  lyFontResolution;
   HDC   hDC;
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   IPresSpaceHandle hps;
   if(presSpaceHandle)
     hps = presSpaceHandle;
   else
     hps = WinGetPS(HWND_DESKTOP);
   hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_VERTICAL_FONT_RES,
                 (long)1, &lyFontResolution);
         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClHeight = (FIXED)(((MAKEFIXED(ulHeight, 0)) / 72 ) *
                      lyFontResolution );

   if (presSpaceHandle == 0)
     WinReleasePS(hps);
#endif
#ifdef IC_WIN
   (ppd->plogfontCl)->lfHeight = ulHeight;

   fontChanged();
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


#ifdef IC_PM
/*------------------------------------------------------------------------------
| IFont::setWidthHeight                                                        |
------------------------------------------------------------------------------*/
IFont& IFont :: setWidthHeight   (unsigned long ulPointSize,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
     return *this;

   long  devRes[2];
         /* Query the device context for the screen and then query */
         /* the resolution of the device for the device context.   */
   IPresSpaceHandle hps;
   if(presSpaceHandle)
      hps = presSpaceHandle;
   else
      hps = WinGetPS(HWND_DESKTOP);
   HDC hDC = GpiQueryDevice(hps);
   DevQueryCaps( hDC, CAPS_HORIZONTAL_FONT_RES,
                 (long)2, devRes);
   if (presSpaceHandle == 0)
     WinReleasePS(hps);

         /* Calculate the size of the character box, based on the  */
         /* point size selected and the resolution of the device.  */
         /* The size parameters are of type FIXED, NOT int.        */
         /*  NOTE: 1 point == 1/72 of an inch.                     */
   ppd->fxClWidth = (FIXED)((MAKEFIXED(ulPointSize, 0) / 72 ) *
                     devRes[0]);
   ppd->fxClHeight = (FIXED)(((MAKEFIXED(ulPointSize, 0)) / 72 ) *
                     devRes[1]);
   return *this;
}
#endif // IC_PM

/*------------------------------------------------------------------------------
| IFont::setCharSize                                                           |
------------------------------------------------------------------------------*/
IFont& IFont :: setCharSize (const ISize& size,
                             const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
#ifdef IC_PM
   setCharWidth(size.width(), presSpaceHandle);
   return setCharHeight(size.height(), presSpaceHandle);
#endif
#ifdef IC_WIN
   (ppd->plogfontCl)->lfHeight = size.height();
   (ppd->plogfontCl)->lfWidth  = size.width();

   fontChanged();
   makeFontMetrics( presSpaceHandle );
   return *this;
#endif
}


/*------------------------------------------------------------------------------
| IFont::setFontAngle                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontAngle     (const IPoint& point,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   ppd->pntClAngle = point;
#ifdef IC_WIN
   double angle = 0.0;
   if (point.x() == 0)
   {
     if (point.y() > 0)
       angle = 90.0;
     else if (point.y() < 0)
       angle = -90.0;
   }
   else
     angle = (double)DEGREETENTHS_TO_RADS *
                            atan((double)point.y() / (double)point.x());

   (ppd->plogfontCl)->lfOrientation = (long)angle;
   fontChanged();
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setFontShear                                                          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontShear     (const IPoint& point,
                                  const IPresSpaceHandle& presSpaceHandle)
{
   if (isBitmapOnly())
   {
     ITHROWLIBRARYERROR(IC_VECTOR_FONT_ONLY_FUNCTION,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);
   }
   copyPrivate();
   ppd->pntClShear = point;
#ifdef IC_WIN
   double angle = 0.0;
   if (point.x() == 0)
   {
     if (point.y() > 0)
       angle = 90.0;
     else if (point.y() < 0)
       angle = -90.0;
   }
   else
     angle = (double)DEGREETENTHS_TO_RADS *
                            atan((double)point.y() / (double)point.x());
   (ppd->plogfontCl)->lfEscapement = (long)angle;
   fontChanged();
   makeFontMetrics();
#endif
   return *this;
}


/*------------------------------------------------------------------------------
| IFont::setFontFromDialog                                                     |
|                                                                              |
| Change the IFont to correspond to the one specified by the fontDlg.          |
------------------------------------------------------------------------------*/
IFont& IFont :: setFontFromDialog (IFontDialog& fontDlg,
                                   const IPresSpaceHandle& presSpaceHandle)
{
   IMODTRACE_DEVELOP("IFont::setFontFromDialog");
   copyPrivate();
#ifdef IC_PM
   delete ppd->pfattrsCl;

   ppd->pfattrsCl = new FATTRS;
   memcpy(ppd->pfattrsCl, &(fontDlg.fData->fAttrs), sizeof(FATTRS));

   if ( ppd->pfattrsCl->fsFontUse & FATTR_FONTUSE_OUTLINE )
   {
      ppd->flClStyles = ppd->flClStyles | IC_VECTOR_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles & ~IC_BITMAP_FONTS_ONLY;
   }
   else
   {
      ppd->flClStyles = ppd->flClStyles & ~IC_VECTOR_FONTS_ONLY;
      ppd->flClStyles = ppd->flClStyles | IC_BITMAP_FONTS_ONLY;
   }

   fontChanged();
   setSizeFromHps(fontDlg.pointSize(), presSpaceHandle);
   checkForMatch(presSpaceHandle);
#endif
#ifdef IC_WIN
   delete ppd->plogfontCl;

   ppd->plogfontCl = new LOGFONT;
   memcpy(ppd->plogfontCl, fontDlg.fData->lpLogFont, sizeof(LOGFONT));

   fontChanged();
   makeFontMetrics( presSpaceHandle );
#endif
   return *this;
}

#ifdef IC_PM
/*------------------------------------------------------------------------------
| IFont::copyWidthsArray                                                       |
|                                                                              |
| Copies the widths array into the specified buffer.                           |
| Used as a performance boost for calculations done for ICanvas layout.        |
------------------------------------------------------------------------------*/
IFont& IFont :: copyWidthsArray(char* buf, unsigned long length)
{
   IASSERTPARM(buf && length <= iCharArrayWidth);
   memcpy(buf, ppd->plClCharWidths, (length * sizeof(long)));

   return *this;
}

/*------------------------------------------------------------------------------
| IFont::fattrs                                                                |
|                                                                              |
| Return a const pointer to the OS/2 FATTRS struct.                            |
------------------------------------------------------------------------------*/
const struct _FATTRS* IFont :: fattrs () const
{
   return ppd->pfattrsCl;
}

/*------------------------------------------------------------------------------
| IFont::fontmetrics                                                           |
|                                                                              |
| Return a const pointer to the OS/2 FONTMETRICS struct.                       |
------------------------------------------------------------------------------*/
const struct _FONTMETRICS* IFont :: fontmetrics () const
{
   return ppd->pfontmetCl;
}
#endif  // IC_PM

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IFont::lpLogFont                                                             |
|                                                                              |
| Return a const pointer to the Windows LOGFONT struct.                        |
------------------------------------------------------------------------------*/
const LOGFONT* IFont :: logfont () const
{
   return ppd->plogfontCl;
}
#endif

/*------------------------------------------------------------------------------
| IFont::copyPrivate                                                           |
|                                                                              |
| Makes a new copy of the fonts private data, decrements the use count for     |
| the old private data and increments the count for the new private data.      |
------------------------------------------------------------------------------*/
IFont& IFont :: copyPrivate( )
{
  if ( ppd->refCount() > 1 )
  {
    IFontPrivateData* newppd = new IFontPrivateData( *ppd );

    // Decrement the ref count for the private data.
    // addRef does not need to be called for the new
    // ifont private data because it is already initialized
    // to 1.
    ppd->removeRef();

    ppd = newppd;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::FaceNameCursor                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::FaceNameCursor::FaceNameCursor ( FontType fontType,
                                        const IPresSpaceHandle& presSpaceHandle)
                                      : index(0), count(0), facenames(0)
{
  IPresSpaceHandle hps(presSpaceHandle);
  HWND hwnd =
#ifdef IC_PM
              HWND_DESKTOP;
#endif
#ifdef IC_WIN
              GetDesktopWindow();
#endif
  if (!presSpaceHandle)
    hps = IGETPS( hwnd );

#ifdef IC_PM
  HDC hdc(GpiQueryDevice(hps));
#endif
#ifdef IC_WIN
  HDC hdc(hps);
#endif
  long hres,vres;
  IQUERYDEVICECAPS(hdc, CAPS_HORIZONTAL_FONT_RES, &hres);
  IQUERYDEVICECAPS(hdc, CAPS_VERTICAL_FONT_RES, &vres);

#ifdef IC_PM
  long numRequired(0);
  long fontCount( GpiQueryFonts( hps,
                                 QF_PRIVATE | QF_PUBLIC,
                                 0,
                                 &numRequired,
                                 (long)sizeof(FONTMETRICS),
                                 0 ));
  if ( fontCount )
  {
    PFONTMETRICS fontmetrics = new FONTMETRICS[fontCount];

    count = GpiQueryFonts( hps,
                           QF_PRIVATE | QF_PUBLIC,
                           0,
                           &fontCount,
                           (long)sizeof(FONTMETRICS),
                           fontmetrics );
    count = fontCount;
    int deviceCount;
    int i;
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ( fontType == FaceNameCursor::both &&
         (((fontmetrics[i].sXDeviceRes == hres) &&
           (fontmetrics[i].sYDeviceRes == vres)) || (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE)))
        deviceCount++;
      else if ( fontType == FaceNameCursor::bitmap &&
              ((fontmetrics[i].sXDeviceRes == hres) &&
               (fontmetrics[i].sYDeviceRes == vres)) &&
              !(fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
      else if (fontType == FaceNameCursor::vector &&
               (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
    } /* endfor */
    facenames = new IString[deviceCount];
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ( fontType == FaceNameCursor::both &&
         (((fontmetrics[i].sXDeviceRes == hres) &&
           (fontmetrics[i].sYDeviceRes == vres)) || (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE)))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
      else if (fontType == FaceNameCursor::bitmap &&
              ((fontmetrics[i].sXDeviceRes == hres) &&
               (fontmetrics[i].sYDeviceRes == vres)) &&
              !(fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
      else if (fontType == FaceNameCursor::vector &&
               (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        facenames[deviceCount] = fontmetrics[i].szFacename;
        deviceCount++;
      }
    } /* endfor */
    delete [] fontmetrics;
    count = deviceCount;
  }
#endif
#ifdef IC_WIN
  FONTCOUNT fontCount = { 0, 0, 0, 0, 0, 0 };

  PFONTINFO pfontT;
  PFACEINFO pfaciFaces = NULL, pfaciT;

  /*------------------------------------------------------------------*/
  /* Enumerate the face names available                               */
  /*------------------------------------------------------------------*/
  EnumFontFamilies( hdc, NULL, (FONTENUMPROC)EnumAllFaces,
                    (LPARAM)&pfaciFaces);

  /*------------------------------------------------------------------*/
  /* If a specific type of font is to be searched for, set a flag     */
  /*  otherwise set both to indicate all fonts                        */
  /*------------------------------------------------------------------*/
  if (fontType == FaceNameCursor::bitmap)
    fontCount.dwType = RASTER_FONTTYPE;
  if (fontType == FaceNameCursor::vector)
    fontCount.dwType = TRUETYPE_FONTTYPE;
  if (fontType == FaceNameCursor::both)
    fontCount.dwType = RASTER_FONTTYPE | TRUETYPE_FONTTYPE;

  /*------------------------------------------------------------------*/
  /* Pass the device resolution in to check against                   */
  /*------------------------------------------------------------------*/
  fontCount.horzRes = hres;
  fontCount.vertRes = vres;

  /*------------------------------------------------------------------*/
  /* enumerate the fonts available for each face name                 */
  /*------------------------------------------------------------------*/
  for (pfaciT = pfaciFaces; pfaciT; pfaciT = pfaciT->pfaciNext)
  {
    EnumFontFamilies( hdc, pfaciT->szFaceName, (FONTENUMPROC)EnumAllFonts,
                      (LPARAM)&fontCount);
  }

  /*------------------------------------------------------------------*/
  /* Allocate the array of strings for the face names to list         */
  /*------------------------------------------------------------------*/
  facenames = new IString[fontCount.matchCount];

  /*------------------------------------------------------------------*/
  /* Free the face name list                                          */
  /*------------------------------------------------------------------*/
  while (pfaciFaces)
  {
    pfaciT = pfaciFaces->pfaciNext;
    delete pfaciFaces;
    pfaciFaces = pfaciT;
  }

  /*------------------------------------------------------------------*/
  /* Transfer the font name list over and then free each item         */
  /*------------------------------------------------------------------*/
  int deviceCount = 0;
  while (fontCount.pFontInfo)
  {
    facenames[deviceCount] = fontCount.pFontInfo->szFaceName;
    deviceCount++;
    pfontT = fontCount.pFontInfo->pfiNext;
    delete fontCount.pFontInfo;
    fontCount.pFontInfo = pfontT;
  }
  /*------------------------------------------------------------------*/
  /* Finally, save the total count of items to cursor over            */
  /*------------------------------------------------------------------*/
  count = deviceCount;
#endif

  if (!presSpaceHandle)
    IRELEASEPS( hwnd, hps );
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::~FacenameCursor                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::FaceNameCursor::~FaceNameCursor ()
{
  if (count)
  {
    delete [] facenames;
  }
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToFirst                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToFirst()
{
  if (count)
  {
    index = 0;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToNext                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToNext()
{
  Boolean foundDuplicate;
  int j(index);
  index++;

  while (index < count)
  {
    foundDuplicate = false;
    for (int i=j;i>=0 && !foundDuplicate;i--)
    {
      if ( facenames[i] == facenames[index] )
        foundDuplicate = true;
    } /* endfor */
    if (!foundDuplicate)
      break;
    index++;
  } /* endwhile */

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToPrevious                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToPrevious()
{

  Boolean foundDuplicate;
  int j(index);
  index--;

  while (index < count)
  {
    foundDuplicate = false;
    for (int i=j;i<count && !foundDuplicate;i++)
    {
      if ( facenames[i] == facenames[index] )
        foundDuplicate = true;
    } /* endfor */
    if (!foundDuplicate)
      break;
    index--;
  } /* endwhile */

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FacenameCursor::setToLast                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::setToLast()
{

  if (count)
  {
    index = count - 1;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::isValid                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::FaceNameCursor::isValid() const
{
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::FaceNameCursor::invalidate                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IFont::FaceNameCursor::invalidate()
{
  index = count;
}

/*------------------------------------------------------------------------------
| IFont::faceNameAt                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IFont::faceNameAt( const FaceNameCursor& faceNameCursor )
{
  return faceNameCursor.facenames[faceNameCursor.index];
}


/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::PointSizeCursor                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::PointSizeCursor::PointSizeCursor ( const char* faceName,
                                        const IPresSpaceHandle& presSpaceHandle)
                                        : index(0), count(0), pointsizes(0)
{
  IPresSpaceHandle hps(presSpaceHandle);
  HWND hwnd =
#ifdef IC_PM
              HWND_DESKTOP;
#endif
#ifdef IC_WIN
              GetDesktopWindow();
#endif
  if (!presSpaceHandle)
    hps = IGETPS( hwnd );

#ifdef IC_PM
  HDC hdc(GpiQueryDevice(hps));
#endif
#ifdef IC_WIN
  HDC hdc(hps);
#endif
  long hres,vres;
  IQUERYDEVICECAPS(hdc, CAPS_HORIZONTAL_FONT_RES, &hres);
  IQUERYDEVICECAPS(hdc, CAPS_VERTICAL_FONT_RES, &vres);

#ifdef IC_PM
  long numRequired(0);
  long fontCount( GpiQueryFonts( hps,
                                 QF_PRIVATE | QF_PUBLIC,
                                 faceName,
                                 &numRequired,
                                 (long)sizeof(FONTMETRICS),
                                 0 ));
  if ( fontCount )
  {
    PFONTMETRICS fontmetrics = new FONTMETRICS[fontCount];

    count = GpiQueryFonts( hps,
                           QF_PRIVATE | QF_PUBLIC,
                           faceName,
                           &fontCount,
                           (long)sizeof(FONTMETRICS),
                           fontmetrics );
    count = fontCount;
    int deviceCount;
    int i;
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ((fontmetrics[i].sXDeviceRes == hres &&
           fontmetrics[i].sYDeviceRes == vres ) ||
          (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
        deviceCount++;
    } /* endfor */
    pointsizes = new long[deviceCount];
    for (i=0,deviceCount=0;i<count;i++)
    {
      if ((fontmetrics[i].sXDeviceRes == hres &&
           fontmetrics[i].sYDeviceRes == vres ) ||
          (fontmetrics[i].fsDefn & FM_DEFN_OUTLINE))
      {
        pointsizes[deviceCount] = fontmetrics[i].sNominalPointSize/10;
        deviceCount++;
      }
    } /* endfor */
    delete [] fontmetrics;
    count = deviceCount;
  }
#endif
#ifdef IC_WIN
  FONTCOUNT fontCount = { 0, 0, 0, 0, 0, 1 };
  PFONTINFO pfontT;

  /*------------------------------------------------------------------*/
  /* Compile a list only for bitmap fonts as vector can be any size   */
  /*------------------------------------------------------------------*/
  fontCount.dwType = RASTER_FONTTYPE;

  /*------------------------------------------------------------------*/
  /* Pass the device resolution in to check against                   */
  /*------------------------------------------------------------------*/
  fontCount.horzRes = hres;
  fontCount.vertRes = vres;

  /*------------------------------------------------------------------*/
  /* Enumerate the sizes available for the face name passed in        */
  /*------------------------------------------------------------------*/
  EnumFontFamilies( hdc, faceName, (FONTENUMPROC)EnumAllFonts,
                    (LPARAM)&fontCount);

  /*------------------------------------------------------------------*/
  /* Allocate the array of longs for the pointsizes to list           */
  /*------------------------------------------------------------------*/
  pointsizes = new long[fontCount.matchCount];

  /*------------------------------------------------------------------*/
  /* Transfer the pointsize list over and then free each item         */
  /*------------------------------------------------------------------*/
  int deviceCount = 0;
  while (fontCount.pFontInfo)
  {
    /*----------------------------------------------------------------*/
    /* Add in reverse order to since linked list is ordered last to   */
    /*  first since it was built that way                             */
    /*----------------------------------------------------------------*/
    deviceCount++;
    pointsizes[fontCount.matchCount-deviceCount] =
                                        fontCount.pFontInfo->iPointSize;
    pfontT = fontCount.pFontInfo->pfiNext;
    delete fontCount.pFontInfo;
    fontCount.pFontInfo = pfontT;
  }
  /*------------------------------------------------------------------*/
  /* Finally, save the total count of items to cursor over            */
  /*------------------------------------------------------------------*/
  count = deviceCount;
#endif

  if (!presSpaceHandle)
    IRELEASEPS( hwnd, hps );
}


/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::~PointSizeCursor                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont::PointSizeCursor::~PointSizeCursor ()
{
  if (count)
  {
    delete [] pointsizes;
  }
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToFirst                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToFirst()
{
  if (count)
  {
    index = 0;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToNext                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToNext()
{
  index++;

  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToPrevious                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToPrevious()
{
  index--;
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::setToLast                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::setToLast()
{

  if (count)
  {
    index = count - 1;
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::isValid                                              |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IFont::PointSizeCursor::isValid() const
{
  if (count && index < count)
  {
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
| IFont::PointSizeCursor::invalidate                                           |
|                                                                              |
| Invalidate the cursor.                                                       |
------------------------------------------------------------------------------*/
void IFont::PointSizeCursor::invalidate()
{
  index = count;
}

/*------------------------------------------------------------------------------
| IFont::pointsizeAt                                                           |
|                                                                              |
| Return the point size for the argument cursor.                               |
------------------------------------------------------------------------------*/
long IFont::pointSizeAt( const PointSizeCursor& pointSizeCursor )
{
  return pointSizeCursor.pointsizes[pointSizeCursor.index];
}


#ifdef IC_WIN
/**********************************************************************/
/* Procedure: EnumAllFonts - Callback function for EnumFonts.         */
/*            Allocates a FONTINFO structure and saves all of the     */
/*            information passed into this function.                  */
/**********************************************************************/
int CALLBACK EnumAllFonts( LPENUMLOGFONT lpelf, LPNEWTEXTMETRIC lpntm,
                           int nFontType, LPARAM lpData )
{
  PFONTINFO pfiNew, pfiNav;
  BOOL found = FALSE;
  int iPointSize = 0;
  FONTCOUNT* pFontCount = (FONTCOUNT*)lpData;
  LPTEXTMETRIC lptm = (LPTEXTMETRIC)lpntm;

  /*------------------------------------------------------------------*/
  /* Add the font to our list if:                                     */
  /*   a) the request was for bitmap fonts and we have a bitmap font  */
  /*      and the device resolution matches that of our device        */
  /*   b) the request was for a vector font and this is not a bitmap  */
  /*      font (therefore it is either a vector or TrueType font)     */
  /*   c) the request was for both (a+b)                              */
  /*------------------------------------------------------------------*/
  if (((pFontCount->dwType & RASTER_FONTTYPE) &&
       (nFontType & RASTER_FONTTYPE) &&
       (pFontCount->horzRes == lptm->tmDigitizedAspectX) &&
       (pFontCount->vertRes == lptm->tmDigitizedAspectY)) ||
      ((pFontCount->dwType & TRUETYPE_FONTTYPE) &&
       !(nFontType & RASTER_FONTTYPE)))
  {
    /*----------------------------------------------------------------*/
    /* Check for duplicate listing.  If point size, then check sizes  */
    /*----------------------------------------------------------------*/
    if (pFontCount->bPointSizes)
    {
      /*--------------------------------------------------------------*/
      /* Calculate the point size for the font enumerated             */
      /*--------------------------------------------------------------*/
      iPointSize = ((1440 / (int)pFontCount->vertRes) *
               (int)(lptm->tmHeight - lptm->tmInternalLeading) + 10)/20;

      /*--------------------------------------------------------------*/
      /* Check if the list currently contains the pointsize found     */
      /*--------------------------------------------------------------*/
      pfiNav = pFontCount->pFontInfo;
      while (!found && (pfiNav))
      {
        /*------------------------------------------------------------*/
        /* If so, set flag to not add again, otherwise keep searching */
        /*------------------------------------------------------------*/
        if (pfiNav->iPointSize == iPointSize)
          found = TRUE;
        else
          pfiNav = pfiNav->pfiNext;
      }
    }
    /*----------------------------------------------------------------*/
    /* Otherwise check facenames for duplicates                       */
    /*----------------------------------------------------------------*/
    else
    {
      /*--------------------------------------------------------------*/
      /* Check if the list currently contains the facename found      */
      /*--------------------------------------------------------------*/
      pfiNav = pFontCount->pFontInfo;
      while (!found && (pfiNav))
      {
        /*------------------------------------------------------------*/
        /* If so, set flag to not add again, otherwise keep searching */
        /*------------------------------------------------------------*/
        if (!(lstrcmp(pfiNav->szFaceName,lpelf->elfLogFont.lfFaceName)))
          found = TRUE;
        else
          pfiNav = pfiNav->pfiNext;
      }
    }

    /*----------------------------------------------------------------*/
    /* Only add to list if not yet in the linked list                 */
    /*----------------------------------------------------------------*/
    if (!found)
    {
      /*--------------------------------------------------------------*/
      /* Allocate a FONTINFO from the heap.  If it fails, punt.       */
      /*--------------------------------------------------------------*/
      if (!(pfiNew = (PFONTINFO) new FONTINFO ))
        return (FALSE);

      /*--------------------------------------------------------------*/
      /* Initialize the FONTINFO structure that was allocated with    */
      /*  the facename passed, the point size calculated from the     */
      /*  font information and set pointer to next block to whole list*/
      /*--------------------------------------------------------------*/
      lstrcpy( pfiNew->szFaceName, lpelf->elfLogFont.lfFaceName );
      pfiNew->iPointSize = ((1440 / (int)pFontCount->vertRes) *
               (int)(lptm->tmHeight - lptm->tmInternalLeading) + 10)/20;
      pfiNew->pfiNext = pFontCount->pFontInfo;

      /*--------------------------------------------------------------*/
      /* Change the list head pointer to the newly allocated link     */
      /*--------------------------------------------------------------*/
      pFontCount->pFontInfo = pfiNew;
      pFontCount->matchCount++;
    }
  }
  return (TRUE);
}


/**********************************************************************/
/* Procedure: EnumAllFaces - Callback function for EnumFonts.         */
/*            Allocates a FONTINFO structure and saves all of the     */
/*            information passed into this function.                  */
/**********************************************************************/
int CALLBACK EnumAllFaces( LPENUMLOGFONT lpelf, LPNEWTEXTMETRIC lpntm,
                           int nFontType, LPARAM lpData )
{
  PFACEINFO pfaciNew;

  /*------------------------------------------------------------------*/
  /* Allocate a FACEINFO from the heap.  If it fails, punt.           */
  /*------------------------------------------------------------------*/
  if (!(pfaciNew = (PFACEINFO) new FACEINFO ))
    return (FALSE);

  /*------------------------------------------------------------------*/
  /* Initialize the FACEINFO                                          */
  /*------------------------------------------------------------------*/
  lstrcpy( pfaciNew->szFaceName, lpelf->elfLogFont.lfFaceName );
  pfaciNew->pfaciNext = *( (PFACEINFO FAR *)lpData );

  /*------------------------------------------------------------------*/
  /* Update the list head pointer                                     */
  /*------------------------------------------------------------------*/
  *((PFACEINFO FAR *) lpData) = pfaciNew;
  return (TRUE);
}
#endif // IC_WIN

