/*******************************************************************************
* FILE NAME: ititle.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ititle.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
   #define INCL_WINSWITCHLIST
   #define INCL_WINFRAMEMGR
   #define INCL_WINSYS
   #include <iwindefs.h>
   #include <string.h>
   #include <stdlib.h>
}
#include <ititle.hpp>
#include <icconst.h>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <inotifev.hpp>
#include <iplatfrm.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iwindow.hpp>


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ITITLE_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| titleBarHwnd                                                                 |
|  Returns FID_TITLEBAR HWND for owner.                                        |
|  Throws an exception if owner doesn't have FID_TITLEBAR.                     |
------------------------------------------------------------------------------*/
static IWindowHandle titleBarHwnd ( IWindow* owner )
{
#ifndef IC_WIN
  IWindowHandle hwnd =
     IWindow::handleWithParent( FID_TITLEBAR, owner->handle() );
  if ( hwnd == 0 )
  {
     ITHROWLIBRARYERROR( IC_NO_TITLEBAR,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }
  return hwnd;
#endif
#ifdef IC_WIN
  // No separate window handle in Windows.  We use this function to
  // validate that there is a title bar and throw if not.
  if ( ( ISTYLEOF( owner->handle() ) & WS_CAPTION ) != WS_CAPTION )
     ITHROWLIBRARYERROR( IC_NO_TITLEBAR,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  return owner->handle();
#endif
}


/*------------------------------------------------------------------------------
| ITitle::ITitle                                                               |
------------------------------------------------------------------------------*/
ITitle :: ITitle ( IWindow*      owner,
                   const char*   objectName,
                   const char*   viewName,
                   unsigned long viewNumber )
  : pwinClOwner( owner ),
    fTitleData( 0 )
{
#ifndef IC_WIN
  // Title bar not a window in Windows.
  // We override some of the IWindow and ITextControl functions in this class
  // because our handle is the same as the owning IFrameWindow handle.
  this->setAutoDestroyWindow( false );
  this->startHandlingEventsFor( titleBarHwnd( this->pwinClOwner ));
#endif

  if ( objectName )
  {
     this->setTitleText( objectName, viewName, viewNumber );
  }
}

/*------------------------------------------------------------------------------
| ITitle::ITitle                                                               |
------------------------------------------------------------------------------*/
ITitle :: ITitle ( IWindow*           owner,
                   const IResourceId& objectName,
                   const IResourceId& viewName,
                   unsigned long      viewNumber )
  : pwinClOwner( owner ),
    fTitleData( 0 )
{
#ifndef   IC_WIN
  // Title bar not a window in Windows.
  this->setAutoDestroyWindow( false);
  this->startHandlingEventsFor( titleBarHwnd( this->pwinClOwner ));
#endif   //IC_WIN

  IString strObjName = objectName.resourceLibrary().loadString( objectName );

  IString strViewName;
  if ( viewName != 0 )
     strViewName = viewName.resourceLibrary().loadString( viewName );

  this->setTitleText( strObjName, strViewName, viewNumber );
}

/*------------------------------------------------------------------------------
| ITitle::~ITitle                                                              |
------------------------------------------------------------------------------*/
ITitle :: ~ITitle ( )
{ }

/*------------------------------------------------------------------------------
| ITitle::objectText                                                           |
------------------------------------------------------------------------------*/
IString ITitle :: objectText ( ) const
{
  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );
  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  IString strObj;
  if ( uLen > 1 )
  {
     CHAR* pszClBuf;                         // To store entire title string.
     pszClBuf = new CHAR[ uLen ];
     *pszClBuf = '\0';

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     strObj = IString( pszClBuf );
     unsigned pos = strObj.indexOf(" - ");   // Locate " - ".
     if ( pos )
        strObj.remove( pos );

     delete [] pszClBuf;
  } /* endif */

  return strObj;
}

/*------------------------------------------------------------------------------
| ITitle::viewText                                                             |
------------------------------------------------------------------------------*/
IString ITitle :: viewText ( ) const
{
  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );

  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  IString strView;

  if ( uLen > 1 )
  {
     char* pszClBuf = new char[ uLen ];

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     IString strObj = IString( pszClBuf );
     unsigned pos = strObj.indexOf(" - ");   // Locate " - ".
     if ( pos )
     {
        strView = strObj.subString( pos + 3 ); // View name.
        pos = strView.lastIndexOf( ':' );      // Get ptr at last ":".
        if ( pos && ( strView.subString(pos+1).isDigits() ) )
        {
           // Remove : and numeric part.
           strView.remove( pos );
        }
     } /* endif */

     delete [] pszClBuf;
  } /* endif */

  return strView;
}

/*------------------------------------------------------------------------------
| ITitle::viewNumber                                                           |
------------------------------------------------------------------------------*/
unsigned long ITitle :: viewNumber ( ) const
{
  unsigned long ulViewNum = 0;

  IWindowHandle hwndTitleBar = titleBarHwnd( this->pwinClOwner );
  unsigned long uLen = IQUERYWINDOWTEXTLENGTH( hwndTitleBar );
  ++uLen;                                    // Add 1 for null terminator.

  if ( uLen > 1 )
  {
     char* pszClBuf = new char[ uLen ];

     IQUERYWINDOWTEXT( this->pwinClOwner->handle(), uLen, (PSZ)pszClBuf );

     IString strObj = IString( pszClBuf );
     unsigned delimPos = strObj.indexOf(" - ");
     unsigned viewPos  = strObj.lastIndexOf( ':' );
     if ( (delimPos) &&                     // Found view delimiter.
          (viewPos > delimPos) &&           // Found view number.
          (viewPos < strObj.length()) )     // : not last char.
     {
        ulViewNum = strObj.subString( viewPos + 1 ).asUnsigned();
     }

     delete [] pszClBuf;
  } /* endif */

  return ulViewNum;
}

/*------------------------------------------------------------------------------
| ITitle::setObjectText                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setObjectText ( const char* objectName )
{
  IString strObj = this->objectText();
  if ( objectName )
  {
     if (strcmp( (char*)strObj, objectName ))
     {
        this->setTitleText( objectName, this->viewText(), this->viewNumber() );
        this->notifyObservers( INotificationEvent( ITitle::objectTextId,
                                                   *this,
                                                   true,
                                                   (char*)objectName ));
     }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setObjectText                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setObjectText ( const IResourceId& objectNameResId )
{
  IString strObjName =
     objectNameResId.resourceLibrary().loadString( objectNameResId );
  if ( strObjName.length() )
     this->setObjectText( strObjName );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewText                                                          |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewText ( const char* viewName )
{
  IString strView = this->viewText();

  if ( strcmp( (char*)strView, viewName ))
  {
     this->setTitleText( this->objectText(), viewName, this->viewNumber() );
     this->notifyObservers( INotificationEvent( ITitle::viewTextId,
                                                *this,
                                                true,
                                                (char*)viewName ));
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewText                                                          |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewText ( const IResourceId& viewNameResId )
{
  IString strViewName =
     (viewNameResId.resourceLibrary()).loadString( viewNameResId );
  if ( strViewName.length() )
     this->setViewText( strViewName );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setViewNumber                                                        |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setViewNumber ( unsigned long viewNum )
{
  /****************************************/
  /* Only change and notify if different. */
  /****************************************/
  if ( viewNum != this->viewNumber() )
  {
     this->setTitleText( this->objectText(), this->viewText(), viewNum );
     this->notifyObservers( INotificationEvent( ITitle::viewNumberId,
                                                *this,
                                                true,
                                                (void*)viewNum ));
  }
  return *this;
}


/*------------------------------------------------------------------------------
| ITitle::setTitleText                                                         |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setTitleText ( const char*   objectName,
                                 const char*   viewName,
                                 unsigned long viewNum )
{
  IString strTitleText;

  // Check to insure a non-NULL string pointer was passed
  if ( objectName )
  {
     strTitleText = objectName;

     // If object name exists, then append view name and number
     if ( *objectName != '\0' )
     {
        if ( viewName && *viewName != '\0' )
           strTitleText += IString( " - " ) + IString( viewName );

        if ( viewNum )
           strTitleText += IString( ":" ) + IString( viewNum );
     }
  }
  this->setText( strTitleText );

#ifndef IC_WIN
  // Update Task List or Window List.

  // Don't need to do anything in Windows since the task list or task bar is
  // automatically updated.

  HSWITCH hsw = WinQuerySwitchHandle( this->pwinClOwner->handle(), (PID)0 );
  if ( hsw )
  {
     SWCNTRL swctl;
     WinQuerySwitchEntry( hsw, &swctl );

     // Length of title cannot be greater than MAXNAMEL (60) + 1,
     // because SWCNTRL.szSwtitle is only that big.
     if ( strTitleText.length() + 1 > MAXNAMEL )
     {
       strTitleText = strTitleText.subString( 1, MAXNAMEL + 1 );
     }
     else if ( !strTitleText.length() )
       strTitleText = " ";

     strcpy( swctl.szSwtitle, strTitleText );
     WinChangeSwitchEntry( hsw, &swctl );
  } /* endif */
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setTitleText                                                         |
------------------------------------------------------------------------------*/
ITitle& ITitle :: setTitleText ( const IResourceId& objectNameResId,
                                 const IResourceId& viewNameResId,
                                 unsigned long      viewNum )
{
  IString strObjName =
     objectNameResId.resourceLibrary().loadString( objectNameResId );

  IString strViewName;
  if ( viewNameResId )
     strViewName = viewNameResId.resourceLibrary().loadString( viewNameResId );
  else
     strViewName = "";

  // Enforcement of requiring an object name string will be done below
  this->setTitleText( strObjName, strViewName, viewNum );

  return *this;
}

/*------------------------------------------------------------------------------
| WIN32 Color Support                                                          |
|                                                                              |
| Windows has no API equivalent for Presentation Parameters.  The following    |
| menu color functions are NOPed for Windows.  The query functions return      |
| an appropriate default IColor object.  The set and reset functions just      |
| return *this.                                                                |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| ITitle::borderColor                                                         |
|                                                                             |
| Returns the border color for the ITitle.                                    |
------------------------------------------------------------------------------*/
IColor ITitle::borderColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_BORDERCOLOR,
                         IGUIColor( IGUIColor::titleLine ));
#else
  return IGUIColor( IGUIColor::titleLine );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::activeColor                                                         |
|                                                                             |
| Returns the active color for the ITitle.                                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_ACTIVECOLOR,
                         IGUIColor( IGUIColor::activeTitleBgnd ));
#else
  return IGUIColor( IGUIColor::activeTitleBgnd );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::inactiveColor                                                       |
|                                                                             |
| Returns the inactive color for the ITitle.                                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_INACTIVECOLOR,
                         IGUIColor( IGUIColor::inactiveTitleBgnd ));
#else
  return IGUIColor( IGUIColor::inactiveTitleBgnd );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::activeTextForegroundColor                                           |
|                                                                             |
| Returns the active text foreground color for the ITitle.                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeTextForegroundColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_ACTIVETEXTFGNDCOLOR,
                         IGUIColor( IGUIColor::activeTitleText ));
#else
  return IGUIColor( IGUIColor::activeTitleText );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::activeTextBackgroundColor                                           |
|                                                                             |
| Returns the active text background color for the ITitle.                    |
------------------------------------------------------------------------------*/
IColor ITitle::activeTextBackgroundColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_ACTIVETEXTBGNDCOLOR,
                         IGUIColor( IGUIColor::activeTitleTextBgnd ));
#else
  return IGUIColor( IGUIColor::activeTitleTextBgnd );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::inactiveTextForegroundColor                                         |
|                                                                             |
| Returns the inactive text foreground color for the ITitle.                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveTextForegroundColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_INACTIVETEXTFGNDCOLOR,
                         IGUIColor( IGUIColor::inactiveTitleText ));
#else
  return IGUIColor( IGUIColor::inactiveTitleText );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::inactiveTextBackgroundColor                                         |
|                                                                             |
| Returns the inactive text background color for the ITitle.                  |
------------------------------------------------------------------------------*/
IColor ITitle::inactiveTextBackgroundColor ( ) const
{
#ifdef   IC_PM
  return IWindow::color( PP_INACTIVETEXTBGNDCOLOR,
                         IGUIColor( IGUIColor::inactiveTitleTextBgnd ));
#else
  return IGUIColor( IGUIColor::inactiveTitleTextBgnd );
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| ITitle::setActiveTextForegroundColor                                        |
|                                                                             |
| Sets the active text foreground color for the ITitle.                       |
------------------------------------------------------------------------------*/
ITitle& ITitle::setActiveTextForegroundColor ( const IColor &color )
{
#ifdef   IC_PM
  IWindow::setColor( PP_ACTIVETEXTFGNDCOLOR, color );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setActiveTextBackgroundColor                                        |
|                                                                             |
| Sets the active text background color for the ITitle.                       |
------------------------------------------------------------------------------*/
ITitle& ITitle::setActiveTextBackgroundColor ( const IColor &color )
{
#ifdef   IC_PM
  IWindow::setColor( PP_ACTIVETEXTBGNDCOLOR, color );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setInactiveTextForegroundColor                                      |
|                                                                             |
| Sets the inactive text foreground color for the ITitle.                     |
------------------------------------------------------------------------------*/
ITitle& ITitle::setInactiveTextForegroundColor ( const IColor &color )
{
#ifdef   IC_PM
  IWindow::setColor( PP_INACTIVETEXTFGNDCOLOR, color );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setInactiveTextBackgroundColor                                      |
|                                                                             |
| Sets the inactive text background color for the ITitle.                     |
------------------------------------------------------------------------------*/
ITitle& ITitle::setInactiveTextBackgroundColor ( const IColor &color )
{
#ifdef   IC_PM
  IWindow::setColor( PP_INACTIVETEXTBGNDCOLOR, color );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetActiveTextForegroundColor                                      |
|                                                                             |
| Resets the active text foreground color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetActiveTextForegroundColor ( )
{
#ifdef   IC_PM
  this->resetColor( PP_ACTIVETEXTFGNDCOLOR );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetActiveTextBackgroundColor                                      |
|                                                                             |
| Resets the active text background color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetActiveTextBackgroundColor ( )
{
#ifdef   IC_PM
  this->resetColor( PP_ACTIVETEXTBGNDCOLOR );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetInactiveTextForegroundColor                                    |
|                                                                             |
| Resets the active text foreground color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetInactiveTextForegroundColor ( )
{
#ifdef   IC_PM
  this->resetColor( PP_INACTIVETEXTFGNDCOLOR );
#endif   //IC_PM
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::resetInactiveTextBackgroundColor                                    |
|                                                                             |
| Resets the active text background color by undoing a previous set.          |
------------------------------------------------------------------------------*/
ITitle& ITitle::resetInactiveTextBackgroundColor ( )
{
#ifdef   IC_PM
  this->resetColor( PP_INACTIVETEXTBGNDCOLOR );
#endif   //IC_PM
  return *this;
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ITitle::setText                                                              |
|                                                                              |
| Set the title text.  Overridden since title is not a window in Windows.      |
| and to skip some of the tests done in ITextControl                           |
------------------------------------------------------------------------------*/
ITitle& ITitle::setText    ( const char* text )
{
  ISETWINDOWTEXT( titleBarHwnd( this->pwinClOwner ), text );
  return *this;
}

/*------------------------------------------------------------------------------
| ITitle::setText                                                              |
|                                                                              |
| Set the title text.  Overridden since title is not a window in Windows.      |
------------------------------------------------------------------------------*/
ITitle& ITitle::setText    ( const IResourceId& textResId )
{
  IString string = textResId.resourceLibrary().loadString( textResId );
   return this->setText( string );
}

/*------------------------------------------------------------------------------
| ITitle::text                                                                 |
|                                                                              |
| Get the title text.  Overridden since title is not a window in Windows.      |
------------------------------------------------------------------------------*/
IString ITitle::text       ( ) const
{
  unsigned long uLen = this->textLength();
  ++uLen;                                    // Add 1 for null terminator.

  IString result;
  if ( uLen > 1 )
  {
     char* pszClBuf = new char[ uLen ];

     IQUERYWINDOWTEXT( titleBarHwnd( this->pwinClOwner ),
                       uLen,
                       (PSZ)pszClBuf );
     result = pszClBuf;
  }
  return result;
}

/*------------------------------------------------------------------------------
| ITitle::textLength                                                           |
|                                                                              |
| Get the title text length.  Overridden since title is not a window in Windows|
------------------------------------------------------------------------------*/
unsigned long ITitle::textLength ( ) const
{
  return IQUERYWINDOWTEXTLENGTH( titleBarHwnd( this->pwinClOwner ) );
}

/*------------------------------------------------------------------------------
| ITitle::owner                                                                |
|                                                                              |
| Return title owner window.  Overridden since title is not a window in Windows|
------------------------------------------------------------------------------*/
IWindow* ITitle::owner( ) const
{
   return pwinClOwner;
}

/*------------------------------------------------------------------------------
| ITitle::isValid                                                              |
|                                                                              |
| True if owner has title.    Overridden since title is not a window in Windows|
------------------------------------------------------------------------------*/
IBase::Boolean ITitle::isValid( ) const
{
  if ( ( pwinClOwner ) &&
       ( ISTYLEOF( pwinClOwner->handle() ) & WS_CAPTION ) )
    return true;
  else
    return false;
}

/*------------------------------------------------------------------------------
| ITitle::handle                                                               |
|                                                                              |
| Returns the handle of the owner window.   This is to allow font and          |
| similar operations to work correctly.                                        |
------------------------------------------------------------------------------*/
IWindowHandle ITitle::handle( ) const
{
   //Throw exception if window handle is not valid window.
   if ( !isValid() ) {
      ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable );
   }
   return pwinClOwner->handle();
}

/*------------------------------------------------------------------------------
| ITitle::nativeRect                                                           |
|                                                                              |
| Returns the rectangle for the title bar.                                     |
------------------------------------------------------------------------------*/
IRectangle  ITitle::nativeRect                ( ) const
{
   // Reference : "How to Draw a Custom Window Caption" in Win32 SDK
   // Knowledge base.
   RECT rc;
   IWindowHandle hTitle = this->handle();
   GetWindowRect( hTitle, &rc );

   unsigned long style = ISTYLEOF( hTitle );
   // size of bitmaps on title bar
   ISize bmpSize( GetSystemMetrics(SM_CXSIZE),
                  GetSystemMetrics(SM_CYSIZE) );
   // size of window border
   ISize borderSize;

   if (style & WS_THICKFRAME )
      {
      // account for thick border
      borderSize = ISize(GetSystemMetrics(SM_CXSIZEFRAME),
                         GetSystemMetrics(SM_CYSIZEFRAME) );
      }
   else
      {
      borderSize = ISize( GetSystemMetrics(SM_CXFIXEDFRAME),
                          GetSystemMetrics(SM_CYFIXEDFRAME) );
      if ( ! ((IPlatform::isWin95()) || (IPlatform::isNTNewShell())) )
        borderSize += ISize( GetSystemMetrics(SM_CXBORDER),
                             GetSystemMetrics(SM_CYBORDER) );
      }

   rc.left += borderSize.width() + 1;  // extra pel pad for title text
   if (style & WS_SYSMENU)
     rc.left += bmpSize.width();
   rc.top += borderSize.height();
   rc.right -= borderSize.width();
   if (style & WS_MAXIMIZEBOX)
     rc.right -= (bmpSize.width() + GetSystemMetrics(SM_CXBORDER));
   if (style & WS_MINIMIZEBOX)
     rc.right -= (bmpSize.width() + GetSystemMetrics(SM_CXBORDER));

   if ( (IPlatform::isWin95()) || (IPlatform::isNTNewShell()) )
   {
      // Account for the close button on the title bar
      if (style & WS_SYSMENU)
         rc.right -= (bmpSize.width() + GetSystemMetrics(SM_CXBORDER));
      // Account for the "whats this?" button on the title bar
      if (GetWindowLong( hTitle, GWL_EXSTYLE ) & WS_EX_CONTEXTHELP)
         rc.right -= (bmpSize.width() + GetSystemMetrics(SM_CXBORDER));
   }
   rc.bottom = rc.top + bmpSize.height() - 1;

   // Map the results (now in screen coordinates) to coordinates
   // of the owner of the ITitle, whose handle is this->handle() since
   // it is not a separate window.
   MapWindowPoints( 0, hTitle, (POINT*)&rc, 2 );

   // Adjust mapping since map points uses client rect for mapping
   unsigned long addYDelta = 0;
   if ( !(style & WS_CHILD) && GetMenu( hTitle ) )
     addYDelta = GetSystemMetrics( SM_CYMENU );
   borderSize.setHeight( borderSize.height() + GetSystemMetrics(SM_CYCAPTION)
                         + addYDelta );
   rc.right  += borderSize.width();
   rc.left   += borderSize.width();
   rc.top    += borderSize.height();
   rc.bottom += borderSize.height();
   return IRectangle(rc);
}

/*------------------------------------------------------------------------------
| ITitle::parentSize                                                           |
|                                                                              |
| Returns the size of the window owning the title bar.                         |
------------------------------------------------------------------------------*/
ISize ITitle::parentSize ( ) const
{
   RECT rect;
   GetWindowRect(this->handle(), &rect);
   ITRACE_ALL( IString("parentSize=") +
               ISize(rect.right - rect.left, rect.bottom - rect.top ).asString() );
   return  ISize(rect.right - rect.left, rect.bottom - rect.top );
}

#endif
